// Default handler.

inherit "roxenlib";

import Sitebuilder;
import MetaData;		// To get the conflict resolution colors.

object parent;

void create( object p )
{
  parent = p;
}

int binaryp()
{
  return 1;
}

object|string|mapping view( object fd, mapping headers, object id,
			    mapping md, object vcfile )
{
  return fd;
}

mapping viewsource(object vcfile, object id)
{
  throw(({"Internal error: viewsource() called in non-text handler.",
	  backtrace()}));
}

// string preview (object vcfile, object id);
// - Define to return a preview that's shown in the file browser.

// string edit (object fd, object id);
// - Define to return an edit form for the file.

// string edit_done (object fd, object id);
// - Define to handle the result of the edit form.

int upload (object vcfile, object id, object vcfd)
{
  return 1;
}

object|string|mapping download (object fd, mapping headers, object id,
				mapping md, object vcfile)
{
  return fd;
}


// Conflict resolution.

int merge_analyze (object(VCFileConflict) vcfc, void|object id)
// Return values: 0: no differences, 1: differences but no conflicts,
// 2: only decided conflicts, 3: undecided conflicts.
{
#ifdef SBDEBUG
  if (!mappingp (vcfc->state)) throw (({"vcfc->state not initialized\n", backtrace()}));
#endif
  mapping(string:mixed) state = vcfc->state;

  if (vcfc->a_data == vcfc->b_data) return 0;
  if (id && (<"a", "b">)[id->variables->data_sel])
    state->data_sel = id->variables->data_sel;
  else if (zero_type (state->data_sel)) state->data_sel = 0;

  if (vcfc->old_data)
    if (vcfc->b_data == vcfc->old_data) {
      if (!state->data_sel) state->data_sel = "a";
      if (id) id->variables->data_sel = state->data_sel;
      return 1;
    }
    else if (vcfc->a_data == vcfc->old_data) {
      if (!state->data_sel) state->data_sel = "b";
      if (id) id->variables->data_sel = state->data_sel;
      return 1;
    }

  if (id) id->variables->data_sel = state->data_sel;
  return state->data_sel ? 2 : 3;
}

string merge_form (object(VCFileConflict) vcfc, object id)
{
#ifdef SBDEBUG
  if (!mappingp (vcfc->state) || zero_type (vcfc->state->data_sel))
    throw (({"vcfc not analyzed or has no differences\n", backtrace()}));
#endif
  string res =
    "There's no support for conflict resolution on this file type.\n"
    "<p><ul>\n";

  string file = vcfc->this_vcfile->abspath (0, 0, 1);

  if (vcfc->old_md &&
      id->misc->sb->ct_data_for (vcfc->old_md["http-content-type"])->downloadp)
    res += "<li>View " + vcfc->state->link_rawview_old + ".\n";

  res +=
    "<li><var type=radio name=data_sel value=a> Use " + vcfc->state->link_rawview_a +
    (vcfc->old_data && vcfc->a_data != vcfc->old_data ? " (changed)" : "") + ".\n"
    "<li><var type=radio name=data_sel value=b> Use " + vcfc->state->link_rawview_b +
    (vcfc->old_data && vcfc->b_data != vcfc->old_data ? " (changed)" : "") + ".\n"
    "</ul>\n";

  return res;
}

string merge_verify (object(VCFileConflict) vcfc, object id)
{
  return merge_analyze (vcfc, id) > 2 &&
    "You must decide which version to use.";
}

string merge_done (object(VCFileConflict) vcfc, object id)
{
  if (merge_analyze (vcfc, id) > 2) return 0;
  switch (vcfc->state->data_sel) {
    case "a": return vcfc->a_data;
    case "b": return vcfc->b_data;
    default: throw (({"Undecided conflict\n", backtrace()}));
  }
}
