// Text handler

inherit "../default/handler.pike";

import Sitebuilder;
import MetaData;		// To get the conflict resolution colors.

int binaryp()
{
  return 0;
}

object|string|mapping view( object fd, mapping headers, object id,
			    mapping md, object vcfile )
{
  string enc = md->charset;
  if(!enc && md->language)
    enc = vcfile->wa->site->charset_for(md->language);
  if(enc)
    headers["content-type"]+="; charset=\""+replace(enc, "\"", "\\\"")+"\"";
  return ::view( fd, headers, id, md, vcfile );
}

mapping viewsource(object vcfile, object id)
{
  object vcres;

  string res;
  if (objectp (vcres = vcfile->view (id)))
    res=vcres->fd->read();
  else
    return 0;

  res=
    "<body bgcolor=white text=black>"
    "<h3>Source for "+html_encode_string(vcfile->name(id))+
    "</h3><table width=90% bgcolor=#eeeeee><tr><td><pre>"+
    html_encode_string(res)+
    "</pre></td></tr></table></body>";
  return http_string_answer(res,"text/html");
}

string edit( object fd, object id )
{
  mapping uservars = (id->misc->sb->mac->id_get_data(id)||([]));

  if(!id->variables->the_file)
    id->variables->the_file = fd->read();
  return
    ("<input type=hidden name=filename value='"+id->variables->filename+"'>"
     "<var parse name=the_file type=text rows="+(uservars->editrows||20)+
     " cols="+(uservars->editcols||80)+" wrap=off"+
//      id->misc->settings->rows+" cols="+
//      id->misc->settings->columns+" wrap="+
//      (id->misc->settings->wordwrapping=="enabled"?"physical":"off")+
     ">");
}

string edit_done(object fd, object id )
{
  string data = replace (id->variables->the_file, "\r\n", "\n");
  if (sizeof (data) && data[-1] != '\n') data += "\n";
  fd->write (data);
}

int upload( object vcfile, object id, object vcfd )
{
  // reopen() comes from VC.CVS.cachefile.TempFile. The primary reason
  // for this solution is that vcfd->upload_end() isn't yet called, so
  // we're safe from races.

  vcfd->fd->reopen ("r");
  string data = vcfd->fd->read();

  data = replace( data, "\r\n", "\n" );
  data = replace( data, "\r", "\n" );

  vcfd->fd->reopen ("cwt");
  vcfd->fd->write (data);

  return 1;
}

// Check user preferences, send file with PC, Mac or Unix \n
// PC: \r\n   Mac: \r    Unix: \n
object|string|mapping download( object fd, mapping headers, object id,
				mapping md, object vcfile )
{
  string nl = ((id->misc->sb->mac->id_get_data(id)||([]))->nl)||"\012";

  return (nl!="\n"? replace(fd->read(),"\n",nl) : fd);
}


// Conflict resolution.

#define CONTEXT_LINES 3
#define BLOCK(text) ("<pre>\n" + html_encode_string (text) + "\n" /*"</pre>"*/)
// Netscape adds a newline when </pre> is the last thing in a table
// cell. Just leaving the <pre> unclosed avoids that.. :P

int merge_analyze (object(VCFileConflict) vcfc, void|object id)
// Return values: 0: no differences, 1: differences but no conflicts,
// 2: only decided conflicts, 3: undecided conflicts.
{
#ifdef SBDEBUG
  if (!mappingp (vcfc->state)) throw (({"vcfc->state not initialized\n", backtrace()}));
#endif
  mapping(string:mixed) state = vcfc->state;

  if (zero_type (state->text_conflicts)) {
    state->text_conflicts =
      vcfc->create_diff (lambda (string s) {return s / "\n";},
			 lambda (array(string) a) {return a * "\n";});
    if (!state->text_conflicts) return 0;
  }

  mapping(string:string) variables = id ? id->variables : ([]);
  state->text_missing = ({});
  if (!state->text_sel)
    state->text_sel = allocate (sizeof (vcfc->diff)); // Allocates a bit too much.
  array(string) sel = state->text_sel;

  for (int nr = 0, i = 0; i < sizeof (vcfc->diff); i++) {
    string|array(string) block = vcfc->diff[i];
    if (arrayp (block)) {
      nr++;
      if ((<"a", "b">)[variables["data_blk" + nr]])
	sel[nr] = variables["data_blk" + nr];
      else {
	if (!sel[nr])
	  if (!block[1])	// New in a, delete in b, or conflict.
	    if (sizeof (block) < 3 || !block[2]) sel[nr] = "a";
	    else if (block[0] == block[2]) sel[nr] = "b";
	    else state->text_missing += ({nr});
	  else if (!block[0])	// New in b, delete in a, or conflict.
	    if (sizeof (block) < 3 || !block[2]) sel[nr] = "b";
	    else if (block[1] == block[2]) sel[nr] = "a";
	    else state->text_missing += ({nr});
	  else			// Change or conflict.
	    if (sizeof (block) < 3 || block[0] != block[2] && block[1] != block[2])
	      state->text_missing += ({nr});
	    else if (block[0] == block[2]) sel[nr] = "b";
	    else sel[nr] = "a";
	if (sel[nr]) variables["data_blk" + nr] = sel[nr];
      }
    }
  }

  return sizeof (state->text_missing) ? 3 : state->text_conflicts;
}

string merge_form (object(VCFileConflict) vcfc, object id)
{
#ifdef SBDEBUG
  if (!mappingp (vcfc->state) || !vcfc->state->text_sel)
    throw (({"vcfc not analyzed or has no differences\n", backtrace()}));
#endif
  array(string) sel = vcfc->state->text_sel;

  string res =
    "<table width=100% cellspacing=0 cellpadding=0 border=0 bgcolor=#000000><tr><td>"
    "<table width=100% cellspacing=0 cellpadding=3 border=0 bgcolor=#eeeeee>\n"
    "<tr valign=bottom><th>&nbsp;&nbsp;#</th>"
    "<th bgcolor=" + color_a + ">" + vcfc->state->msg_a_column + "</th>"
    "<th bgcolor=" + color_b + ">" + vcfc->state->msg_b_column + "</th>"
    "<th align=left width=99%>Text</th></tr>\n"
    "<tr></tr>\n";

  for (int nr = 0, i = 0; i < sizeof (vcfc->diff); i++) {
    string|array(string) block = vcfc->diff[i];
    if (stringp (block)) {	// Common text.
      array(string) parts = block / "\n";
      int notfirst = i > 0, notlast = i < sizeof (vcfc->diff) - 1;
      if (sizeof (parts) > (CONTEXT_LINES << (notfirst && notlast))) {
	res += "<tr>"
	  "<td>&nbsp;</td>"
	  "<td bgcolor=" + color_a + ">&nbsp;</td>"
	  "<td bgcolor=" + color_b + ">&nbsp;</td>\n"
	  "<td><table width=100% cellspacing=0 cellpadding=0 border=0>\n";
	if (notfirst)
	  res += "<tr><td>" + BLOCK (parts[..CONTEXT_LINES - 1] * "\n") +
	    "</td></tr>\n";
	res += "<tr><td align=center><b>//</b></td></tr>\n";
	if (notlast)
	  res += "<tr><td>" + BLOCK (parts[sizeof (parts) - CONTEXT_LINES..] * "\n") +
	    "</td></tr>\n";
	res += "</table></td></tr>\n";
      }
      else
	res += "<tr>"
	  "<td>&nbsp;</td>"
	  "<td bgcolor=" + color_a + ">&nbsp;</td>"
	  "<td bgcolor=" + color_b + ">&nbsp;</td>\n"
	  "<td>" + BLOCK (block) + "</td></tr>\n";
    }
    else {
      nr++;
      string cl = sizeof (block) > 2 && block[0] == block[2] ? color_old : color_a;
      string cr = sizeof (block) > 2 && block[1] == block[2] ? color_old : color_b;
      string cn = cl == color_old || sizeof (block) < 3 && !block[0] ? cr :
	(cr == color_old || sizeof (block) < 3 && !block[1] ? cl : color_conflict);

      if (!block[1])		// New in a, delete in b, or conflict.
	res += "<tr>"
	  "<td" + (cn == color_conflict ?
		   " bgcolor=" + color_conflict : "") + " align=right>"
	  "<a name=data_blk" + nr + ">" + nr + "</a></td>"
	  "<td bgcolor=" + cl + ">"
	  "<var type=radio name=data_blk" + nr + " value=a>" +
	  (cl == color_old ? "&nbsp;Old&nbsp;" : "") + "</td>"
	  "<td bgcolor=" + cr + ">"
	  "<var type=radio name=data_blk" + nr + " value=b>&nbsp;Del&nbsp;</td>\n"
	  "<td bgcolor=" + cl + ">" + BLOCK (block[0]) + "</td></tr>\n";

      else if (!block[0])	// New in b, delete in a, or conflict.
	res += "<tr>"
	  "<td" + (cn == color_conflict ?
		   " bgcolor=" + color_conflict : "") + " align=right>"
	  "<a name=data_blk" + nr + ">" + nr + "</a></td>"
	  "<td bgcolor=" + cl + ">"
	  "<var type=radio name=data_blk" + nr + " value=a>&nbsp;Del&nbsp;</td>"
	  "<td bgcolor=" + cr + ">"
	  "<var type=radio name=data_blk" + nr + " value=b>" +
	  (cr == color_old ? "&nbsp;Old&nbsp;" : "") + "</td>\n"
	  "<td bgcolor=" + cr + ">" + BLOCK (block[1]) + "</td></tr>\n";

      else			// Change or conflict.
	res += "<tr>"
	  "<td" + (cn == color_conflict ?
		   " bgcolor=" + color_conflict : "") + " align=right rowspan=2>"
	  "<a name=data_blk" + nr + ">" + nr + "</a></td>"
	  "<td bgcolor=" + cl + ">"
	  "<var type=radio name=data_blk" + nr + " value=a>" +
	  (cl == color_old ? "&nbsp;Old&nbsp;" : "") + "</td>"
	  "<td bgcolor=" + cl + ">&nbsp;</td>\n"
	  "<td bgcolor=" + cl + ">" + BLOCK (block[0]) + "</td></tr>\n"
	  "<tr>"
	  "<td bgcolor=" + cr + ">&nbsp;</td>"
	  "<td bgcolor=" + cr + ">"
	  "<var type=radio name=data_blk" + nr + " value=b>" +
	  (cr == color_old ? "&nbsp;Old&nbsp;" : "") + "</td>\n"
	  "<td bgcolor=" + cr + ">" + BLOCK (block[1]) + "</td></tr>\n";
    }

    res += "<tr></tr>\n";
  }

  return res + "</table></td></tr></table>\n";
}

string merge_verify (object(VCFileConflict) vcfc, object id)
{
  if (merge_analyze (vcfc, id) < 3) return 0;

  if (sizeof (vcfc->state->text_missing) == 1)
    return "Conflict #" + vcfc->state->text_missing[0] + " must be decided.";
  else
    return "Conflicts " + String.implode_nicely (
      Array.map (vcfc->state->text_missing, lambda (int i) {return "#" + i;})
    ) + " must be decided.";
}

string merge_done (object(VCFileConflict) vcfc, object id)
{
  if (merge_analyze (vcfc, id) > 2) return 0;

  array(string) new = allocate (sizeof (vcfc->diff));
  array(string) sel = vcfc->state->text_sel;

  for (int nr = 0, i = 0; i < sizeof (new); i++) {
    string|array(string) block = vcfc->diff[i];
    if (stringp (block)) new[i] = block;
    else switch (sel[++nr]) {
      case "a": new[i] = block[0]; break;
      case "b": new[i] = block[1]; break;
      default: throw (({"Undecided conflict\n", backtrace()}));
    }
  }

  return (new - ({0})) * "\n";
}
