// This is a Roxen module.

// Use an AC database as Roxen user database.

string cvs_version = "$Id: ac_userdb.pike,v 1.7 1999/07/01 11:59:37 mast Exp $";

#include <module.h>
inherit "module";
inherit "roxenlib";
inherit "aclib";

object config;


// Connect to an acdb.

string config_sel_name (object conf)
{
  return conf->name +
    (conf->query_name() != conf->name ? " (" + conf->query_name() + ")" : "");
}

mapping(string:object) ac_list()
{
  mapping acvar = roxen->query_var ("AC");
  mapping(object:object) loaders = acvar && acvar->loaders;
  if (!loaders) return ([]);

  mapping(string:object) list = ([]);
  foreach (indices (loaders), object conf)
    list[config_sel_name (conf)] = conf;
  return list;
}

array(object) ac_get()
{
  mapping acvar = roxen->query_var ("AC");
  if (!acvar || !acvar->loaders) return 0;
  if (QUERY (ac_config) == "all")
    return values (acvar->loaders);
  else {
    object ac_config = ac_list()[QUERY (ac_config)];
    if (!ac_config) return 0;
    return ({acvar->loaders[ac_config]});
  }
}


// Module interface.

void create (object conf)
{
  if (!conf) return;
  config = conf;

  mapping(string:object) ac_sel_list = ac_list();
  string ac_sel = config_sel_name (config);
  if (!ac_sel_list[ac_sel]) ac_sel = "none";

  defvar ("ac_config", ac_sel,
	  "AC database", TYPE_STRING_LIST,
	  "The server containing the AC database to use. The 'all' alternative "
	  "authenticates against all databases in all virtual servers.",
	  ({"none"}) + indices (ac_sel_list) + ({"all"}));
}

string status()
{
  variables["ac_config"][VAR_MISC] = ({"none"}) + indices (ac_list()) + ({"all"});
  return ac_get() ?
    (QUERY (ac_config) == "all" ? "Using AC databases in all servers." :
     "Using AC database in server " + QUERY (ac_config) + ".") :
    "<b>Not connected to any AC database.</b>";
}

array register_module()
{
  return ({MODULE_AUTH, "AC: User Database",
	   "Uses an Access Control database as the source for users and "
	   "passwords to the Roxen user database. This is typically "
	   "needed to get proper authentication with the FTP protocol. "
	   "This module can access an AC database in any virtual server, "
	   "or all databases in all servers.",
	   0, 1});
}

void start (int when, object conf)
{
  config = conf;
}

// MODULE_AUTH interface.

// FIXME: Add some cache?

array|int auth (string *auth, object req_id)
{
  array(object) acs = ac_get();
  if (!acs) return 0;
  acs = Array.filter (acs, lambda (object ac) {return ac->acdb;});

  array(object(Identity)) ids = ({});
  foreach (acs, object ac) {
    array(object(Identity)) ac_ids = ({});
    foreach (indices (ac->acauth_modules), object mod)
      if (mod && mod->global_passive_auth)
	foreach (mod->get_protocols(), string prot)
	  if (prot == req_id->prot[..sizeof(prot)-1]) {
	    array(object(Identity)) list =
	      ac->acdb->try_call (mod, "global_passive_auth", ac->acdb, req_id);
	    if (list) ids += list;
	  }
    if (sizeof (ac_ids)) ids += ac_ids;
    else {
      object(Identity) id = ac->acdb->check_backdoor (req_id);
      if (id) ids += ({id});
    }
  }
  if (!sizeof (ids)) return 0;

  string *info;
  string handle;
  foreach (ids, object(Identity) id) {
    string *res = id->get_userinfo();
    if (!info) info = res, handle = id->handle();
    else if (handle != id->handle() || !equal (info, res)) {
      // Awkward situation. This seems like the best we can do.
      handle = (auth[1] / ":")[0];
      info = ({handle, "*", "", "", "", "", ""});
      break;
    }
  }

  req_id->misc->uid = info[2];
  req_id->misc->gid = info[3];
  req_id->misc->gecos = info[4];
  req_id->misc->home = info[5];
  req_id->misc->shell = info[6];
  return ({1, handle, 0});
}

string *userinfo (string handle)
{
  array(object) acs = ac_get();
  if (!acs) return 0;

  array(object(Identity)) ids = ({});
  foreach (acs, object ac)
    if (ac->acdb) {
      object(Identity) id = ac->acdb->identities->find_by_handle (handle);
      if (id) ids += ({id});
    }
  if (!sizeof (ids)) return 0;

  string *info;
  foreach (ids, object(Identity) id) {
    string *res = id->get_userinfo();
    if (!info) info = res;
    else if (!equal (info, res)) {
      // Awkward situation. This seems like the best we can do.
      info = ({handle, "*", "", "", "", "", ""});
      break;
    }
  }
  if (info[0] != handle) info = ({handle}) + info[1..];

  return info;
}

string *user_from_uid (int uid)
{
  array(object) acs = ac_get();
  if (!acs) return 0;

  multiset(object(Identity)) ids = (<>);
  foreach (acs, object ac)
    if (ac->acdb) {
      object(IdAuth) ia = ac->acdb->id_auths->find_by_auth_data (AUTH_FS_UID, uid, 0);
      if (ia) ids[ia->identity()] = 1;
    }
  if (!sizeof (ids)) return 0;

  string *info;
  string handle;
  foreach (indices (ids), object(Identity) id) {
    string *res = id->get_userinfo();
    if (!info) info = res, handle = id->handle();
    else if (handle != id->handle() || !equal (info, res)) {
      // Awkward situation. This seems like the best we can do.
      info = ({"ambiguous", "*", "", "", "", "", ""});
      break;
    }
  }
  if (info[0] != handle) info = ({handle}) + info[1..];

  return info;
}

string *userlist()
{
  array(object) acs = ac_get();
  if (!acs) return 0;

  array(string) handles = ({});
  foreach (acs, object ac)
    if (ac->acdb)
      handles += ac->acdb->identities->handles();

  // FIXME: Remove dups?
  return handles;
}
