// This is a Roxen module.

// Loader and administration UI to the access control system.

string cvs_version = "$Id: acadm.pike,v 1.107 1999/08/26 16:24:47 mast Exp $";

#include <config.h>
#include <module.h>
inherit "module";
inherit "roxenlib";
inherit "aclib";

import Array;
import String;

object config;

object(ModuleAC) mac;
constant PP_ROOT = "/";

multiset(object) acauth_modules = 0;
// FIXME: This doesn't account for remote modules.

mapping(int:array(string)) init_id_authmods = ([ID_TYPE_USER: ({"acintern_passwd"})]);
// Mapping that for each identity type holds a list of string
// identifiers for authentication modules for which IdAuth objects
// should be created for new identities.

array(string) init_pc_authmods = ({"acauth_httpbasic"});
// Ordered list of initial authentication modules that should be added
// to new pclasses.


// Database I/O.

object(AC_DB) online_acdb = 0;

array(object) backends = ({});

string last_io_error;		// Used by the file backend.
int failed_connections;		// Used by the remote backend.
object remote;			// The remote backend object.

object online_db, acdb;		// Compatibility measures.

object(AC_DB) create_acdb (object conf)
{
  object(AC_DB) acdb = AC_DB (conf);
  acdb->superuser = lambda() {return QUERY (superuser);};
  return acdb;
}

void init_acdb (object(AC_DB) acdb)
{
  acdb->identities->everyone();
  acdb->identities->init_authmods =
    mkmapping (indices (init_id_authmods),
	       map (values (init_id_authmods),
		    lambda (string *arr) {
		      return map (arr,
				  lambda (string name) {
				    return config->modules[name]->enabled ||
				      config->modules[name]->master;
				  });
		    }));
  acdb->pclasses->init_authmods = map (init_pc_authmods,
				       lambda (string name) {
					 return config->modules[name]->enabled ||
					   config->modules[name]->master;
				       });
  online_acdb = acdb;
  online_db = class {object(AC_DB) acdb;}();

  // Compatibility measures.
  online_db->acdb = acdb;
  this_object()->acdb = acdb;
}


// Admin UI.

static constant BODY_COLOR =
  "bgcolor=white text=black link=black vlink=black alink=red",
  MSG_BOX_BG = "#eeeeee",
  UNIT_GIF = "<img src=internal-roxen-unit height=1 width=1 alt=\"\">",
  ERRMSG_BEG = "<font color=red><strong>", ERRMSG_END = "</strong></font>",
  DELMSG_BEG = "<font color=red>", DELMSG_END = "</font>";

static constant
  ID_HDRHDR_BG =	"#88b4bc",
  ID_HDR_BG =		"#a8d4dc",
  ID_TYPEHDR_BG =	({"#88bcb4", ID_HDRHDR_BG}),
  ID_TYPE_BG =		({"#a8dcd4", ID_HDR_BG}),
  ID_MEM_BG =		({"#b4b4f0", "#94d094", "#94d094", "#94d094",
			  "#b4f0b4", "#b4f0b4", "#b4f0b4", "#b4f0b4",
			  "#b4b4f0", "#b4b4f0", "#b4b4f0", "#b4b4f0",
			  0, 0, 0, 0,
			  "#c0d8c0", "#c0d8c0"}),
  ID_MEM_SYM =		({"-", "a", "m", "am",
			  "M", "Ma", "Mm", "Mam",
			  "(-)", "(a)", "(m)", "(am)",
			  0, 0, 0, 0,
			  "&lt;", "a&lt;"});

static constant
  PERM_HDRHDR_BG =	ID_HDRHDR_BG,
  PERM_HDR_BG =		ID_HDR_BG,
  PERM_CLSHDR_BG =	"#80c4b4",
  PERM_CLS_BG =		"#a0e4d4",
  PERM_BG =		({"#b4b4f0", "#b4f0b4", "#f0b4b4"}),
  PERM_UNSET_SYM =	({"-", "-", "-"}),
  PERM_SET_SYM =	({"N", "R", "W"}),
  PERM_IMPL_SYM =	({"n", "r", "w"});

static constant
  RSC_CLS_EMPTY =	"empty",
  RSC_HDR_BG =		"#acb498",
  RSC_PCLASS_BG =	"#ccd4b8",
  RSC_PPOINT_BG =	"#c4dcb8",
  RSC_NEW_BG =		"#a0a0b8",
  RSC_MHDR_BG =		"#9c9c70",
  RSC_MPCLASS_BG =	"#bcbc90",
  RSC_MPPOINT_BG =	"#b4c490",
  RSC_MNEW_BG =		"#907890";

static constant NO_UNDO_NOTICE =
  "<em>This action can not be undone.</em>\n";

static constant DESC_ID_EVERYONE =
  "<em>This is a special group that can't be deleted. If a request fails "
  "all authentication attempts it will get this identity.\n"
  "Every newly created identity is also a member of this group by default.\n"
  "This means that all resources that this identity has permissions for are "
  "publicly accessible.</em>\n";

static constant DESC_ID_DELETE =
  "All authentication information associated with this identity will also "
  "be deleted.\n"
  "If it is a group, identities that are members of it will not be deleted, "
  "just their memberships.\n"
  "\n<p>" + NO_UNDO_NOTICE;

static constant DESC_IMP_AUTH =
  "The permission has no effect because there are no active authentication "
  "modules in common between the identity and the protection class.\n"
  "To solve the problem, activate more authentication modules below and "
  "make sure that the authentication data permits login.\n";

static constant DESC_PC_REALM =
  "The realm specifies the set of resources where the same authentication "
  "applies.\n"
  "It is also often used to describe the resources to the user.\n";

static constant DESC_PC_DELETE =
  "The permission settings associated with this class will be lost.\n"
  "\n<p>" + NO_UNDO_NOTICE;

static constant DESC_PP_DELETE =
  "If this is the only protection point in the protection class, the class "
  "will be deleted too, along with all permissions that's been set on it.\n"
  "You can avoid this by moving the protection point to a new protection "
  "class before you delete it.\n"
  "\n<p>" + NO_UNDO_NOTICE;

#ifdef DEVELOPER
static constant DESC_PP_DEVEL_DELETE =
  "You're about to delete a protection point.\n"
  "This is normally permitted only by the module that registered it, but "
  "since you run in developer mode you may override.\n"
  "<strong>You better know what you're doing; this action might render the "
  "module completely FUBAR, and there is no way to add a protection point "
  "from here.</strong>\n"
  "\n<p>" + NO_UNDO_NOTICE;
#endif

static constant DESC_PROT_MOVE =
  "The \"Move\" button moves all the protection points marked by the "
  "checkboxes to the class selected by the radio buttons.\n";

static constant CONFIG_DESC =
  "Contains the AC system used by other modules to authenticate users to "
  "resources.";

static constant DESC_AC_INTRO =
  "This interface allows viewing and editing of the AC database used "
  "for authenticating users and resolving permissions.\n"
  "A terse description of the key concepts is given below for reference.\n"
  "\n<ul>\n"
  "\n<p><li><em>Identities</em> are given to incoming requests.\n"
  "An identity may be either a <em>user</em> or a <em>group</em> of other "
  "users and groups.\n"
  "An identity may be authenticated by zero or more authentication "
  "modules.\n"
  "There is a special group identity, normally called \"" + EVERYONE_NAME +
  "\", which is given to all requests that isn't successfully "
  "authenticated.\n"
  "Every identity is by default a member of this group when created.\n"
  "\n<p><li><em>Permissions</em> are set for identitities to access "
  "resources.\n"
  "It may be either <em>none</em> (the default), <em>read</em>, or "
  "<em>write</em>.\n"
  "If a permission is not assigned directly, it defaults to the highest "
  "permission inherited from any of the groups the identity is a (direct or "
  "indirect) member of.\n"
  "\n<p><li><em>Resources</em> are the information items provided by Roxen "
  "modules that are protected by the AC system.\n"
  "For each piece of data being protected, a <em>protection point</em> is "
  "registered.\n"
  "A number of protection points forms a <em>protection class</em>, for "
  "which permissions may be set.\n"
  "When a protection point is registered it will normally get a class of "
  "its own.\n"
  "Every protection class contains a list of authentication modules that "
  "will be run when its data is accessed.\n"
  "The first module in the list that successfully authenticates the request "
  "determines the identity.\n"
  "If no module in the list is successful, the special \"" + EVERYONE_NAME +
  "\" identity is used.\n"
  "\n</ul>\n"
  "\n<p>Note that the effective permission for a request is determined both "
  "by the permission settings and the list of active authentication "
  "modules in the protection class.\n"
  "The permission matrix will reflect this and warn for permission settings "
  "that doesn't have any effect.\n";

static constant DESC_DB_PURGE =
  "This will remove all references to modules that are not successfully "
  "loaded at the moment.\n"
  "This includes protection points belonging to the modules, or "
  "authentication data used by missing authentication modules.\n"
  "\n<p>" + NO_UNDO_NOTICE;

static constant
  TAG_ENTER = "_e", TAG_RETURN = "_r", TAG_GOTO = "_g", TAG_STATE = "_s",
  TAG_CHNGUSER = "_chg";

#define URL_STATE(pre) ((pre) + TAG_STATE + "=" + state->seq + \
			(sizeof (state->id) ? "," + state->id : ""))
#define FORM_STATE ("<input type=hidden name=" + TAG_STATE + " value=" + state->seq + \
		    (sizeof (state->id) ? "," + state->id : "") + ">\n")

class State
{
  string id;
  int timestamp;

  string *nestlist = ({});	// Stack of canonical URLs of visited pages.
  string last;			// Canonical URL of current page.
  int seq = 1;			// Sequence number.
  int seqstate = 0;		// 0: in sequence, 1: reloaded, 2: out of sequence.

  object(AC_DB) edb;
  multiset(int) unfolds = (<>);
  int change_user = 0;		// 1 if we're changing user by forcing a new login.
  int id_page;			// Current page in the Identities list.
}

static mapping(string:object(State)) states = ([]);

static object(State) get_state (object req_id, object(AC_DB) acdb)
{
  string userid = 0;
  int userseq = 0;
  object(State) state = 0;

  if (req_id->variables[TAG_STATE])
    sscanf (req_id->variables[TAG_STATE], "%d,%s", userseq, userid);

  if (userid)
    state = states [userid];
  else if (req_id->cookies["RoxenUserID"])
    state = states [req_id->cookies["RoxenUserID"]];

  if (!state || state->edb != acdb) {
    state = State();
    state->edb = acdb;
    if (userid) {
      state->id = userid;
      states[state->id] = state;
    }
    else if (req_id->cookies["RoxenUserID"]) {
      state->id = "";
      states[req_id->cookies["RoxenUserID"]] = state;
    }
    else {
      random_seed (sizeof (states) + time (1));
      int i = random (65535);
      while (states[(string) i]) i = random (65525);
      state->id = (string) i;
      states[state->id] = state;
    }
  }

  state->seqstate = 2;
  if (userseq)
    if (state->seq == userseq) state->seqstate = 0;
    else if (state->seq == userseq + 1) state->seqstate = 1;
  state->seq = userseq + 1;

  state->timestamp = time (1);	// Not really race safe, but it's pretty harmless.

  return state;
}

static void garb_states()
{
  int garbtime = time (1) - 60*60;
  foreach (indices (states), string userid)
    if (states[userid]->timestamp < garbtime) m_delete (states, userid);
  call_out (garb_states, 60*60);
}

class LocPath
{
  inherit "roxenlib";

  string *upath = ({});		// Handled URL path elements.
  string *dpath;		// URL path elements left to handle.
  string up;			// Canonical URL of parent page.
  string *uplist = ({});	// List of canonical URLs of all parent pages.
  string here;			// Canonical URL of current page.

  object(State) state;
  object req_id;

  int tabchoice = -1;

  void copy (object path)
  {
    upath = copy_value (path->upath);
    dpath = copy_value (path->dpath);
    up = path->up;
    uplist = copy_value (path->uplist);
    here = path->here;
    state = path->state;
    req_id = path->req_id;
  }

  void create (object|array(string) path, void|string mountpoint,
	       void|object(State) init_state, void|object init_req_id)
  {
    if (query_num_arg() == 1) copy (path);
    else {
      dpath = path;
      up = here = mountpoint;
      if (here[-1] == '/') here = here[..sizeof (here) - 2];
      state = init_state;
      req_id = init_req_id;
    }
  }

  string|array(string) head (void|int levels)
  {
    if (!levels)
      return sizeof (dpath) && dpath[0];
    else
      return sizeof (dpath) >= levels && dpath[..levels-1];
  }

  int has_head (void|int levels)
  {
    return levels ? sizeof (dpath) >= levels : sizeof (dpath);
  }

  string|array(string) descend (void|int levels)
  {
    int single = !levels;
    if (single) levels = 1;
    int n = sizeof (dpath);
    string *path = dpath[..levels-1];
    while (sizeof (path) < levels) path += ({""});
    upath += path;
    dpath = dpath[levels..];
    if (sizeof (uplist)) uplist = ({up = here}) + uplist;
    else uplist = ({up});
    here += "/" + path * "/";
    return single ? n && path[0] : path[..n-1];
  }

  static inline string form_vars (void|string|array(string)|mapping(string:string) vars)
  {
    string s = FORM_STATE;
    if (vars) {
      if (stringp (vars)) vars = vars / "&";
      array n, v;
      if (mappingp (vars)) n = indices (vars), v = values (vars);
      else {
	array t = Array.map (vars, `/, "=");
	n = column (t, 0);
	v = Array.map (t, lambda (array a) {return a[1..] * "=";});
      }

      for (int i = 0; i < sizeof (n); i++)
	s += "<input type=hidden name=" + html_encode_tag_value (n[i]) +
	  " value=" + html_encode_tag_value (v[i]) + ">\n";
    }
    return s;
  }

  static inline string url_vars (void|string|array(string)|mapping(string:string) vars,
				 void|string extra)
  {
    string s = "";
    if (vars) {
      if (stringp (vars)) s = vars;
      else if (mappingp (vars))
	foreach (indices (vars), string n) s += "&" + n + "=" + vars[n];
      else s = vars * "&";
    }
    if (extra) s += "&" + extra;
    s += URL_STATE("&");
    if (s[..0] == "&") s = s[1..];
    return sizeof (s) ? "?" + s : "";
  }

  mixed error_back()
  // Returns a redirect that is appropriate when the URL is malformed.
  // Should not be used to handle other error conditions.
  {
    return http_redirect (up + "?" + TAG_GOTO + "=" + URL_STATE("&"), req_id);
  }

  mixed redir_back()
  {
    uplist = uplist[1..];
    return http_redirect (up + URL_STATE("?"), req_id);
  }

  mixed redir_revisit (void|string|array(string)|mapping(string:string) vars)
  {
    return http_redirect (here + url_vars (vars), req_id);
  }

  string url_revisit (void|string|array(string)|mapping(string:string) vars)
  {
    return html_encode_tag_value (here + url_vars (vars));
  }

  string form_revisit (void|string|array(string)|mapping(string:string) vars)
  {
    return "<form action=" + html_encode_tag_value (here) + ">\n" +
      form_vars (vars);
  }

  mixed redir_replace (string|array(string) dir,
		       void|string|array(string)|mapping(string:string) vars)
  {
    if (arrayp (dir)) dir = dir * "/";
    return http_redirect (up + "/" + dir + url_vars (vars), req_id);
  }

  string url_replace (string|array(string) dir,
		      void|string|array(string)|mapping(string:string) vars)
  {
    if (arrayp (dir)) dir = dir * "/";
    return html_encode_tag_value (up + "/" + dir + url_vars (vars));
  }

  string form_replace (string|array(string) dir,
		       void|string|array(string)|mapping(string:string) vars)
  {
    if (arrayp (dir)) dir = dir * "/";
    return "<form action=" + html_encode_tag_value (up + "/" + dir) + ">\n" +
      form_vars (vars);
  }

  mixed redir_goto (string|array(string) dir,
		    void|string|array(string)|mapping(string:string) vars)
  {
    if (arrayp (dir)) dir = dir * "/";
    return http_redirect (dir + url_vars (vars, TAG_GOTO + "="), req_id);
  }

  string url_goto (string|array(string) dir,
		   void|string|array(string)|mapping(string:string) vars)
  // Kludge: Uses internal encoding (since it's used only with config_tablist).
  {
    if (arrayp (dir)) dir = dir * "/";
    return "'" + replace (dir + url_vars (vars, TAG_GOTO + "="), "'", "'\"'\"'") + "'";
  }

  string form_goto (string|array(string) dir,
		    void|string|array(string)|mapping(string:string) vars)
  {
    if (arrayp (dir)) dir = dir * "/";
    return "<form action=" + html_encode_tag_value (dir) + ">\n" +
      form_vars (vars) + "<input type=hidden name=" + TAG_GOTO + " value=>\n";
  }

  mixed redir_enter (string|array(string) dir,
		     void|string|array(string)|mapping(string:string) vars)
  {
    if (arrayp (dir)) dir = dir * "/";
    return http_redirect (dir + url_vars (vars, TAG_ENTER + "="), req_id);
  }

  string url_enter (string|array(string) dir,
		    void|string|array(string)|mapping(string:string) vars)
  {
    if (arrayp (dir)) dir = dir * "/";
    return html_encode_tag_value (dir + url_vars (vars, TAG_ENTER + "="));
  }

  string form_enter (string|array(string) dir,
		     void|string|array(string)|mapping(string:string) vars)
  {
    if (arrayp (dir)) dir = dir * "/";
    return "<form action=" + html_encode_tag_value (dir) + ">\n" +
      form_vars (vars) + "<input type=hidden name=" + TAG_ENTER + " value=>\n";
  }

  mixed redir_return (void|string|array(string)|mapping(string:string) vars)
  {
    string to;
    if (sizeof (state->nestlist)) to = state->nestlist[0];
    else if (sizeof (uplist)) to = uplist[0];
    else to = up;
    return http_redirect (to + url_vars (vars, TAG_RETURN + "="), req_id);
  }

  string url_return (void|string|array(string)|mapping(string:string) vars)
  {
    string to;
    if (sizeof (state->nestlist)) to = state->nestlist[0];
    else if (sizeof (uplist)) to = uplist[0];
    else to = up;
    return html_encode_tag_value (to + url_vars (vars, TAG_RETURN + "="));
  }

  string form_return (void|string|array(string)|mapping(string:string) vars)
  {
    string to;
    if (sizeof (state->nestlist)) to = state->nestlist[0];
    else if (sizeof (uplist)) to = uplist[0];
    else to = up;
    return "<form action=" + html_encode_tag_value (to) + ">\n" +
      form_vars (vars) + "<input type=hidden name=" + TAG_RETURN + " value=>\n";
  }
}

static mixed identities_view
  (object(LocPath) path, object(State) state, object req_id);
static mixed permissions_view
  (object(LocPath) path, object(State) state, object req_id);
static mixed resources_view
  (object(LocPath) path, object(State) state, object req_id);

static constant TAB_ID = 0;
static constant TAB_PERM = 1;
static constant TAB_RSC = 2;

static constant tabs = ({"Identities", "Permissions", "Resources"});
static array tab_handlers = ({identities_view, permissions_view, resources_view});

static object handlechrs = Regexp ("^[_0-9A-Za-z---]+$");

string http_encode_literal (string f)
// More thorough than http_encode_string.
{
  return replace (f, ({"\000", "\t", "\n", "\r", " ", "\"", "#", "%", "&", "'",
		       "+", "/", ":", ";", "<", "=", ">", "?", "@"}),
		  ({"%00", "%09", "%0a", "%0d", "%20", "%22", "%23", "%25", "%26", "%27",
		    "%2b", "%2f", "%3a", "%3b", "%3c", "%3d", "%3e", "%3f", "%40"}));
}

#if !constant (trim)
string trim (string s)
{
  sscanf (reverse (s), "%*[ \t]%s", s);
  sscanf (reverse (s), "%*[ \t]%s", s);
  return s;
}
#endif

string get_authmodname (object module, object acdb)
// Returns the descriptive name of an authentication module.
{
    string name = acdb->modulefullname (module);
    if (name && name[..3] == "AC: ") name = name[4..];
    return name;
}

static string header (object(State) state, string user)
{
  return "<html>\n<head><title>Access Control</title>\n"
    "<meta http-equiv=Expires content=0>\n"
    "</head>\n"
    "\n<body " + BODY_COLOR + ">\n"
    "\n<table width=100% border=0 cellpadding=0 cellspacing=0>\n"
    "<tr><td valign=bottom align=left>"
    "<img border=0 src=internal-roxen-roxen-icon-gray alt=\"\"></td>\n"
    "<td>&nbsp;</td>\n"
    "<td width=100% height=39><table width=100% border=0 cellpadding=0 cellspacing=0>\n"
    "<tr><td width=100% align=right valign=center height=28>"
    "User: " + user + " &nbsp;&nbsp; "
    "<a href=" + html_encode_tag_value (state->last + "?" +
					TAG_CHNGUSER + "=" + URL_STATE("&")) + ">"
    "Change</a></td></tr>\n"
    "<tr><td bgcolor=#003366 width=100% align=right height=12>"
    "<font color=white size=-2>Access Control Administrator Interface&nbsp;&nbsp;</font>"
    "</td></tr>\n"
    "</table></td></tr>\n"
    "</table><br>\n";
}

// The following "message boxes" are encapsulated in arrays to get
// them unaltered to find_file.

static array error_message (string msg, int backup,
			    object(LocPath) path, object(State) state)
{
  if (!backup) state->nestlist = ({path->here}) + state->nestlist;
  return ({"\n" + path->form_return() +
	   "<div align=center>\n"
	   "<table bgcolor=black border=0 cellspacing=0 cellpadding=1>\n"
	   "<tr><td><table bgcolor=" + MSG_BOX_BG + " border=0 "
	   "cellspacing=0 cellpadding=5>\n"
	   "<tr><td>\n"
	   "<font size=+2>" + ERRMSG_BEG + msg + ERRMSG_END + "</font>\n" +
	   "\n<p><div align=center>\n"
	   "<input type=submit value=\" Ok \">\n"
	   "</div>\n"
	   "</td></tr>\n"
	   "</table></td></tr>\n"
	   "</table>\n"
	   "</div>\n"
	   "</form>\n"});
}

static array confirm_message (string title, string msg,
			      string ok_val, string cancel_val,
			      object(LocPath) path, object(State) state)
{
  return ({"\n" + path->form_revisit() +
	   "<div align=center>\n"
	   "<table width=80% bgcolor=black border=0 cellspacing=0 cellpadding=1>\n"
	   "<tr><td><table width=100% bgcolor=" + MSG_BOX_BG +
	   " border=0 cellspacing=0 cellpadding=5>\n"
	   "<tr><td>\n"
	   "<font size=+2><b>" + title + "</b></font>\n"
	   "\n<hr size=1 noshade>\n\n" + msg +
	   "\n<p><div align=center>\n"
	   "<input type=submit name=" + ok_val + " value=\" Ok \">\n"
	   "<input type=submit name=" + cancel_val + " value=\" Cancel \">\n"
	   "</div>\n"
	   "</td></tr>\n"
	   "</table></td></tr>\n"
	   "</table>\n"
	   "</div>\n"
	   "</form>\n"});
}

static mixed id_auths_subview
  (object(LocPath) path, object(State) state, object req_id, object id)
{
  int access = mac->access_perm (PP_ROOT, req_id);
  object(IdAuth) ia = 0;

  object(IdAuth) *ias = state->edb->id_auths->find_by_id (id);

  // Parse and check.

  string dir = path->descend();

  if (dir && dir != "" && access == PERM_WRITE) {
    string s, error = 0;
    int iaid;
    sscanf (dir, "%D%s", iaid, s);
    if (s != "") return path->error_back();

    if (!iaid && req_id->variables["add"])
      if (id->type() == ID_TYPE_GROUP)
	error = "The identity is a group";
      else {
	string name = req_id->variables["module"];
	if (!name) return path->redir_back();
	object am = state->edb->findmodule (name);
	if (!acauth_modules[am])
	  error = "Authentication module \"" + html_encode_string (name) +
	    "\" doesn't exist";
	else if (zero_type (
	  ia = state->edb->try_call (am, "id_auth_init",
				     state->edb->id_auths->new (am, id))))
	  error = "The authentication module doesn't support explicit addition";
	else
	  if (ia) {
	    state->edb->id_auths->add (ia);
	    return path->redir_back();
	  }
	  else error = "Failed to add this authentication method to the identity";
      }

    else {
      ia = state->edb->id_auths[iaid];
      if (!ia || search (ias, ia) == -1) return path->error_back();

      if (req_id->variables["__del"]) {
	object am = ia->auth_module();
	int rem_ok = 1;
	if (am && zero_type (rem_ok = state->edb->try_call (am, "id_auth_destroy", ia)))
	  error = "The authentication module doesn't support explicit deletion";
	else
	  if (rem_ok) {
	    state->edb->id_auths->remove (ia);
	    return path->redir_back();
	  }
	  else error = "Failed to remove this authentication method from the identity";
      }
      else if (!req_id->variables["__ena"] == !ia->disabled())
	ia->set_disabled (!req_id->variables["__ena"]);
    }

    if (error) return error_message (error, 1, path, state);
  }

  // Make the page.

  if (id->type() == ID_TYPE_GROUP && access > PERM_READ) access = PERM_READ;

  string res;
  if (!sizeof (ias))
    res = "\n<h3>No authentication methods</h3>\n";
  else {
    res = "\n<h3>Authentication methods</h3>\n\n";

    string *res_ok = ({}), *res_del = ({});
    object(IdAuth) *ias_ok = ({}), *ias_del = ({});
    foreach (ias, object(IdAuth) iia) {
      string r;
      int changeable = 1, deletable = 1;

      object iam = iia->auth_module();
      if (!iam ||
	  zero_type (r = state->edb->try_call (iam, "id_auth_view",
					       iia, iia == ia && access == PERM_WRITE ?
					       req_id->variables : ([]), access))) {
	changeable = 0;
	r = ERRMSG_BEG + "Authentication module " +
	  (iia->auth_modname() ? html_encode_string (iia->auth_modname()) : "") +
	  " not found.\nIt may have been deleted." + ERRMSG_END + "\n";
      }
      else {
	deletable = state->edb->has_function (iam, "id_auth_destroy");
	if (!r) r = "No information available.\n";
	else if (iia == ia && sizeof (ia->table()->dirty)) return path->redir_back();
	r = "<dt><table width=100% border=0>\n"
	  "<tr><td><b>" + html_encode_string (iia->authmod_desc()) +
	  (iia->working() ? "" : " (authentication not possible)") + "</b></td>" +
	  (id->type() != ID_TYPE_GROUP ?
	   "<td align=right>" +
	   (access == PERM_WRITE ?
	    "<input type=checkbox name=__ena value=1" +
	    (iia->disabled() ? "" : " checked") + "> Enabled" :
	    (iia->disabled() ? "Disabled" : "Enabled")) + "</td>" :
	   "") + "</tr>\n"
	  "</table>\n"
	  "<dd>" + r;
      }

      r = "<table width=100% border=0><tr><td>" + // Netscape 4 kludge.
	path->form_replace ((string) iia->id()) +
	"<dl>" + r;
      if (access == PERM_WRITE) {
	if (changeable) r += "<input type=submit value=\" Change \">\n";
	if (deletable) r += "<input type=submit name=__del value=\" Delete \">\n";
      }
      r += "</dl>\n"
	"</form>"
	"</td></tr></table>\n"; // Netscape 4 kludge.

      if (iia->auth_module()) res_ok += ({r}), ias_ok += ({iia});
      else res_del += ({r}), ias_del += ({iia});
    }

    sort (ias_ok, res_ok);
    sort (ias_del, res_del);
    res += (res_ok + res_del) * "\n";
  }

  if (id->type() != ID_TYPE_GROUP)
    if (!sizeof (acauth_modules))
      res += "\n<p><b>There are no authentication modules installed "
	"that can be added</b>\n";
    else {
      res += "\n<p>"
	"<table width=100% border=0><tr><td>" + // Netscape 4 kludge.
	path->form_replace ("0") +
	"<b>" + (access == PERM_WRITE ? "Add a new authentication method:" :
		 "Available authentication modules:") + "</b> <select name=module>\n";
      object *mods = filter (indices (acauth_modules),
			     lambda (object mod, object(AC_DB) edb) {
			       return edb->has_function (mod, "id_auth_init");
			     }, state->edb);
      string *names = map (mods, get_authmodname, state->edb);
      sort (names, mods);
      for (int i = 0; i < sizeof (mods); i++)
	res += "<option value=" +
	  html_encode_tag_value (state->edb->modulename (mods[i])) + ">" +
	  html_encode_string (names[i]) + "\n";
      res += "</select>\n";
      if (access == PERM_WRITE)
	res += "<input type=submit name=add value=\" Add \">\n";
      res += "</form>"
	"</td></tr></table>\n";	// Netscape 4 kludge.
    }

  return res;
}

static mixed pclass_auths_subview
  (object(LocPath) path, object(State) state, object req_id, object(ProtClass) pc)
{
  int access = mac->access_perm (PP_ROOT, req_id);

  object *ams = pc->authmods();

  // Parse and check.

  string dir = path->descend();

  if (dir && dir != "" && access == PERM_WRITE) {
    string error = 0;

    if (dir == "0" && req_id->variables["add"]) {
      string name = req_id->variables["module"];
      if (!name) return path->redir_back();
      object am = state->edb->findmodule (name);
      if (!acauth_modules[am])
	error = "Authentication module \"" + html_encode_string (name) +
	  "\" doesn't exist";
      else if (!state->edb->has_function (am, "passive_auth"))
	error = "Authentication module \"" + html_encode_string (name) +
	  "\" can't do authentication";
      else if (search (ams, am) != -1)
	error = "The authentication module is already activated";
      else {
	int pos = req_id->variables["pos"] ?
	  (int) req_id->variables["pos"] : sizeof (ams);
	pc->add_authmod (am, pos);
	return path->redir_back();
      }
    }

    else if (req_id->variables["rem"]) {
      int pos = pc->authmod_pos (dir);
      if (pos == -1)
	error = "The authentication module isn't activated";
      else {
	pc->remove_authmod (pos);
	return path->redir_back();
      }
    }

    if (error) return error_message (error, 1, path, state);
  }

  // Make the page.

  string res;
  if (!sizeof (ams))
    res = "\n<h3>No activated authentication modules</h3>\n";
  else {
    res = "\n<h3>Active authentication modules</h3>\n"
      "<table width=100% border=0 cellspacing=0 cellpadding=2>\n";

    for (int pos = 0; pos < sizeof (ams); pos++) {
      object am = ams[pos];
      res += "<tr><td>";
      if (!am)
	res += ERRMSG_BEG + "Module \"" + html_encode_string (pc->authmod_name (pos)) +
	  "\" not found. It may have been deleted." + ERRMSG_END;
      else res += html_encode_string (get_authmodname (am, state->edb)) + "</td>"
	     "<td>" + am->get_protocols() * ", ";
      if (access == PERM_WRITE)
	res += "</td><td align=right><a href=" +
	  path->url_replace (http_encode_literal (pc->authmod_name (pos)), "rem=") +
	  ">" + DELMSG_BEG + "Remove" + DELMSG_END + "</a>";
      res += "</td></tr>\n";
    }

    res += "</table>\n";
  }

  {
    object *mods = filter (indices (acauth_modules),
			   lambda (object mod, object(AC_DB) edb) {
			     return edb->has_function (mod, "passive_auth");
			   }, state->edb) - ams;
    if (!sizeof (mods))
      res += "\n<p><b>No more modules to add</b>\n";
    else {
      res += "\n<p>"
	"<table width=100% border=0><tr><td>" + // Netscape 4 kludge.
	path->form_replace ("0") +
	"<b>" + (access == PERM_WRITE ? "Add a new authentication module:" :
		 "Available authentication modules:") + "</b> <select name=module>\n";
      string *names = map (mods, get_authmodname, state->edb);
      sort (names, mods);
      for (int i = 0; i < sizeof (mods); i++)
	res += "<option value=" +
	  html_encode_tag_value (state->edb->modulename (mods[i])) +
	  ">" + html_encode_string (names[i]) + "\n";
      res += "</select>\n";
      if (access == PERM_WRITE)
	res += "<input type=submit name=add value=\" Add \">\n";

      res += "</form>"
	"</td></tr></table>\n"; // Netscape 4 kludge.
    }
  }

  return res;
}

static mixed identity_view
  (object(LocPath) path, object(State) state, object req_id)
{
  int access = mac->access_perm (PP_ROOT, req_id);
  int idid;
  object id = 0;

  multiset errors = (<>);

  // Parse and check.

  string dir = path->descend();

  {
    string s;
    sscanf (dir, "%D%s", idid, s);
    if (s != "") return 0;
    id = state->edb->identities[idid];
  }

  string cmd;

  int type = -1;
  int src_auth_id = -1;
  string handle = 0;
  string name = 0;
  int params = 0;

  if (path->has_head()) cmd = "set";
  else {
    foreach (({"ret", "set", "del", "delok", "newgroup", "newuser", 0}), cmd)
      if (req_id->variables[cmd]) break;
    if (!cmd) cmd = "set";

    if (access == PERM_WRITE) {
      type = search (ID_TYPE_TEXTS, req_id->variables["type"]);
      if (req_id->variables["src_auth"])
	sscanf (req_id->variables["src_auth"], "%d", src_auth_id);
      handle = req_id->variables["handle"];
      name = req_id->variables["name"];
      params = type >= 0 || src_auth_id >= 0 || handle || name;
    }
  }

  string error = 0;
  int err_back = 0;

  switch (cmd) {
    case "newgroup":
    case "newuser":
      if (id)
	error = "Can't replace an existing identity", err_back = 1;
      else if (idid)
	error = "Can't create an identity with explicit ID", err_back = 1;
      break;

    case "set":
      if (idid && !id)
	error = "The identity does not exist", err_back = 1;
      else if (id && type != -1 && type != id->type())
	error = "Can't change the type of an existing identity";
      break;

    case "del":
    case "delok":
      if (!id)
	error = "The identity does not exist", err_back = 1;
      else if (id == state->edb->identities->everyone())
	error = "This is a special group that can't be deleted";
      break;

    case "ret":
    default:
      return path->redir_return();
  }

  if (error) return error_message (error, err_back, path, state);

  object(IdAuth) src_auth = id && id->src_auth();
  object(IdAuth) *src_auths = id &&
    filter (id->auths(), lambda (object(IdAuth) ia, object(AC_DB) edb) {
			   return !ia->disabled() && ia->auth_module() &&
			     edb->has_function (ia->auth_module(), "id_update_userinfo");
			 }, state->edb);
  if (src_auth) src_auths |= ({src_auth});

  if (access != PERM_WRITE) {
    if (id) {
      type = id->type();
      src_auth_id = id->src_auth() && id->src_auth()->id();
      handle = id->handle();
      name = id->name();
    }
  }

  else switch (cmd) {
    case "newgroup":
    case "newuser":
      type = cmd == "newgroup" ? ID_TYPE_GROUP : ID_TYPE_USER;

    case "set":
      if (type >= 0) {
	if (!id) id = state->edb->identities->new (type);
      }
      else if (id) type = id->type();

      if (type == -1 && id) type = id->type();

      if (id)
	if (src_auth_id > 0 && (!src_auth || src_auth_id != src_auth->id())) {
	  object(IdAuth) ia = state->edb->id_auths[src_auth_id];
	  if (ia && search (src_auths, ia) >= 0) id->set_src_auth (src_auth = ia);
	}
	else if (src_auth_id == 0 && src_auth) id->set_src_auth (src_auth = 0);

      if (id && src_auth) {
	id->update_userinfo();
	handle = id->handle();
	name = id->name();
      }
      else {

	if (handle) {
	  if (handle == "") handle = 0;
	  if (handle && !handlechrs->match (handle)) errors["handle_chrs"] = 1;
	  else {
	    object(Identity) other = state->edb->identities->find_by_handle (handle);
	    if (other && other != id) errors["handle_taken"] = 1;
	    else if (id && handle != id->handle())
	      id->set_handle (handle);
	  }
	}
	else if (id) handle = id->handle();

	if (name) {
	  if (name == "") name = 0;
	  if (id && name != id->name())
	    id->set_name (name);
	}
	else if (id) name = id->name();
      }

      if (!idid && id && !sizeof (errors) && params) {
	state->edb->identities->add (id);
	state->edb->identities->init_auth_methods (id);
      }

      if (!sizeof (errors) && params)
	return path->redir_replace ((string) id->id());

      break;

    case "del":
      return confirm_message (
	"Delete " + (id->name() ? ID_TYPE_TEXTS[id->type()] + ": " +
		     html_encode_string (id->name()) :
		     "nameless " + ID_TYPE_TEXTS[id->type()] + " (ID " + idid + ")"),
	DESC_ID_DELETE, "delok", "ret", path, state);

    case "delok":
      state->edb->identities->remove (id);
      if (sizeof (state->nestlist) && state->nestlist[0] == path->here)
	// Do not return to the page editing the deleted record.
	state->nestlist = state->nestlist[1..];
      return path->redir_return();
  }

  // Make the page.

  string ia_res = "";
  if (idid && id) {
    object(LocPath) orig_path = LocPath (path);
    mixed r = id_auths_subview (path, state, req_id, id);
    if (!stringp (r)) return r;
    ia_res = r;
    path->copy (orig_path);
  }

  string res;
  if (!idid) res = "\n<h2>New " + ID_TYPE_TEXTS[type] + "</h2>\n";
  else res = "\n<h2>" +
	 (name ? capitalize (ID_TYPE_TEXTS[type]) + ": " +
	  html_encode_string (name) :
	  "Nameless " + ID_TYPE_TEXTS[type] + " (ID " + idid + ")") +
	 "</h2>\n";

  if (id && id == state->edb->identities->everyone())
    res += "\n<p>" + DESC_ID_EVERYONE;

  res += "\n<p>";
  if (access == PERM_WRITE)
    res += "<table width=100% border=0><tr><td>" + // Netscape 4 kludge.
      path->form_revisit ((["type":ID_TYPE_TEXTS[type],
				 TAG_ENTER + "_var": "del"]));

  if (access != PERM_WRITE || id && src_auth)
    res += "Name: " + (name ? html_encode_string (name) : "") + "<br>\n"
      "Username: " + (handle ? html_encode_string (handle) : "") + "<br>\n";
  else {
    res += "Name: <input type=text name=name size=50" +
      (name ? " value=" + html_encode_tag_value (name) : "") + "><br>\n";

    res += "Username: <input type=text name=handle size=20" +
      (handle ? " value=" + html_encode_tag_value (handle) : "") + ">";
    {
      string *errs = ({});
      if (errors["handle_taken"]) errs += ({"must be unique if specified"});
      if (errors["handle_chrs"]) errs += ({"may only contain alphanumerical characters"});
      if (sizeof (errs))
	res += " " + ERRMSG_BEG + capitalize (errs * ", ") + "." + ERRMSG_END;
      res += "<br>\n";
    }
  }

  if (access != PERM_WRITE) {
    if (id && src_auth)
      res += "Controlled by: " +
	html_encode_string (src_auth->authmod_desc() ||
			    "(unknown authentication method)") + "<br>\n";
  }
  else if (src_auths && sizeof (src_auths)) {
    res += "Controlled by: <select name=src_auth>\n"
      "<option value=0" + (src_auth ? "" : " selected") + ">manually\n";
    foreach (sort (src_auths), object(IdAuth) ia)
      res += "<option value=" + html_encode_tag_value ((string) ia->id()) +
	(src_auth == ia ? " selected" : "") + ">" +
	html_encode_string (ia->authmod_desc() ||
			    "(unknown authentication method)") + "\n";
    res += "</select><br>\n";
  }

  if (access == PERM_WRITE) {
    res += "\n<p>"
      "<input type=submit name=set value=\" " + (idid ? "Change" : "Add") + " \">\n";
    if (idid && id != state->edb->identities->everyone())
      res += "<input type=submit name=del value=\" Delete \">\n";
    res += "</form>"
      "</td></tr></table>\n";	// Netscape 4 kludge.
  }

  return res + ia_res;
}

static mixed identities_view
  (object(LocPath) path, object(State) state, object req_id)
{
  int access = mac->access_perm (PP_ROOT, req_id);

  // Parse.

  if (path->has_head()) return identity_view (path, state, req_id);

  string cmd;
  foreach (({"add", "rem", "clr", 0}), cmd)
    if (req_id->variables[cmd]) break;

  if (cmd) {
    string s, error = 0;
    int idid, gid, err_back = 0;
    object(Identity) id, g;
    if (sscanf (req_id->variables[cmd], "%D,%D%s", idid, gid, s) != 3 || sizeof (s) ||
	!(id = state->edb->identities[idid]) || !(g = state->edb->identities[gid]) ||
	g->type() != ID_TYPE_GROUP)
      return path->redir_revisit();

    if (access == PERM_WRITE)
      switch (cmd) {
	case "add":
	  if (g->member (id))
	    error ="Can't create a cyclic membership";
	  else {
	    if (!id->incl_member (g))
	      id->add_incl_group (g);
	    return path->redir_revisit();
	  }
	  break;

	case "rem":
	  if (!id->excl_member (g))
	    id->add_excl_group (g);
	  return path->redir_revisit();

	case "clr":
	  if (id->incl_member (g))
	    id->remove_incl_group (g);
	  else if (id->excl_member (g))
	    id->remove_excl_group (g);
	  return path->redir_revisit();
      }

    if (error) return error_message (error, err_back, path, state);
  }

  if (req_id->variables->page) state->id_page = (int) req_id->variables->page;

  // Make the page.

  string res = "";

  array groups = sort (state->edb->identities->find_by_type (ID_TYPE_GROUP));
  array users = sort (state->edb->identities->find_by_type (ID_TYPE_USER));
  array ids;

  if (!QUERY(pagelength) || sizeof (groups) + sizeof (users) <= QUERY(pagelength)) {
    ids = groups + users;
    foreach (ids, object(Identity) id) id->update_userinfo (1);
  }

  else {
    if (state->id_page > sizeof (users) / QUERY(pagelength) + 1) state->id_page = 0;
    res += "\n<p>Page: ";
    if (state->id_page == 0) res += "<b>Groups</b>";
    else res += "<a href=\"" + tabs[TAB_ID] + "?page=0\">Groups</a>";

    for (int p = 1, i = 0; i < sizeof (users); p++, i += QUERY(pagelength)) {
      object(Identity) idfirst = users[i],
	idlast = users[min (i + QUERY(pagelength), sizeof (users)) - 1];
      string first = idfirst->name() || idfirst->handle();
      if (!first) first = (string) idfirst->id(); else first = first[..2];
      string last = idlast->name() || idlast->handle();
      if (!last) last = (string) idlast->id(); else last = last[..2];
      string s = replace (trim (first) + "-" + trim (last), " ", "&nbsp;");
      if (p == state->id_page) s = "<b>" + s + "</b>";
      else s = "<a href=\"" + tabs[TAB_ID] + "?page=" + p + "\">" + s + "</a>";
      res += " " + s;
    }

    if (state->id_page) {
      ids = users[(state->id_page - 1) * QUERY(pagelength) ..
		  state->id_page * QUERY(pagelength) - 1];
      foreach (groups + ids, object(Identity) id) id->update_userinfo (1);
    }
    else {
      ids = groups;
      foreach (groups, object(Identity) id) id->update_userinfo (1);
    }
  }

  res += "\n<br clear=all><p><table border=0 cellspacing=1 cellpadding=2>\n";

  res += "<tr><td></td><th bgcolor=" + ID_HDRHDR_BG + " colspan=" + sizeof (groups) + ">"
    "Memberships</th></tr>\n";
  res += "<tr valign=bottom><td></td>\n";
  foreach (groups, object(Identity) gid) {
    res += "<th bgcolor=" + ID_HDR_BG + ">";
    string s = "<gtext href=" + path->url_enter (tabs[TAB_ID] + "/" + gid->id()) +
      " bg=" + ID_HDR_BG + " nfont=lucida 4 rotate=90";
    if (gid->name()) res += s + ">" + gid->name() + "</gtext>";
    else res += "<em>" + s + " italic>Nameless (ID\240" + gid->id() + ")</gtext></em>";
    res += "</th>\n";
  }
  res += "</tr>\n";

  {
    int idtype = -1;
    foreach (ids, object(Identity) id) {

      if (id->type() != idtype) {
	idtype = id->type();
	res += "<tr bgcolor=" + ID_TYPEHDR_BG[idtype] + " align=left><th>" +
	  capitalize (ID_TYPE_TEXTS[idtype]) + "s</th>" +
	  strmult ("<td>&nbsp;</td>", sizeof (groups)) + "</tr>\n";
      }

      res += "<tr bgcolor=" + ID_TYPE_BG[idtype] + " align=center>"
	"<td align=left nowrap>"
	"<a href=" + path->url_enter (tabs[TAB_ID] + "/" + id->id()) + ">" +
	(id->name() ? id->name() : "<em>Nameless (ID " + id->id() + ")</em>") + "</a>";
      if (id->handle()) res += " <em>(" + id->handle() + ")</em>";
      res += "</td>";

      foreach (groups, object(Identity) gid) {
	int rel = (id == gid || gid->member (id)) && 16;
	if (!rel) {
	  if (id->excl_member (gid)) rel |= 8;
	  if (id->incl_member (gid)) rel |= 4;
	  if (id->indirect_member_ignore_excl (gid)) rel |= 2;
	}
	if (id->auto_member_ignore_excl (gid)) rel |= 1;
	string s = ID_MEM_SYM[rel];
	if (rel < 16 && access == PERM_WRITE)
	  s = "<a href=\"" + tabs[TAB_ID] + "?" +
	    (rel & 4 ? "rem" : (rel & 8 ? "clr" : "add")) + "=" +
	     id->id() + "," + gid->id() +
	     URL_STATE("&") + "\">" + s + "</a>";
	res += "<td bgcolor=" + ID_MEM_BG[rel] + ">" + s + "</td>";
      }
      res += "</tr>\n";
    }
    res += "</table>\n";
  }

  if (access == PERM_WRITE)
    res += "\n<p>"
      "<table width=100% border=0><tr><td>" + // Netscape 4 kludge.
      path->form_enter (path->here + "/0") +
      "<input type=submit name=newgroup value=\" New group \">\n"
      "<input type=submit name=newuser value=\" New user \">\n"
      "</form>"
      "</td></tr></table>\n";	// Netscape 4 kludge.

  return res;
}

static mixed auth_impossible_helper
  (object(LocPath) path, object(State) state, object req_id)
{
  int access = mac->access_perm (PP_ROOT, req_id);
  int perm;
  int idid, pcid;
  object(Identity) id;
  object(ProtClass) pc;

  // Parse.

  string dir = path->descend();

  if (access != PERM_WRITE) return 0;

  {
    string s;
    if (sscanf (dir, "%D,%D,%s", idid, pcid, s) != 3 ||
	!(id = state->edb->identities[idid]) || !(pc = state->edb->pclasses[pcid]) ||
	(perm = search (PERM_TEXTS, s)) == -1)
      return 0;
  }

  string cmd = 0;
  foreach (({"cancel", "ok", 0}), cmd)
    if (req_id->variables[cmd]) break;

  switch (cmd) {
    case "cancel":
      return path->redir_return();

    case "ok":
      return path->redir_return (([PERM_TEXTS[perm]: idid + "," + pcid,
				   "force": ""]));
  }

  // Make the page.

  string res;

  mapping orig_vars = req_id->variables;
  object(LocPath) orig_path = LocPath (path);

  {
    res = "\n<hr size=1 noshade>\n"
      "\n<h2>" + (id->name() ? capitalize (ID_TYPE_TEXTS[id->type()]) + ": " +
		  html_encode_string (id->name()) :
		  "Nameless " + ID_TYPE_TEXTS[id->type()] + " (ID " + idid + ")") +
      "</h2>\n";

    if (path->head() != "id") req_id->variables = ([]), path->dpath = ({"id"});
    path->descend();
    mapping r = id_auths_subview (path, state, req_id, id);
    req_id->variables = orig_vars;
    if (!stringp (r)) return r;
    res += r;
  }
  path = LocPath (orig_path);

  {
    res += "\n<hr size=1 noshade>\n"
      "\n<h2>" + (pc->name() ? "Protection class: " + html_encode_string (pc->name()) :
		  "Nameless protection class (ID " + pcid + ")") + "</h2>\n";

    if (path->head() != "pc") req_id->variables = ([]), path->dpath = ({"pc"});
    path->descend();
    mapping r = pclass_auths_subview (path, state, req_id, pc);
    req_id->variables = orig_vars;
    if (!stringp (r)) return r;
    res += r;
  }
  path->copy (orig_path);

  string intro;

  if (state->edb->possible_auth (id, pc))
    intro = "\n<h2>Authentication is possible</h2>\n";
  else
    intro = "\n<h2>No authentication possible</h2>\n"
      "\n<p>" + DESC_IMP_AUTH;

  intro += "\n<p>"
    "<table width=100% border=0><tr><td>" + // Netscape 4 kludge.
    path->form_revisit();

  intro += "<em>Setting " + PERM_DESC[perm] + " permission for " +
    (id->name() ? ID_TYPE_TEXTS[id->type()] +
     " \"" + html_encode_string (id->name()) + "\"" :
     "nameless " + ID_TYPE_TEXTS[id->type()] + " (ID " + idid + ")") + " on " +
    (pc->name() ? "protection class \"" + html_encode_string (pc->name()) + "\"" :
     "nameless protection class (ID " + pcid + ")") + ".</em><br>\n";

  intro += "\n<p>"
    "<input type=submit name=cancel value=\" Cancel \">\n"
    "<input type=submit name=ok value=\" Ok \">\n"
    "</form>"
    "</td></tr></table>\n";	// Netscape 4 kludge.

  return intro + res;
}

static mixed permissions_view
  (object(LocPath) path, object(State) state, object req_id)
{
  int access = mac->access_perm (PP_ROOT, req_id);

  // Parse.

  if (path->has_head()) return auth_impossible_helper (path, state, req_id);

  string cmd;
  foreach (PERM_TEXTS + ({"clr", 0}), cmd)
    if (req_id->variables[cmd]) break;

  if (cmd) {
    int perm = search (PERM_TEXTS, cmd);
    int idid, pcid;
    object(Identity) id;
    object(ProtClass) pc;
    string s;
    if (sscanf (req_id->variables[cmd], "%D,%D%s", idid, pcid, s) != 3 || sizeof (s) ||
	!(id = state->edb->identities[idid]) || !(pc = state->edb->pclasses[pcid]))
      return path->redir_revisit();

    if (access == PERM_WRITE) {
      if (perm != -1 && !req_id->variables["force"] &&
	  id->type() != ID_TYPE_GROUP && !state->edb->possible_auth (id, pc))
	return path->redir_enter (path->here + "/" + idid + "," + pcid + "," +
				  PERM_TEXTS[perm]);
      state->edb->set_perm (id, pc, perm);
      return path->redir_revisit();
    }
  }

  // Make the page.

  string res = "\n<br clear=all><p><table border=0 cellspacing=1 cellpadding=2>\n";

  object(Identity) *ids = sort (state->edb->identities->this_values());
  object(Identity) *grps = ({});
  for (int i = 0; i < sizeof (ids);) {
    ids[i]->update_userinfo (1);
    if (ids[i]->type() == ID_TYPE_GROUP)
      grps += ({ids[i]}), ids = ids[..i-1] + ids[i+1..];
    else i++;
  }

  res += "<tr><td></td><th bgcolor=" + PERM_HDRHDR_BG + " colspan=" + sizeof (grps) + ">"
    "Groups</th><td></td></tr>\n";
  res += "<tr valign=bottom><td></td>\n";
  foreach (grps, object(Identity) grp) {
    res += "<th bgcolor=" + PERM_HDR_BG + ">";
    string s = "<gtext href=" + path->url_enter (tabs[TAB_ID] + "/" + grp->id()) +
      " bg=" + PERM_HDR_BG + " nfont=lucida 4 rotate=90";
    if (grp->name()) res += s + ">" + grp->name() + "</gtext>";
    else res += "<em>" + s + " italic>Nameless (ID\240" + grp->id() + ")</gtext></em>";
    res += "</th>\n";
  }
  res += "<td></td></tr>\n";
  res += "<tr align=left><th bgcolor=" + PERM_CLSHDR_BG + ">Protection classes</th>" +
    strmult ("<td bgcolor=" + PERM_CLSHDR_BG + ">&nbsp;</td>", sizeof (grps)) +
    "<td></td></tr>\n";

  foreach (sort (state->edb->pclasses->this_values()), object(ProtClass) pc) {
    res += "<tr align=center>"
      "<td bgcolor=" + PERM_CLS_BG + " align=left nowrap>"
      "<a href=" + path->url_enter (tabs[TAB_RSC] + "/c" + pc->id()) + ">" +
      (pc->name() ? pc->name() : "<em>Nameless (ID " + pc->id() + ")</em>") + "</a></td>";

    foreach (grps, object(Identity) grp) {
      //int active = grp->type() == ID_TYPE_GROUP || state->edb->possible_auth (grp, pc);
      int dp = state->edb->direct_perm (grp, pc), p = -1;
      if (/*!active ||*/ dp == -1) p = state->edb->simulated_perm (grp, pc);
      res += "<td bgcolor=" + PERM_BG[p != -1 ? p : dp] + ">";
      for (int i = 0; i < sizeof (PERM_TEXTS); i++) {
	string s;
	if (i == p) s = PERM_IMPL_SYM[i];
	else if (/*active &&*/ i == dp) s = PERM_SET_SYM[i];
	else s = PERM_UNSET_SYM[i];
	//if (!active && i == dp) s = ERRMSG_BEG + "(" + s + ")" + ERRMSG_END;
	if (access == PERM_WRITE)
	  s = "<a href=\"" + tabs[TAB_PERM] + "?" +
	    (i == dp ? "clr" : PERM_TEXTS[i]) + "=" +
	    grp->id() + "," + pc->id() + URL_STATE("&") + "\">" + s + "</a>";
	if (i) res += "&nbsp;";
	res += s;
      }
      res += "</td>";
    }

    string *ss = ({});
    foreach (ids, object(Identity) id) {
      int dp = state->edb->direct_perm (id, pc);
      if (dp != -1)
	ss += ({"<a href=" + path->url_enter (tabs[TAB_ID] + "/" + id->id()) + ">" +
		(id->name() ? id->name() : "<em>Nameless (ID " + id->id() + ")</em>") +
		"</a>: " + PERM_SET_SYM[dp]});
    }
    res += "<td>" + ss * ", " + "</td></tr>\n";
  }
  res += "</table>\n";

  return res;
}

static mixed pclass_view
  (object(LocPath) path, object(State) state, object req_id)
{
  int access = mac->access_perm (PP_ROOT, req_id);
  int pcid;
  object(ProtClass) pc = 0;

  // Parse and check.

  string dir = path->descend();

  {
    string s;
    sscanf (dir, "%*c%D%s", pcid, s);
    if (s != "") return 0;
    pc = state->edb->pclasses[pcid];
  }

  string cmd;

  string name = 0;
  string realm = 0;
  int realm_class = 0;
  int params = 0;
  int own_realm = 0;

  if (path->has_head()) cmd = "set";
  else {
    foreach (({"ret", "set", "del", "delok", 0}), cmd)
      if (req_id->variables[cmd]) break;
    if (!cmd) cmd = "set";

    if (access == PERM_WRITE) {
      name = req_id->variables["name"];
      realm = req_id->variables["realm"];
      realm_class = (int) req_id->variables["realm_class"];
      params = !!(name || realm || realm_class);

      if (zero_type (req_id->variables["own_realm"])) {
	if (!params) own_realm = !pc || pc->realm_class() == pc->id();
      }
      else own_realm = (int) req_id->variables["own_realm"];
    }
  }

  string error = 0;
  int err_back = 0;

  switch (cmd) {
    case "set":
      if (pcid && !pc)
	error = "The protection class does not exist", err_back = 1;
      break;

    case "del":
    case "delok":
      if (!pc)
	error = "The protection class does not exist", err_back = 1;
      else if (sizeof (pc->ppoints()))
	error = "The protection class must be empty to be deleted";
      break;

    case "ret":
    default:
      return path->redir_return();
  }

  if (error) return error_message (error, err_back, path, state);

  if (access != PERM_WRITE) {
    if (pc) {
      name = pc->name();
      realm = pc->realm();
      realm_class = pc->realm_class();
      own_realm = realm_class == pc->id();
    }
  }

  else switch (cmd) {
    case "set":
      if (!pc && params) pc = state->edb->pclasses->new();

      if (name) {
	if (name == "") name = 0;
	if (pc && name != pc->name()) pc->set_name (name);
      }
      else if (pc) name = pc->name();

      if (realm && own_realm) {
	if (realm == "") realm = 0;
	if (pc && realm != pc->realm()) pc->set_realm (realm);
      }
      else if (pc) realm = pc->realm();

      if (!pcid && pc && params)
	state->edb->pclasses->add (pc);

      if (pc) {
	if (own_realm) realm_class = pc->id();
	if (realm_class && realm_class != pc->realm_class() &&
	    state->edb->pclasses[realm_class])
	  pc->set_realm_class (realm_class);
	else realm_class = pc->realm_class();
      }

      if (params) return path->redir_replace ("c" + pc->id(), "own_realm=" + own_realm);
      break;

    case "del":
      return confirm_message (
	"Delete " + (pc->name() ? "protection class: " + html_encode_string (pc->name()) :
		     "nameless protection class (ID " + pcid + ")"),
	DESC_PC_DELETE, "delok", "ret", path, state);

    case "delok":
      state->edb->pclasses->remove (pc);
      if (sizeof (state->nestlist) && state->nestlist[0] == path->here)
	// Do not return to the page editing the deleted record.
	state->nestlist = state->nestlist[1..];
      return path->redir_return();
  }

  // Make the page.

  string ia_res = "";
  if (pcid && pc) {
    object(LocPath) orig_path = LocPath (path);
    mixed r = pclass_auths_subview (path, state, req_id, pc);
    if (!stringp (r)) return r;
    ia_res = r;
    path->copy (orig_path);
  }

  string res;
  if (!pcid) res = "\n<h2>New protection class</h2>\n";
  else res = "\n<h2>" + (pc->name() ?
			 "Protection class: " + html_encode_string (pc->name()) :
			 "Nameless protection class (ID " + pcid + ")") + "</h2>\n";

  res += "\n<p>";
  if (access != PERM_WRITE) {
    res += "Name: " + (name ? html_encode_string (name) : "") + "<br>\n";
    res += "Realm: " + (realm ? html_encode_string (realm) : "<em>Nameless</em>");
    if (pc && realm_class != pc->id()) {
      object(ProtClass) rc = state->edb->pclasses[realm_class];
      res += (rc->name() ? ", <em>from</em> " + html_encode_string (rc->name()) :
	      ", <em>from nameless protection class ID " + realm_class);
    }
    res += "<br>\n";
  }
  else {
    res += "<table width=100% border=0><tr><td>" + // Netscape 4 kludge.
      path->form_revisit (([TAG_ENTER + "_var": "del"]));

    res += "Name: <input type=text name=name size=50" +
      (name ? " value=" + html_encode_tag_value (name) : "") + "><br>\n";

    res += "Realm: <input type=checkbox name=own_realm value=1 " +
      (own_realm ? " checked" : "") + "> Own ";
    if (own_realm || !pc) {
      res += "<input type=text name=realm size=50" +
	(realm ? " value=" + html_encode_tag_value (realm) : "") + "><br>\n";
    }
    else {
      res += "<select name=realm_class>\n";
      if (realm_class == pc->id())
	res += "<option value=" + pc->id() + " selected>" +
	  "Own realm" + (realm ? ": " + html_encode_string (realm) : "") + "\n";
      object(ProtClass) *rcs = sort (state->edb->pclasses->find_realm_classes() - ({pc}));
      foreach (rcs, object(ProtClass) rc)
	res += "<option value=" + rc->id() +
	  (realm_class == rc->id() ? " selected" : "") + ">" +
	  (rc->realm() ? html_encode_string (rc->realm()) : "Nameless realm") +
	  (rc->name() ? " (" + html_encode_string (rc->name()) + ")" : "") + "\n";
      res += "</select><br>\n";
    }
    res += DESC_PC_REALM + "<br>\n";

    res += "\n<p>"
      "<input type=submit name=set value=\" " + (pcid ? "Change" : "Add") + " \">\n";
    if (pcid && !sizeof (pc->ppoints()))
      res += "<input type=submit name=del value=\" Delete \">\n";
    res += "</form>"
      "</td></tr></table>\n";	// Netscape 4 kludge.
  }

  if (pcid && pc) {
    object(ProtPoint) *pps = pc->ppoints();
    if (!sizeof (pps))
      res += "\n<h3>No protection points</h3>\n";
    else {
      res += "\n<h3>Protection points</h3>\n"
	"<table width=100% border=0 cellspacing=0 cellpadding=2>\n";
      foreach (sort (pps), object(ProtPoint) pp) {
	string name = get_authmodname (pp->own_module(), state->edb);
	if (!name) name = ERRMSG_BEG + "no owner" + ERRMSG_END;
	res += "<tr><td><a href=" + path->url_enter (path->up + "/p" + pp->id()) + ">" +
	  (pp->name() ? html_encode_string (pp->name()) :
	   "<em>Nameless (ID " + pp->id() + ")</em>") + "</a></td>"
	  "<td align=right><em>" + name + "</em></td></tr>\n";
      }
      res += "</table>\n";
    }
  }

  return res + ia_res;
}

static mixed ppoint_view
  (object(LocPath) path, object(State) state, object req_id)
{
  int access = mac->access_perm (PP_ROOT, req_id);
  int ppid;
  object(ProtPoint) pp = 0;

  // Parse and check.

  string dir = path->descend();

  {
    string s;
    sscanf (dir, "%*c%D%s", ppid, s);
    if (s != "") return 0;
    pp = state->edb->ppoints[ppid];
  }

  string cmd;

  string name = 0;
  int params = 0;

  foreach (({"ret", "set", "del", "delok", 0}), cmd)
    if (req_id->variables[cmd]) break;
  if (!cmd) cmd = "set";

  if (access == PERM_WRITE) {
    name = req_id->variables["name"];
    params = !!name;
  }

  string error = 0;
  int err_back = 0;

  switch (cmd) {
    case "set":
    case "del":
    case "delok":
      if (!pp)
	error = "The protection point does not exist", err_back = 1;
      break;

    case "ret":
    default:
      return path->redir_return();
  }

  if (error) return error_message (error, err_back, path, state);

  if (access != PERM_WRITE)
    name = pp->name();

  else switch (cmd) {
    case "set":
      if (name) {
	if (name == "") name = 0;
	if (name != pp->name()) pp->set_name (name);
      }
      else name = pp->name();

      if (params) return path->redir_revisit();
      break;

    case "del":
      if (pp->deletable() || !pp->own_module())
	return confirm_message (
	  "Delete " +
	  (pp->name() ? "protection point: " + html_encode_string (pp->name()) :
	   "nameless protection point (ID " + ppid + ")"),
	  DESC_PP_DELETE, "delok", "ret", path, state);
      else
#ifdef DEVELOPER
	return confirm_message (
	  "Delete " +
	  (pp->name() ? "protection point: " + html_encode_string (pp->name()) :
	   "nameless protection point (ID " + ppid + ")"),
	  DESC_PP_DEVEL_DELETE, "delok", "ret", path, state);
#else
        break;
#endif

    case "delok":
#ifndef DEVELOPER
      if (pp->deletable() || !pp->own_module()) {
#endif
	state->edb->ppoints->remove (pp);
	if (sizeof (state->nestlist) && state->nestlist[0] == path->here)
	  // Do not return to the page editing the deleted record.
	  state->nestlist = state->nestlist[1..];
	return path->redir_return();
#ifndef DEVELOPER
      }
#endif
  }

  // Make the page.

  string res = "\n<h2>" +
    (pp->name() ? "Protection point: " + html_encode_string (pp->name()) :
     "Nameless protection point (ID " + ppid + ")") + "</h2>\n";

  res += "\n<p>";
  if (pp->own_module())
    res += "Owning module: " + get_authmodname (pp->own_module(), state->edb) + "<br>\n";
  else {
    string str = pp->own_modname();
    res += ERRMSG_BEG + "Owning module \"" + (str ? html_encode_string (str) : "") +
      "\" not found.\nIt may have been removed." + ERRMSG_END + "<br>\n";
  }

  res += "\n<p>";
#ifdef DEVELOPER
  res += "Handle: " + (pp->handle() ? "\"" + pp->handle() + "\"" : pp->id()) + "<br>\n";
#endif
  if (access != PERM_WRITE)
    res += "Name: " + (name ? html_encode_string (name) : "") + "<br>\n";
  else {
    res += "<table width=100% border=0><tr><td>" + // Netscape 4 kludge.
      path->form_revisit (([TAG_ENTER + "_var": "del"]));
    res += "Name: <input type=text name=name size=50" +
      (name ? " value=" + html_encode_tag_value (name) : "") + "><br>\n";

    res += "\n<p>"
      "<input type=submit name=set value=\" Change \">\n";
    if (ppid && (pp->deletable() || !pp->own_module()))
      res += "<input type=submit name=del value=\" Delete \">\n";
#ifdef DEVELOPER
    else if (ppid) res += "<input type=submit name=del value=\" (Delete) \">\n";
#endif
    res += "</form>"
      "</td></tr></table>\n";	// Netscape 4 kludge.
  }

  return res;
}

static mixed resources_view
  (object(LocPath) path, object(State) state, object req_id)
{
  int access = mac->access_perm (PP_ROOT, req_id);

  // Parse.

  if (req_id->variables["new"])
    return path->redir_enter (tabs[TAB_RSC] + "/c0"); // A bit ugly.
  if (path->has_head()) {
    switch (path->head()[..0]) {
      case "c": return pclass_view (path, state, req_id);
      case "p": return ppoint_view (path, state, req_id);
      default: return 0;
    }
  }

  if (req_id->variables["unfold"]) {
    int pcid = (int) req_id->variables["unfold"];
    if (!pcid) {
      state->unfolds = (<>);
      foreach (state->edb->pclasses->this_values(), object(ProtClass) pc)
	state->unfolds[pc->id()] = 1;
    }
    else state->unfolds[pcid] = 1;
  }
  if (req_id->variables["fold"]) {
    int pcid = (int) req_id->variables["fold"];
    if (!pcid) state->unfolds = (<>);
    else state->unfolds[pcid] = 0;
  }

  if (req_id->variables["move"] && access == PERM_WRITE) {
    object(ProtPoint) *pps = ({});
    if (req_id->variables["pp"])
      foreach (req_id->variables["pp"] / "\0", string s) {
	object(ProtPoint) pp = state->edb->ppoints[(int) s];
	if (pp) pps += ({pp});
      }

    object(ProtClass) pc;
    int pcid;
    if (req_id->variables["pc"]) {
      pcid = (int) req_id->variables["pc"];
      if (!pcid && sizeof (pps)) {
	pc = state->edb->pclasses->new();
	state->edb->pclasses->add (pc);
      }
      else pc = state->edb->pclasses[pcid];
    }

    if (pc && sizeof (pps)) {
      foreach (pps, object(ProtPoint) pp) pp->set_pclass (pc);
      state->unfolds[pc->id()] = 1;
    }

    return path->redir_revisit();
  }

  // Make the page.

  string res = "\n<br clear=all><p>" + path->form_revisit();

  int a_fold = 0, an_unfold = 0;
  string tbl = "<table width=100% border=0 cellspacing=0 cellpadding=2>\n"
    "<tr><td colspan=2></td><th bgcolor=" + RSC_MHDR_BG + ">Move</th>"
    "<th bgcolor=" + RSC_HDR_BG + " colspan=2 align=left width=95%>Description</th>"
    "</tr>\n";

  foreach (sort (state->edb->pclasses->this_values()), object(ProtClass) pc) {
    object(ProtPoint) *pps = pc->ppoints();

    tbl += "<tr></tr>\n<tr bgcolor=" + RSC_PCLASS_BG + ">";

    if (sizeof (pps) == 1 && pc->name() && pc->name() == pps[0]->name()) {
      object(ProtPoint) pp = pps[0];
      string name = get_authmodname (pp->own_module(), state->edb);
      if (!name) name = ERRMSG_BEG + "no owner" + ERRMSG_END;

      tbl += "<td colspan=2>&nbsp;</td>"
	"<td bgcolor=" + RSC_MPCLASS_BG + " align=center>" +
	(access == PERM_WRITE ?
	 "<input type=radio name=pc value=" + pc->id() + ">"
	 "<input type=checkbox name=pp value=" + pp->id() + ">" : "&nbsp;") + "</td>";

      tbl += "<td><a href=" + path->url_enter (tabs[TAB_RSC] + "/c" + pc->id()) + ">" +
	html_encode_string (pc->name()) + "</a>" +
	(pc->realm() ? ", <em>realm:</em> " + html_encode_string (pc->realm()) : "") +
	"</td>"
	"<td align=right><em>" + name + "</em>";
      if (access == PERM_WRITE && (pp->deletable() || !pp->own_module()))
	tbl += " <a href=" +
	  path->url_enter (tabs[TAB_RSC] + "/p" + pp->id(), "del=") + ">" +
	  DELMSG_BEG + "Delete" + DELMSG_END + "</a>";
#ifdef DEVELOPER
      else if (access == PERM_WRITE)
	tbl += " <font size=-1>(<a href=" +
	  path->url_enter (tabs[TAB_RSC] + "/p" + pp->id(), "del=") + ">" +
	  DELMSG_BEG + "Delete" + DELMSG_END + "</a>)</font>";
      tbl += " <font size=-1>(" +
	(pp->handle() ? "\"" + pp->handle() + "\"" : pp->id()) +
	")</font>";
#endif
      tbl += "</td></tr>\n";
    }

    else {
      if (!sizeof (pps))
	tbl += "<td align=center colspan=2>" + RSC_CLS_EMPTY + "</td>";
      else if (state->unfolds[pc->id()])
	tbl += "<td align=center colspan=2>"
	  "<a href=" + path->url_revisit ("fold=" + pc->id()) + ">"
	  "<img border=0 width=20 height=20 src=internal-roxen-fold alt=\"\\/\"></a>"
	  "</td>";
      else
	tbl += "<td align=right>" + sizeof (pps) + "</td>"
	  "<td align=left><a href=" + path->url_revisit ("unfold=" + pc->id()) + ">"
	  "<img border=0 width=20 height=20 src=internal-roxen-unfold alt=\"--\"></a>"
	  "</td>";

      tbl += "<td bgcolor=" + RSC_MPCLASS_BG + " align=center>" +
	(access == PERM_WRITE ? "<input type=radio name=pc value=" + pc->id() + ">" :
	 "&nbsp;") + "</td>";

      tbl += "<td" + (sizeof (pps) ? " colspan=2" : "") + ">"
	"<a href=" + path->url_enter (tabs[TAB_RSC] + "/c" + pc->id()) + ">" +
	(pc->name() ? html_encode_string (pc->name()) :
	 "<em>Nameless (ID " + pc->id() + ")</em>") + "</a>" +
	(pc->realm() ? ", <em>realm:</em> " + html_encode_string (pc->realm()) : "") +
	"</td>";

      if (!sizeof (pps) && access == PERM_WRITE)
	tbl += "<td align=right>"
	  "<a href=" + path->url_enter (tabs[TAB_RSC] + "/c" + pc->id(), "del=") + ">" +
	  DELMSG_BEG + "Delete" + DELMSG_END + "</a></td>";
      tbl += "</tr>\n";

      if (state->unfolds[pc->id()]) {
	foreach (sort (pps), object(ProtPoint) pp) {
	  string name = get_authmodname (pp->own_module(), state->edb);
	  if (!name) name = ERRMSG_BEG + "no owner" + ERRMSG_END;

	  tbl += "<tr bgcolor=" + RSC_PPOINT_BG + "><td colspan=2></td>"
	    "<td bgcolor=" + RSC_MPPOINT_BG + " align=center>" +
	    (access == PERM_WRITE ?
	     "<input type=checkbox name=pp value=" + pp->id() + ">" : "&nbsp") + "</td>";

	  tbl += "<td>"
	    "<a href=" + path->url_enter (tabs[TAB_RSC] + "/p" + pp->id()) + ">" +
	    (pp->name() ? html_encode_string (pp->name()) :
	     "<em>Nameless (ID " + pp->id() + ")</em>") + "</a></td>";

	  tbl += "<td align=right><em>" + name + "</em>";
	  if (access == PERM_WRITE && (pp->deletable() || !pp->own_module()))
	    tbl += " <a href=" +
	      path->url_enter (tabs[TAB_RSC] + "/p" + pp->id(), "del=") + ">" +
	      DELMSG_BEG + "Delete" + DELMSG_END + "</a>";
#ifdef DEVELOPER
	  else if (access == PERM_WRITE)
	    tbl += " <font size=-1>(<a href=" +
	      path->url_enter (tabs[TAB_RSC] + "/p" + pp->id(), "del=") + ">" +
	      DELMSG_BEG + "Delete" + DELMSG_END + "</a>)</font>";
	  tbl += " <font size=-1>(" +
	    (pp->handle() ? "\"" + pp->handle() + "\"" : pp->id()) +
	    ")</font>";
#endif
	  tbl += "</td></tr>\n";
	}
	an_unfold = 1;
      }
      else a_fold = 1;
    }
  }

  tbl += "<tr><td height=0></td></tr>\n"
    "<tr bgcolor=" + RSC_NEW_BG + "><td colspan=2>&nbsp;</td>"
    "<td bgcolor=" + RSC_MNEW_BG + " align=center>" +
    (access == PERM_WRITE ? "<input type=radio name=pc value=0 checked>" : "&nbsp;") +
    "</td><td colspan=2><em>Move to a new protection class</em></td></tr>\n"
    "</table>\n";

  if (a_fold)
    res += "<a href=" + path->url_revisit ("unfold=0") + ">Unfold all</a>\n";
  if (an_unfold) res += "<a href=" + path->url_revisit ("fold=0") + ">Fold all</a>\n";
  res += "<br>" + tbl;

  if (access == PERM_WRITE)
    res += "\n<p>"
      "<table width=100% border=0><tr><td>" // Netscape 4 kludge.
      "<input type=submit name=new value=\" New class \">\n"
      "<input type=submit name=move value=\" Move \"><br>\n" +
      DESC_PROT_MOVE +
      "</td></tr></table>\n";	// Netscape 4 kludge.
  res += "</form>\n";

  return res;
}

static mixed root_view
  (object(LocPath) path, object(State) state, object req_id)
{
  int mchoice = -1;
  string here = path->here, res;

  string dir = path->has_head() && path->descend();
  if (dir && sizeof (dir)) {
    mchoice = search (tabs, dir);
    if (mchoice == -1) return 0;
    mixed r = tab_handlers[mchoice] (path, state, req_id);
    if (!stringp (r)) return r;
    res = r;
  }
  else {
    int access = mac->access_perm (PP_ROOT, req_id);

    if (access == PERM_WRITE) {
      // Parse db maintenance options.
      if (req_id->variables["purge"])
	return confirm_message ("Purge removed modules", DESC_DB_PURGE,
				"purgeok", "ret", path, state);
      else if (req_id->variables["purgeok"]) {
	state->edb->purge_missing_modules();
	return path->redir_revisit();
      }
      else if (req_id->variables["ret"])
	return path->redir_revisit();
    }

    // Make the page.

    res = "\n<p>" + DESC_AC_INTRO;

    if (access == PERM_WRITE) {
      res += "\n<h2>Database maintenance</h2>\n\n<ul>";

      res += "\n<p><li><a href=" + path->url_revisit ("purge=") + ">"
	"Purge references to removed modules from the database.</a>\n";

      res += "</ul>\n";
    }
  }

  string menu = "<config_tablist>\n";
  for (int i = 0; i < sizeof (tabs); i++)
    menu += "<tab href=" + path->url_goto (here + "/" + tabs[i]) +
      (i == mchoice ? " selected" : "") +
      (req_id->misc->colorscheme?(" bgcolor="+req_id->misc->colorscheme->bgcolor):
       "")+
      ">" + tabs[i] + "</tab>\n";
  menu += "</config_tablist><br>\n"
    "<back>\n";

  return menu + res;
}

int acadm_root_perm (object req_id)
{
  return mac->access_perm (PP_ROOT, req_id);
}

mixed root_access (object req_id, void|string realm)
{
  int access = mac->access_try (PP_ROOT, req_id);

  // Backdoor auth.
  if (QUERY (backdoor) && req_id->realauth) {
    string u, p;
    if (sscanf (req_id->realauth, "%s:%s", u, p) == 2 &&
	u == roxen->QUERY (ConfigurationUser) &&
	crypt (p, roxen->QUERY (ConfigurationPassword))) {
      mac->access_set (PP_ROOT, req_id, PERM_WRITE);
      req_id->misc->acadm_backdoor = 1;
      access = PERM_WRITE;
    }
  }

  if (access == PERM_NONE) {
    mapping res = mac->access_throw (PP_ROOT, req_id, 0, realm);
    if (!res && QUERY (backdoor))
      // Not configured properly; let the backdoor show.
      return http_auth_required ("Access Control");
    return res;
  }

  return 0;
}

string|mapping page_body (object req_id, string mountpoint, void|object(AC_DB) acdb,
			  void|object(State) in_state)
{
  acdb = acdb || online_acdb;
  object(State) state = in_state || get_state (req_id, acdb);
  if (sizeof (mountpoint) && mountpoint[-1] != '/') mountpoint += "/";

  object maclock = acdb && mac->lock_to_thread();

  string err = "";

  if (last_io_error)
    err = "\n" + ERRMSG_BEG +
      "<i>I/O error:</i> " + last_io_error + ERRMSG_END + "\n<p>";

  if (QUERY (remote) && sizeof (QUERY (connect)) && !remote->server) {
    array s = (QUERY (connect) / ":") + ({0});
    s[1] = (int) s[1] || QUERY (server_port);
    remote->client_on (s[0], s[1]);
    if (!remote->server)
      err += "\n" + ERRMSG_BEG + "<i>Could not connect to remote AC database server. "
	"Any changes made now will be lost when the server responds again.</i>" +
	ERRMSG_END + "\n<p>";
  }

  mixed res;

  if (!acdb || !acdb->initialized) {
    res = "\n<h2>" + ERRMSG_BEG + "No AC database is active" +
      ERRMSG_END + "</h2>\n";
  }

  else {
    {
      // Handle page entry/return.
      string *new_nestlist = state->nestlist;
      if (req_id->variables[TAG_ENTER] ||
	  req_id->variables[TAG_ENTER + "_var"] &&
	  req_id->variables[req_id->variables[TAG_ENTER + "_var"]]) {
	// Entering a nested page.
	if (state->last) new_nestlist = ({state->last}) + state->nestlist;
      }
      else if (req_id->variables[TAG_RETURN] ||
	       req_id->variables[TAG_RETURN + "_var"] &&
	       req_id->variables[req_id->variables[TAG_RETURN + "_var"]]) {
	// Returning from a nested page.
	if (sizeof (state->nestlist)) new_nestlist = state->nestlist[1..];
      }
      else if (req_id->variables[TAG_GOTO] ||
	       req_id->variables[TAG_GOTO + "_var"] &&
	       req_id->variables[req_id->variables[TAG_GOTO + "_var"]]) {
	// Go to a page (ie reset stack to canonical pagelist).
	new_nestlist = ({});
      }
      if (state->seqstate == 0) // Update if in sequence.
	state->nestlist = new_nestlist;
      else if (state->seqstate != 1)
	// Reset if out of sequence, but not on reloads.
	state->nestlist = ({});
    }
    int set_nestlist = !sizeof (state->nestlist);

    object(LocPath) path;
    {
      // Can't use loc; we must first split on "/", then decode.
      // Otherwise we might split on "%2f".
      string s = req_id->raw_url;
      int i = search (s, ";");
      if (i == -1) i = search (s, "?");
      if (i != -1) s = s[..i-1];
      path = LocPath (
	Array.map (s / "/", http_decode_string)[sizeof (mountpoint / "/") - 1..],
	mountpoint, state, req_id);
    }

    //acdb->begin(); // Unnecessary since the mac does this above.

    res = root_view (path, state, req_id);

    if (set_nestlist) state->nestlist = path->uplist;
    state->last = path->here;

    acdb->commit();
    if (!res || mappingp (res)) return res;
    else if (arrayp (res)) res = res[0]; // Errors are encapsulated in arrays.
  }

  res = err + res;

  {
    string back =
      state->last != mountpoint && sizeof (state->nestlist) ?
      "<a href=" + html_encode_tag_value (
	state->nestlist[0] + "?" + TAG_RETURN + "=" + URL_STATE("&")) + ">"
      "<img src=/internal-roxen-back alt=\"[Back]\" align=left hspace=0 border=0>"
      "</a>" : "&nbsp;";
    string *parts = res / "<back>";
    if (sizeof (parts) > 1) res = parts * back;
    else res = back + "\n\n" + res;
  }

  return res;
}


// Module interface.

int file_not_used() {return !sizeof (QUERY (db_filename));}
int remote_not_enabled() {return !QUERY (remote);}
int remote_not_server() {return !QUERY (remote) || !QUERY (server);}

void create (object conf)
{
  defvar ("db_filename", "", "Database file",
	  TYPE_FILE,
	  "The file containing the AC database. The AC system is disabled "
	  "when this is empty and remote access is disabled.");
  defvar ("db_binary", 0, "Store database in binary format",
	  TYPE_FLAG,
	  "Whether to store the database in text or binary format. It's by "
	  "default stored in a somewhat readable text format to make manual "
	  "recovery possible if the database should ever get corrupted. "
	  "However, if the database is large (i.e. more than a couple of "
	  "thousand users) it may take too long to parse the text format. "
	  "A database can always be read regardless of text or binary "
	  "format, so it's possible to change this setting later.",
	  0, file_not_used);
  defvar ("mountpoint", "/acadm/", "Administrator interface location",
	  TYPE_LOCATION,
	  "The place in the server namespace to put the administrator "
	  "interface for the AC database. The interface may be disabled by "
	  "leaving this empty.");
  defvar ("backdoor", 1, "Administrator interface backdoor",
	  TYPE_FLAG,
	  "When active, the user and password for the configuration "
	  "interface also gives full privileges in the AC administrator "
	  "interface.\n"
	  "<p>This works as a backdoor. Under normal circumstances the AC "
	  "system performs access control even to its own admin interface, "
	  "so users are granted access to it in the same way as any other "
	  "module that uses AC. The configuration user and password doesn't "
	  "even exist within AC. So, this is only a way to solve the "
	  "chicken-and-egg problem when the AC is not yet configured. (It "
	  "might also come in handy if the administrator by mistake should "
	  "remove his own access. ;) It may be deactivated later for "
	  "security reasons if so desired.");
  defvar ("superuser", 0, "All resources backdoor",
	  TYPE_FLAG | VAR_MORE,
	  "When enabled, the user and password for the configuration "
	  "interface also gives full privileges for all protected "
	  "resources.\n"
	  "<p>This is a backdoor to everything that this AC module protects. "
	  "It's typically useful when setting up systems that embeds AC "
	  "until they are in working order. It's also useful to get access "
	  "everywhere in complex setups without having to sort out how the "
	  "security system is configured. However, since the user doesn't "
	  "exist in the AC database, you might get backtraces in "
	  "applications that doesn't expect this.\n"
	  "<p>Since the backdoor is tested before anything else, it's "
	  "probably a good idea to choose an uncommon username in the "
	  "configuration interface that's less likely to clash with a "
	  "user in the AC database.\n"
	  "<p>A note on security: With this enabled you get a protection "
	  "model with a superuser that can do anything anywhere. This is "
	  "like every UNIX system; it works well and is handy, but it "
	  "becomes more important to protect this user. Since the login will "
	  "work in all situations, even over low security connections "
	  "(typically when using the HTTP basic authentication scheme over "
	  "the internet), you'd better use it with care, or turn this "
	  "feature off.");
  defvar ("pagelength", 25, "Limit of the number of entries per page",
	  TYPE_INT,
	  "Maximum number of entries to list per page in the administrator "
	  "interface. This is currently implemented for the user list under "
	  "the Identities tab.");

  defvar ("remote", 0, "Remote access: Enabled",
	  TYPE_FLAG,
	  "Enable synchronization of the AC database with other AC modules "
	  "in Roxen servers over the network.\n"
	  "<p>A not on security: The communication protocol does NOT use "
	  "encryption, so the AC database, which typically contain sensitive "
	  "information, will be transmitted in cleartext. The handshake "
	  "does however use a simple challenge-response mechanism, so the "
	  "remote access secret is not transmitted over the network. "
	  "So even though the network traffic may be eavesdropped, that "
	  "information isn't enough to connect rogue clients to alter the "
	  "database. This is no protection against man-in-the-middle "
	  "attacks, though.");
  defvar ("server", 0, "Remote access: Server enabled",
	  TYPE_FLAG,
	  "Whether to run as a server or not.",
	  0, remote_not_enabled);
  defvar ("server_port", 3747, "Remote access: Server listen port",
	  TYPE_INT,
	  "The port to listen to at the server side. It's the database at "
	  "the server side that is the authorative source.",
	  0, remote_not_server);
  defvar ("server_interface", "ANY", "Remote access: Server listen interface",
	  TYPE_STRING,
	  "The interface to listen to.",
	  0, remote_not_server);
  defvar ("connect", "", "Remote access: Server to connect to",
	  TYPE_STRING,
	  "The server to which this AC database should act as a slave. The "
	  "format is <tt><i>host</i>[:<i>port</i>]</tt>, where port defaults "
	  "to the server listen port.\n"
	  "<p>You don't need to specify a database file when using this, but "
	  "if you do, AC will use the local copy when it can't connect to "
	  "the server.\n"
	  "<p>WARNING: Do not connect to a server in the same Roxen server; "
	  "it can cause hangs.",
	  0, remote_not_enabled);
  defvar ("secret", "", "Remote access: Shared secret",
	  TYPE_TEXT,
	  "This string is the shared secret between all connected AC "
	  "servers. I.e. it must be the same for the clients to be allowed "
	  "access to a server. You typically cut'n'paste this string to "
	  "all other AC modules that should get access (you are using https "
	  "for the configuration interface, aren't you?).\n"
	  "<p>The secret is never passed over the network; it's only used "
	  "for a simple challenge-response handshake. Note that the "
	  "security against rogue clients depends on this secret; keep in "
	  "mind that it's stored in cleartext in the configuration files of "
	  "all connected Roxen servers.\n"
	  "<p>The sample string was generated from reasonably good "
	  "randomness when this module was added to the server. There are no "
	  "restrictions on the length or format.",
	  0, remote_not_enabled);
  defvar ("remote_id", "", "Remote access: Remote ID",
	  TYPE_STRING | VAR_MORE,
	  "A unique ID among the connected AC modules. This is used in "
	  "module references on other servers to identify the AC modules "
	  "in this server. It may only contain the characters 0-9, A-Z, a-z "
	  "and '_'. If left empty it'll be assigned a unique ID on first "
	  "connect. Normally it should never be changed.",
	  0, remote_not_enabled);
}

string check_variable (string var, mixed val)
{
  if (var == "pagelength" && val < 0)
    return "The limit can't be negative.";
  else if (var == "remote_id" && !Regexp ("^[0-9A-Za-z_]*$")->match (val))
    return "The remote ID can only contain 0-9, A-Z, a-z and '_'.";
  return 0;
}

string format_netaddr (string addr, void|int noport)
{
  if (!addr) return "(unknown host)";
  array(string) s = addr / " ";
  if (noport) s[1] = ""; else s[1] = ":" + s[1];
#if constant (gethostbyaddr)
  array host = gethostbyaddr (s[0]);
  return arrayp (host) ? host[0] + s[1] : s * "";
#else
  return s * "";
#endif
}

string status()
{
  string ret;

  if (!online_acdb) ret = "<p><b>No AC database active.</b>\n";
  else {
    ret = "<p>Access Control system is running.\n";

    ret += "\n<h4>Currently connected modules</h4>\n\n";
    online_acdb->conn_mods -= ({0});
    if (sizeof (online_acdb->conn_mods)) {
      ret += "<table border=0 cellspacing=0 cellpadding=2>\n"
	"<tr><th align=left>ID</th><th align=left>Name</th></tr>\n";
      foreach (online_acdb->conn_mods, object mod)
	ret += "<tr><td>" + html_encode_string (online_acdb->modulename (mod)) + "</td>"
	  "<td>" + html_encode_string (get_authmodname (mod, online_acdb)) +
	  "</td></tr>\n";
      ret += "</table>\n\n";
    }
    else ret += "<em>None.</em>\n\n";

    if (last_io_error)
      ret += "<p><b>I/O error</b>: " + last_io_error + "\n";
  }

  if (remote) {
    if (remote->want_client) {
      string addr;
      catch (addr = format_netaddr (remote->server_conn->con->query_address (1)));
      string name;
      catch (name = remote->server->name());
      if (!name)
	ret += "<p><b>Connection failed to remote AC database" +
	  (addr ? " on " + html_encode_string (addr) : "") + ".</b>\n";
      else
	ret += "<p>Replicating from remote AC database in virtual server " +
	  html_encode_string (name) + " on " + html_encode_string (addr) + ".\n";
    }

    if (remote->want_server) {
      if (remote->remsrvr) {
	ret += "\n<h4>Connected remote AC database clients</h4>\n\n";
	array(object) clients = remote->clients;
	if (sizeof (clients)) {
	  ret += "<table border=0 cellspacing=0 cellpadding=2>\n"
	    "<tr><th align=left>Host</th><th align=left>Virtual server name</th></tr>\n";
	  for (int i = 0; i < sizeof (clients); i++) {
	    string addr = "?";
	    catch (addr = format_netaddr (
	      remote->client_conns[i]->con->query_address(), 1));
	    string name;
	    catch (name = clients[i]->name());
	    name = name ? html_encode_string (name) : "<b>Connection failed</b>";
	    ret += "<tr><td>" + html_encode_string (addr) + "</td>"
	      "<td>" + name + "</td></tr>\n";
	  }
	  ret += "</table>\n\n";
	}
	else ret += "<em>None.</em>\n\n";
      }
      else
	ret += "<p><b>Failed to open listen port for remote access server.</b>\n";

      ret += "<p>" + failed_connections +
	" failed connections since last module (re)load.\n";
    }
  }

  return ret;
}

array register_module()
{
  return ({MODULE_LOCATION | MODULE_SECURITY, "Access Control",
	   CONFIG_DESC,
	   0, 1});
}

int save_config;

void set_remote_id (string str)
{
  set ("remote_id", str);
  save_config = 1;
}

void init_acadm (object(AC_DB) acdb)
{
  init_acdb (acdb);

  foreach (values (config->modules), mapping moddata)
    if (moddata->copies)
      foreach (values (moddata->copies), object mod) {
	if (mod->get_auth_types) mod->ac_connect();
      }
    else if (moddata->enabled && moddata->enabled->get_auth_types)
      moddata->enabled->ac_connect();

  // Register the ppoints for this module.
  mac->pp_create (PP_ROOT, "Access Control", 0, "Access Control");

  if (remote && QUERY (server))
    remote->server_on (QUERY (server_interface), QUERY (server_port));

  if (save_config) call_out (config->save, 0);
}

void start (int when, object conf)
{
  save_config = 0;
  config = conf;
  module_dependencies (config, uniq (`+ (({"configtablist", "graphic_text"}),
					 @values (init_id_authmods),
					 init_pc_authmods)));

  if (!sizeof (QUERY (secret))) {
    string rnd, secret = "";
    if (!catch (rnd = Crypto.randomness.reasonably_random()->read (64))) {
      for (int l = 1, r = 0; l <= 4; l++) {
	for (; r < l * 16; r++) secret += sprintf ("%c", rnd[r] % ('~' - '!') + '!');
	secret += "\n";
      }
      set ("secret", secret);
      save_config = 1;
    }
  }

  if (online_acdb) {
#ifdef THREADS
    // Won't touch mac here, so we don't have to observe its locking order.
    object lock = online_acdb->lock();
#endif
    foreach (backends, object backend) destruct (backend);
    backends = ({});
    destruct (online_acdb);
  }

  states = ([]);
  garb_states();

  mapping acvar = roxen->query_var ("AC");
  if (!acvar) {
    acvar = ([]);
    roxen->set_var ("AC", acvar);
  }
  acvar->ModuleAC = ModuleAC;
  acvar->ModuleACDB = ModuleACDB;
  //acvar->init_acdb = init_acdb;
  if (!acvar->loaders) acvar->loaders = ([]);

  if (acvar->loaders[config] && acvar->loaders[config] != this_object())
    report_error ("AC: An access database is already registered for this config\n");
  else {
    acvar->loaders[config] = this_object();
    acauth_modules = (<>);
  }

  mac = mac || ModuleAC (config, this_object());

  add_constant ("PERM_N_A", PERM_N_A);
  add_constant ("PERM_NONE", PERM_NONE);
  add_constant ("PERM_READ", PERM_READ);
  add_constant ("PERM_WRITE", PERM_WRITE);

  foreach (backends, object backend) destruct (backend);
  backends = ({});

  object(AC_DB) acdb = create_acdb (config);

  // Init remote first to set fix remote ID in module lookup in acdb.
  if (QUERY (remote))
    backends += ({
      remote = ((program) "acbackend_remote") (this_object(), acdb, init_acadm,
					       QUERY (secret), QUERY (remote_id))
    });

  // Read local db file.
  if (sizeof (QUERY (db_filename))) {
    backends += ({
      ((program) "acbackend_file") (this_object(), acdb, init_acadm,
				    QUERY (db_filename), QUERY (db_binary))
    });
    if (acdb->last_read_binary != QUERY (db_binary))
      backends[-1]->commit();	// Rewrite file in correct format.
  }

  // Request update from remote server after reading local db file
  // to override it.
  if (remote && sizeof (QUERY (connect))) {
    array s = (QUERY (connect) / ":") + ({0});
    s[1] = (int) s[1] || QUERY (server_port);
    remote->client_on (s[0], s[1]);
  }
}

void stop()
{
  foreach (backends, object backend) destruct (backend);
  backends = ({});

  remove_call_out (garb_states);

  mapping acvar = roxen->query_var ("AC");
  if (acvar && acvar->loaders) m_delete (acvar->loaders, config);

  if (online_acdb) {
#ifdef THREADS
    // Won't touch mac here, so we don't have to observe its locking order.
    object lock = online_acdb->lock();
#endif
    destruct (online_acdb);
  }
}

// MODULE_LOCATION interface.

string query_location()
{
  return sizeof (QUERY (mountpoint)) && QUERY (mountpoint);
}

mapping find_file (string loc, object req_id)
{
  if (!online_acdb || !sizeof (QUERY (mountpoint))) return 0;
  if (req_id->method != "GET" && req_id->method != "POST") return 0;

  mixed res = root_access (req_id);
  if (res) return res;

  object(State) state = get_state (req_id, online_acdb);

  if (req_id->variables[TAG_CHNGUSER]) {
    state->change_user = !state->change_user;
    if (state->change_user)
      return mac->access_throw (PP_ROOT, req_id);
  }

  res = page_body (req_id, QUERY (mountpoint), online_acdb, state);

  if (!stringp (res)) return res;

  res = header (state, req_id->misc->acadm_backdoor ? "Using backdoor" :
		mac->id_get_name (req_id)) + res + "\n</body></html>\n";

  return http_string_answer (parse_rxml (res, req_id)) |
    (["extra_heads": (["Expires": http_date (0),
		       "Last-Modified": http_date (time (1))])]);
}
