// This is a Roxen module.

string cvs_version = "$Id: acauth_httpbasic.pike,v 1.14 1999/07/01 08:43:06 mast Exp $";

#include <config.h>
#include <module.h>
inherit "acauthmodule";
inherit "roxenlib";

int successes = 0;
int failures = 0;
int cache_hits = 0;
int cache_misses = 0;

class IdAuthCache
{
  int timestamp;
  string cryptwd;		// This is kept hashed for paranoic reasons..
  int|object(IdAuth) ia;
}
static mapping(string:array(object(IdAuthCache))) ia_cache = ([]);
// Cache of looked up handles.

#ifdef THREADS
static object(Thread.Mutex) ia_cache_lock = Thread.Mutex();
#endif

int|object(IdAuth) cached_auth_passwd (object(AC_DB) db, string user, string passwd)
{
  if (!QUERY (cache_enabled)) return auth_std_passwd (db, user, passwd);

#ifdef THREADS
  object lock = ia_cache_lock->lock();
#endif
  array(object(IdAuthCache)) iacs = ia_cache[user];
  if (iacs) {
    object(IdAuthCache) iac;
    foreach (iacs, object(IdAuthCache) i)
      if (crypt (passwd, i->cryptwd)) {iac = i; break;}
    if (iac) {
      if (iac->ia && iac->timestamp > time (1) - QUERY(cache_ttl) && iac->ia->table()) {
	cache_hits++;
	return iac->ia;
      }
      destruct (iac);
      ia_cache[user] -= ({0});
    }
  }
#ifdef THREADS
  lock = 0;
#endif
  cache_misses++;

  int|object(IdAuth) ia = auth_std_passwd (db, user, passwd);
  if (!objectp (ia)) return ia;

#ifdef THREADS
  lock = ia_cache_lock->lock();
#endif
  object(IdAuthCache) iac;
  if (iacs = ia_cache[user])
    foreach (iacs, object(IdAuthCache) i)
      if (crypt (passwd, i->cryptwd)) {iac = i; break;}
  if (!iac) {
    iac = IdAuthCache();
    if (iacs) ia_cache[user] += ({iac});
    else ia_cache[user] = ({iac});
  }
  iac->ia = ia;
  iac->timestamp = time (1);
  iac->cryptwd = crypt (passwd);

  return ia;
}

// Module interface.

string status()
{
  string res =
    "<h4>Authentications since last module reload</h4>\n\n"
    "Successful: " + successes + "<br>\n"
    "Failed: " + failures + "\n";
  if (QUERY (cache_enabled))
    res +=
      "<h4>Password lookup cache status</h4>\n\n"
      "Number of entries: " + sizeof (ia_cache) + "<br>\n"
      "Misses since last module reload: " + cache_misses + "<br>\n"
      "Hits since last module reload: " + cache_hits + "\n";
  return res;
}

array register_module()
{
  return ({MODULE_SECURITY, "AC: Basic HTTP and FTP authentication",
	   "Performs authentication with passwords using the HTTP "
	   "basic authentication scheme. It also works for FTP "
	   "connections.\n"
	   "\n"
	   "<p>Note that unless SSL is used, this method has very "
	   "low security; the password is sent effectively unencrypted "
	   "over the network.\n"
	   "\n"
	   "<p>Note also that the AC system is capable of authenticating "
	   "users by calling AC modules in other servers across the "
	   "network. In that case, the password may be sent unencrypted "
	   "to the other server.",
	   0, 1});
}

void create()
{
  defvar ("cache_enabled", 0, "Cache results of password lookups",
	  TYPE_FLAG,
	  "Whether to cache the results of user/password lookups for a "
	  "limited time. This is normally only needed when you check "
	  "passwords in a remote AC module.");
  defvar ("cache_ttl", 600, "Number of seconds password lookups are cached",
	  TYPE_INT,
	  "The amount of time user/password lookups are cached. You can "
	  "flush the cache by reloading this module.",
	  0, lambda() {return !QUERY (cache_enabled);});
}

// Auth interface.

constant auth_types = auth_types_std_passwd;
constant protocols = ({"HTTP", "FTP"});
constant replicable = 1;

object(Identity) passive_auth (object(ProtPoint) pp, object req_id)
{
  string user, passwd;

  if (!req_id->realauth || sscanf (req_id->realauth, "%s:%s", user, passwd) != 2)
    return 0;

  // There's currently no mechanism to limit the set of identities
  // based on the realm of a ppoint, so we auth against the whole db.
  object(AC_DB) db = pp->db();

  int|object(IdAuth) ia = cached_auth_passwd (db, user, passwd);

  if (objectp (ia)) {
    successes++;
    return ia->identity();
  }

  if (!ia) return db->identities->everyone();
  failures++;
  return 0;
}

array(object(Identity)) global_passive_auth (object(AC_DB) db, object req_id)
{
  string user, passwd;

  if (!req_id->realauth || sscanf (req_id->realauth, "%s:%s", user, passwd) != 2)
    return ({});

  int|object(IdAuth) ia = cached_auth_passwd (db, user, passwd);
  if (objectp (ia)) {
    successes++;
    return ({ia->identity()});
  }

  if (ia) failures++;
  return ({});
}

mixed active_auth (object(ProtPoint) pp, object req_id, string realm, string msg)
{
  return http_auth_required (realm, msg);
}
