#!NOMODULE
// Program to be inherited by all AC authentication modules.

// $Id: acauthmodule.pike,v 1.21 1999/07/01 04:42:08 mast Exp $

#include <module.h>
inherit "module";
inherit "aclib";

object config;

object ac_get()
{
  mapping acvar = roxen->query_var ("AC");
  return acvar && acvar->loaders && acvar->loaders[config];
}

void ac_connect()
{
  object ac = ac_get();
  if (ac && ac->acauth_modules) ac->acauth_modules[this_object()] = 1;
}

void ac_disconnect()
{
  object ac = ac_get();
  if (ac) {
    if (ac->acauth_modules) ac->acauth_modules[this_object()] = 0;
    object(AC_DB) acdb = ac->online_acdb;
    if (acdb && acdb->initialized) acdb->flush_all_caches();
  }
}

object(AC_DB) online_acdb()
{
  object ac = ac_get();
  return ac && ac->online_acdb;
}

// Module interface.

string status()
{
  return !ac_get() && "<b>The AC system is not active.</b>";
}

void start (int when, object conf)
{
  module_dependencies (config = conf, ({"acadm"}));
  ac_connect();
}

void stop()
{
  ac_disconnect();
}

// Auth interface.

// array(int) auth_types = ({});
// This must be set to the auth types (e.g. AUTH_PASSWD_CRYPT) this
// module supports. It's reasonable to leave it empty if the module
// both collects auth data and does actual authentication with it in a
// purely internal way. The existence of this identifier signifies
// that this is an AC authentication module.

// object(Identity) passive_auth (object(ProtPoint) pp, object req_id);
//
// If this function is defined, this auth module is able to perform
// authentication: The given request req_id is being identified on the
// given protection point pp. The function should search for and
// return the identity that the request can be authenticated as. It
// should return 0 if a specific identity was identified but the
// credentials are incorrect, and db->identities->everyone() if noone
// could be identified.

// mixed active_auth (object(ProtPoint) pp, object req_id,
//                    string realm, string msg)
//
// If this function is defined, this auth module is able to perform
// active authentication: The given request req_id doesn't have
// sufficient permission on the given protection point pp. The
// function should return an appropriate answer or perform the
// necessary dialogue according to the current protocol to try to
// validate the request. The (final) response from the user must then
// be handled in passive_auth.
//
// The string realm is taken from a protection class and describes the
// realm of resources the user tried to access. If msg is defined, it
// is some kind of error message explaining why the request was denied
// access to the resource. Both of these strings should be included in
// appropriate ways in the response if possible.

// array(object(Identity)) global_passive_auth (object(AC_DB) db, object req_id);
//
// If defined, this function should identify the request req_id on any
// identity in the given AC database. It should return an array of all
// matching identities.
//
// This function is typically used by the MODULE_AUTH interface in the
// AC system to identify requests against the online database.

// array(string) protocols;
//
// If passive_auth or active_auth is defined, this must also be
// defined. It's an array containing the names of the protocols this
// module can handle. Valid names include "HTTP/1.0", "HTTP/0.9",
// "FTP", "GOPHER" and "NFSv2". A matching prefix suffices, e.g.
// specify "HTTP" to match any HTTP version.

// int replicable;
//
// If this is nonzero, the module doesn't have any state outside that
// stored in the AC database. This is used in distributed databases;
// instead of making remote calls to the proper module instance, a
// local copy of the module will always be instantiated and used
// instead.

// void id_update_userinfo (object(IdAuth) ia)
//
// If this function is defined, this auth module is able to update the
// user info (e.g. full name and user name) of an identity which has
// it as an authentication method. It should do this update to the
// identity for the given IdAuth object.

// void id_handle_changed (object(IdAuth) ia)
//
// If this function is defined, it's called whenever the handle in the
// identity for the given IdAuth object changes.

// object(IdAuth) id_auth_init (object(IdAuth) ia)
//
// If this function is defined, this auth module supports
// authentication of explicitly given identities. It must do whatever
// initialization necessary of the given IdAuth object (which already
// has the corresponding identity filled in). It should not add the
// object to the AC database; the caller handles that. If it's
// successful, it returns the same object, otherwise it returns 0.

// int id_auth_destroy (object(IdAuth) ia)
//
// If id_auth_init is defined, this function should also be defined to
// support explicit removal of this authentication method from an
// identity. It must do whatever cleanup necessary and return nonzero
// to allow the IdAuth object to be removed, zero otherwise. It should
// not remove the object from the AC database; the caller handles
// that.

// string id_auth_view (object(IdAuth) ia, mapping(string:string) variables, int access)
//
// If this function is defined, it's called by the admin UI to display
// or change the variables connected to the given IdAuth object. It
// should return a rxml string containing a form or just the status of
// the auth method if it doesn't support manual editing or the user
// has read-only access. access is set to the permission the user has
// to this data, it's either PERM_READ or PERM_WRITE. variables
// contains the user response for a previously generated form. Do not
// use form variables beginning with `_' here.
//
// FIXME: This UI interface stinks. It'll hopefully be replaced by a
// defvar-style work soon.

// array(string) id_get_userinfo (object(IdAuth) ia)
//
// If defined, this function is called to get an auth module-style
// userinfo array, i.e. ({handle, cryptwd, uid, gid, fullname,
// homedir, shell}), about the given IdAuth object (which this module
// has created earlier). Fields the module has no info on should be
// set to zero (don't fill in stuff from the Identity object).

// Helpers.

constant auth_types_std_passwd = ({AUTH_PASSWD_CRYPT, AUTH_NT});
// The auth types that auth_std_passwd() below handles.

int|object(IdAuth) auth_std_passwd (object(AC_DB) db, string user, string passwd)
// Checks if the given user and password is valid with any of the
// standard auth data types that uses a user/password tuple. Returns
// the IdAuth object of a successful match, 1 if some user matched but
// passwd was wrong, 0 otherwise.
{
  object(IdAuths) iatbl = db->id_auths;

  object(IdAuth) ia = iatbl->search_by_secret (
    AUTH_PASSWD_CRYPT, user, lambda (string cryptwd, string passwd) {
			       return crypt (passwd, cryptwd);
			     }, passwd);
  if (ia) return ia;

  ia = iatbl->search_with_testfn (AUTH_NT, user, 0, passwd);
  if (ia) return ia;

  if (iatbl->sizeof_secrets (AUTH_PASSWD_CRYPT, user) ||
      iatbl->sizeof_secrets (AUTH_NT, user)) return 1;

  return 0;
}

// Functions for accessing variables remotely.
array(int) get_auth_types() {return this_object()->auth_types;}
array(string) get_protocols() {return this_object()->protocols;}
