#!NOMODULE
// Plain file database backend.

// $Id: acbackend_file.pike,v 1.3 1999/07/23 19:45:11 mast Exp $

#include <config.h>
#include <module.h>
inherit "module";
inherit "aclib";

import Stdio;

inherit ACDBBackend;

object acadm;
static string file;
static int binary;
static int mtime, last_stat_time = 0;

void read_db (void|int is_initialized)
{
  string last_io_error = 0;

  object(FILE) fd = FILE();

  if (fd->open (file, "r")) {
    if (!is_initialized) acdb->begin (begin, 1);
    last_io_error = acdb->read_file (fd);
    if (fd->errno())
      last_io_error = "Error reading DB " + file + ": " + strerror (fd->errno());
    else if (last_io_error)
      last_io_error = "Error reading DB " + file + ": " + last_io_error;
  }

  else if (fd->open (file, "cxw")) {
    if (!is_initialized) acdb->begin (begin, 1);
    acdb->write_file (fd, binary);
    if (fd->errno())
      last_io_error = "Error creating new DB " + file + ": " + strerror (fd->errno());
    else report_notice ("AC: A new database file was created\n");
  }

  else {
    last_io_error = "Error opening DB " + file + " for read: " + strerror (fd->errno());
    fd = 0;
  }

#ifndef __NT__
  if (!last_io_error) mtime = fd->stat()[3];
#endif
  if (fd) fd->close();
#ifdef __NT__
  if (!last_io_error) {array(int) stat = file_stat (file); mtime = stat && stat[3];}
#endif

  if (last_io_error) {
    report_error ("AC: " + last_io_error + "\n");
    if (fd) acdb->erase();
  }
  else acdb->commit (commit);

  acadm->last_io_error = last_io_error;
}

void create (object _acadm, object(AC_DB) _acdb,
	     function(object(AC_DB):void) _init_callback,
	     string _file, void|int _binary)
{
  acadm = _acadm;
  ACDBBackend::create (_acdb);
  file = _file;
  binary = _binary;
  read_db();
  if (acdb->initialized) _init_callback (acdb);
}

void begin (void|int fromscratch)
{
  if (!file || !sizeof (file)) {
    acadm->last_io_error = "Error reading DB: No file specified";
    report_error ("AC: " + acadm->last_io_error + "\n");
    return;
  }

  if (!fromscratch && last_stat_time == time (1)) return;

  last_stat_time = time (1);
  int *stat = file_stat (file);
  if (fromscratch || !stat || stat[3] > mtime) read_db (fromscratch);
}

int commit()
{
  string last_io_error = 0;

  object(FILE) fd = FILE();

  if (fd->open (file + ".w", "ctw")) {
    acdb->write_file (fd, binary);
    if (fd->errno()) {
      fd->close();
      last_io_error = "Error writing DB " + file + ": " + strerror (fd->errno());
      rm (file + ".w");
    }
    else {
#ifndef __NT__
      mtime = fd->stat()[3];
#endif
      fd->close();
#ifdef __NT__
      array(int) stat = file_stat (file);
      mtime = stat && stat[3];
      rm(file+"~");
#endif	
      if (!mv (file, file + "~") || !mv (file + ".w", file))
	last_io_error = "Error renaming DB " + file + ": " + strerror (fd->errno());
    }
  }
  else
    last_io_error = "Error opening DB " + file + " for write: " +
      strerror (fd->errno());

  if (last_io_error) report_error ("AC: " + last_io_error + "\n");

  acadm->last_io_error = last_io_error;
  return !last_io_error;
}
