// This is a Roxen module.

string cvs_version = "$Id: acimport_acdb.pike,v 1.9 1999/08/25 20:19:58 mast Exp $";

#include <module.h>
inherit "acauthmodule";
inherit "roxenlib";

import Array;
import Stdio;

string query_name();
void id_update_userinfo (object(IdAuth) ia);
void id_handle_changed (object(IdAuth) ia);

#define DATA_ID		0
#define DATA_HANDLE	1
#define DATA_NAME	2

int lastimport = 0, readusers, readgroups, impusers, impgroups, readpws, imppws;


// ACDB access.

object(IdAuth) update_id (object(AC_DB) acdb,
			  mapping(int:object(IdAuth)) oldias, object(Identity) impid)
{
  object(IdAuth) ia = oldias[impid->id()];
  object(Identity) id;

  if (!ia) {
    ia = acdb->id_auths->new (this_object());

    if (impid->handle() &&
	(QUERY (sethandle) == "Both" ||
	 QUERY (sethandle) == (impid->type() == ID_TYPE_USER ? "Users" : "Groups")))
      id = acdb->identities->find_by_handle (impid->handle());
    else if (impid->name()) {
      object(Identity) *ids = filter (acdb->identities->find_by_name (impid->name() ||
								      impid->handle()),
				      lambda (object(Identity) id, int type) {
					return id->type() == type;
				      }, impid->type());
      if (sizeof (ids) == 1 && !ids[0]->handle()) id = ids[0];
    }
    if (id && id->type() == impid->type() &&
	search (id->authmods(), this_object()) < 0) {
      // Attach the user/group to an existing identity of the same
      // type if it has the same handle or the same name if both
      // handles are 0 and isn't under control of this importer
      // earlier.
      ia->set_identity (id);
      acdb->id_auths->add (ia);
#ifdef MODULE_DEBUG
      report_debug (query_name() + ": Attaching " + (impid->handle() || impid->id()) +
		    " to " + ID_TYPE_TEXTS[id->type()] + " " +
		    (id->handle() || id->id()) + "\n");
#endif
    }
    else {
      id = acdb->identities->new (impid->type());
      acdb->identities->add (id);
      ia->set_identity (id);
      acdb->id_auths->add (ia);
      id->set_src_auth (ia);
#ifdef MODULE_DEBUG
      report_debug (query_name() + ": Creating new " + ID_TYPE_TEXTS[id->type()] +
		    " for " + (impid->handle() || impid->id()) + "\n");
#endif
    }
  }
  else id = ia->identity();

  array data = ({impid->id(), impid->handle(), impid->name()});
  if (!equal (data, ia->data)) {
    ia->data = data, ia->set_dirty();
    if (id->src_auth() == ia) id_update_userinfo (ia);
  }

  id_handle_changed (ia);

  return ia;
}

void update_acdb (object(AC_DB) acdb, object(AC_DB) impdb, object(Identity) *ids)
{
  mapping(object(Identity):object(IdAuth)) ias = ([]);

  mapping(int:object(IdAuth)) oldias = ([]);
  foreach (acdb->id_auths->find_by_authmod (this_object()), object(IdAuth) ia)
    oldias[ia->data[DATA_ID]] = ia;

  lastimport = time (1);
  impusers = impgroups = imppws = 0;

  foreach (ids, object(Identity) id) {
    if (id->type() == ID_TYPE_USER) impusers++; else impgroups++;
    ias[id] = update_id (acdb, oldias, id);
  }

  foreach (values (oldias) - values (ias), object(IdAuth) ia)
    acdb->id_auths->remove (ia);

  foreach (ids, object(Identity) id) {
    object(IdAuth) ia = ias[id];
    multiset(object(Identity)) oldgroups = ia->groups();
    multiset(object(Identity)) newgroups = (<>);
    foreach (indices (id->direct_groups()), object(Identity) g)
      if (ias[g]) newgroups[ias[g]->identity()] = 1;
    newgroups -= (<0>);

    foreach (indices (oldgroups - newgroups), object(Identity) id)
      ia->remove_group (id);
    foreach (indices (newgroups - oldgroups), object(Identity) id)
      ia->add_group (id);
  }
}


// Getting data.

array(string) split_filter_fields (string filter)
{
  string *qsplit = filter / "\"";
  string *fields = ({""});

  for (int q = 0, inquote = 0; q < sizeof (qsplit); q++, inquote = !inquote) {
    string field = qsplit[q];
    while (q < sizeof (qsplit) - 1 && qsplit[q][-1] == '\\')
      field += "\"" + qsplit[++q];

    if (inquote) fields[-1] += "\0" + field; // \0 is a flag for parse_filter_field.
    else {
      string *fsplit = (field - " " - "\t" - "\n") / ",";
      fields[-1] += fsplit[0];
      fields += fsplit[1..];
    }
  }

  return fields;
}

array parse_filter_field (string field)
{
  mixed pat;
  int invflag, atflag, nameflag = 0;

  invflag = sizeof (field) && field[0] == '^';
  if (invflag) field = field[1..];
  atflag = sizeof (field) && field[0] == '@';
  if (atflag) field = field[1..];
  if (!invflag) {		// May come after @.
    invflag = sizeof (field) && field[0] == '^';
    if (invflag) field = field[1..];
  }

  int i, dummy;
  if (sscanf (field, "%d%c", i, dummy) == 1)
    pat = i;
  else {
    nameflag = sizeof (field) && field[0] == 0;
    if (nameflag) field = field[1..];
    if (search (field, "\0") >= 0)
      report_warning (query_name() + ": Extra quotes in pattern: " +
		      replace (field, ({"\0", "\""}), ({"\"", "\\\""})) + "\n");
    field -= "\0";
    pat = field;
  }

  return ({pat, invflag, atflag, nameflag});
}

object(Identity) *apply_filter (object(AC_DB) impdb)
{
  string *flt = split_filter_fields (QUERY (importids));
  multiset(object(Identity)) left = mkmultiset (impdb->identities->this_values());
  object(Identity) *res = ({});

  readusers = readgroups = readpws = 0;
  for (int i = 0; i < sizeof (flt); i++) {
    array arr = parse_filter_field (flt[i]);
    mixed pat = arr[0];
    int rem = arr[1], chkgrp = arr[2], name = arr[3];

    foreach (indices (left), object(Identity) id) {
      int match = 0;
      foreach (chkgrp ? indices (id->groups()) : ({id}), object(Identity) tst)
	if (intp (pat)) {
	  if (pat == tst->id() || (string) pat == tst->handle()) {match = 1; break;}
	}
	else if (name) {
	  if (glob (pat, tst->name() || "")) {match = 1; break;}
	}
	else if (glob (pat, tst->handle() || "")) {match = 1; break;}
      if (match) {
	left[id] = 0;
	if (id->type() == ID_TYPE_USER) readusers++; else readgroups++;
	if (!rem) res += ({id});
      }
    }
  }

  foreach (indices (left), object(Identity) id)
    if (id->type() == ID_TYPE_USER) readusers++; else readgroups++;

  return res;
}

int acdb_file_mtime;

void get_data (object(AC_DB) acdb)
{
#ifdef MODULE_DEBUG
  //report_debug (query_name() + ": Updating\n");
#endif

  if (sizeof (QUERY (acdb_file)))
  {
    int mtime = file_stat (QUERY (acdb_file))[3];
    if (mtime != acdb_file_mtime) {
      object(AC_DB) impdb = AC_DB();
      impdb->initialize();

      object(FILE) fd = FILE();
      string err;
      {
	if (fd->open (QUERY (acdb_file), "r")) {
	  err = impdb->read_readable (fd);
	  if (fd->errno())
	    err = "Error reading ACDB " + QUERY (acdb_file) + ": " +
	      strerror (fd->errno());
	  else if (err)
	    err = "Error reading ACDB " + QUERY (acdb_file) + ": " + err;
	}

	else {
	  err = "Error opening ACDB " + QUERY (acdb_file) + " for read: " +
	    strerror (fd->errno());
	  fd = 0;
	}

	if (fd) fd->close();
      }

      if (err) report_error (query_name() + ": " + err + "\n");
      else {
#ifdef THREADS
	object dblock = acdb->lock();
#endif
	acdb->begin();
	object df = acdb->disable_flush();
	update_acdb (acdb, impdb, apply_filter (impdb));
	acdb->commit();
      }
      destruct (impdb);
    }
  }
}


// Module interface.

string status()
{
  string res;
  if (!lastimport) res = "<p>No successful import done yet.\n";
  else
    res = "<p>Last successful import: " + ctime (lastimport) + "<br>\n"
      "<table border=0>\n"
      "<tr><td></td><th>Read</th><th>Imported</th></gr>\n"
      "<tr><td>Users</td>"
      "<td align=right>" + readusers + "</td>"
      "<td align=right>" + impusers + "</td></tr>\n"
      "<tr><td>Groups</td>"
      "<td align=right>" + readgroups + "</td>"
      "<td align=right>" + impgroups + "</td></tr>\n"
      "<tr><td>Passwords</td>"
      "<td align=right>" + readpws + "</td>"
      "<td align=right>" + imppws + "</td></tr>\n"
      "</table>\n";
  return res;
}

array register_module()
{
  return ({MODULE_SECURITY, "AC: AC database user import",
	   "Imports users and groups from another AC database. "
	   "The imported data is updated automatically in periodic "
	   "intervals. Reload the module to force an update.",
	   0, 0});
}

void create()
{
  defvar ("acdb_file", "", "AC database file",
	  TYPE_FILE,
	  "The AC database file to read. It may be in use simultaneously "
	  "by another AC instance.");
  defvar ("importids", "^" + EVERYONE_ID + ",*", "Identities to import",
	  TYPE_STRING,
	  "This is a pattern that specifies which identities to import.\n"
	  "\n<p>It consists of fields separated by '<code>,</code>'. Each "
	  "field is either a username, a fullname surrounded by "
	  "'<code>\"</code>', or a numeric ID. The glob characters "
	  "'<code>?</code>' and '<code>*</code>' may be used in usernames "
	  "and fullnames. Quotes are escaped by '<code>\\</code>'. "
	  "Any field may be prefixed by '<code>@</code>' to also match all "
	  "identities that are members (directly or indirectly) of a given "
	  "group identity. The list of fields is searched from left to "
	  "right and the first field that matches the group in question "
	  "determines whether it should be included or not. It's not "
	  "included if the field is prefixed by '<code>^</code>', otherwise "
	  "it is. Whitespace is ignored anywhere except inside quotes.");
// Not yet implemented
//   defvar ("import_passwd", 0, "Import passwords",
// 	  TYPE_FLAG | VAR_MORE,
// 	  "Whether to import passwords. These will then work for "
// 	  "authentication in this AC database.");
//   defvar ("import_uid_gid", 0, "Import UIDs and GIDs",
// 	  TYPE_FLAG | VAR_MORE,
// 	  "Whether to import UIDs and GIDs. These are imported to the AC "
// 	  "database by the OS user importer module and is only used by "
// 	  "special authentication modules such as the NFS server.");
  defvar ("sethandle", "Both", "Set usernames",
	  TYPE_MULTIPLE_STRING | VAR_MORE,
	  "Whether to set the username from the imported data. You may not "
	  "want this, because the username is unique in the AC database, "
	  "so you might not want to have that namespace \"polluted\".",
	  ({"No", "Users", "Groups", "Both"}));
// Idea:
//   defvar ("basegroup", 0, "Base group for imports",
// 	  TYPE_FOO,
// 	  "All imported groups and users will become members of this group "
// 	  "if specified. Normally the importer will attach the imported "
// 	  "information to existing identities with the same usernames. "
// 	  "Setting this group will limit this to identities which are "
// 	  "members (directly or indirectly) to this group.");
  defvar ("update", 60, "Minutes between periodic updates",
	  TYPE_INT | VAR_MORE,
	  "This specifies the interval in minutes between periodic "
	  "updates of the imported information. 0 disables this feature.");
  defvar ("update_on_load", 0, "Update immediately on load",
	  TYPE_FLAG | VAR_MORE,
	  "Set this if you want the database to be updated when the "
	  "server is started. If the database is large, this causes "
	  "an extra delay till the server is ready to handle connections. "
	  "If it's disabled, the database will be updated first after a "
	  "full update period has passed.");
}

string query_name()
{
  return "AC: ACDB user import: " +
    (sizeof (QUERY (acdb_file)) ? QUERY (acdb_file) : "inactive");
}

static mixed periodic_update_callout;

void periodic_update()
{
  object(AC_DB) acdb = online_acdb();
  if (acdb) get_data (acdb);
  remove_call_out (periodic_update_callout);
  if (QUERY (update))
    periodic_update_callout = call_out (periodic_update, QUERY (update) * 60);
}

void start (int when, object conf)
{
  ::start (when, conf);

  acdb_file_mtime = 0;

  string key = get_modname (this_object()) + "_loaded";
  object(AC_DB) acdb = online_acdb();
  if (acdb && QUERY (update_on_load) || roxen->query_var (key)) get_data (acdb);
  roxen->set_var (key, 1); // To detect reloads.

  remove_call_out (periodic_update_callout);
  if (QUERY (update))
    periodic_update_callout = call_out (periodic_update, QUERY (update) * 60);
}


// Auth interface.

array(int) auth_types = ({});

void id_update_userinfo (object(IdAuth) ia)
{
  object(Identity) id = ia->identity();

  string handle = ia->data[DATA_HANDLE];
  if (QUERY (sethandle) == "No" ||
      QUERY (sethandle) == (id->type() != ID_TYPE_USER ? "Users" : "Groups"))
    handle = 0;
  else if (handle) {
    object(Identity) other = id->table()->find_by_handle (handle);
    if (other && other != id) handle = 0;
  }
  if (id->handle() != handle) id->set_handle (handle);

  string name = ia->data[DATA_NAME];
  if (id->name() != name) id->set_name (name);
}

void id_handle_changed (object(IdAuth) ia)
// Awkward name in this case..
{
}

string id_auth_view (object(IdAuth) ia, mapping(string:string) var, int access)
{
  return "Imported info: "
    "Username " + (ia->data[DATA_HANDLE] ?
		   "\"" + html_encode_string (ia->data[DATA_HANDLE]) + "\"" :
		   "none") +
    ", name " + (ia->data[DATA_NAME] ?
		 "\"" + html_encode_string (ia->data[DATA_NAME]) + "\"" :
		 "none") + "<br>\n";
}

array(string) id_get_userinfo (object(IdAuth) ia)
{
  return ({ia->data[DATA_HANDLE], 0, 0, 0, ia->data[DATA_NAME], 0, 0});
}
