// This is a Roxen module.

string cvs_version = "$Id: acimport_osuser.pike,v 1.32 1999/09/02 23:48:44 mast Exp $";

#include <config.h>
#include <module.h>
inherit "acauthmodule";
inherit "roxenlib";

#if !constant (trim)
string trim (string s)
{
  if (stringp (s)) {
    sscanf (reverse (s), "%*[ \t]%s", s);
    sscanf (reverse (s), "%*[ \t]%s", s);
  }
  return s;
}
#endif

import Array;
import Stdio;

// Fairly weak check of password for portability.
#define CRYPTWD_CHECK(cryptwd) \
  (!!(cryptwd) && sizeof (cryptwd) >= 10 && \
   search ((cryptwd), "*") < 0 && search ((cryptwd), "!") < 0)

string query_name();
void id_update_userinfo (object(IdAuth) ia);
void id_handle_changed (object(IdAuth) ia);

#define DATA_HANDLE	0
#define DATA_CRYPTWD	1
#define DATA_UID	2
#define DATA_GID	3
#define DATA_FULLNAME	4
#define DATA_HOMEDIR	5
#define DATA_COMMENT	6
#define DATA_SHELL	7

int lastimport = 0, readusers, readgroups, impusers, impgroups, readpws, imppws;
string dcsrv;


// ACDB access.

object(IdAuth) update_id (object(AC_DB) acdb,
			  mapping(string:object(IdAuth)) oldias, array data)
{
  int is_user = intp (data[DATA_UID]);
  string handle = data[DATA_HANDLE];
  string fullname = data[DATA_FULLNAME];

  while (sizeof (data) <= DATA_SHELL) data += ({0});

  object(IdAuth) ia = oldias[is_user + "\0" + (handle||fullname)];
  object(Identity) id;

  fullname = trim (data[DATA_FULLNAME]);
  if (!fullname || !sizeof (fullname)) fullname = data[DATA_FULLNAME] = 0;
  else if (QUERY (strip_gecos)) {
    int i = search (fullname, ",");
    if (i >= 0) {
      fullname = data[DATA_FULLNAME] = trim (fullname[..i-1]);
      if (!data[DATA_COMMENT]) data[DATA_COMMENT] = fullname[i+1..];
    }
  }
  {
    string homedir = data[DATA_HOMEDIR];
    if (homedir && !sizeof (homedir)) data[DATA_HOMEDIR] = 0;
  }
  data[DATA_COMMENT] = trim (data[DATA_COMMENT]);
  {
    string shell = data[DATA_SHELL];
    if (shell && !sizeof (shell)) data[DATA_SHELL] = 0;
  }

  if (!ia) {
    ia = acdb->id_auths->new (this_object());

    if (handle &&
	(QUERY (sethandle) == "Both" ||
	 QUERY (sethandle) == (is_user ? "Users" : "Groups")))
      id = acdb->identities->find_by_handle (handle);
    else if (fullname || handle) {
      array(object(Identity)) ids =
	filter (acdb->identities->find_by_name (fullname || handle),
		lambda (object(Identity) id, int type) {
		  return id->type() == type;
		}, is_user ? ID_TYPE_USER : ID_TYPE_GROUP);
      if (sizeof (ids) == 1 && !ids[0]->handle()) id = ids[0];
    }

    if (id && id->type() == (is_user ? ID_TYPE_USER : ID_TYPE_GROUP) &&
	search (id->authmods(), this_object()) < 0) {
      // Attach the user/group to an existing identity of the same
      // type if it has the same handle, or the same name if both
      // handles are 0, and isn't under control of this importer
      // earlier.
      ia->set_identity (id);
      acdb->id_auths->add (ia);
#ifdef MODULE_DEBUG
      report_debug (query_name() + ": Attaching " + (handle || fullname) + " to " +
		    ID_TYPE_TEXTS[id->type()] + " " + (id->handle() || id->id()) + "\n");
#endif
    }
    else {
      id = acdb->identities->new (is_user ? ID_TYPE_USER : ID_TYPE_GROUP);
      acdb->identities->add (id);
      ia->set_identity (id);
      acdb->id_auths->add (ia);
      id->set_src_auth (ia);
#ifdef MODULE_DEBUG
      report_debug (query_name() + ": Creating new " + ID_TYPE_TEXTS[id->type()] +
		    " for " + (handle || fullname) + "\n");
#endif
    }
  }
  else id = ia->identity();

  if (!equal (data, ia->data)) {
    ia->data = data, ia->set_dirty();
    if (id->src_auth() == ia) id_update_userinfo (ia);
  }

  id_handle_changed (ia);

  return ia;
}

void update_acdb (object(AC_DB) acdb, mapping(array:array) mbs,
		  mapping(mixed:array) uids, mapping(mixed:array) gids)
{
  mapping(array:object(IdAuth)) ias = ([]);

  mapping(string:object(IdAuth)) oldias = ([]);
  foreach (acdb->id_auths->find_by_authmod (this_object()), object(IdAuth) ia)
    oldias[(ia->identity()->type() == ID_TYPE_USER) + "\0" +
	   (ia->data[DATA_HANDLE] || ia->data[DATA_FULLNAME])] = ia;

  lastimport = time (1);
  impusers = impgroups = imppws = 0;

  if (QUERY (collapse_groups))
    foreach (indices (mbs), array data)
      if (intp (data[DATA_GID])) {
	array member = sizeof (mbs[data]) == 1 && mbs[data][0];
	if (mbs[member] && data[DATA_HANDLE] == member[DATA_HANDLE] &&
	    !CRYPTWD_CHECK (data[DATA_CRYPTWD])) {
	  member[DATA_GID] = data[DATA_GID];
	  m_delete (mbs, data);
	}
      }

  foreach (indices (mbs), array data) {
    if (intp (data[DATA_UID])) impusers++;
    if (intp (data[DATA_GID])) impgroups++;
    ias[data] = update_id (acdb, oldias, data);
    if (ias[data]->has_auth_data (AUTH_PASSWD_CRYPT)) imppws++;
  }

  foreach (values (oldias) - values (ias), object(IdAuth) ia)
    acdb->id_auths->remove (ia);

  foreach (indices (mbs), array data)
    if (intp (data[DATA_UID])) {
      object(IdAuth) ia = ias[data];
      multiset(object(Identity)) oldgroups = ia->groups();
      multiset(object(Identity)) newgroups = (<>);
      foreach (mbs[data], mixed g)
	if (ias[gids[g]]) newgroups[ias[gids[g]]->identity()] = 1;
      newgroups -= (<0>);

      foreach (indices (oldgroups - newgroups), object(Identity) id)
	ia->remove_group (id);
      foreach (indices (newgroups - oldgroups), object(Identity) id)
	ia->add_group (id);
    }
}

#ifdef __NT__
array apply_filter_nt (array users, array groups, array lgroups)
{
  array(array|string) usrflt = QUERY (importusers) / ",";
  array(array|string) grpflt = QUERY (importgroups) / ",";
  array(int) remusr = allocate (sizeof (usrflt)),
    usrchkgrp = allocate (sizeof (usrflt)),
    remgrp = allocate (sizeof (grpflt));

  for (int i = 0; i < sizeof (usrflt); i++) {
    array res = parse_filter_field (usrflt[i]);
    usrflt[i] = res[0], remusr[i] = res[1], usrchkgrp[i] = res[2];
  }
  for (int i = 0; i < sizeof (grpflt); i++) {
    array res = parse_filter_field (grpflt[i]);
    grpflt[i] = res[0], remgrp[i] = res[1];
  }

  readusers = readgroups = 0;

  array(array) usrgrps = allocate (sizeof (users));
  mapping(string:array) grs, lgrs;
  for (int i = 0; i < sizeof (users); i++) {
    int remove = 1;
    readusers++;
    for (int j = 0; j < sizeof (usrflt); j++) {
      array chkdata;
      if (usrchkgrp[j]) {
	if (!usrgrps[i]) {
	  if (!grs) {
	    grs = mkmapping (column (groups, DATA_FULLNAME), groups);
	    lgrs = mkmapping (column (lgroups, DATA_FULLNAME), lgroups);
	  }
	  usrgrps[i] = rows (grs, users[i][DATA_COMMENT + 1]) +
	    rows (lgrs, users[i][DATA_COMMENT + 2]);
	}
	chkdata = usrgrps[i];
      }
      else chkdata = ({users[i]});
      int match = 0;
      foreach (chkdata, array tst)
	if (arrayp (usrflt[j])) {
	  mixed idno = intp (tst[DATA_UID]) ? tst[DATA_UID] : tst[DATA_GID];
	  if (intp (idno) &&
	      (!intp (usrflt[j][0]) || usrflt[j][0] <= idno) &&
	      (!intp (usrflt[j][1]) || usrflt[j][1] >= idno)) {match = 1; break;}
	}
	else if (glob (usrflt[j], intp (tst[DATA_UID]) ?
		       tst[DATA_HANDLE] : tst[DATA_FULLNAME])) {match = 1; break;}
      if (match) {remove = remusr[j]; break;}
    }
    if (remove) users[i] = 0;
  }

  array(array) allgroups = groups + lgroups;
  for (int i = 0; i < sizeof (allgroups); i++) {
    array data = allgroups[i];
    int remove = 1;
    readgroups++;
    for (int j = 0; j < sizeof (grpflt); j++) {
      if (arrayp (grpflt[j])) {
	mixed idno = data[DATA_GID];
	if (intp (idno) &&
	    (!intp (grpflt[j][0]) || grpflt[j][0] <= idno) &&
	      (!intp (grpflt[j][1]) || grpflt[j][1] >= idno))
	  {remove = remgrp[j]; break;}
      }
      else
	if (glob (grpflt[j], data[DATA_FULLNAME]))
	  {remove = remgrp[j]; break;}
    }
    if (remove) {
      if (i >= sizeof (groups)) lgroups[i - sizeof (groups)] = 0;
      else groups[i] = 0;
    }
  }

  return ({users - ({0}), groups - ({0}), lgroups - ({0})});
}

void update_ntusers (object(AC_DB) acdb, array users,
		     array groups, array lgroups)
{
  mapping(string:object(IdAuth)) grs = ([]), lgrs = ([]);

  mapping(array:object(IdAuth)) ias = ([]);

  mapping(string:object(IdAuth)) oldias = ([]);
  foreach (acdb->id_auths->find_by_authmod (this_object()), object(IdAuth) ia)
    oldias[(ia->identity()->type() == ID_TYPE_USER) + "\0" +
	   (ia->data[DATA_HANDLE]||ia->data[DATA_FULLNAME])] = ia;

  [users, groups, lgroups] = apply_filter_nt (users, groups, lgroups);

  lastimport = time (1);
  impusers = impgroups = imppws = 0;

  foreach (users, array data) {
    impusers++;
    ias[data] = update_id (acdb, oldias, data[..DATA_COMMENT]);
  }

  foreach (groups, array data) {
    impgroups++;
    grs[data[DATA_FULLNAME]] = ias[data] =
      update_id (acdb, oldias, data[..DATA_COMMENT]);
  }

  foreach (lgroups, array data) {
    impgroups++;
    lgrs[data[DATA_FULLNAME]] = ias[data] =
      update_id (acdb, oldias, data[..DATA_COMMENT]);
  }

  foreach (values (oldias) - values (ias), object(IdAuth) ia)
    acdb->id_auths->remove (ia);

  foreach (users+groups+lgroups, array data)
  {
    object(IdAuth) ia = ias[data];
    multiset(object(Identity)) oldgroups = ia->groups();
    multiset(object(Identity)) newgroups = (<>);
    foreach (data[DATA_COMMENT+1], string g)
      if (grs[g]) newgroups[grs[g]->identity()] = 1;
    foreach (data[DATA_COMMENT+2], string lg)
      if (lgrs[lg]) newgroups[lgrs[lg]->identity()] = 1;
    newgroups -= (<0>);
    
    foreach (indices (oldgroups - newgroups), object(Identity) id)
      ia->remove_group (id);
    foreach (indices (newgroups - oldgroups), object(Identity) id)
      ia->add_group (id);
  }
}

static array resolve_ntuserdata(string srv, array users)
{
  users = Array.filter(users, lambda(array a) {
				return a[6]&system.UF_NORMAL_ACCOUNT;
			      });
  foreach(indices(users), int i) {
    array udata = users[i];
    array groups, lgroups;
    if (catch (groups = system.NetUserGetGroups(srv,udata[0],0)))
      groups = ({});
    if (catch (lgroups = system.NetUserGetLocalGroups(srv,udata[0],0)))
      lgroups = ({});
    users[i] = rows(udata, ({0,1,24,25,9,4,5})) + ({groups, lgroups});
  }
  return users;
}

static array resolve_ntgroupdata(string srv, array groups)
{
  foreach(indices(groups), int i) {
    array gdata = groups[i];
    groups[i] = ({0, 0, "", gdata[2], gdata[0], 0, gdata[1]})+({({}),({})});
  }
  return groups;
}

static array resolve_ntlgroupdata(string srv, array lgroups, array groups)
{
  mapping(string:array) gmap = mkmapping(column(groups, 4), groups);
  foreach(indices(lgroups), int i) {
    array lgdata = lgroups[i];
    lgroups[i] = ({0, 0, "", "", lgdata[0], 0, lgdata[1]})+({({}),({})});
    array lmem;
    if (catch (lmem = system.NetLocalGroupGetMembers(srv, lgdata[0], 1)))
      lmem = ({});
    foreach(rows(gmap,
		 column(Array.filter(lmem,
				     lambda(array a) {
				       return (<system.SidTypeGroup,
						system.SidTypeWellKnownGroup>)
					 [a[1]];
				     }), 2))-({0}), array gg)
      gg[DATA_COMMENT+2]+=({lgdata[0]});
  }
  return lgroups;
}
#endif

// Getting data.

// Data structures: A user or group is represented by an array
// containing data about it.
// o. mbs maps identity data arrays to arrays of handles/UIDs for
//    every group the user is a member of, or vice versa if the identity
//    is a group.
// o. uids maps handles to either a data array in mbs' index, or to an
//    array of the handles/UIDs that user is a member of so far.
// o. gids does the reverse for groups, but also maps GIDs.

static object(Regexp) sane = Regexp ("^[-*?_0-9A-Za-z---]*$");

array parse_filter_field (string field)
{
  mixed pat;
  int invflag, atflag;

  field = trim (field);

  invflag = sizeof (field) && field[0] == '^';
  if (invflag) field = field[1..];
  atflag = sizeof (field) && field[0] == '@';
  if (atflag) field = field[1..];
  if (!invflag) {		// May come after @.
    invflag = sizeof (field) && field[0] == '^';
    if (invflag) field = field[1..];
  }

  int l = 0, h = -1, hl, hh, dummy;
  string ls, hs;
  if (sscanf (field, "[%s..%s]%c", ls, hs, dummy) == 2 &&
      (hl = sscanf (ls, "%d%c", l, dummy)) <= 1 &&
      (hh = sscanf (hs, "%d%c", h, dummy)) <= 1)
    pat = ({hl ? l : "", hh ? h : ""});
  else if (sscanf (field, "[%d%c", l, dummy) == 2 && dummy == ']')
    pat = ({l, l});
  else {
    if (!sane->match (field))
      report_warning (query_name() + ": Suspect glob pattern: \"" + field + "\"\n");
    pat = field;
  }

  return ({pat, invflag, atflag});
}

void apply_filter (mapping(array:array) mbs, mapping(mixed:array) gids)
{
  array(array|string) usrflt = QUERY (importusers) / ",";
  array(array|string) grpflt = QUERY (importgroups) / ",";
  array(int) remusr = allocate (sizeof (usrflt)),
    usrchkgrp = allocate (sizeof (usrflt)),
    remgrp = allocate (sizeof (grpflt));

  for (int i = 0; i < sizeof (usrflt); i++) {
    array res = parse_filter_field (usrflt[i]);
    usrflt[i] = res[0], remusr[i] = res[1], usrchkgrp[i] = res[2];
  }
  for (int i = 0; i < sizeof (grpflt); i++) {
    array res = parse_filter_field (grpflt[i]);
    grpflt[i] = res[0], remgrp[i] = res[1];
  }

  readusers = readgroups = readpws = 0;
  foreach (indices (mbs), array data) {
    array(array|string) flt;
    int remove = 1;
    array(int) rem;
    if (intp (data[DATA_UID])) flt = usrflt, rem = remusr, readusers++;
    else flt = grpflt, rem = remgrp, readgroups++;
    if (CRYPTWD_CHECK (data[DATA_CRYPTWD])) readpws++;
    for (int i = 0; i < sizeof (flt); i++) {
      int match = 0;
      foreach (flt == usrflt && usrchkgrp[i] ? rows (gids, mbs[data]) : ({data}),
	       array tst)
	if (mbs[tst]) {
	  if (arrayp (flt[i])) {
	    int idno = intp (tst[DATA_UID]) ? tst[DATA_UID] : tst[DATA_GID];
	    if ((!intp (flt[i][0]) || flt[i][0] <= idno) &&
		(!intp (flt[i][1]) || flt[i][1] >= idno)) {match = 1; break;}
	  }
	  else if (glob (flt[i], tst[DATA_HANDLE])) {match = 1; break;}
	}
      if (match) {remove = rem[i]; break;}
    }
    if (remove)
      if (flt == usrflt) m_delete (mbs, data);
      else data[DATA_UID] = "remove"; // Remove groups in a separate pass.
  }
  foreach (indices (mbs), array data)
    if (data[DATA_UID] == "remove") m_delete (mbs, data);
}

void resolve_pwdata (array(array) pwdata, mapping(array:array) mbs,
		     mapping(mixed:array) uids, mapping(mixed:array) gids)
{
  foreach (pwdata, array pwent)
    if (mbs[uids[pwent[0]]])
      report_warning (query_name() + ": Duplicate user " + pwent[0] +
		      " UID " + pwent[2] + "\n");
    else {
      if (mbs[gids[pwent[3]]]) mbs[gids[pwent[3]]] += ({pwent});
      else if (gids[pwent[3]]) gids[pwent[3]] += ({pwent});
      else gids[pwent[3]] = ({pwent});

      mbs[pwent] = ({pwent[3]});
      pwent[3] = "";
      if (uids[pwent[0]]) mbs[pwent] += uids[pwent[0]];
      uids[pwent[0]] = pwent;
    }
}

void resolve_grdata (array(array) grdata, mapping(array:array) mbs,
		     mapping(mixed:array) uids, mapping(mixed:array) gids)
{
  foreach (grdata, array grent)
    if (mbs[gids[grent[0]]] || mbs[gids[grent[2]]])
      report_warning (query_name() + ": Duplicate group " + grent[0] +
		      " GID " + grent[2] + "\n");
    else {
      array data = ({grent[0], grent[1], "", grent[2], 0, 0, 0, 0});
      mbs[data] = gids[grent[2]] ? gids[grent[2]] : ({});
      if (gids[grent[0]]) mbs[data] += gids[grent[0]];
      gids[grent[2]] = gids[grent[0]] = data;

      foreach (grent[3], string user) {
	mbs[data] += ({user});
	if (mbs[uids[user]]) mbs[uids[user]] += ({grent[0]});
	else if (uids[user]) uids[user] += ({grent[0]});
	else uids[user] = ({grent[0]});
      }
    }
}

void resolve_spdata (array(array) spdata, mapping(array:array) mbs,
		     mapping(mixed:array) uids)
{
  int now = time (1) / (60*60*24); // Days since Jan 1, 1970.
  foreach (spdata, array spent) {
    array data = uids[spent[0]];
    if (mbs[data]) {
      string cryptwd = spent[1];
      if (spent[7] > 0 && spent[7] <= now) cryptwd = "";
      else if (spent[2] >= 0 && spent[4] >= 0 && spent[6] >= 0 &&
	       spent[2] + spent[4] + spent[6] <= now) cryptwd = "";
      data[DATA_CRYPTWD] = cryptwd;
    }
  }
}

void resolve_sgdata (array(array) sgdata, mapping(array:array) mbs,
		     mapping(mixed:array) gids)
{
  foreach (sgdata, array sgent) {
    array data = gids[sgent[0]];
    if (mbs[data]) data[DATA_CRYPTWD] = sgent[1];
  }
}

array(array) parse_passwd (string file)
{
  array(array) pwdata = ({});
  int lineno = 0;
  foreach (file / "\n", string line) {
    lineno++;
    array(string) fields = line / ":";
    int dummy, uid, gid;
    if (sizeof (fields) < 6 ||	// Don't bother about login shell.
	!sizeof (fields[0]) ||
	sscanf (fields[2], "%d%c", uid, dummy) != 1 ||
	sscanf (fields[3], "%d%c", gid, dummy) != 1) {
      if (sizeof (fields) != 1 || sizeof (fields[0]))
	report_error (query_name() + ": "
		      "Parse error in passwd data line " + lineno + "\n");
    }
    else
      // There seems to be some strange syntax that stores expire
      // information in the password field with a comma separated
      // list. Maybe it should be parsed or removed. Now it's neither.
      pwdata += ({({fields[0], fields[1], uid, gid, fields[4], fields[5], 0,
		    sizeof (fields) > 6 && fields[6]})});
  }
  return pwdata;
}

array(array) parse_group (string file)
{
  array(array) grdata = ({});
  int lineno = 0;
  foreach (file / "\n", string line) {
    lineno++;
    array(string) fields = line / ":";
    int dummy, gid;
    if (sizeof (fields) < 4 || !sizeof (fields[0]) ||
	sscanf (fields[2], "%d%c", gid, dummy) != 1) {
      if (sizeof (fields) != 1 || sizeof (fields[0]))
	report_error (query_name() + ": "
		      "Parse error in group data record " + lineno + "\n");
    }
    else grdata += ({({fields[0], fields[1], gid, fields[3] / ","})});
  }
  return grdata;
}

array(array) parse_shadow (string file)
{
  array(array) spdata = ({});
  int lineno = 0;
  foreach (file / "\n", string line) {
    lineno++;
    array(string) fields = line / ":";
    int dummy, lastchg = -1, max = -1, inactive = -1, expire = 0;
    if (sizeof (fields) < 2 || !sizeof (fields[0]) ||
	sizeof (fields) >= 8 && ( // Ignore uninteresting fields.
	  sscanf (fields[2], "%d%c", lastchg, dummy) > 1 ||
	  sscanf (fields[4], "%d%c", max, dummy) > 1 ||
	  sscanf (fields[6], "%d%c", inactive, dummy) > 1 ||
	  sscanf (fields[7], "%d%c", expire, dummy) > 1)) {
      if (sizeof (fields) != 1 || sizeof (fields[0]))
	report_error (query_name() + ": "
		      "Parse error in shadow data line " + lineno + "\n");
    }
    else spdata += ({({fields[0], fields[1], lastchg, 0, max, 0, inactive, expire})});
  }
  return spdata;
}

array(array) parse_gshadow (string file)
{
  array(array) sgdata = ({});
  int lineno = 0;
  foreach (file / "\n", string line) {
    lineno++;
    array(string) fields = line / ":";
    if (sizeof (fields) < 2 || !sizeof (fields[0])) { // Ignore uninteresting fields.
      if (sizeof (fields) != 1 || sizeof (fields[0]))
	report_error (query_name() + ": "
		      "Parse error in gshadow data line " + lineno + "\n");
    }
    else sgdata += ({({fields[0], fields[1]})});
  }
  return sgdata;
}

// FIXME: Ability to split large updates into a series of call outs.

int passwd_file_mtime, group_file_mtime, passwd_shadow_mtime, group_shadow_mtime;

void get_data (object(AC_DB) acdb)
{
  if ((<"file", "shadow">)[QUERY (method)]) {
    int reread = 0;
    array(int) stat;
    reread =
      !sizeof (QUERY (passwd_file)) || !(stat = file_stat (QUERY (passwd_file))) ||
      stat[3] != passwd_file_mtime || reread;
    passwd_file_mtime = stat && stat[3];
    reread =
      !sizeof (QUERY (group_file)) || !(stat = file_stat (QUERY (group_file))) ||
      stat[3] != group_file_mtime || reread;
    group_file_mtime = stat && stat[3];
    if (QUERY (method) == "shadow") {
      reread =
	!sizeof (QUERY (passwd_shadow)) || !(stat = file_stat (QUERY (passwd_shadow))) ||
	stat[3] != passwd_shadow_mtime || reread;
      passwd_shadow_mtime = stat && stat[3];
      reread =
	!sizeof (QUERY (group_shadow)) || !(stat = file_stat (QUERY (group_shadow))) ||
	stat[3] != group_shadow_mtime || reread;
      group_shadow_mtime = stat && stat[3];
    }
    if (!reread) {
#ifdef MODULE_DEBUG
      //report_debug (query_name() + ": No changes\n");
#endif
      return;
    }
  }

#ifdef MODULE_DEBUG
  //report_debug (query_name() + ": Updating\n");
#endif

#ifdef THREADS
  object dblock = acdb->lock();
#endif
  acdb->begin();
  object df = acdb->disable_flush();

  mapping(array:array) mbs = ([]);
  mapping(mixed:array) uids = ([]);
  mapping(mixed:array) gids = ([]);

  switch (QUERY (method)) {
    case "file":
    case "shadow":
      if (sizeof (QUERY (passwd_file))) {
	object(File) passwd = File (QUERY (passwd_file), "r");
	if (passwd->errno()) {perror ("Error opening " + QUERY (passwd_file)); return;}
	resolve_pwdata (parse_passwd (passwd->read()), mbs, uids, gids);
	if (passwd->errno()) {perror ("Error reading " + QUERY (passwd_file)); return;}
      }

      if (sizeof (QUERY (group_file))) {
	object(File) group = File (QUERY (group_file), "r");
	if (group->errno()) {perror ("Error opening " + QUERY (group_file)); return;}
	resolve_grdata (parse_group (group->read()), mbs, uids, gids);
	if (group->errno()) {perror ("Error reading " + QUERY (group_file)); return;}
      }

      if (QUERY (method) == "shadow") {
	if (sizeof (QUERY (passwd_shadow))) {
	  string spfile;
	  {
#if constant(geteuid)
	    object privs = getuid() != geteuid() &&
	      Privs ("Getting user shadow passwords");
#endif
	    object(File) shadow = File (QUERY (passwd_shadow), "r");
	    if (shadow->errno())
	      {perror ("Error opening " + QUERY (passwd_shadow)); return;}
	    spfile = shadow->read();
	    if (shadow->errno())
	      {perror ("Error reading " + QUERY (passwd_shadow)); return;}
#if constant(geteuid)
	    if (privs) destruct (privs);
#endif
	  }
	  resolve_spdata (parse_shadow (spfile), mbs, uids);
	}

	if (sizeof (QUERY (group_shadow))) {
	  string sgfile;
	  {
#if constant(geteuid)
	    object privs = getuid() != geteuid() &&
	      Privs ("Getting group shadow passwords");
#endif
	    object(File) gshadow = File (QUERY (group_shadow), "r");
	    if (gshadow->errno())
	      {perror ("Error opening " + QUERY (group_shadow)); return;}
	    sgfile = gshadow->read();
	    if (gshadow->errno())
	      {perror ("Error reading " + QUERY (group_shadow)); return;}
#if constant(geteuid)
	    if (privs) destruct (privs);
#endif
	  }
	  resolve_sgdata (parse_gshadow (sgfile), mbs, gids);
	}
      }
      break;

#ifdef __NT__
    case "NetUser":
      string srv;
      array err;
      if (sizeof (QUERY (domain)) &&
	  (err = catch (srv = dcsrv = system.NetGetAnyDCName (0, QUERY (domain)))))
	report_error (err[0]);
      else {
	array ggroups, uenum = ({}), genum = ({}), lgenum = ({});
	if (err = catch {
	  uenum = system.NetUserEnum(srv, 3, system.FILTER_NORMAL_ACCOUNT);
	  genum = system.NetGroupEnum(srv, 2);
	  lgenum = system.NetLocalGroupEnum(srv, 1);
	}) report_error (err[0]);
	update_ntusers(acdb, resolve_ntuserdata(srv, uenum),
		       (ggroups=resolve_ntgroupdata(srv, genum)),
		       resolve_ntlgroupdata(srv, lgenum, ggroups));
	acdb->commit();
      }

      return;
#else

    case "ypcat":
      if (sizeof (QUERY (passwd_ypcat_cmd)))
	resolve_pwdata (parse_passwd (Process.popen (QUERY (passwd_ypcat_cmd))),
			mbs, uids, gids);
      if (sizeof (QUERY (group_ypcat_cmd)))
	resolve_grdata (parse_group (Process.popen (QUERY (group_ypcat_cmd))),
			mbs, uids, gids);
      break;

    case "niscat":
      {
	string pwfile, grfile;
	{
	  object privs = getuid() != geteuid() && Privs ("Getting NIS+ user database");
	  if (sizeof (QUERY (passwd_niscat_cmd)))
	    pwfile = Process.popen (QUERY (passwd_niscat_cmd));
	  if (sizeof (QUERY (group_niscat_cmd)))
	    grfile = Process.popen (QUERY (group_niscat_cmd));
	  if (privs) destruct (privs);
	}
	resolve_pwdata (parse_passwd (pwfile), mbs, uids, gids);
	resolve_grdata (parse_group (grfile), mbs, uids, gids);
      }
      break;
#endif

#if constant (getpwent)
    case "getpwent":
      {
	array(array) pwdata = ({}), grdata = ({});
	{
	  object privs = getuid() != geteuid() && Privs ("Getting system user database");
	  pwdata = get_all_users();
	  for (int i = 0; i < sizeof (pwdata); i++)
	    pwdata[i] += ({pwdata[i][6]}), pwdata[i][6] = 0;
#if constant (get_all_groups)
	  grdata = get_all_groups();
#endif
	  if (privs) destruct (privs);
	}
	resolve_pwdata (pwdata, mbs, uids, gids);
	resolve_grdata (grdata, mbs, uids, gids);
      }
      break;
#endif

    default:
      report_error ("AC: User import: Unknown method " + QUERY (method) + "\n");
      return;
  }

  apply_filter (mbs, gids);
  update_acdb (acdb, mbs, uids, gids);
  acdb->commit();
}


// Module interface.

string status()
{
  string res;
  if (!lastimport) res = "<p>No successful import done yet.\n";
  else {
    res = "<p>Last successful import: " + ctime (lastimport) + "<br>\n"
      "<table border=0>\n"
      "<tr><td></td><th>Read</th><th>Imported</th></gr>\n"
      "<tr><td>Users</td>"
      "<td align=right>" + readusers + "</td>"
      "<td align=right>" + impusers + "</td></tr>\n"
      "<tr><td>Groups</td>"
      "<td align=right>" + readgroups + "</td>"
      "<td align=right>" + impgroups + "</td></tr>\n";
    if (QUERY (method) != "NetUser")
      res += "<tr><td>Passwords</td>"
	"<td align=right>" + readpws + "</td>"
	"<td align=right>" + imppws + "</td></tr>\n";
    res += "</table>\n";
#ifdef __NT__
    if (QUERY (domain))
      res += "<p>Reading domain from domain controller " + dcsrv + ".\n";
#endif
  }
  return res;
}

array register_module()
{
  return ({MODULE_SECURITY, "AC: OS user import",
	   "Imports users and groups from the OS using various methods. "
	   "The imported data is updated automatically in periodic "
	   "intervals. Reload the module to force an update.",
	   0, 0});
}

int method_is_not_file_or_shadow()
  {return QUERY (method) != "file" && QUERY (method) != "shadow";}

int method_is_not_shadow()
  {return QUERY (method) != "shadow";}

int method_is_not_ypcat()
  {return QUERY (method) != "ypcat";}

int method_is_not_niscat()
  {return QUERY (method) != "niscat";}

int method_is_not_NetUser()
  {return QUERY (method) != "NetUser";}

int method_is_NetUser()
  {return QUERY (method) == "NetUser";}

void create()
{
  defvar ("method",
#ifdef __NT__
	  "NetUser",
#else
#if constant (getpwent) && constant (getgrent)
	  "getpwent",
#else
	  "file",
#endif
#endif
	  "Import method",
	  TYPE_STRING_LIST,
	  "The method used to get the user and group info from the OS:\n"
	  "<p><dl>\n"
	  "<dt><em>file</em><dd>Read users, groups and passwords from "
	  "<code>/etc/passwd</code>, <code>/etc/group</code> style files.\n"
	  "<dt><em>shadow</em><dd>As <em>file</em>, but passwords are also "
	  "read from <code>/etc/shadow</code>, <code>/etc/gshadow</code> "
	  "style files.\n"
#ifdef __NT__
	  "<dt><em>NetUser</em><dd>Read users and group info from Windows NT "
	  "account database. The user account running Roxen needs to have "
	  "the \"Act as part of the operating system\" privilege to be able "
	  "to check passwords.\n"
#else
	  "<dt><em>ypcat</em><dd>As <em>file</em>, but <code>ypcat</code> "
	  "is used in nonprivileged mode to get the files.\n"
	  "<dt><em>niscat</em><dd>As <em>file</em>, but <code>niscat</code> "
	  "is used in privileged mode to get the files.\n"
#endif
#if constant (getpwent)
#if constant (getgrent)
	  "<dt><em>getpwent</em><dd>Read users, groups and passwords using "
	  "the <code>getpwent</code> and <code>getgrent</code> system calls.\n"
#else
	  "<dt><em>getpwent</em><dd>Read users and passwords using the "
	  "<code>getpwent</code> system call.\n"
#endif
#endif
	  "</dl>\n"
	  "<p>For the <em>file</em> and <em>shadow</em> methods, updates "
	  "will only take place when the files has changed.",
	  ({"file", "shadow", 
#ifdef __NT__
	    "NetUser",
#else
	    "ypcat", "niscat",
#endif
#if constant (getpwent)
	    "getpwent"
#endif
	  }));
  defvar ("passwd_file", "/etc/passwd", "User database file",
	  TYPE_FILE,
	  "The file containing a <code>/etc/passwd</code> style user "
	  "database. "
	  "Nothing is imported if left empty.",
	  0, method_is_not_file_or_shadow);
  defvar ("group_file", "/etc/group", "Group database file",
	  TYPE_FILE,
	  "The file containing a <code>/etc/group</code> style group "
	  "database. "
	  "Nothing is imported if left empty.",
	  0, method_is_not_file_or_shadow);
  defvar ("passwd_shadow", "/etc/shadow", "User shadow password file",
	  TYPE_FILE,
	  "The file containing a <code>/etc/shadow</code> style user "
	  "password database. Expire dates are checked, but not stored. "
	  "So if a user is imported valid, it will remain valid until "
	  "the next import. "
	  "Nothing is imported if left empty.",
	  0, method_is_not_shadow);
  defvar ("group_shadow", "/etc/gshadow", "Group shadow password file",
	  TYPE_FILE,
	  "The file containing a <code>/etc/gshadow</code> style group "
	  "password database. Currently, neither group memberships nor "
	  "group administrators are imported from this file. "
	  "Nothing is imported if left empty.",
	  0, method_is_not_shadow);
  defvar ("domain", "", "Domain",
	  TYPE_STRING,
	  "The domain to read accounts from. The user account running Roxen "
	  "needs to be able to browse accounts in this domain, which means "
	  "it needs to be a member of the \"Administrators\" or \"Account "
	  "operators\" groups in the specified domain. Leave blank to use "
	  "local database.",
	  0, method_is_not_NetUser);
  defvar ("passwd_ypcat_cmd", "ypcat passwd", "Command to read the user database",
	  TYPE_STRING,
	  "This command is run without special privileges. It should "
	  "produce a <code>/etc/passwd</code> style user database on "
	  "standard output. "
	  "Nothing is imported if left empty.",
	  0, method_is_not_ypcat);
  defvar ("group_ypcat_cmd", "ypcat group", "Command to read the group database",
	  TYPE_STRING,
	  "This command is run without special privileges. It should "
	  "produce a <code>/etc/group</code> style group database on "
	  "standard output. "
	  "Nothing is imported if left empty.",
	  0, method_is_not_ypcat);
  defvar ("passwd_niscat_cmd", "niscat passwd.org_dir",
	  "Command to read the user database",
	  TYPE_STRING,
	  "This command is run with special privileges. It should "
	  "produce a <code>/etc/passwd</code> style user database on "
	  "standard output. "
	  "Nothing is imported if left empty.",
	  0, method_is_not_niscat);
  defvar ("group_niscat_cmd", "niscat groups.org_dir",
	  "Command to read the group database",
	  TYPE_STRING,
	  "This command is run with special privileges. It should "
	  "produce a <code>/etc/group</code> style group database on "
	  "standard output. "
	  "Nothing is imported if left empty.",
	  0, method_is_not_niscat);

  defvar ("importgroups",
#ifdef __NT__
	  "*",
#else
	  "[100..59999]",
#endif
	  "Groups to import",
	  TYPE_STRING,
	  "This is a pattern that specifies which groups to import.\n"
	  "\n<p>It consists of fields separated by '<code>,</code>'. Each "
	  "field is either a group name or a numeric range of group ids. "
	  "The glob characters '<code>?</code>' and '<code>*</code>' may be "
	  "used in group names. A numeric range of GIDs is written "
	  "<code>[<i>gid</i>]</code> or <code>[<i>from</i>..<i>to</i>]</code>, "
	  "where either <i>from</i> or <i>to</i> may be left out. The list "
	  "of fields is searched from left to right and the first field that "
	  "matches the group in question determines whether it should be "
	  "included or not. It's not included if the field is prefixed by "
	  "'<code>^</code>', otherwise it is. Whitespace is ignored between "
	  "fields.");
  defvar ("importusers",
#ifdef __NT__
	  "*",
#else
	  "[100..59999]",
#endif
	  "Users to import",
	  TYPE_STRING,
	  "This is a pattern that specifies which users to import. "
	  "In addition to the syntax described in <em>Groups to import</em>, "
	  "any field may be prefixed by '<code>@</code>'. In that case, "
	  "the field is matched against the groups the user in question "
	  "is a member of. For example, to import all users belonging to the "
	  "group 'local', use '@local' here.");
  defvar ("disable_crypt", 0, "Disable passwords",
	  TYPE_FLAG | VAR_MORE,
	  "The imported passwords will not be active for authentication "
	  "inside AC.", 0, method_is_NetUser);
  defvar ("disable_uid_gid", 0, "Disable UIDs and GIDs",
	  TYPE_FLAG | VAR_MORE,
	  "The imported UIDs and GIDs will not be active for authentication "
	  "inside AC.");
  defvar ("sethandle", "Users", "Set usernames",
	  TYPE_MULTIPLE_STRING | VAR_MORE,
	  "Whether to set the username from the imported data. You may not "
	  "want this, because the username is unique in the AC database, "
	  "so you might not want to have that namespace \"polluted\". "
	  "This will also have the effect that authentication won't work "
	  "right away - usernames have to be set explicitly first.",
	  ({"No", "Users", "Groups", "Both"}));
  defvar ("collapse_groups", 1, "Collapse user groups",
	  TYPE_FLAG | VAR_MORE,
	  "Whether to join groups with users for the common special case "
	  "where every user has an own group: A group is joined with a "
	  "user if they have the same username, the user is the only "
	  "member of the group, and the group does not have a valid "
	  "password.", 0, method_is_NetUser);
// Idea:
//   defvar ("basegroup", 0, "Base group for imports",
// 	  TYPE_FOO,
// 	  "All imported groups and users will become members of this group "
// 	  "if specified. Normally the importer will attach the imported "
// 	  "information to existing identities with the same usernames. "
// 	  "Setting this group will limit this to identities which are "
// 	  "members (directly or indirectly) to this group.");
  defvar ("strip_gecos", 1, "Strip finger info from the GECOS (fullname) field",
	  TYPE_FLAG | VAR_MORE,
	  "Strip everyting after the first ',' character from the GECOS "
	  "field of the user database.", 0, method_is_NetUser);
  defvar ("update", "60", "Update period or time",
	  TYPE_STRING,
	  "If this is a single integer, it specifies the interval in minutes "
	  "between periodic updates of the imported information. If it's "
	  "a time value on the form <tt>hh:mm</tt>, it specifies the time "
	  "of day to do the update. 0 disables periodic updates.");
  defvar ("update_on_load", 0, "Update immediately on load",
	  TYPE_FLAG | VAR_MORE,
	  "Set this if you want the database to be updated when the "
	  "server is started. If the database is large, this causes "
	  "an extra delay till the server is ready to handle connections.");
}

string check_variable (string var, mixed val)
{
  if (var == "update")
    if (sscanf (val, "%*d%*c") != 1 && sscanf (val, "%*d:%*d%*c") != 2)
      return "Invalid format";
  return 0;
}

string query_name()
{
  string res = "AC: OS user import: ";
  string m = QUERY (method), f, g;

  switch (m) {
    case "file": case "shadow": f = "passwd_file", g = "group_file"; break;
#ifdef __NT__
    case "NetUser": return res + "NetUser";
#else
    case "ypcat": m = 0; f = "passwd_ypcat_cmd", g = "group_ypcat_cmd"; break;
    case "niscat": m = 0; f = "passwd_niscat_cmd", g = "group_niscat_cmd"; break;
#endif
#if constant (getpwent)
    case "getpwent": return res + "getpwent";
#endif
    default: return res + "unimplemented method " + m;
  }

  if (m) res += m + " ";
  if (sizeof (query (f))) res += query (f);
  else if (sizeof (query (g))) res += query (g) + " (groups only)";
  else if (m == "shadow" && sizeof (QUERY (passwd_shadow)))
    res += QUERY (passwd_shadow) + " (passwords only)";
  else if (m == "shadow" && sizeof (QUERY (group_shadow)))
    res += QUERY (group_shadow) + " (group passwords only)";
  else res += "(no source)";

  return res;
}

static mixed periodic_update_callout;

void periodic_update (void|int skip)
{
  if (!skip) {
    object(AC_DB) acdb = online_acdb();
    if (acdb) get_data (acdb);
  }

  remove_call_out (periodic_update_callout);
  if (QUERY (update)) {
    int m;
    if (sscanf (QUERY (update), "%d%*c", m) == 1)
      periodic_update_callout = call_out (periodic_update, m * 60);

    else {
      int h;
      sscanf (QUERY (update), "%d:%d", h, m);

      // Get the local time as a unix time integer.
      mapping(string:int) t = localtime (time());
      t->timezone = 0;
      t->hour += t->isdst;

      m = ((h * 60 + m) * 60 - mktime (t)) % (24*60*60);
      if (m <= 1) m += 24*60*60; // Just to be safe.
      periodic_update_callout = call_out (periodic_update, m);
    }
  }
}

void start (int when, object conf)
{
  if (intp (QUERY (update))) set ("update", (string) QUERY (update));

  ::start (when, conf);

  passwd_file_mtime = group_file_mtime = passwd_shadow_mtime = group_shadow_mtime = 0;

  string key = get_modname (this_object()) + "_loaded";
  object(AC_DB) acdb = online_acdb();
  if (acdb && (QUERY (update_on_load) || roxen->query_var (key))) get_data (acdb);
  roxen->set_var (key, 1); // To detect reloads.

  periodic_update (1);
}


// Auth interface.

// This is the authentication test function used with AUTH_NT.
#if constant (system.LogonUser)
int test_nt_passwd (object(IdAuth) ia, string user, array args, string passwd)
{
  return !!system.LogonUser (user, args[0], passwd, @args[1..]);
}
#else
private static int has_warned_no_nt = 0;
int test_nt_passwd (object(IdAuth) ia, string user, array args, string passwd)
{
  if (!has_warned_no_nt) {
    report_warning ("AC: A user with NT system authentication "
		    "couldn't be authenticated since the server "
		    "isn't running on NT.\n");
    has_warned_no_nt = 1;
  };
  return 0;
}
#endif

array(int) auth_types = ({AUTH_PASSWD_CRYPT, AUTH_FS_UID, AUTH_FS_GID, AUTH_NT});

void id_update_userinfo (object(IdAuth) ia)
{
  object(Identity) id = ia->identity();

  string handle = ia->data[DATA_HANDLE];
  if (QUERY (sethandle) == "No" ||
      QUERY (sethandle) == (id->type() != ID_TYPE_USER ? "Users" : "Groups"))
    handle = 0;
  else if (handle) {
    object(Identity) other = id->table()->find_by_handle (handle);
    if (other && other != id) handle = 0;
  }
  if (id->handle() != handle) id->set_handle (handle);

  string fullname = ia->data[DATA_FULLNAME] || ia->data[DATA_HANDLE];
  if (id->name() != fullname) id->set_name (fullname);
}

void id_handle_changed (object(IdAuth) ia)
// Awkward name in this case..
{
  string handle = ia->identity()->handle();
  string cryptwd = ia->data[DATA_CRYPTWD];
  mixed uid = ia->data[DATA_UID], gid = ia->data[DATA_GID];

  int working = 0;

  if (QUERY (disable_uid_gid) || !intp (uid)) {
    if (ia->has_auth_data (AUTH_FS_UID)) ia->auth_remove (AUTH_FS_UID);
  }
  else {
    if (!ia->has_auth_data (AUTH_FS_UID) || ia->auth_handle (AUTH_FS_UID) != uid)
      ia->set_auth_data (AUTH_FS_UID, uid, 0);
    working = 1;
  }

  if (QUERY (disable_uid_gid) || !intp (gid)) {
    if (ia->has_auth_data (AUTH_FS_GID)) ia->auth_remove (AUTH_FS_GID);
  }
  else {
    if (!ia->has_auth_data (AUTH_FS_GID) || ia->auth_handle (AUTH_FS_GID) != gid)
      ia->set_auth_data (AUTH_FS_GID, gid, 0);
    working = 1;
  }

  if (QUERY (disable_crypt) || !handle || !CRYPTWD_CHECK (cryptwd)) {
    if (ia->has_auth_data (AUTH_PASSWD_CRYPT)) ia->auth_remove (AUTH_PASSWD_CRYPT);
  }
  else {
    if (!ia->has_auth_data (AUTH_PASSWD_CRYPT) ||
	ia->auth_handle (AUTH_PASSWD_CRYPT) != handle ||
	ia->auth_secret (AUTH_PASSWD_CRYPT) != cryptwd)
      ia->set_auth_data (AUTH_PASSWD_CRYPT, handle, cryptwd);
    working = 1;
  }

  if (QUERY (method) != "NetUser") {
    if (ia->has_auth_data (AUTH_NT)) ia->auth_remove (AUTH_NT);
  }
  else {
#ifdef __NT__
    ia->set_auth_data (AUTH_NT, handle,
		       ({ sizeof (QUERY (domain)) && QUERY (domain),
			  system.LOGON32_LOGON_NETWORK,
			  system.LOGON32_PROVIDER_DEFAULT }),
		       "test_nt_passwd");
    working = 1;
#endif
  }

  if (working != ia->working()) ia->set_working (working);
}

string id_auth_view (object(IdAuth) ia, mapping(string:string) var, int access)
{
  string cryptwd = ia->data[DATA_CRYPTWD];
  string res = "Imported info: "
    "Username " + (ia->data[DATA_HANDLE] ?
		   "\"" + html_encode_string (ia->data[DATA_HANDLE]) + "\"" :
		   "none") +
    ", fullname " + (ia->data[DATA_FULLNAME] ?
		     "\"" + html_encode_string (ia->data[DATA_FULLNAME]) + "\"" :
		     "none");
  if (intp (ia->data[DATA_UID])) res += ", UID " + ia->data[DATA_UID];
  if (intp (ia->data[DATA_GID])) res += ", GID " + ia->data[DATA_GID];
  if (ia->data[DATA_HOMEDIR])
    res += ", homedir <code>" + html_encode_string (ia->data[DATA_HOMEDIR]) + "</code>";
  if (sizeof (ia->data) > DATA_SHELL && stringp (ia->data[DATA_SHELL]))
    res += ", shell <code>" + html_encode_string (ia->data[DATA_SHELL]) + "</code>";
  if (sizeof(ia->data) > DATA_COMMENT && stringp(ia->data[DATA_COMMENT]))
    res += ", comment: <i>"+html_encode_string (ia->data[DATA_COMMENT])+"</i>";
  res += "<br>\n";
  if (stringp(cryptwd))
    res += "Password: \"<code>" + html_encode_string (cryptwd) + "</code>\"" +
      (!CRYPTWD_CHECK (cryptwd) ? " (invalid)" : "") + "<br>\n";

  if (intp (ia->data[DATA_GID]) && !ia->has_auth_data (AUTH_FS_GID))
    res = "Authentication with GID " +
      (QUERY (disable_uid_gid) ? "disabled" : "not possible") + ".<br>\n" + res;
  if (intp (ia->data[DATA_UID]) && !ia->has_auth_data (AUTH_FS_UID))
    res = "Authentication with UID " +
      (QUERY (disable_uid_gid) ? "disabled" : "not possible") + ".<br>\n" + res;
  if (QUERY (method) != "NetUser" && !ia->has_auth_data (AUTH_PASSWD_CRYPT))
    res = "Authentication with password " +
      (QUERY (disable_crypt) ? "disabled" : "not possible") + ".<br>\n" + res;

  return res;
}

array(string) id_get_userinfo (object(IdAuth) ia)
{
  array(string) info = ia->data[..DATA_HOMEDIR] + ({ia->data[DATA_SHELL]});
  info[DATA_UID] = stringp (info[DATA_UID]) ? 0 : (string) info[DATA_UID];
  info[DATA_GID] = stringp (info[DATA_GID]) ? 0 : (string) info[DATA_GID];
  return info;
}
