// This is a Roxen module.

string cvs_version = "$Id: acintern_passwd.pike,v 1.8 1999/06/29 23:12:17 mast Exp $";

#include <module.h>
inherit "acauthmodule";
inherit "roxenlib";

import Array;


// Module interface.

string status()
{
  return 0;
}

array register_module()
{
  return ({MODULE_SECURITY, "AC: Internal password",
	   "Lets you set passwords on users and groups from the "
	   "administrator interface. These passwords are entirely "
	   "internal in the Access Control system.",
	   0, 1});
}


// Auth interface.

constant auth_types = ({AUTH_PASSWD_CRYPT});
constant replicable = 1;

object(IdAuth) id_auth_init (object(IdAuth) ia)
{
  ia->data = ({0});
  ia->set_dirty();
  return ia;
}

int id_auth_destroy (object(IdAuth) ia) {return 1;}

void id_handle_changed (object(IdAuth) ia)
{
  string handle = ia->identity()->handle();
  string cryptwd = ia->data && ia->data[0];
  int works = !!handle && !!cryptwd && sizeof (cryptwd);
  if (works) ia->set_auth_data (AUTH_PASSWD_CRYPT, handle, cryptwd);
  else ia->auth_remove (AUTH_PASSWD_CRYPT);
  ia->set_working (works);
}

void id_auth_set_cryptwd (object(IdAuth) ia, string cryptwd)
// This is a kludgy hook used by ModuleAC::id_create().
{
  ia->data[0] = cryptwd;
  ia->set_dirty();
  id_handle_changed (ia);
}

#define ERRMSG_BEG "<font color=red><strong>"
#define ERRMSG_END "</strong></font>"

string id_auth_view (object(IdAuth) ia, mapping(string:string) variables, int access)
{
  if (!ia->data)
    // Compatibility kludge.
    ia->data = ({ia->auth_secret (AUTH_PASSWD_CRYPT)}), ia->set_dirty();

  string handle = ia->identity()->handle();
  string cryptwd = ia->data[0] || "";
  string res = "";

  if (access != PERM_WRITE) {
    res = "Current password: \"" + html_encode_string (cryptwd) + "\"<br>\n";
  }
  else {
    string
      passwd1 = variables["passwd1"] ? variables["passwd1"] : "",
      passwd2 = variables["passwd2"] ? variables["passwd2"] : "";
    string newcryptwd = variables["cryptwd"];

    if (sizeof (passwd1) || sizeof (passwd2)) {
      if (passwd1 != passwd2) {
	res += ERRMSG_BEG "The two passwords doesn't match - nothing changed"
	  ERRMSG_END "<br>\n";
	newcryptwd = 0;
      }
      else newcryptwd = crypt (passwd1); // FIXME: Use platform indep hash.
    }

    if (newcryptwd) id_auth_set_cryptwd (ia, cryptwd = newcryptwd);

    res += "Password: <input type=password name=passwd1 size=20> "
      "Again: <input type=password name=passwd2 size=20><br>\n"
      "Encrypted: <input name=cryptwd value=" +
      html_encode_tag_value (cryptwd) + " size=20> "
      "Use this to change the encrypted password directly.<br>\n";
  }

  if (!sizeof (cryptwd))
    res = "Authentication not possible because the password is not set.<br>\n" + res;

  if (!handle)
    res = "Authentication not possible because the username is not set.<br>\n" + res;

  return res;
}

array(string) id_get_userinfo (object(IdAuth) ia)
{
  return ({0, ia->data[0], 0, 0, 0, 0, 0});
}
