#!NOMODULE
// Access control database library with a simple text file implementation.

// $Id: aclib.pike,v 1.121 1999/09/03 01:25:53 mast Exp $

// You're probably most interested in ModuleAC.

#if constant (roxen)
#include <config.h>
#include <module.h>
#else
#define report_error werror
#define report_warning werror
#define report_notice werror
#define report_debug werror
#endif


// Some common tools. Maybe these should be moved to a general tool
// library (eg roxenlib).

class ReadableDecode
// Reading and writing data in a readable text file format.
{
  import Array;

  string encode (mixed data)
  // The returned string is in pike syntax and contains no \n or \r.
  {
    if (intp (data) || floatp (data)) return (string) data;
    else if (stringp (data))
      return "\"" +
	replace (data, ({"\"", "\0", "\b", "\t", "\n", "\r", "\\"}),
		 ({"\\\"", "\\0", "\\b", "\\t", "\\n", "\\r", "\\\\"})) + "\"";
    else if (arrayp (data))
      return "({" + map (data, encode) * ", " + "})";
    else if (mappingp (data))
      return "([" + map (indices (data), lambda (mixed i, mapping data) {
					   return encode (i) + ":" + encode (data[i]);
					 }, data) * ", " + "])";
    else if (multisetp (data))
      return "(<" + map (indices (data), encode) * ", " + ">)";
    else {
      report_error ("ReadableDecode: Can't save a function or object\n");
      return "0";
    }
  }

  static object(Regexp) start = Regexp
    ("^(([-+0-9][-+0-9a-fx.]*)"
     "|\"(([^\"\\\\]|\\\\.)*)\"|([_A-Za-z---][_0-9A-Za-z---]*)"
     "|\\([ \t\r\n]*(.))(.*)$");
  static object(Regexp) end = Regexp ("^(.[ \t\r\n]*\\))(.*)$");

  array decode (string str)
  // Returns an array containing the unread part of the string and the
  // value. If there was a parse error, the returned string begins at
  // that point and no value is returned. Not a very efficient
  // implementation if str is long.
  {
    sscanf (str, "%*[ \t\r\n]%s", str);
    string *tok = start->split (str);
    if (!tok) return ({str});
    else if (tok[1]) {
      int i;
      float f;
      string e;
      if (sscanf (tok[1], "%D%s", i, e) == 2 && e == "")
	return ({tok[6], i});
      else if (sscanf (tok[1], "%f%s", f, e) == 2 && e == "")
	return ({tok[6], f});
      else return ({tok[0] + tok[6]});
    }
    else if (tok[2])
      return ({tok[6], replace (tok[2],
				({"\\\"", "\\0", "\\b", "\\t", "\\n", "\\r", "\\\\"}),
				({"\"", "\0", "\b", "\t", "\n", "\r", "\\"}))});
    else if (tok[4])
      return ({tok[6], tok[4]});
    else if (tok[5]) {
      int t = (['{':0, '[':1, '<':2])[tok[5][0]];
      if (zero_type (t) == 1) return ({tok[5] + tok[6]});
      array|mapping|multiset d =
	t == 0 ? ({}) :
	t == 1 ? ([]) :
	(<>);
      str = tok[6];

      while (sscanf (str, "%*[ \t\r\n]%s", str),
	     !(tok = end->split (str))) {
	array r = decode (str);
	if (sizeof (r) == 1) return r;

	switch (t) {
	  case 0: d += ({r[1]}); break;
	  case 1:
	    mixed i = r[1];
	    sscanf (r[0], "%*[ \t\r\n]%s", str);
	    if (sizeof (str) == 0 || str[0] != ':') return ({str});
	    r = decode (str[1..]);
	    if (sizeof (r) == 1) return r;
	    d[i] = r[1];
	    break;
	  default:
	    d[r[1]] = 1;
	}

	sscanf (r[0], "%*[ \t\r\n]%s", str);
	if (sizeof (str) >= 1 && str[0] == ',')
	  str = str[1..];
	else {
	  tok = end->split (str);
	  break;
	}
      }

      if (!tok) return ({str});
      else if (tok[0][0] != ({'}', ']', '>'})[t]) return ({tok[0] + tok[1]});
      else return ({tok[1], d});
    }
    else return ({tok[0] + tok[6]});
  }
}


// A simple table of records implementation.

// A table is indexed by integers. A valid index is always nonzero.

class Table
{
  static object _db;
  static mapping(int:object) _table = ([]);
  static int _unused_id = 1;
  static program _rec_prg;

  constant ac_table = "ac_table";
  constant name = 0;

  multiset(int) dirty = (<>);
  // IDs of records in which the persistent info has changed (includes
  // IDs of new and deleted records).

  void create (object db, program p) {_db = db; _rec_prg = p;}
  object db() {return _db;}
  object new (mixed ... args) {return _rec_prg (this_object(), @args);}

  object `[](int i) {return _table[i];}
  object *this_values() {return values (_table);} // Proper name scoping would be nice.
  int *this_indices() {return indices (_table);}
  int this_sizeof() {return sizeof (_table);}

  int add (object rec, void|int id)
  {
    int i = id || _unused_id;
#ifdef DEBUG
    int si;
#endif
    if (zero_type (_table[i]) != 1)
      report_error ("Table " + name + ": Record ID " + i + " already in use\n");
    else if (i == 0)
      report_error ("Table: Record ID 0 is invalid\n");
#ifdef DEBUG
    else if (zero_type (si = search (_table, rec)) != 1)
      throw (({"Table " + name + ": Record ID " + si + " already in table\n",
	       backtrace()}));
#endif
    else {
      rec->really_set_id_for_this_record (i);
      if (i >= _unused_id) _unused_id = i + 1;
      _table[i] = rec;
      dirty[i] = 1;
      return 1;
    }
    return 0;
  }

  int remove (object rec, void|function preremove_fun, void|mixed ... args)
  {
    if (zero_type (_table[rec->id()]) == 1) return 0;
#ifdef DEBUG
    if (_table[rec->id()] != rec)
      throw (({"Table " + name + ": Not same record ID " + rec->id() +
	       " in table for removal\n", backtrace()}));
#endif

    if (preremove_fun) preremove_fun (@args);

    int i = rec->id();
    m_delete (_table, i);
    dirty[i] = 1;
    return 1;
  }

  void is_added (object tbl, object rec)
  // Should be called when a record has been added to a table that
  // records in this table may want to refer to.
  {
    values (_table)->is_added (tbl, rec);
  }

  void is_removed (object tbl, object rec)
  // Should be called when a record is about to be removed from a
  // table that records in this table may refer to.
  {
    values (_table)->is_removed (tbl, rec);
  }

  static int _flush_disabled = 0;

  void flush_caches()
  {
    if (!_flush_disabled) values (_table)->flush_caches();
  }

  object disable_flush()
  {
    _flush_disabled++;
    return class {
      void destroy () {_flush_disabled--, flush_caches();}
    }();
  }

  int check_consistency()
  {
    int r = 1;
    foreach (indices (_table), int i) {
      if (i != _table[i]->id()) {
	report_error ("Table " + name + ": ID mismatch - " + i + " in table, " +
		      _table[i]->id() + " in record\n");
	r = 0;
      }
      if (i >= _unused_id) {
	report_error ("Table " + name + ": Unused ID " + i + ">=" + _unused_id +
		      " in use\n");
	r = 0;
      }
      if (!_table[i]->check_consistency (this_object())) r = 0;
    }
    return r;
  }

  mapping(int:array) dump_values (void|array ids)
  {
    if (ids) {
      array vals = allocate (sizeof (ids));
      for (int i = 0; i < sizeof (ids); i++)
	if (_table[ids[i]]) vals[i] = _table[ids[i]]->dump_values();
      return mkmapping (ids, vals);
    }
    else
      return mkmapping (indices (_table), values (_table)->dump_values());
  }

  void assign_values (mapping(int:array) val)
  {
    foreach (indices (val), int i)
      if (_table[i]) {
	if (!val[i]) remove (_table[i]);
	else if (!_table[i]->assign_values (val[i], 1))
	  report_error ("Table " + name + ": Couldn't decode value "
			"for record " + i + "\n");
      }
      else if (val[i]) {
	object r = _rec_prg (this_object());
	// Do not use new above; we want a really "clean" object.
	if (!r->assign_values (val[i]))
	  report_error ("Table " + name + ": Couldn't decode value "
			"for record " + i + "\n");
	array err = catch (add (r, i));
	if (err) report_error ("Table " + name + ": Couldn't add new record " + i +
			       ": " + err[0]);
	r->set_dirty();
      }
  }

  void write_readable (object file)
  {
    foreach (sort (indices (_table)), int i) {
      // The sort is not necessary, but it's nicer if the lines don't
      // jump around all the time. If it's a performance problem it
      // may be removed.
      object r = _table[i];
      file->write (r->encode_readable());
      file->write ("\n");
    }
  }

  int|string read_readable (object file, void|int line)
  // Returns the line number of the last line read if successful, an
  // error message string otherwise.
  {
    string l;
    int m = 0;
    for (line++; (l = file->gets()) && l != ""; line++) {
      object r = _rec_prg (this_object());
      // Do not use new above; we want a really "clean" object.
      int i = r->decode_readable (l);
      if (!i) return name + ":" + line + ": Parse error\n";
      if (i > m) m = i;
      array err = catch {add (r, i);};
      if (err) return name + ":" + line + ": " + err[0];
      r->set_dirty();
    }
    _unused_id = m + 1;
    dirty = (<>);
    return line;
  }
}

class Record
{
  static object _table;
  static int _id = 0;

  constant ac_record = "ac_record";

  void create (object tbl) {_table = tbl;}

  object table() {return _table;}
  object db() {return _table->db();}
  int id() {return _id;}	// Atomic access assumed in ModuleAC.

  int set_dirty()
  {
    if (_table && _id) _table->dirty[_id] = 1;
  }

  void really_set_id_for_this_record (int i)
  {
#ifdef DEBUG
    if (_id != 0) report_error ("Record: ID " + _id + " in " + (_table && _table->name) +
				" already has an ID set\n");
#endif
    _id = i;
  }

  void is_added (object tbl, object rec) {}
  // A record was added to the specified table.

  void is_removed (object tbl, object rec) {}
  // A record is about to be removed from the specified table.

  void flush_caches() {}	// Should be overidden by classes that do caching.

  int check_consistency (object tbl)
  {
    if (db() != tbl->db()) {
      report_error ("Record: ID " + id() + " in " + (_table && _table->name) +
		    " is in different db from its table\n");
      return 0;
    }
    return 1;
  }
}

class RecordPtr
{
  static object _table = 0;
  static int _id = 0;
  static object _rec = 0;

  int id() {return _id;}

  object record()
  {
    if (!_rec && _id) {
      _rec = _table[_id];
      if (!_rec) _id = 0;
    }
    return _rec;
  }

  int set (int|object rec, void|object tbl)
  {
    int r = 1;
    if (objectp (rec)) {
      _table = rec->table();
#ifdef DEBUG
      if (objectp (tbl) && _table != tbl) {
	report_error ("RecordPtr: Record ID " + rec->id() +
		      " does not belong to explicitly specified table\n");
	r = 0;
      }
      else if (_table && _table[rec->id()] != rec) {
	report_error ("RecordPtr: Record ID " + rec->id() +
		      " is not a member of the table for this pointer\n");
	r = 0;
      }
#endif
      _id = rec->id(), _rec = rec;
    }
    else {
      if (objectp (tbl)) _table = tbl;
#ifdef DEBUG
      if (!_table) {
	report_error ("RecordPtr: Setting record ID " + rec + " in unknown table\n");
	r = 0;
      }
#endif
      _id = rec, _rec = 0;
    }
    return r;
  }

  void create (object tbl, void|int|object rec)
  {
    set (rec, tbl);
  }
}


// Global constants.

// These should be accessible from both aclib and ModuleAC.

#define GLOB_CONST \
  /* Standard auth data types. See IdAuth._auth_data. */ \
  /*       Auth data type	(H)andle, (S)ecret, (T)est function extra args	*/ \
  \
  constant AUTH_PASSWD_CRYPT=0;	/* H: string username				*/ \
				/* S: string crypt()'ed password		*/ \
				/* T: not used					*/ \
  \
  constant AUTH_FS_UID = 1;	/* H: int filesystem uid			*/ \
				/* S: 0 (not used)				*/ \
				/* T: not used					*/ \
  \
  constant AUTH_FS_GID = 2;	/* H: int filesystem gid			*/ \
				/* S: 0 (not used)				*/ \
				/* T: not used					*/ \
  \
  constant AUTH_NT = 3;		/* H: string username				*/ \
				/* S: array(string|int) containing:		*/ \
				/*    0: string domain or 0			*/ \
				/*    1: optional int logon type		*/ \
				/*    2: optional int logon provider		*/ \
				/* T: string plaintext password			*/ \
  \
  constant AUTH_TYPES_COUNT = 4; \
  \
  /* Types of identities. */ \
  constant ID_TYPE_N_A = -1;	/* Values <= this one are never valid types. */ \
  constant ID_TYPE_USER = 0; \
  constant ID_TYPE_GROUP = 1; \
  \
  constant ID_TYPE_TEXTS = ({"user", "group"}); \
  \
  /* Permissions. */ \
  constant PERM_N_A = -1;	/* Values <= this one are never valid permissions. */ \
  constant PERM_NONE = 0; \
  constant PERM_READ = 1; \
  constant PERM_WRITE = 2;	/* Implies PERM_READ. */ \
  \
  constant PERM_TEXTS = ({"n", "r", "w"}); \
  constant PERM_DESC = ({"no", "read", "write"});

GLOB_CONST;


// Table records.

// General tip: Only change a value in a record if it's certain that
// the value is different; there are all sorts of caches that
// shouldn't be flushed unnecessarily.

#ifdef AC_PARSE_DEBUG
#  define RETURN_ERR(tbl, id) \
     throw (({"Couldn't parse " + tbl + " record " + id +"\n", backtrace()}))
#else
#  define RETURN_ERR(tbl, id) return 0
#endif

class IdAuth
// Record that binds authentication data to an identity for a certain
// authentication module.
{
  static string _auth_modname = 0; // Name of the owning auth module.
  static inherit RecordPtr: _identity; // The identity this record applies to.

  inherit Record;

  static int _disabled = 0;
  // Nonzero means this authentication record is explicitly disabled
  // by the user. The module may update the auth data, but it will
  // never be used for actual authentification.

  static int _working = 0;
  // This flag indicates that the record contains data that enables
  // the auth module to authenticate the identity. In other words,
  // this field should be 0 when the data inhibits the identity (e.g.
  // when there's a "*" in a password field) and 1 otherwise (e.g.
  // when a password field is empty or contains a valid encrypted
  // password).

  static mapping (int:array) _auth_data = ([]);
  // Contains standardized authentication data, i.e. things like
  // passwords that may be collected by some modules and used for
  // actual authentication by some other.
  //
  // The indices are integers taken from the set of AUTH_* constants
  // and describes the type of auth data. The values are tuples where
  // the first element is some kind of identity-unique handle, e.g. an
  // account name, and the second is the authentication "secret"
  // (which hopefully isn't sensitive enough that it actually has to
  // be kept secret), e.g. the encrypted account password. Exactly
  // what the handles and secrets are is specified by the auth data
  // type.
  //
  // The third argument, if any, is the name of a function to call in
  // the module to do the actual authentication test if the auth data
  // type requires that. It will receive at least three arguments: The
  // first is this IdAuth object, the second and third are the handle
  // and secret in this tuple. It will normally receive extra
  // arguments with the actual credentials to be tested, e.g. a
  // plaintext password. Exactly what these are is specified by the
  // auth data type.
  //
  // FIXME: Currently, the case where several IdAuth objects contains
  // the same handle and secret is not handled cleanly.

  static multiset(int) _group_ids = (<>);
  // The group memberships for the identity that's been set by the
  // module. The IDs of those groups are stored here.

  array data = 0;
  // Data used internally by the authentication module, e.g.
  // non-standard auth data. The array may not contain objects,
  // functions or programs. set_dirty() should always be called
  // whenever this is changed.

  static object _auth_mod = 0;	// Cache for the auth module object.
  static multiset(object) _groups = 0; // Cache for memberships.

  void create (object tbl, object auth_mod, object id)
  {
    Record::create (tbl);
    if (tbl) {
      _auth_modname = db()->modulename (auth_mod);
      _auth_mod = auth_mod;
      if (id && id->type() == ID_TYPE_GROUP)
	throw (({"Groups can not have auth methods\n", backtrace()}));
      _identity::create (db()->identities, id);
      if (id) id->flush_caches();
    }
    else {
      // We're creating a faked IdAuth. The id is probably fake too.
      _identity::create (0, id);
    }
  }

  int set_dirty()
  {
    object id = _identity::record();
    if (id) id->check_userinfo();
    Record::set_dirty();
  }

  string auth_modname() {return _auth_modname;}

  object auth_module()
  {
    if (!_auth_mod) _auth_mod = db()->findmodule (_auth_modname);
    return _auth_mod;
  }

  int set_auth_module (object mod)
  {
    _auth_modname = db()->modulename (mod);
    _auth_mod = mod;
    set_dirty();
    return !!_auth_modname;
  }

  object identity() {return _identity::record();}
  void set_identity (object id)
  {
    object oldid = _identity::record();
    _identity::set (id, db()->identities);
    if (table() && table()[Record::id()] == this_object()) {
      if (oldid) oldid->check_src_auth(), oldid->update_auths (this_object());
      if (id) id->update_auths (this_object());
    }
    set_dirty();
  }

  int disabled() {return _disabled;}
  void set_disabled (int disabled)
  {
    _disabled = disabled;
    if (_identity::record())
      _identity::record()->flush_caches(), _identity::record()->check_src_auth();
    if (Record::id())
      table()->update_auth_data_map_all (this_object());
    set_dirty();
  }

  int working() {return _working;}
  void set_working (int working)
  {
    _working = working;
    if (_identity::record()) _identity::record()->flush_caches();
    set_dirty();
  }

  int active() {return !_disabled && _working;}

  int *auth_types() {return indices (_auth_data);}
  int has_auth_data (int type) {return !!_auth_data[type];}
  mixed auth_handle (int type) {return _auth_data[type] && _auth_data[type][0];}
  mixed auth_secret (int type) {return _auth_data[type] && _auth_data[type][1];}
  string auth_testfn (int type) {return _auth_data[type] && _auth_data[type][2];}

  void set_auth_data (int type, mixed handle, mixed secret, void|string testfn)
  {
    array o = _auth_data[type];
    _auth_data[type] = ({handle, secret, testfn});
    if (Record::id())
      table()->update_auth_data_map (type, o && o[0], o && o[1], this_object());
    set_dirty();
  }

  void auth_remove (int type)
  {
    array o = _auth_data[type];
    m_delete (_auth_data, type);
    if (Record::id())
      table()->update_auth_data_map (type, o && o[0], o && o[1], this_object());
    set_dirty();
  }

  multiset(object) groups()
  {
    if (!_groups)
      _groups = mkmultiset (rows (db()->identities, indices (_group_ids))) - (<0>);
    return _groups;
  }

  void add_group (object g)
  {
#ifdef DEBUG
    if (!g || !functionp (g->type) || g->type() != ID_TYPE_GROUP)
      throw (({"AClib: The object is not a group\n", backtrace()}));
    if (!db()->identities[g->id()])
      throw (({"AClib: The identity " + (g->handle() || g->id()) +
	       " is not in the right table\n", backtrace()}));
#endif

    _group_ids[g->id()] = 1;
    if (_groups) _groups[g] = 1;
    set_dirty();

    if (_identity::record()) _identity::record()->check_groups();
  }

  void remove_group (object g)
  {
#ifdef DEBUG
    if (!g || !functionp (g->type) || g->type() != ID_TYPE_GROUP)
      throw (({"AClib: The object is not a group\n", backtrace()}));
    if (!db()->identities[g->id()])
      throw (({"AClib: The identity " + (g->handle() || g->id()) +
	       " is not in the right table\n", backtrace()}));
#endif

    _group_ids[g->id()] = 0;
    if (_groups) _groups[g] = 0;
    set_dirty();

    if (_identity::record()) _identity::record()->check_groups();
  }

  string authmod_desc()
  {
    string name = db()->modulefullname (auth_module());
    if (name && name[..3] == "AC: ") name = name[4..];
    return name;
  }

  int `< (object o)
  // Sorts on module, then on identity, then on record ID.
  {
    return
      auth_modname() < o->auth_modname() ||
      auth_modname() == o->auth_modname() && (
	identity() < o->identity() ||
	identity() == o->identity() && Record::id() < o->id());
  }

  void flush_caches() {_groups = 0;}

  array dump_values()
  {
    return ({
      Record::id(),
      db()->moduleremotename (_auth_modname),
      _identity::id(), _working, data, _disabled,
      _auth_data, _group_ids
    });
  }

  string encode_readable()
  {
    return Record::id() + ": " +
      db()->readable->encode (_auth_modname) + ", " +
      _identity::id() + ", " +
      _working + ", " +
      db()->readable->encode (data) + ", " +
      _disabled + ", " +
      db()->readable->encode (_auth_data) + ", " +
      db()->readable->encode (_group_ids);
  }

  int assign_values (array v, void|int update)
  {
    if (!arrayp (v) || !sizeof (v)) RETURN_ERR ("IdAuth", "(unknown)");
    int id = v[0];
    if (update && id != Record::id())
      throw (({"AClib: Wrong id in update.\n", backtrace()}));

    if (sizeof (v) < 5 ||
	!stringp (v[1]) ||
	!intp (v[2]) ||
	!intp (v[3]) ||
	!arrayp (v[4]) && v[4] != 0 ||
	sizeof (v) >= 6 && !intp (v[5]) ||
	sizeof (v) >= 7 && !mappingp (v[6]) ||
	sizeof (v) >= 8 && !multisetp (v[7])) RETURN_ERR ("IdAuth", id);
    if (sizeof (v) >= 7)
      foreach (indices (v[6]), mixed i)
	if (!intp (i) || i < 0 || i >= AUTH_TYPES_COUNT ||
	    !arrayp (v[6][i]) || sizeof (v[6][i]) < 2) RETURN_ERR ("IdAuth", id);
    if (sizeof (v) >= 8)
      foreach (indices (v[7]), mixed e) if (!intp (e)) RETURN_ERR ("IdAuth", id);

    // Compatibility kludge.
    if (search (v[1], "#") < 0) v[1] += "#0";

    _auth_modname = db()->modulelocalname (v[1]);
    _auth_mod = 0;
    _identity::set (v[2], db()->identities);
    _working = v[3];
    data = v[4];
    if (sizeof (v) >= 6) _disabled = v[5];
    if (sizeof (v) >= 7) _auth_data = v[6];
    if (sizeof (v) >= 8) _group_ids = v[7];

    // Compatibility kludges.
    foreach (indices (_auth_data), int type)
      while (sizeof (_auth_data[type]) < 3) _auth_data[type] += ({0});
    if (_auth_data[AUTH_NT] && !_auth_data[AUTH_NT][2])
      _auth_data[AUTH_NT][2] = "test_nt_passwd";

    if (update) flush_caches(), set_dirty();

    return id;
  }

  int decode_readable (string s)
  {
    int id;
    if (sscanf (s, "%D%*[ \t]:%s", id, s) != 3) RETURN_ERR ("IdAuth", "(unknown)");
    array v = db()->readable->decode ("({" + s + "})");
    if (sizeof (v) == 1) RETURN_ERR ("IdAuth", id);
    return assign_values (({id}) + v[1]);
  }

  // The following functions assume that the record is in a table.

  void is_removed (object tbl, object rec)
  {
    if (tbl == db()->identities)
      if (rec == identity())
	db()->id_auths->remove (this_object());
      else {
	_group_ids[rec->id()] = 0;
	_groups = 0;
	set_dirty();
      }
  }

  void module_is_removed (object mod)
  {
    if (mod && mod == auth_module()) table()->remove (this_object());
  }

  void purge_missing_modules()
  {
    if (!auth_module()) table()->remove (this_object());
  }
}

// Special identities.
static constant EVERYONE_ID = 1;
constant EVERYONE_NAME = "Everyone";
static constant PRIVILEGED_ID = -1; // Also used in IdAuths.privileged().
constant PRIVILEGED_NAME = "Privileged";

class Identity
// An identity. This may be a user, a group of users or something
// else that should be subject to access control.
{
  inherit Record;

  static int _type;
  static string _handle = 0;	// This is a short handle. It's either 0 or unique.

  static multiset(int) _incl_group_ids = (<>);
  static multiset(int) _excl_group_ids = (<>);
  // Structures to keep track of the IDs of the groups this one is a
  // member of. _incl_group_ids are the memberships explicitly added
  // by the administrator, and _excl_group_ids those explicitly
  // removed. They are mutually exclusive. The automatic memberships
  // added by auth modules are kept in the connected IdAuth records.

  static inherit RecordPtr: _src_auth;
  // The IdAuth that controls this identity, or 0 if none. This means
  // that the module is free to change the user info at any time. If
  // the identity only has one IdAuth object and that one is
  // _src_auth, then the identity is said to be completely controlled
  // by that IdAuth, and it will be removed if the IdAuth is.

  // Misc user info.
  static string _name = 0;

  static mapping(string:mixed) _data = ([]);
  // The data stored in this identity by different modules. Indexed by
  // module name.

  static multiset(object) _auto_groups = 0;
  // Cache of the group memberships added by auth modules.

  static multiset(object) _direct_groups = 0;
  static multiset(object) _indirect_groups = 0;
  // Caches of groups this identity is direct resp indirect member of.

  static object(IdAuth) *_auths = ({});
  // Cache of connected id_auths. If _src_auth is set, it's always a
  // member when this is an array. We can initialize this to ({})
  // since an IdAuth must always have an Identity to be added. Note
  // that this cache is not flushed and rebuilt in the normal fashion.

  static object *_active_authmods = 0;
  // Cache of the auth modules corresponding to the _auths that are
  // active.

  static int _has_enabled_authmods = 0;
  // Cache for has_enabled_authmods().

  static string *_userinfo = 0;
  // Cache for the auth module style userinfo array.

  void create (object tbl, int type) {Record::create (tbl), _type = type;}

  int type() {return _type;}	// Atomic access assumed in ModuleAC.

  string handle() {return _handle;} // Atomic access assumed in ModuleAC.

  object *auths();
  int set_handle (string handle)
  {
    if (handle && Record::id() && table()->find_by_handle (handle)) {
      report_error ("AClib: Can't set duplicate identity handle " + handle + "\n");
      return 0;
    }

    string o = _handle;
    _handle = handle;
    if (Record::id()) {
      table()->update_handle_map (o, this_object());
      foreach (auths(), object(IdAuth) ia)
	db()->try_call (ia->auth_module(), "id_handle_changed", ia);
    }
    set_dirty();
    return 1;
  }

  string name() {return _name;} // Atomic access assumed in ModuleAC.
  void set_name (string name) {_name = name; set_dirty();}

  mixed data (object module)
  {
    return _data[db()->modulename (module)];
  }

  int set_data (object module, mixed data)
  {
    string modname = db()->modulename (module);
    if (!modname) return 0;
    _data[modname] = data;
    set_dirty();
    return 1;
  }

  void update_userinfo (void|int not_remote)
  {
    object(IdAuth) ia = _src_auth::record();
    object authmod;
    if (ia && !ia->disabled() && (authmod = ia->auth_module()) &&
	(!not_remote || !objectp (authmod->variables)))
      db()->try_call (authmod, "id_update_userinfo", ia);
  }

  int `< (object o)
  {
    string n, on, h, oh;
    return
      Record::id() == EVERYONE_ID || o->id() != EVERYONE_ID &&
      (_type > o->type() || _type == o->type() &&
       ((n = name()) && !(on = o->name()) ||
	n && on && n < on ||
	n == on &&
	((h = handle()) && !(oh = o->handle()) ||
	 !h && !oh && Record::id() < o->id() ||
	 h && oh && h < oh ||
	 h == oh && Record::id() < o->id())));
  }

  void flush_caches()
  {
    _auto_groups = _direct_groups = _indirect_groups = 0;
    _active_authmods = 0; _has_enabled_authmods = 0;
  }

  array dump_values()
  {
    return ({
      Record::id(), _type, _name, _handle, _incl_group_ids, _data, _src_auth::id(),
      _excl_group_ids
    });
  }

  string encode_readable()
  {
    return Record::id() + ": " +
      ID_TYPE_TEXTS[_type] + ", " +
      db()->readable->encode (_name) + ", " +
      db()->readable->encode (_handle) + ", " +
      db()->readable->encode (_incl_group_ids) + ", " +
      db()->readable->encode (_data) + ", " +
      _src_auth::id() + ", " +
      db()->readable->encode (_excl_group_ids);
  }

  int assign_values (array v, void|int update)
  {
    if (!arrayp (v) || !sizeof (v)) RETURN_ERR ("Identity", "(unknown)");
    int id = v[0];
    if (update && id != Record::id())
      throw (({"AClib: Wrong id in update.\n", backtrace()}));

    if (sizeof (v) < 5 ||
	!stringp (v[1]) && !intp (v[1]) ||
	!stringp (v[2]) && v[2] != 0 ||
	!stringp (v[3]) && v[3] != 0 ||
	!multisetp (v[4]) ||
	sizeof (v) >= 6 && !mappingp (v[5]) ||
	sizeof (v) >= 7 && !intp (v[6]) ||
	sizeof (v) >= 8 && !multisetp (v[7])) RETURN_ERR ("Identity", id);
    foreach (indices (v[4]), mixed e) if (!intp (e)) RETURN_ERR ("Identity", id);
    int type = stringp (v[1]) ? search (ID_TYPE_TEXTS, v[1]) : v[1];
    if (type < 0 || type >= sizeof (ID_TYPE_TEXTS)) RETURN_ERR ("Identity", id);
    if (sizeof (v) >= 8)
      foreach (indices (v[7]), mixed e) if (!intp (e)) RETURN_ERR ("Identity", id);

    _type = type;
    _name = v[2];
    _handle = v[3];
    _incl_group_ids = v[4];
    if (sizeof (v) >= 6) _data = v[5];
    if (sizeof (v) >= 7) _src_auth::set (v[6], db()->id_auths);
    if (sizeof (v) >= 8) _excl_group_ids = v[7];

    if (update) flush_caches(), set_dirty();

    return id;
  }

  int decode_readable (string s)
  {
    int id;
    if (sscanf (s, "%D%*[ \t]:%s", id, s) != 3) RETURN_ERR ("Identity", "(unknown)");
    array v = db()->readable->decode ("({" + s + "})");
    if (sizeof (v) == 1) RETURN_ERR ("Identity", id);
    return assign_values (({id}) + v[1]);
  }

  // The following functions assume that the record is in a table.

  void check_indirect_groups() {_indirect_groups = 0;}

  void check_groups()
  {
    _auto_groups = _direct_groups = 0;
    table()->this_values()->check_indirect_groups();
  }

  void check_src_auth()
  {
    if (search (auths(), _src_auth::record()) < 0 || _src_auth::record()->disabled())
      _src_auth::set (0);
  }

  object(IdAuth) *auths()
  {
    return _auths;
  }

  object(IdAuth) *active_auths()
  {
    object(IdAuth) *ias = ({});
    foreach (auths(), object(IdAuth) ia)
      if (ia->active()) ias += ({ia});
    return ias;
  }

  object *authmods()
  {
    object *mods = column (auths(), "auth_module")();
    mapping(object:object) m = mkmapping (mods, mods);
    m_delete (m, 0);
    return indices (m);
  }

  object *active_authmods()
  {
    if (!_active_authmods) {
      object *mods = column (active_auths(), "auth_module")();
      mapping(object:object) m = mkmapping (mods, mods);
      m_delete (m, 0);
      _active_authmods = indices (m);
    }
    else _active_authmods -= ({0});
    return _active_authmods;
  }

  int has_enabled_authmods()
  {
    if (!_has_enabled_authmods)
      _has_enabled_authmods = search (column (auths(), "disabled")(), 0) >= 0;
    return _has_enabled_authmods;
  }

  object(IdAuth) src_auth() {return _src_auth::record();}
  void set_src_auth (object(IdAuth) src_auth)
  {
#ifdef DEBUG
    if (src_auth && (src_auth->disabled() || search (auths(), src_auth) < 0))
      throw (({"AClib: Can't set IdAuth " + src_auth->id() +
	       " to source for identity " + id() + "\n", backtrace()}));
#endif
    _src_auth::set (src_auth, db()->id_auths);
    set_dirty();
  }

  multiset(object) auto_groups()
  {
    if (!_auto_groups) _auto_groups = `|((<>), @auths()->groups());
    return _auto_groups;
  }

  multiset(object) direct_groups()
  {
    if (!_direct_groups)	// Constructed in indirect_groups too.
      _direct_groups = (auto_groups() |
			mkmultiset (rows (table(), indices (_incl_group_ids)))) -
	mkmultiset (rows (table(), indices (_excl_group_ids))) - (<0>);
    return _direct_groups;
  }

  multiset(object) indirect_groups (void|multiset(object) exclude)
  {
    // Handle circularity. Although forbidden normally, it may occur
    // as a result of automatic memberships.
    if (exclude)
      if (exclude[this_object()]) return (<>);
      else exclude[this_object()] = 1;
    else exclude = (<this_object()>);

    if (!_indirect_groups) {
      multiset(object) excl = mkmultiset (rows (table(), indices (_excl_group_ids)));
      if (!_direct_groups)	// Construct _direct_groups now when we got excl already.
	_direct_groups = (auto_groups() |
			  mkmultiset (rows (table(), indices (_incl_group_ids)))) -
	  excl - (<0>);
      _indirect_groups = (<>);
      foreach (indices (_direct_groups), object g) {
	_indirect_groups |= g->indirect_groups (exclude);
	_indirect_groups |= g->direct_groups();
      }
      _indirect_groups -= excl;
    }
    return _indirect_groups;
  }

  multiset(object) groups() {return indirect_groups() | direct_groups();}
  int member (object g) {return indirect_groups()[g] || direct_groups()[g];}

  int incl_member (object g) {return _incl_group_ids[g->id()];}
  int excl_member (object g) {return _excl_group_ids[g->id()];}
  int auto_member (object g) {return !_excl_group_ids[g->id()] && auto_groups()[g];}
  int auto_member_ignore_excl (object g) {return auto_groups()[g];}
  int direct_member (object g) {return direct_groups()[g];}
  int direct_member_ignore_excl (object g)
    {return _incl_group_ids[g->id()] || auto_groups()[g];}
  int indirect_member (object g) {return indirect_groups()[g];}
  int indirect_member_ignore_excl (object g)
  {
    if (_excl_group_ids[g->id()]) {
      foreach (indices (auto_groups()) + rows (table(), indices (_incl_group_ids)),
	       object h)
	if (h && !_excl_group_ids[h->id()] && h->member (g)) return 1;
      return 0;
    }
    return indirect_groups()[g];
  }

  multiset(object) members()
  // Returns the empty array for users. Note: Currently an expensive
  // operation.
  {
    multiset(object) m = (<>);
    if (_type == ID_TYPE_GROUP)
      foreach (table()->this_values(), object i) m[i] = i->member (this_object());
    return m;
  }

  int add_incl_group (object g)
  // Removes the group from _excl_group_ids if it's there. Checks for
  // cyclic membership. Returns true if successful.
  {
#ifdef DEBUG
    if (!g || !functionp (g->type) || g->type() != ID_TYPE_GROUP)
      throw (({"AClib: The object is not a group\n", backtrace()}));
    if (!table()[g->id()])
      throw (({"AClib: The identity " + (g->handle() || g->id()) +
	       " is not in this table\n", backtrace()}));
#endif

    if (g == this_object() || g->member (this_object()) || _incl_group_ids[g->id()])
      return 0;

    _excl_group_ids[g->id()] = 0;
    _incl_group_ids[g->id()] = 1;
    set_dirty();

    if (_direct_groups) _direct_groups[g] = 1;
    table()->this_values()->check_indirect_groups();

    return 1;
  }

  int remove_incl_group (object g)
  {
#ifdef DEBUG
    if (!g || !functionp (g->type) || g->type() != ID_TYPE_GROUP)
      throw (({"AClib: The object is not a group\n", backtrace()}));
    if (!table()[g->id()])
      throw (({"AClib: The identity " + (g->handle() || g->id()) +
	       " is not in this table\n", backtrace()}));
#endif

    if (!_incl_group_ids[g->id()]) return 0;

    _incl_group_ids[g->id()] = 0;
    set_dirty();

    if (_direct_groups && !auto_groups()[g]) _direct_groups[g] = 0;
    table()->this_values()->check_indirect_groups();

    return 1;
  }

  int add_excl_group (object g)
  // Removes the group from _incl_group_ids if it's there. Checks for
  // cyclic membership (it's meaningless to exclude a group that is a
  // member of this one). Returns true if successful.
  {
#ifdef DEBUG
    if (!g || !functionp (g->type) || g->type() != ID_TYPE_GROUP)
      throw (({"AClib: The object is not a group\n", backtrace()}));
    if (!table()[g->id()])
      throw (({"AClib: The identity " + (g->handle() || g->id()) +
	       " is not in this table\n", backtrace()}));
#endif

    if (g == this_object() || g->member (this_object()) || _excl_group_ids[g->id()])
      return 0;

    _incl_group_ids[g->id()] = 0;
    _excl_group_ids[g->id()] = 1;
    set_dirty();

    if (_direct_groups) _direct_groups[g] = 0;
    table()->this_values()->check_indirect_groups();

    return 1;
  }

  int remove_excl_group (object g)
  {
#ifdef DEBUG
    if (!g || !functionp (g->type) || g->type() != ID_TYPE_GROUP)
      throw (({"AClib: The object is not a group\n", backtrace()}));
    if (!table()[g->id()])
      throw (({"AClib: The identity " + (g->handle() || g->id()) +
	       " is not in this table\n", backtrace()}));
#endif

    if (!_excl_group_ids[g->id()]) return 0;

    _excl_group_ids[g->id()] = 0;
    set_dirty();

    if (_direct_groups && auto_groups()[g]) _direct_groups[g] = 1;
    table()->this_values()->check_indirect_groups();

    return 1;
  }

  void check_userinfo() {_userinfo = 0;}

  string *get_userinfo()
  // Ambiguity resolution rules in case several IdAuths gives userinfo
  // data:
  // 1. The handle and name of this identity are used if they are set.
  // 2. If the controlling IdAuth gives userinfo for a field, it's used.
  // 3. If exactly one other IdAuth gives userinfo for a field, it's used.
  {
    if (!_userinfo) {
      {
	object(IdAuth) ia = _src_auth::record();
	if (!ia || ia->disabled() || !ia->auth_module() ||
	    !(_userinfo = db()->try_call (ia->auth_module(), "id_get_userinfo", ia)))
	  _userinfo = ({0}) * 7;
      }
      if (handle()) _userinfo[0] = handle();
      if (name()) _userinfo[4] = name();
      if (search (_userinfo, 0) >= 0) {
	int *flag = ({0}) * 7;
	foreach (auths(), object(IdAuth) ia) {
	  string *res;
	  if (!ia->disabled() && ia->auth_module() &&
	      (res = db()->try_call (ia->auth_module(), "id_get_userinfo", ia)))
	    for (int i = 0; i < 7; i++)
	      if (res[i])
		if (!_userinfo[i] && !flag[i])
		  _userinfo[i] = res[i], flag[i] = 1;
		else if (flag[i])
		  _userinfo[i] = 0;
	}
      }
      if (!_userinfo[1]) _userinfo[1] = "*";
      foreach (({0, 2, 3, 4, 5, 6}), int i)
	if (!_userinfo[i]) _userinfo[i] = "";
    }
    return _userinfo;
  }

  void update_auths (object(IdAuth) ia)
  {
    if (ia->identity() == this_object()) _auths |= ({ia});
    else _auths -= ({ia});
    // flush_caches() inlined.
    _auto_groups = _direct_groups = _indirect_groups = 0;
    _active_authmods = 0; _has_enabled_authmods = 0;
  }

  void is_removed (object tbl, object rec)
  {
    if (tbl == db()->identities) {
      if (auto_groups()[rec]) check_groups();
      if (_incl_group_ids[rec->id()]) remove_incl_group (rec);
      if (_excl_group_ids[rec->id()]) remove_excl_group (rec);
    }
    else if (tbl == db()->id_auths) {
      check_groups();
      update_auths (rec);
      if (rec == _src_auth::record()) {
	if (sizeof (_auths))
	  // Avoid changing the identity from automatic to manual if possible.
	  _src_auth::set (_auths[0]);
	else _src_auth::set (0);
	set_dirty();
      }
    }
  }
}

class ProtClass
// A group of protection points that is treated as one entity that
// identities are given permissions to. It also contains the list of
// active auth modules.
{
  inherit Record;

  static mapping(int:int) _perms = ([]); // Mapping from identity IDs to permissions.
  static string *_authmod_names = ({}); // Ordered list of authentication module names.

  static string _name = 0;
  // Used in the UI. If zero and there's exactly one ppoint in this
  // class, name() in that ppoint is used.

  static string|int _realm = 0;
  // The realm sent in http authentication responses. If an int, the
  // ID of the pclass containing the realm string. No more than one
  // indirection is allowed.

  static object *_active_authmods = 0;
  // Cache of the auth module list. Bogus entries are removed in this cache.

  static string _realmstr = 0;
  // Cache of the realm string.

  static object *_ppoints = 0;
  // Cache of the ppoints belonging to this pclass.

  void create (object tbl) {Record::create (tbl);}

  // The permission functions should not be used externally; they
  // might be moved in the future. Use the functions in AC_DB instead.

  int perm (object(Identity) id)
  // Returns the directly set permission or -1 if none.
  {
#ifdef DEBUG
    if (db()->identities[id->id()] != id)
      throw (({"AClib: Not a valid identity for permission query\n", backtrace()}));
#endif
    return zero_type (_perms[id->id()]) ? -1 : _perms[id->id()];
  }

  int set_perm (object(Identity) id, int perm)
  // A perm of -1 removes any permission.
  {
#ifdef DEBUG
    if (db()->identities[id->id()] != id)
      throw (({"AClib: Not a valid identity to be given a permission\n", backtrace()}));
#endif
    if (perm != -1) {
      if (zero_type (_perms[id->id()]) == 1 || perm != _perms[id->id()])
	_perms[id->id()] = perm, set_dirty();
    }
    else
      if (zero_type (_perms[id->id()]) != 1) m_delete (_perms, id->id()), set_dirty();
    return 1;
  }

  mapping(int:int) perms() {return copy_value (_perms);}
  void set_perms (mapping(int:int) perms) {_perms = perms; set_dirty();}

  string real_name() {return _name;}
  string real_realm() {return stringp (_realm) && _realm;}

  string name();
  void set_name (string newname)
  {
    _name = 0;
    _name = newname == "" || db() && newname == name() ? 0 : newname;
    set_dirty();
  }

  string authmod_name (int pos)
  {
    return _authmod_names[pos];
  }

  string *authmod_names() {return copy_value (_authmod_names);}

  void set_authmod_names (string *authmod_names)
  {
    _authmod_names = authmod_names;
    _active_authmods = 0;
    set_dirty();
  }

  object *authmods()
  // Excludes modules that can't be resolved.
  {
    object *am = ({});
    int missing = 0;

    foreach (_authmod_names, string name) {
      object mod = db()->findmodule (name);
      if (!mod) {
	missing = 1;
#ifdef DEBUG
	report_warning ("AClib: Auth module " + name + " not found\n");
#endif
      }
      am += ({mod});
    }

    return am;
  }

  object *active_authmods()
  {
    if (!_active_authmods) {
      object *am = ({});
      int missing = 0;

      foreach (_authmod_names, string name) {
	object mod = db()->findmodule (name);
	if (!mod) {
	  missing = 1;
#ifdef DEBUG
	  report_warning ("AClib: Auth module " + name + " not found\n");
#endif
	}
	else am += ({mod});
      }

      if (missing) return am;	// Don't cache if there are modules missing.
      else _active_authmods = am;
    }
    else _active_authmods -= ({0});

    return _active_authmods;
  }

  int authmod_pos (object|string mod)
  {
    return stringp (mod) ? search (_authmod_names, mod) : search (authmods(), mod);
  }

  void add_authmod (object mod, int pos)
  // Give negative pos to add to end.
  {
    string name = db()->modulename (mod);

#ifdef DEBUG
    if (!name)
      throw (({"AClib: Couldn't get name of auth module to add to pclass\n",
	       backtrace()}));
    if (search (_authmod_names, name) != -1)
      throw (({"AClib: Can't add an auth module to a pclass twice\n", backtrace()}));
#endif

    if (pos < 0) _authmod_names += ({name});
    else _authmod_names = _authmod_names[..pos-1] + ({name}) + _authmod_names[pos..];
    _active_authmods = 0;
    set_dirty();
  }

  void remove_authmod (int pos)
  {
    _authmod_names = _authmod_names[..pos-1] + _authmod_names[pos+1..];
    _active_authmods = 0;
    set_dirty();
  }

  void flush_caches() {_active_authmods = 0; _realmstr = 0; _ppoints = 0;}

  array dump_values()
  {
    return ({
      Record::id(), _name, _realm, _perms,
      Array.map (_authmod_names, db()->moduleremotename)
    });
  }

  string encode_readable()
  {
    return Record::id() + ": " +
      db()->readable->encode (_name) + ", " +
      db()->readable->encode (_realm) + ", "
      "([" + Array.map (indices (_perms),
			lambda (int id) {
			  return id + ":" + PERM_TEXTS[_perms[id]];
			}) * ", " + "]), " +
      db()->readable->encode (_authmod_names);
  }

  int assign_values (array v, void|int update)
  {
    if (!arrayp (v) || !sizeof (v)) RETURN_ERR ("ProtClass", "(unknown)");
    int id = v[0];
    if (update && id != Record::id())
      throw (({"AClib: Wrong id in update.\n", backtrace()}));

    if (sizeof (v) < 5 ||
	!stringp (v[1]) && v[1] ||
	!stringp (v[2]) && !intp (v[2]) && v[2] ||
	!mappingp (v[3]) ||
	!arrayp (v[4])) RETURN_ERR ("ProtClass", id);
    foreach (v[4], mixed e) if (!stringp (e)) RETURN_ERR ("ProtClass", id);

    mapping perms = ([]);
    foreach (indices (v[3]), mixed i) {
      if (!intp (i)) RETURN_ERR ("ProtClass", id);
      int perm = stringp (v[3][i]) ? search (PERM_TEXTS, v[3][i]) : v[3][i];
      if (perm < 0 || perm >= sizeof (PERM_TEXTS)) RETURN_ERR ("ProtClass", id);
      perms[i] = perm;
    }

    // Compatibility kludge.
    foreach (indices (v[4]), int i) if (search (v[4][i], "#") < 0) v[4][i] += "#0";

    _name = v[1];
    _realm = v[2];
    _perms = perms;
    _authmod_names = Array.map (v[4], db()->modulelocalname);

    if (update) flush_caches(), set_dirty();

    return id;
  }

  int decode_readable (string s)
  {
    int id;
    if (sscanf (s, "%D%*[ \t]:%s", id, s) != 3) RETURN_ERR ("ProtClass", "(unknown)");
    array v = db()->readable->decode ("({" + s + "})");
    if (sizeof (v) == 1) RETURN_ERR ("ProtClass", id);
    return assign_values (({id}) + v[1]);
  }

  // The following functions assume that the record is in a table.

  string realm()
  {
    if (!_realmstr)
      if (stringp (_realm) || !_realm) _realmstr = _realm;
      else _realmstr = table()[_realm]->real_realm();
    return _realmstr;
  }

  void set_realm (string realm)
  {
    _realmstr = _realm = realm;
    set_dirty();
    table()->this_values()->realm_has_changed (id(), id());
  }

  void realm_has_changed (int from, int to)
  {
    if (_realm == from) {
      if (_realm != to) _realm = to, set_dirty();
      _realmstr = 0;
    }
  }

  int realm_class() {return stringp (_realm) || !_realm ? id() : _realm;}

  int set_realm_class (int pid)
  // Checks for cyclic setting.
  {
    object o = table()[pid];

#ifdef DEBUG
    if (!o) throw (({"AClib: Invalid realm class " + pid + "\n", backtrace()}));
#endif

    if (o->realm_class() == id()) return 0;

    table()->realm_class_has_changed();
    if (o == this_object()) set_realm (realm());
    else {
      table()->this_values()->realm_has_changed (id(), o->realm_class());
      _realm = o->realm_class();
      _realmstr = 0;
    }
    set_dirty();
    return 1;
  }

  object *ppoints()
  {
    if (!_ppoints) _ppoints = db()->ppoints->find_by_pclass (this_object());
    return _ppoints;
  }

  string name()
  {
    return _name ? _name : sizeof (ppoints()) == 1 && _ppoints[0]->name();
  }

  void pre_change_ppoints()
  {
    if (!_name && sizeof (ppoints()) == 1) _name = name(), set_dirty();
  }

  void post_change_ppoints()
  {
    if (sizeof (ppoints()) == 1 && _name == _ppoints[0]->name()) _name = 0, set_dirty();
  }

  int `< (object o)
  {
    string r, or, n, on;
    int rc, orc;
    return
      (r = realm()) && !(or = o->realm()) ||
      r && or && r < or ||
      r == or &&
      ((rc = realm_class()) < (orc = o->realm_class()) ||
       rc == orc &&
       (rc == Record::id() ||
	orc != o->id() &&
	((n = name()) && !(on = o->name()) ||
	 n && on && n < on ||
	 n == on && Record::id() < o->id())));
  }

  void is_removed (object tbl, object rec)
  {
    if (tbl == db()->identities) {
      if (zero_type (_perms[rec->id()]) != 1)
	m_delete (_perms, rec->id());
    }
    else if (tbl == db()->ppoints) {
      int i = search (ppoints(), rec);
      if (i >= 0)
	// Do not run pre_change_ppoints here; we want to keep the
	// name to be able to detect a "silly" pclass in
	// ProtPoints.remove().
	_ppoints = _ppoints[..i-1] + _ppoints[i+1..];
    }
    else if (tbl == table() && _realm == rec->id()) {
      _realmstr = _realm = rec->realm();
      set_dirty();
      rec->set_realm_class (id());
    }
  }

  void module_is_removed (object mod)
  {
    if (mod) {
      int i;
      while ((i = search (authmods(), mod)) >= 0) {
	_authmod_names = _authmod_names[..i-1] + _authmod_names[i+1..];
	_active_authmods = 0;
      }
    }
  }

  void purge_missing_modules()
  {
    for (int i; i < sizeof (_authmod_names); i++)
      if (!db()->findmodule (_authmod_names[i]))
	_authmod_names = _authmod_names[..i-1] + _authmod_names[i+1..];
  }
}

class ProtPoint
// This is the object used by modules to get access control on
// whatever they wish.
{
  inherit RecordPtr: _pclass; // The class this ppoint belongs to.
  static string _own_modname = 0; // Name of the owning module or 0 if global.
  static string _handle = 0;	// This is a short handle.
  // It's either 0 or unique among the ppoints for the owning module.
  // It's not viewable or changeable from the UI; it's purpose is to
  // make it easier to access the ppoint in the owning module.
  static int _deletable = 0;
  // True if the AC admin interface should let the user delete this ppoint.

  inherit Record;

  static string _name = 0;
  // This is parsed with parse_with_module_vars before it's shown to
  // the user.

  static mixed _data = 0;
  // Arbitrary data stored in this ppoint by the module.

  static object _own_mod = 0;	// Cache for the owning module object.

  static int _parsed_name_timeout = 0;
  static string _parsed_name;
  // Timed cache for the parsed name.

  void create (object tbl, object own_mod, void|int deletable)
  {
    Record::create (tbl);
    _own_modname = db()->modulename (own_mod);
    _own_mod = own_mod;
    _pclass::create (db()->pclasses);
    _deletable = deletable;
  }

  object(ProtClass) pclass() {return _pclass::record();}

  void set_pclass (object(ProtClass) pc)
  {
    object(ProtClass) old_pc = _pclass::record();
    if (pc) pc->pre_change_ppoints();
    if (old_pc) old_pc->pre_change_ppoints();
    _pclass::set (pc, db()->pclasses);
    if (pc) pc->flush_caches(), pc->post_change_ppoints();
    if (old_pc) old_pc->flush_caches(), old_pc->post_change_ppoints();
    set_dirty();
  }

  int deletable() {return _deletable;}
  void set_deletable (int deletable)
  {
    _deletable = deletable;
    set_dirty();
  }

  string real_name() {return _name;}

  void set_name (string name)
  {
    if (_pclass::record() && _pclass::record()->name() == name)
      _pclass::record()->set_name (0);
    _name = name;
    set_dirty();
  }

  string handle() {return _handle;}

  int set_handle (string handle)
  {
    if (handle && Record::id() && table()->find_by_handle (handle)) {
      report_error ("AClib: Can't set duplicate ppoint handle " + handle +
		    " for module " + _own_modname + "\n");
      return 0;
    }

    string o = _handle;
    _handle = handle;
    if (Record::id()) table()->update_handle_map (o, this_object());
    set_dirty();
    return 1;
  }

  string own_modname() {return _own_modname;}

  object own_module()
  {
    if (!_own_mod) _own_mod = db()->findmodule (_own_modname);
    return _own_mod;
  }

  mixed data() {return _data;}
  int set_data (mixed data) {_data = data; set_dirty();}

  void flush_caches() {_own_mod = 0;}

  array dump_values()
  {
    return ({
      Record::id(), db()->moduleremotename (_own_modname), _handle, _name, _pclass::id(),
      _data, _deletable
    });
  }

  string encode_readable()
  {
    return Record::id() + ": " +
      db()->readable->encode (_own_modname) + ", " +
      db()->readable->encode (_handle) + ", " +
      db()->readable->encode (_name) + ", " +
      _pclass::id() + ", " +
      db()->readable->encode (_data) + ", " +
      db()->readable->encode (_deletable);
  }

  int assign_values (array v, void|int update)
  {
    if (!arrayp (v) || !sizeof (v)) RETURN_ERR ("ProtPoint", "(unknown)");
    int id = v[0];
    if (update && id != Record::id())
      throw (({"AClib: Wrong id in update.\n", backtrace()}));

    if (sizeof (v) < 5 ||
	!stringp (v[1]) ||
	!stringp (v[2]) && v[2] != 0 ||
	!stringp (v[3]) && v[3] != 0 ||
	!intp (v[4]) ||
	sizeof (v) >= 7 && !intp (v[6])) RETURN_ERR ("ProtPoint", id);

    // Compatibility kludge.
    if (search (v[1], "#") < 0) v[1] += "#0";

    _own_modname = db()->modulelocalname (v[1]);
    _own_mod = 0;
    _handle = v[2];
    _name = v[3];
    _pclass::set (v[4], db()->pclasses);
    if (sizeof (v) >= 6) _data = v[5];
    if (sizeof (v) >= 7) _deletable = v[6];

    if (update) flush_caches(), set_dirty();

    return id;
  }

  int decode_readable (string s)
  {
    int id;
    if (sscanf (s, "%D%*[ \t]:%s", id, s) != 3) RETURN_ERR ("ProtPoint", "(unknown)");
    array v = db()->readable->decode ("({" + s + "})");
    if (sizeof (v) == 1) RETURN_ERR ("ProtPoint", id);
    return assign_values (({id}) + v[1]);
  }

  static string parse_with_module_vars (string str)
  // Parses $-escapes in str, substituting in values from the
  // variables of the owning module. Expressions of the form "$foo$"
  // is replaced with the current value of the module variable foo. If
  // there's no such variable, the function named foo in the module is
  // called with this object as the argument and the returned value is
  // used. "$$" is replaced with "$".
  {
    string *split = str / "$";
    string res = "";

    for (int i; i < sizeof (split); i++) {
      res += split[i];
      if (++i < sizeof (split))
	if (split[i] == "") res += "$";
	else {
	  string r;
	  object mod = own_module();
	  if (!catch {
	    if (zero_type (r = db()->try_call (mod, "query", split[i]))) throw();
	  })
	    res += (string) r;
	  else if (mod && !zero_type (r = db()->try_call (mod, split[i], this_object())))
	    res += (string) r;

	  else res += "$" + split[i] + "$";
	}
    }

    return res;
  }

  // The following functions assume that the record is in a table.

  string name()
  {
    if (_name) {
      if (_parsed_name_timeout <= time (1)) {
	_parsed_name_timeout = time (1) + 30;
	_parsed_name = parse_with_module_vars (_name);
      }
      return _parsed_name;
    }
    else return 0;
  }

  int `< (object o)
  {
    string n, on;
    return
      pclass() < o->pclass() ||
      pclass() == o->pclass() && (
	(n = name()) && !(on = o->name()) ||
	n && on && n < on ||
	n == on && Record::id() < o->id());
  }

  void is_removed (object tbl, object rec)
  {
    if (tbl == db()->pclasses && rec == pclass()) {
      report_warning ("AClib: Removal of a nonempty pclass "
		      "causes a new one to be created for ppoint ID " +
		      Record::id() + "\n");
      db()->ppoints->add_default_pclass (rec);
    }
  }

  void module_is_removed (object mod)
  {
    if (mod && mod == own_module()) table()->remove (this_object());
  }

  void purge_missing_modules()
  {
    if (!own_module()) table()->remove (this_object());
  }
}


// Tables.

// FIXME: Optimize searches.

class IdAuths
{
  inherit Table;

  constant name = "id_auths";

  static array(mapping(mixed:array)) _auth_data =
    allocate (AUTH_TYPES_COUNT);
  // Contains the enabled auth data collected from all IdAuth objects.
  // For each auth data type, there's a mapping indexed by the
  // handles. For each handle, there's an array of the form
  // ({secret_1, testfnname_1, testfn_1, idauth_1, secret_2,
  // testfnname_2, testfn_2, idauth_2, ...}). For each secret_n, the
  // idauth_n is the IdAuth object defining this handle/secret pair.
  // The testfnname_n is the name of the test function and testfn_n is
  // the actual function resolved from it.

#define AUTH_DATA_SECRET 0
#define AUTH_DATA_TESTFNNAME 1
#define AUTH_DATA_TESTFN 2
#define AUTH_DATA_IDAUTH 3
#define AUTH_DATA_RECLEN 4

  static object(IdAuth) _privileged = 0;

  object(IdAuth) privileged()
  // Returns an IdAuth that has the identity returned by
  // Identity.privileged(). It's only valid to take the identity from
  // it, e.g. it's not a record in any table and it's not saved, but
  // it has a valid ID number.
  {
    if (!_privileged) {
      _privileged = IdAuth (0, 0, db()->identities->privileged());
      _privileged->really_set_id_for_this_record (PRIVILEGED_ID);
    }
    return _privileged;
  }

  static void _auth_data_add (int type, mixed handle,
			      mixed secret, object(IdAuth) rec, void|string testfn)
  {
    if (!_auth_data[type])
      _auth_data[type] = ([handle:({secret, testfn, 0, rec})]);
    else if (!_auth_data[type][handle])
      _auth_data[type][handle] = ({secret, testfn, 0, rec});
    else
      _auth_data[type][handle] += ({secret, testfn, 0, rec});
  }

  static int _auth_data_find (int type, mixed handle, mixed secret)
  {
    if (!_auth_data[type] || !_auth_data[type][handle]) return -1;
    array arr = _auth_data[type][handle];
    for (int i = 0; i < sizeof (arr); i += AUTH_DATA_RECLEN)
      if (arr[i + AUTH_DATA_SECRET] == secret) return i;
    return -1;
  }

  static void _auth_data_remove (int type, mixed handle, mixed secret)
  {
    int i = _auth_data_find (type, handle, secret);
    if (i >= 0) _auth_data[type][handle] =
		  _auth_data[type][handle][..i-1] +
		  _auth_data[type][handle][i + AUTH_DATA_RECLEN ..];
  }

  void create (object db) {Table::create (db, IdAuth);}

  int add (object(IdAuth) rec, mixed ... args)
  {
#ifdef DEBUG
    if (!rec->identity())
      throw (({"AClib: Can't add IdAuth record without a valid identity\n",
	       backtrace()}));
#endif

    if (!Table::add (rec, @args)) return 0;
    rec->identity()->update_auths (rec);

    if (!rec->disabled())
      foreach (rec->auth_types(), int type)
	_auth_data_add (type, rec->auth_handle (type), rec->auth_secret (type), rec,
			rec->auth_testfn (type));

    db()->identities->flush_caches();
    return 1;
  }

  int remove (object(IdAuth) rec)
  {
    object(Identity) id = rec->identity(), remid = id;
    if (id && (id->src_auth() != rec || sizeof (id->auths()) != 1)) remid = 0;

    int *types = rec->auth_types();
    mixed *handles = Array.map (types, rec->auth_handle);
    mixed *secrets = Array.map (types, rec->auth_secret);

    if (Table::remove (rec, lambda (object(Identity) id, object(IdAuth) rec) {
			      rec->set_identity (0);
			      if (id) id->is_removed (this_object(), rec);
			    }, id, rec)) {
      if (remid) db()->identities->remove (remid);
      for (int i = 0; i < sizeof (types); i++)
	_auth_data_remove (types[i], handles[i], secrets[i]);
      return 1;
    }
    return 0;
  }

  object(IdAuth) *find_by_id (object(Identity) id)
  // Obsolete.
  {
    return id->auths();
  }

  object(IdAuth) *find_by_authmod (object authmod)
  {
    object(IdAuth) *ias = ({});
    foreach (this_values(), object(IdAuth) ia)
      if (ia->auth_module() == authmod) ias += ({ia});
    return ias;
  }

  object(IdAuth) *find_by_authmod_id (object authmod, object(Identity) id)
  {
    object(IdAuth) *ias = ({});
    foreach (id->auths(), object(IdAuth) ia)
      if (ia->auth_module() == authmod) ias += ({ia});
    return ias;
  }

  object(IdAuth) *find_active_by_authmod_id (object authmod, object(Identity) id)
  {
    object(IdAuth) *ias = ({});
    foreach (id->auths(), object(IdAuth) ia)
      if (ia->active() && ia->auth_module() == authmod) ias += ({ia});
    return ias;
  }

  object(IdAuth) find_by_auth_data (int type, mixed handle, mixed secret)
  {
    int i = _auth_data_find (type, handle, secret);
    return i < 0 ? 0 : _auth_data[type][handle][i + AUTH_DATA_IDAUTH];
  }

  int sizeof_secrets (int type, mixed handle)
  {
#if constant (roxen)
    int n = db()->superuser() && type == AUTH_PASSWD_CRYPT &&
      handle == roxen->QUERY (ConfigurationUser);
#else
    int n = 0;
#endif
    if (!_auth_data[type] || !_auth_data[type][handle]) return n;
    return n + sizeof (_auth_data[type][handle]) / AUTH_DATA_RECLEN;
  }

  object(IdAuth) search_by_secret (
    int type, mixed handle, function(mixed,mixed...:int) testfn, mixed ... args)
  {
#if constant (roxen)
    if (db()->superuser() && type == AUTH_PASSWD_CRYPT &&
	handle == roxen->QUERY (ConfigurationUser) &&
	testfn (roxen->QUERY (ConfigurationPassword), @args)) {
      return privileged();
    }
#endif

    if (!_auth_data[type] || !_auth_data[type][handle]) return 0;
    array arr = _auth_data[type][handle];
    for (int i = 0; i < sizeof (arr); i += AUTH_DATA_RECLEN)
      if (testfn (arr[i + AUTH_DATA_SECRET], @args)) return arr[i + AUTH_DATA_IDAUTH];
    return 0;
  }

  object(IdAuth) search_with_testfn (
    int type, mixed handle, function(mixed,mixed...:int) testfn, mixed ... args)
  // Like search_by_secret(), but testfn is only used as fallback if
  // the auth_data doesn't specify a test function. It takes the same
  // arguments as the test function (see IdAuth._auth_data).
  {
#if constant (roxen)
    if (db()->superuser() && type == AUTH_PASSWD_CRYPT &&
	handle == roxen->QUERY (ConfigurationUser) &&
	testfn (roxen->QUERY (ConfigurationPassword), @args)) {
      return privileged();
    }
#endif

    if (!testfn) testfn = lambda() {return 0;};
    if (!_auth_data[type] || !_auth_data[type][handle]) return 0;
    array arr = _auth_data[type][handle];
    string|function|object fn;
    array err;
    for (int i = 0; i < sizeof (arr) && !err; i += AUTH_DATA_RECLEN) {
      int retry;
      do {
	// try_call() stuff here.
	object(IdAuth) ia = arr[i + AUTH_DATA_IDAUTH];
	if (arr[i + AUTH_DATA_TESTFNNAME]) {
	  if (!(fn = arr[i + AUTH_DATA_TESTFN])) {
	    object mod = ia->auth_module();
	    if (!mod) {
	      report_error ("AClib: Couldn't do authentication test "
			    "with nonexisting module " + ia->auth_modname() + "\n");
	      fn = testfn;
	    }
	    else {
	      fn = mod[arr[i + AUTH_DATA_TESTFNNAME]];
	      if (!functionp (fn) &&
		  !(objectp (mod->variables) && objectp (fn))) // Remote module.
		throw (({"AClib: There's no auth test function " +
			 arr[i + AUTH_DATA_TESTFNNAME] + " in module " +
			 ia->auth_modname() + "\n", backtrace()}));
	      arr[i + AUTH_DATA_TESTFN] = fn;
	    }
	  }
	}
	else fn = testfn;
	if (err = catch {
	  if (fn (ia, handle, arr[i + AUTH_DATA_SECRET], @args)) return ia;
	})
	  if (objectp (fn) && !retry) {
	    catch (destruct (fn->ctx->other[fn->objectid]));
	    arr[i + AUTH_DATA_TESTFN] = 0;
	    retry = 1;
	  }
	  else break;
	else retry = 0;
      }
      while (retry);
    }

    if (err) {
      if (objectp (fn)) {
	report_error ("AClib: Failed to do authentication check in remote AC module: " +
		      err[0]);
#ifdef AC_REMOTE_DEBUG
	roxen_perror (describe_backtrace (err));
#endif
      }
      else throw (err);
    }

    return 0;
  }

  void update_auth_data_map (int type, mixed oldhandle, mixed oldsecret,
			     object(IdAuth) rec)
  {
    if (oldhandle) _auth_data_remove (type, oldhandle, oldsecret);
    if (!rec->disabled() && rec->has_auth_data (type))
      _auth_data_add (type, rec->auth_handle (type), rec->auth_secret (type), rec,
		      rec->auth_testfn (type));
  }

  void update_auth_data_map_all (object(IdAuth) rec)
  {
    foreach (rec->auth_types(), int type)
      _auth_data_remove (type, rec->auth_handle (type), rec->auth_secret (type));
    if (!rec->disabled())
      foreach (rec->auth_types(), int type)
	_auth_data_add (type, rec->auth_handle (type), rec->auth_secret (type), rec,
			rec->auth_testfn (type));
  }
}

class Identities
{
  inherit Table;

  constant name = "identities";

  static object(Identity) _everyone = 0, _privileged = 0;
  static mapping(string:object(Identity)) _handles = ([]);

  mapping(int:array(object)) init_authmods = ([]);
  // Mapping that for each identity type holds a list of
  // authentication modules for which IdAuth objects should be created
  // for new identities.

  void create (object db) {Table::create (db, Identity);}

  object(Identity) everyone();

  object(Identity) privileged()
  // Returns a fake identity that will always get PERM_WRITE
  // permission on any check. It's only valid as an argument to
  // AC_DB.perm() and AC_DB.simulated_perm(), e.g. it's not a record
  // in any table and it's not saved, but it has a valid ID number.
  {
    if (!_privileged) {
      // The existence of the table object is used as flag in various
      // ModuleAC functions. It should not actually be used.
      _privileged = Identity (Table (db(), Identity), ID_TYPE_USER);
      _privileged->really_set_id_for_this_record (PRIVILEGED_ID);
      _privileged->set_name (PRIVILEGED_NAME);
    }
    return _privileged;
  }

  object(Identity) new (mixed ... args)
  {
    object(Identity) o = Table::new (@args);
    o->add_incl_group (everyone());
    return o;
  }

  int add (object(Identity) rec, mixed ... args)
  {
    if (rec->handle() && zero_type (_handles[rec->handle()]) == 0) {
      report_error ("AClib: An identity handle " + rec->handle() + " already exist\n");
      return 0;
    }

    if (!Table::add (rec, @args)) return 0;
    if (rec->handle()) _handles[rec->handle()] = rec;
    return 1;
  }

  void init_auth_methods (object(Identity) rec)
  {
    if (init_authmods [rec->type()])
      foreach (init_authmods [rec->type()], object am)
	if (!am)
	  report_error ("AClib: Couldn't find an initial auth module for identity\n");
	else {
	  object(IdAuth) ia;
	  if (zero_type (
	    ia = db()->try_call (am, "id_auth_init", db()->id_auths->new (am, rec))))
	    report_error ("AClib: Initial auth module "
			  "doesn't support explicit addition of identitites\n");
	  else if (!ia)
	    report_error ("AClib: Failed to add initial auth module to identity\n");
	  else db()->id_auths->add (ia);
	}
  }

  string *handles()
  {
    return indices (_handles);
  }

  object(Identity) *find_by_type (int type)
  {
    object(Identity) *ids = ({});
    foreach (this_values(), object(Identity) id)
      if (id->type() == type) ids += ({id});
    return ids;
  }

  object(Identity) find_by_handle (string handle)
  {
    return _handles[handle];
  }

  object(Identity) *find_by_name (string name)
  {
    object(Identity) *ids = ({});
    foreach (this_values(), object(Identity) id)
      if (id->name() == name) ids += ({id});
    return ids;
  }

  void update_handle_map (string oldname, object(Identity) rec)
  {
    if (oldname) m_delete (_handles, oldname);
    if (rec->handle()) _handles[rec->handle()] = rec;
  }

  object(Identity) everyone()
  {
    if (!_everyone) {
      _everyone = `[](EVERYONE_ID);
      if (!_everyone) {
	_everyone = Table::new (ID_TYPE_GROUP);
	_everyone->set_name (EVERYONE_NAME);
	if (add (_everyone, EVERYONE_ID))
	  report_notice ("AClib: An everyone identity was created\n");
	else
	  report_error ("AClib: Couldn't add an everyone identity\n");
      }
    }
    return _everyone;
  }

  int remove (object(Identity) rec)
  {
    if (rec->id() == EVERYONE_ID) {
      report_error ("AClib: Can't remove the everyone identity\n");
      return 0;
    }

    string handle = rec->handle();
    if (!Table::remove (rec, lambda (object(Identity) rec) {
			       db()->id_auths->is_removed (this_object(), rec);
			       db()->pclasses->is_removed (this_object(), rec);
			       is_removed (this_object(), rec);
			     }, rec)) return 0;
    if (handle) m_delete (_handles, handle);
    return 1;
  }
}

class ProtClasses
{
  inherit Table;

  constant name = "pclasses";

  static object(ProtClass) *_realm_classes = 0;
  // Cache of the realm class representatives.

  object *init_authmods = ({}); // Ordered list of initial authentication modules.

  void create (object db) {Table::create (db, ProtClass);}

  int add (object(ProtClass) rec, mixed ... args)
  {
    if (Table::add (rec, @args)) {
      if (!sizeof (rec->authmod_names()))
	foreach (init_authmods, object am) rec->add_authmod (am, -1);
      return 1;
    }
    return 0;
  }

  object(ProtClass) *find_realm_classes()
  {
    if (!_realm_classes) {
      multiset(int) rcids = (<>);
      foreach (this_values(), object(ProtClass) pc) rcids[pc->realm_class()] = 1;
      _realm_classes = rows (this_object(), indices (rcids));
    }
    return _realm_classes;
  }

  void realm_class_has_changed() {_realm_classes = 0;}

  int remove (object(ProtClass) rec)
  {
    return Table::remove (rec, lambda (object(ProtClass) rec) {
				 db()->ppoints->is_removed (this_object(), rec);
				 if (search (find_realm_classes(), rec) >= 0)
				   this_values()->is_removed (this_object(), rec);
			       }, rec);
  }
}

class ProtPoints
{
  inherit Table;

  constant name = "ppoints";

  static mapping(string:object(ProtPoint)) _handles = ([]);
  // Records are rec->own_modname() + "\0" + rec->handle(): rec

  void create (object db) {Table::create (db, ProtPoint);}

  int add_default_pclass (object(ProtPoint) rec)
  {
    object(ProtClass) pc = db()->pclasses->new();
    if (db()->pclasses->add (pc)) {
      rec->set_pclass (pc);
      pc->set_realm (rec->name());
      return 1;
    }
    else
      throw (({"AClib: Couldn't create default pclass for ppoint ID " + rec->id() + "\n",
	       backtrace()}));
  }

  int add (object(ProtPoint) rec, mixed ... args)
  // Makes a new pclass for rec if it hasn't one set.
  {
    string hkey = rec->handle() ? rec->own_modname() + "\0" + rec->handle() : 0;
    if (hkey && zero_type (_handles[hkey]) != 1) {
      report_error ("AClib: A ppoint handle " + rec->handle() +
		    " already exist for module " + rec->own_modname() + "\n");
      return 0;
    }

    if (Table::add (rec, @args)) {
      if (!rec->pclass()) add_default_pclass (rec);
      if (hkey) _handles[hkey] = rec;
      return 1;
    }
    return 0;
  }

  object(ProtPoint) find_by_handle (object own_mod, string handle)
  {
    return _handles[db()->modulename (own_mod) + "\0" + handle];
  }

  object(ProtPoint) *find_by_pclass (object(ProtClass) pclass)
  // Not cached; use pclass->ppoints() instead.
  {
    object(ProtPoint) *pps = ({});
    foreach (this_values(), object(ProtPoint) pp)
      if (pp->pclass() == pclass) pps += ({pp});
    return pps;
  }

  object(ProtPoint) *find_by_ownmod (object own_mod)
  {
    object(ProtPoint) *pps = ({});
    foreach (this_values(), object(ProtPoint) pp)
      if (pp->own_module() == own_mod) pps += ({pp});
    return pps;
  }

  void update_handle_map (string oldname, object(ProtPoint) rec)
  {
    if (oldname) m_delete (_handles, rec->own_modname() + "\0" + oldname);
    if (rec->handle()) _handles[rec->own_modname() + "\0" + rec->handle()] = rec;
  }

  int remove (object(ProtPoint) rec)
  {
    string hkey = rec->handle() ? rec->own_modname() + "\0" + rec->handle() : 0;
    object(ProtClass) pc = rec->pclass();

    if (!Table::remove (rec, pc && pc->is_removed, this_object(), rec)) return 0;

    if (pc && !pc->real_name() && !sizeof (pc->ppoints()))
      // Remove containing pclass if it's empty and doesn't have an own name.
      db()->pclasses->remove (pc);

    if (hkey) m_delete (_handles, hkey);
    return 1;
  }
}


// Module interface.

class ModuleAC
// This is the interface for modules that uses protection points. It's
// designed for use with untrusted modules, so the available
// operations on the AC database are therefore limited. Every module
// gets its own object of this class. Due to the caching done here,
// the module must use the same single object during its lifetime. It
// is thread safe.
//
// Note: Protection points may be accessed either through ID ints or
// handle strings. A valid ID is never 0. Accessing through ID is
// somewhat faster, but handles are more convenient because they may
// be choosed freely. Thus, with systematic naming, the calling module
// doesn't have to keep track of them itself.
{
  // Implementation note: This class is severely bloated because lots
  // of code is duplicated in every function. It actually does very
  // little; simply wraps calls to the database inside handlers for
  // the ppoint cache and for the "loose" thread safe connection to
  // the AC database.

  GLOB_CONST;

#ifdef THREADS
  // Used to lock the _ppoints cache only.
  static inherit Thread.Mutex;
#endif

  static object _config = 0;
  // The configuration containing the AC database or 0 if global. Not
  // used if a get_db function is given to create().

  static function(void|int:object) _get_db =
#if constant (roxen)
    lambda (void|int silent) {
      mapping acvar = roxen->query_var ("AC");
      object acmod = acvar && acvar->loaders && acvar->loaders[_config];
      object acdb = acmod && acmod->online_acdb;
      if (!acdb && !silent && !acmod)
	// Do not show if the only problem is that there's no online
	// db; we're probably just not initialized yet.
	report_error ("ModuleAC: Couldn't connect to the AC database\n");
      return acdb;
    };
#else
    0;
#endif
  // The method used to find the AC_DB object.

  static object _module = 0;
  // The owning module. Note that this module is not necessarily a
  // member of _config.

  static object _db = 0;
  static int _last_db_nr = 0;
  // Cache of the active AC database and it's state sequence number.

  static string _misc_id_key, _misc_perm_key;
  // The keys in req_id->misc->common in which to store the resolved
  // identity object and the resolved permission, respectively. Set by
  // db_connect(); if one is set, both are set.

#define REQIDNR(req_id) ((req_id)->misc->common->ac_req_nr)
#define REQIDID(req_id) ((req_id)->misc->common[_misc_id_key])
#define REQIDPERM(req_id) ((req_id)->misc->common[_misc_perm_key])
#define NEWREQIDNR (++all_constants()->_mac_last_req_id_nr)

  class PPData
  {
    object(ProtPoint) pp;
    int id;
    string handle;
    int lastperm = PERM_NONE;
    int req_id_nr = -1;
    int idid = 0;
  }

  class FakeIdentity		// Used in place of Identity for unidentified requests.
  {
    constant fake = 1;
    int id() {return -1;}
  }

  static mapping(int|string:object(PPData)) _ppoints = ([]);
  // Cache of resolved ppoints belonging to this module. Each ppoint
  // is always indexed by ID and by handle if it has one. Beware that
  // the ppoint objects cached here may not have a table if the db
  // file has been reloaded.

  static int _all_ppoints = 0;  // True if _ppoints is known to contain all ppoints.

#ifdef MAC_DEBUG
#  if constant (cpp)
#    define MAC_DEBUG_MSG(fn, msg) debug_msg (#fn, (msg))
#  else
#    define MAC_DEBUG_MSG(fn, msg) debug_msg ("fn", (msg))
#  endif
  static void debug_msg (string fn, string msg)
  {
    report_debug ("ModuleAC:" +
		  (_misc_id_key ? _misc_id_key[10..] : "(acdb nr not resolved)") +
		  ": " + fn + " (" +
		  Array.map (backtrace()[-2][3..],
			     lambda (mixed arg) {
			       if (intp (arg)) return (string) arg;
			       else if (stringp (arg)) return "\"" + arg + "\"";
			       else return sprintf ("%t", arg);
			     }) * ", " +
			     "): " + msg + "\n");
  }
#else
#  define MAC_DEBUG_MSG(fn, msg)
#endif

  static object db_connect (void|int silent)
  {
    if (!_db)
      // _db may be destroyed from time to time (eg when the AC loader
      // module is reloaded).
      _db = _get_db (silent);

    if (_db) {
      int db_nr = _db->nr();
      if (db_nr != _last_db_nr) {
	_misc_id_key = "ac_req_id_" + db_nr;
	_misc_perm_key = "ac_req_perm_" + db_nr;
	_last_db_nr = db_nr;
      }
    }

    return _db;
  }

  static mixed db_begin (void|int silent)
  {
    if (!db_connect (silent)) return 0;

    _db->conn_mods |= ({_module});

#ifdef THREADS
    object dblock = _db->lock();
    _db->begin();
    return _db->initialized && dblock;
#else
    _db->begin();
    return _db->initialized;
#endif
  }

  static object(PPData) get_ppdata (int|string ppoint)
  {
    object(PPData) ppdata = _ppoints[ppoint];
    if (ppdata)
      if (ppdata->pp) return ppdata;
      else {
	m_delete (_ppoints, ppdata->id);
	if (ppdata->handle) m_delete (_ppoints, ppdata->handle);
      }

    return 0;
  }

  static object(PPData) make_ppdata (int|string ppoint)
  {
    if (!_db) return 0;

#ifdef MAC_DEBUG
    if (!_module)
      throw (({"ModuleAC: Owning module not set.\n", backtrace()}));
    if (!_db->modulename (_module))
      throw (({"ModuleAC: Couldn't find the module name of the owning object.\n",
	       backtrace()}));
#endif

    object(ProtPoint) pp;
    if (stringp (ppoint)) {
      pp = _db->ppoints->find_by_handle (_module, ppoint);
      if (!pp) return 0;
    }
    else {
      pp = _db->ppoints[ppoint];
      if (!pp) return 0;
      if (pp->own_module() != _module) {
	report_error ("ModuleAC: Module " + _db->modulename (_module) +
		      " tried to use ppoint " + ppoint + " belonging to " +
		      pp->own_modname() + "\n");
	return 0;
      }
    }

    object(PPData) ppdata = get_ppdata (ppoint) || PPData();
    ppdata->pp = pp;
    ppdata->id = pp->id();
    ppdata->handle = pp->handle();
    _ppoints[ppdata->id] = ppdata;
    if (ppdata->handle) _ppoints[ppdata->handle] = ppdata;
    return ppdata;
  }

  /*static*/ object _locking_thread;	// For internal use only!
  /*static*/ object _thislock, _dblock; // For internal use only!
  // Strict locking order: _thislock before _dblock.

  class ModuleAC_Lock {
    static object _mac, _db, _thislock, _dblock;
    void create (object mac, object db, object thislock, object dblock)
      {_mac = mac, _db = db, _thislock = thislock, _dblock = dblock;}
    void destroy()
    {
      if (_dblock) {
	if (_db) _db->commit();
	else report_warning ("ModuleAC: Internal error - "
			     "changes in AC database might be lost.\n");
	catch (destruct (_dblock));
      }
      catch (destruct (_thislock));
      catch (_mac->_locking_thread = 0);
    }
  }

  object(ModuleAC_Lock) lock_to_thread()
  // Locks this ModuleAC and the database to the current thread for
  // doing several operations. Returns 0 on error, ie if the db isn't
  // accessible, and a lock object otherwise.
  {
    object thislock;
#ifdef THREADS
    thislock = lock();
#else
    if (_thislock) throw (({"ModuleAC: Recursive lock\n", backtrace()}));
    thislock = class {}();
#endif

    mixed dblock;
    if (!(dblock = db_begin())) {
      MAC_DEBUG_MSG (lock_to_thread, "couldn't connect to ACDB");
      return 0;
    }

    _thislock = thislock;

#ifdef THREADS
    _dblock = dblock;
    _locking_thread = this_thread();
#else
    _dblock = class {}();
    _locking_thread = class {}();
#endif

    return ModuleAC_Lock (this_object(), _db, _thislock, _dblock);
  }

#ifdef THREADS

#  ifdef MAC_DEBUG
#    define THISLOCK(thislock) \
       object thislock = _locking_thread != this_thread() ? lock() : _thislock; \
       if (!thislock) throw (({"ModuleAC: _locking_thread set but not _thislock\n", \
			       backtrace()}))
#  else
#    define THISLOCK(thislock) \
       object thislock = _locking_thread != this_thread() ? lock() : _thislock
#  endif

#  define DBLOCK (_locking_thread != this_thread() ? db_begin() : _dblock)
#  define DBCOMMIT(_db) (_locking_thread != this_thread() && _db->commit())

#else

#  define THISLOCK(thislock)
#  define DBLOCK (_locking_thread ? _dblock : db_begin())
#  define DBCOMMIT(_db) (_locking_thread || _db->commit())

#endif

  void create (object config, object module,
	       void|function(void:object/*(AC_DB)*/) get_db)
  // config is the object of the configuration containing the AC
  // database. module is the module this object should serve. Note
  // that config is not necessarily the configuration containing the
  // module.
  //
  // A nonstandard method of getting the AC database may be given with
  // the get_db function. If specified, it should on call give the
  // AC_DB object to use, or 0 if it couldn't be found. The function
  // should work when arbitrary modules have been reloaded. If get_db
  // is specified, config isn't used for anything internally.
  {
    if (get_db) _get_db = get_db;
    else {
      if (!config)
	// Global AC is not in use yet.
	report_error ("ModuleAC: Can't create without configuration object.\n");
      _config = config;
    }

    if (!module)
      throw (({"ModuleAC: Can't create without module object.\n", backtrace()}));

    if (db_connect (1)) {
      if (search (_db->conn_mods, module) >= 0)
	throw (({"ModuleAC: Can't create duplicate for a module.\n", backtrace()}));
      _db->conn_mods += ({module});
    }

    _module = module;
  }

  void destroy()
  {
    if (db_connect (1))
      // This catch prevents a "Can't access global variables in
      // destructed object" bug which is hard to reproduce.
      catch (_db->conn_mods -= ({_module}));
  }

  object config() {return _config;}
  function(void:object) get_db() {return _get_db;}
  object module() {return _module;}

  int module_is_removed()
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) return 0;
    if (_module) _db->module_is_removed (_module);
    _db->commit();
    return !!_module;
  }

  // Handling protection points.

  int pp_create (void|string handle, void|string ppname,
		 void|string pcname, void|string realm)
  // Creates a new protection point and returns its ID or 0 if
  // something went wrong (a valid ID is always != 0). Note that
  // ppoints are always persistent, so remove_pp should be called when
  // it isn't used anymore. If a handle is given and there's a
  // matching ppoint, the ID of that one is returned and no new is
  // created. If a new ppoint was created, it's name is set to the
  // given one if it's specified. If a new pclass was created, it's
  // name and realm are set to the given ones if they're specified.
  {
    THISLOCK (thislock);
    mixed dblock = 0;

    if (stringp (handle)) {
      object(PPData) ppdata = get_ppdata (handle);
      if (!ppdata || !ppdata->pp->table()) {
	if (!(dblock = DBLOCK)) {
	  MAC_DEBUG_MSG (pp_create, "couldn't connect to ACDB");
	  return 0;
	}
	ppdata = make_ppdata (handle);
      }
      if (ppdata) {
	MAC_DEBUG_MSG (pp_create, "old ppoint " + ppdata->id + " exists");
	return ppdata->id;
      }
    }

    if (!dblock && !(dblock = DBLOCK)) {
      MAC_DEBUG_MSG (pp_create, "couldn't connect to ACDB");
      return 0;
    }
    object(ProtPoint) pp = ProtPoint (_db->ppoints, _module);

    if (stringp (handle)) pp->set_handle (handle);
    if (ppname) pp->set_name (ppname);
    _db->ppoints->add (pp);
    if (pcname) pp->pclass()->set_name (pcname);
    if (realm) pp->pclass()->set_realm (realm);
    _db->commit();

    object(PPData) ppdata = PPData();
    ppdata->pp = pp;
    ppdata->id = pp->id();
    ppdata->handle = handle;
    _ppoints[ppdata->id] = ppdata;
    if (handle) _ppoints[handle] = ppdata;

    MAC_DEBUG_MSG (pp_create, "new ppoint " + ppdata->id + " created");
    return ppdata->id;
  }

  int pp_clone (int|string ppoint, void|string handle,
		void|string ppname, void|object mac)
  // Creates a new protection point in a new protection class in the
  // same realm class and with same permissions and auth modules as
  // the given one. Returns the ID of the new ppoint or 0 if something
  // failed. ppname names the new ppoint and defaults to the same name
  // as the given ppoint. mac is the ModuleAC object in which ppoint
  // will be looked up and defaults to this object.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_clone, "couldn't connect to ACDB");
      return 0;
    }

    if (handle && _db->ppoints->find_by_handle (_module, handle)) {
      MAC_DEBUG_MSG (pp_clone, "old ppoint exists");
      return 0;
    }

    object(ProtPoint) srcpp = 0;

    if (mac)
      if (stringp (ppoint))
	srcpp = _db->ppoints->find_by_handle (mac->module(), ppoint);
      else {
	srcpp = _db->ppoints[ppoint];
	if (srcpp && srcpp->own_module() != mac->module()) srcpp = 0;
      }
    else {
      object(PPData) ppdata = get_ppdata (ppoint);
      if (ppdata && ppdata->pp->table() || (ppdata = make_ppdata (ppoint)))
	srcpp = ppdata->pp;
    }

    if (!srcpp) {
      MAC_DEBUG_MSG (pp_clone, "unknown ppoint");
      return 0;
    }

    object(ProtPoint) pp = ProtPoint (_db->ppoints, _module);

    if (handle) pp->set_handle (handle);
    pp->set_name (ppname ? ppname : srcpp->name());
    _db->ppoints->add (pp);
    pp->pclass()->set_realm_class (srcpp->pclass()->realm_class());
    pp->pclass()->set_perms (srcpp->pclass()->perms());
    pp->pclass()->set_authmod_names (srcpp->pclass()->authmod_names());
    _db->commit();

    object(PPData) ppdata = PPData();
    ppdata->pp = pp;
    ppdata->id = pp->id();
    ppdata->handle = handle;
    _ppoints[ppdata->id] = ppdata;
    if (handle) _ppoints[handle] = ppdata;

    MAC_DEBUG_MSG (pp_clone, "new ppoint " + ppdata->id + " created");
    return ppdata->id;
  }

  int pp_remove (int|string ppoint)
  // Removes the protection point. Returns true if successful.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_remove, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (pp_remove, "unknown ppoint");
      return 0;
    }

    _db->ppoints->remove (ppdata->pp);
    destruct (ppdata->pp);
    _db->commit();

    m_delete (_ppoints, ppdata->id);
    if (ppdata->handle) m_delete (_ppoints, ppdata->handle);

    MAC_DEBUG_MSG (pp_remove, "ppoint " + ppdata->id + " removed");
    return 1;
  }

  int pp_get_id (int|string ppoint)
  // Returns the ID of the protection point if it exists, 0 otherwise.
  // Accesses the AC database if necessary.
  {
    THISLOCK (thislock);

    object(PPData) ppdata = get_ppdata (ppoint);
    if (!ppdata || !ppdata->pp->table()) {
      mixed dblock = DBLOCK;
      if (!dblock) {
	MAC_DEBUG_MSG (pp_get_id, "couldn't connect to ACDB");
	return 0;
      }

      ppdata = make_ppdata (ppoint);
      return ppdata ? ppdata->id : 0;
    }
    else return ppdata->id;
  }

  string pp_get_handle (int|string ppoint)
  // Returns the handle of the protection point if it exists and has
  // one, 0 otherwise. Accesses the AC database if necessary.
  {
    THISLOCK (thislock);

    object(PPData) ppdata = get_ppdata (ppoint);
    if (!ppdata || !ppdata->pp->table()) {
      mixed dblock = DBLOCK;
      if (!dblock) {
	MAC_DEBUG_MSG (pp_get_handle, "couldn't connect to ACDB");
	return 0;
      }

      ppdata = make_ppdata (ppoint);
      if (ppdata) return ppdata->handle;
      else {
	MAC_DEBUG_MSG (pp_get_handle, "ppoint not found");
	return 0;
      }
    }
    else return ppdata->handle;
  }

  int pp_set_handle (int|string ppoint, string newhandle)
  // Returns true if successful. The handle is deleted if newhandle is
  // 0.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_set_handle, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (pp_set_handle, "unknown ppoint");
      return 0;
    }

    if (newhandle &&
	_db->ppoints->find_by_handle (_module, newhandle)) {
      MAC_DEBUG_MSG (pp_set_handle, "handle already taken");
      return 0;
    }

    MAC_DEBUG_MSG (pp_set_handle, "changed from " + ppdata->handle);
    ppdata->pp->set_handle (newhandle);
    _db->commit();

    if (ppdata->handle) m_delete (_ppoints, ppdata->handle);
    ppdata->handle = newhandle;
    if (newhandle) _ppoints[newhandle] = ppdata;

    return 1;
  }

  int pp_get_deletable (int|string ppoint)
  // Returns the deletable flag of the protection point if it exists,
  // 0 otherwise. Accesses the AC database.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_get_deletable, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if (!ppdata || !ppdata->pp->table()) ppdata = make_ppdata (ppoint);

    if (ppdata) return ppdata->pp->deletable();
    else {
      MAC_DEBUG_MSG (pp_get_deletable, "ppoint not found");
      return 0;
    }
  }

  int pp_set_deletable (int|string ppoint, int deletable)
  // Returns true if successful.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_set_deletable, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (pp_set_deletable, "unknown ppoint");
      return 0;
    }

    MAC_DEBUG_MSG (pp_set_deletable, "changed from " + ppdata->pp->deletable());
    ppdata->pp->set_deletable (deletable);
    _db->commit();
    return 1;
  }

  string pp_get_name (int|string ppoint)
  // Returns the unparsed name of the protection point if it exists
  // and has one, 0 otherwise. Accesses the AC database.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_get_name, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if (!ppdata || !ppdata->pp->table()) ppdata = make_ppdata (ppoint);

    if (ppdata) return ppdata->pp->real_name();
    else {
      MAC_DEBUG_MSG (pp_get_name, "ppoint not found");
      return 0;
    }
  }

  int pp_set_name (int|string ppoint, string name)
  // Returns true if successful.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_set_name, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (pp_set_name, "unknown ppoint");
      return 0;
    }

    MAC_DEBUG_MSG (pp_set_name, "changed from " + ppdata->pp->real_name());
    ppdata->pp->set_name (name);
    _db->commit();
    return 1;
  }

  mixed pp_get_data (int|string ppoint)
  // Returns the arbitrary data stored in the protection point or 0 if
  // it doesn't exist or there isn't any data in it. Accesses the AC
  // database if necessary. It's not threadsafe to change this
  // destructively. If changed, pp_set_data() must be called
  // afterwards.
  {
    THISLOCK (thislock);

    object(PPData) ppdata = get_ppdata (ppoint);
    if (!ppdata || !ppdata->pp->table()) {
      mixed dblock = DBLOCK;
      if (!dblock) {
	MAC_DEBUG_MSG (pp_get_data, "couldn't connect to ACDB");
	return 0;
      }

      ppdata = make_ppdata (ppoint);
      if (ppdata) return ppdata->pp->data();
      else {
	MAC_DEBUG_MSG (pp_get_data, "ppoint not found");
	return 0;
      }
    }
    else return ppdata->pp->data();
  }

  int pp_set_data (int|string ppoint, mixed data)
  // Stores the given data in the protection point for later access.
  // The data may be anything except objects, functions or programs.
  // Returns true if successful.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_set_data, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (pp_set_data, "unknown ppoint");
      return 0;
    }

    MAC_DEBUG_MSG (pp_set_data, "ok");
    ppdata->pp->set_data (data);
    _db->commit();

    return 1;
  }

  int *pp_get_ids()
  // Returns the IDs of all protection points belonging to this module
  // or 0 if there are problems accessing the AC database.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_get_ids, "couldn't connect to ACDB");
      return 0;
    }

#ifdef MAC_DEBUG
    if (!_module)
      throw (({"ModuleAC: Owning module not set.\n", backtrace()}));
    if (!_db->modulename (_module))
      throw (({"ModuleAC: Couldn't find the module name of the owning object.\n",
	       backtrace()}));
#endif

    if (_all_ppoints) {
      int *ppidxs = ({});
      foreach (indices (_ppoints), int|string ppidx)
	if (intp (ppidx) && _ppoints[ppidx]->pp) ppidxs += ({ppidx});
      return ppidxs;
    }

    object(ProtPoint) *pps = _db->ppoints->find_by_ownmod (_module);
    foreach (pps, object(ProtPoint) pp)
      if (!_ppoints[pp->id()]) {
	object(PPData) ppdata = PPData();
	ppdata->pp = pp;
	ppdata->id = pp->id();
	ppdata->handle = pp->handle();
	_ppoints[ppdata->id] = ppdata;
	if (ppdata->handle) _ppoints[ppdata->handle] = ppdata;
      }
    _all_ppoints = 1;

    return pps->id();
  }

  string *pp_get_handles()
  // Returns the handles of all protection points belonging to this
  // module that has such or 0 if there are problems accessing the AC
  // database.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pp_get_handles, "couldn't connect to ACDB");
      return 0;
    }

#ifdef MAC_DEBUG
    if (!_module)
      throw (({"ModuleAC: Owning module not set.\n", backtrace()}));
    if (!_db->modulename (_module))
      throw (({"ModuleAC: Couldn't find the module name of the owning object.\n",
	       backtrace()}));
#endif

    if (_all_ppoints) {
      string *ppidxs = ({});
      foreach (indices (_ppoints), int|string ppidx)
	if (stringp (ppidx) && _ppoints[ppidx]->pp) ppidxs += ({ppidx});
      return ppidxs;
    }

    object(ProtPoint) *pps = _db->ppoints->find_by_ownmod (_module);
    foreach (pps, object(ProtPoint) pp)
      if (!_ppoints[pp->id()]) {
	object(PPData) ppdata = PPData();
	ppdata->pp = pp;
	ppdata->id = pp->id();
	ppdata->handle = pp->handle();
	_ppoints[ppdata->id] = ppdata;
	if (ppdata->handle) _ppoints[ppdata->handle] = ppdata;
      }
    _all_ppoints = 1;

    return pps->handle() - ({0});
  }

  // Accessing protection classes.

  string pc_get_name (int|string ppoint)
  // Returns the name of the protection class of ppoint. This is 0 if
  // the pclass uses the parsed name of the ppoint.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pc_get_name, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if (!ppdata || !ppdata->pp->table()) ppdata = make_ppdata (ppoint);

    if (ppdata) return ppdata->pp->pclass()->real_name();
    else {
      MAC_DEBUG_MSG (pc_get_name, "ppoint not found");
      return 0;
    }
  }

  int pc_set_name (int|string ppoint, string name)
  // Returns true if successful. If the name is 0 and the pclass
  // contains only one ppoint, the pclass will use the dynamically
  // parsed name of the ppoint.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pc_set_name, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (pc_set_name, "unknown ppoint");
      return 0;
    }

    MAC_DEBUG_MSG (pc_set_name, "changed from " + ppdata->pp->pclass()->real_name());
    ppdata->pp->pclass()->set_name (name);
    _db->commit();
    return 1;
  }

  string pc_get_realm (int|string ppoint)
  // Returns the realm of the protection class of ppoint.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pc_get_realm, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if (!ppdata || !ppdata->pp->table()) ppdata = make_ppdata (ppoint);

    if (ppdata) return ppdata->pp->pclass()->realm();
    else {
      MAC_DEBUG_MSG (pc_get_realm, "ppoint not found");
      return 0;
    }
  }

  int pc_set_realm (int|string ppoint, string realm)
  // Returns true if successful. The protection class will get its own
  // realm class.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pc_set_realm, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (pc_set_realm, "unknown ppoint");
      return 0;
    }

    object(ProtClass) pc = ppdata->pp->pclass();
    MAC_DEBUG_MSG (pc_set_realm, "changed from " +
		   (pc->realm_class() != pc->id() ? "realm class " + pc->realm_class() :
		    pc->realm()));
    pc->set_realm (realm);
    _db->commit();
    return 1;
  }

  int pc_get_realm_class (int|string ppoint)
  // Returns the realm class of the protection class of ppoint. All
  // pclasses that share the same realm will return the same unique
  // number, which is never 0.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pc_get_realm_class, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if (!ppdata || !ppdata->pp->table()) ppdata = make_ppdata (ppoint);

    if (ppdata) return ppdata->pp->pclass()->realm_class();
    else {
      MAC_DEBUG_MSG (pc_get_realm_class, "ppoint not found");
      return 0;
    }
  }

  int pc_set_realm_class (int|string ppoint, int realm_class)
  // Returns true if successful. It's only guaranteed to work with
  // numbers returned from pc_get_realm_class().
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (pc_set_realm_class, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (pc_set_realm_class, "unknown ppoint");
      return 0;
    }

    object(ProtClass) pc = ppdata->pp->pclass();
    MAC_DEBUG_MSG (pc_set_realm_class, "changed from " +
		   (pc->realm_class() != pc->id() ? "realm class " + pc->realm_class() :
		    pc->realm()));
    pc->set_realm_class (realm_class);
    _db->commit();
    return 1;
  }

  // Testing permissions.

  int access_try (int|string ppoint, object req_id, void|int force)
  // Authenticates the request req_id on the given protection point
  // and returns the permission. Does passive authentication only if
  // necessary, i.e. if the identity and permission hasn't been
  // resolved before for this request on this ppoint. Always does
  // authentication if force is nonzero. If there's an error,
  // PERM_NONE is returned.
  {
    THISLOCK (thislock);

#ifdef MAC_DEBUG
    if (!req_id)
      throw (({"ModuleAC: Didn't get a request ID object\n", backtrace()}));
#endif

    if (!req_id->misc->common) req_id->misc->common = ([]);

    object(PPData) ppdata = get_ppdata (ppoint);
    if (ppdata) {
      object(Identity)|object(FakeIdentity) id;
      if (!force && ppdata->req_id_nr == REQIDNR (req_id) &&
	  _misc_id_key && (id = REQIDID (req_id)) && ppdata->idid == id->id()) {
	MAC_DEBUG_MSG (access_try,
		       "[req:" + ppdata->req_id_nr + "] " +
		       (id->db ?
			"identity " + (id->handle() || id->name() || id->id()) :
			"unidentified request") +
		       " cached perm " + PERM_DESC[ppdata->lastperm]);
	return REQIDPERM (req_id) = ppdata->lastperm;
      }
    }

    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (access_try, "couldn't connect to ACDB");
      if (_misc_perm_key) REQIDPERM (req_id) = PERM_NONE;
      return PERM_NONE;
    }

    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (access_try, "unknown ppoint");
      if (_misc_perm_key) REQIDPERM (req_id) = PERM_NONE;
      return PERM_NONE;
    }

    ppdata->req_id_nr = REQIDNR (req_id) || (REQIDNR (req_id) = NEWREQIDNR);
    object(Identity) id = REQIDID (req_id) = _db->passive_auth (ppdata->pp, req_id);
    ppdata->idid = id->id();
    ppdata->lastperm = REQIDPERM (req_id) = _db->perm (id, ppdata->pp);
    MAC_DEBUG_MSG (access_try,
		   "[req:" + ppdata->req_id_nr + "] resolved identity " +
		   (id->handle() || id->name() || id->id()) +
		   " perm " + PERM_DESC[ppdata->lastperm]);

    return ppdata->lastperm;
  }

  mixed access_throw (int|string ppoint, object req_id,
		      void|string msg, void|string realm)
  // Should be called if the request doesn't have sufficient
  // permission for a protection point. Returns a server reply
  // appropriate for the current protocol that should be passed on.
  // Does active authentication if applicable. The message string, if
  // given, is included in the reply as the reason for the failure (if
  // the auth method supports it). If realm is given, it overrides the
  // realm for the protection point. 0 is returned if there's an error
  // or no appropriate reply could be constructed.
  {
    THISLOCK (thislock);

#ifdef MAC_DEBUG
    if (!req_id)
      throw (({"ModuleAC: Didn't get a request ID object\n", backtrace()}));
#endif

    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (access_throw, "couldn't connect to ACDB");
      return 0;
    }

    object(PPData) ppdata = get_ppdata (ppoint);
    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (access_throw, "unknown ppoint");
      return 0;
    }

    mixed res = _db->active_auth (ppdata->pp, req_id, msg, realm);
    MAC_DEBUG_MSG (access_throw, res ? "ok" : "returned 0");
    return res;
  }

  int access_perm (int|string ppoint, object req_id)
  // Returns the permission for the identity the request req_id is
  // currently authenticated as on the given protection point. This
  // differs from access_try() in that this function does not try to
  // resolve the identity for ppoint. You normally want to use
  // access_try() when you're checking the permission on a resource
  // you're about to return; this function is mainly useful to do
  // auxiliary tests on other ppoints once the identity has been
  // established. If there's an error, PERM_NONE is returned.
  {
    THISLOCK (thislock);

#ifdef MAC_DEBUG
    if (!req_id)
      throw (({"ModuleAC: Didn't get a request ID object\n", backtrace()}));
#endif

    if (!req_id->misc->common) req_id->misc->common = ([]);

    object(PPData) ppdata = get_ppdata (ppoint);
    if (ppdata) {
      object(Identity)|object(FakeIdentity) id;
      if (ppdata->req_id_nr == REQIDNR (req_id) &&
	  _misc_id_key && (id = REQIDID (req_id)) && ppdata->idid == id->id()) {
	MAC_DEBUG_MSG (access_perm,
		       "[req:" + ppdata->req_id_nr + "] " +
		       (id->db ?
			"identity " + (id->handle() || id->name() || id->id()) :
			"unidentified request") +
		       " cached perm " + PERM_DESC[ppdata->lastperm]);
	return ppdata->lastperm;
      }
    }

    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (access_perm, "couldn't connect to ACDB");
      return PERM_NONE;
    }

    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (access_perm, "unknown ppoint");
      return PERM_NONE;
    }

    ppdata->req_id_nr = REQIDNR (req_id) || (REQIDNR (req_id) = NEWREQIDNR);

    object(Identity)|object(FakeIdentity) id = REQIDID (req_id);
    if (id && id->db) {
      ppdata->idid = id->id();
      if (!id->table()) id = _db->identities[ppdata->idid];
    }
    else id = 0;
    if (id) {
      ppdata->lastperm = _db->perm (id, ppdata->pp);
      MAC_DEBUG_MSG (access_perm,
		     "[req:" + ppdata->req_id_nr + "] identity " +
		     (id->handle() || id->name() || id->id()) +
		     " perm " + PERM_DESC[ppdata->lastperm]);
    }
    else {
      id = REQIDID (req_id) = FakeIdentity();
      ppdata->idid = id->id();
      ppdata->lastperm = PERM_NONE;
      MAC_DEBUG_MSG (access_perm,
		     "[req:" + ppdata->req_id_nr + "] unidentified request");
    }

    return ppdata->lastperm;
  }

  int access_chain_try (int|string ppoint, object req_id, void|int force)
  // Does chained authenticaton of the request req_id on the given
  // protection point and returns the permission. It works like
  // access_try(), but if the request already has a resolved identity
  // it won't be checked again, instead the permission for that
  // identity on the given ppoint will be returned (similar to
  // access_perm()). The returned permission is never more privileged
  // than the previously resolved one for the request (regardless of
  // what ppoint that permission was resolved for). This feature is
  // used to "chain" several ppoints protecting the same resource,
  // something that generally should be avoided, because it introduces
  // hidden dependencies, is often unnecessary and slower. However,
  // it's needed in some special situations, e.g. when you have two
  // independent sets of ppoints whose conjunction specifies the
  // permissions. So there!
  {
    THISLOCK (thislock);

#ifdef MAC_DEBUG
    if (!req_id)
      throw (({"ModuleAC: Didn't get a request ID object\n", backtrace()}));
#endif

    if (!req_id->misc->common) req_id->misc->common = ([]);

    object(PPData) ppdata = get_ppdata (ppoint);
    if (ppdata) {
      object(Identity)|object(FakeIdentity) id;
      if (!force && ppdata->req_id_nr == REQIDNR (req_id) &&
	  _misc_id_key && (id = REQIDID (req_id)) && ppdata->idid == id->id()) {
	int perm = min (REQIDPERM (req_id), ppdata->lastperm);
	MAC_DEBUG_MSG (access_chain_try,
		       "[req:" + ppdata->req_id_nr + "] " +
		       (id->db ?
			"identity " + (id->handle() || id->name() || id->id()) :
			"unidentified request") +
		       " cached perm " + PERM_DESC[ppdata->lastperm] +
		       " chained from " + PERM_DESC[REQIDPERM (req_id)] +
		       " to " + PERM_DESC[perm]);
	return REQIDPERM (req_id) = perm;
      }
    }

    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (access_chain_try, "couldn't connect to ACDB");
      if (_misc_perm_key) REQIDPERM (req_id) = PERM_NONE;
      return PERM_NONE;
    }

    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (access_chain_try, "unknown ppoint");
      if (_misc_perm_key) REQIDPERM (req_id) = PERM_NONE;
      return PERM_NONE;
    }

    ppdata->req_id_nr = REQIDNR (req_id) || (REQIDNR (req_id) = NEWREQIDNR);

    int perm;
    object(Identity)|object(FakeIdentity) id = REQIDID (req_id);
    if (id && id->db) {
      ppdata->idid = id->id();
      if (!id->table()) id = _db->identities[ppdata->idid];
    }
    else id = 0;
    if (id) {
      ppdata->lastperm = _db->perm (id, ppdata->pp);
      perm = min (REQIDPERM (req_id), ppdata->lastperm);
      MAC_DEBUG_MSG (access_chain_try,
		     "[req:" + ppdata->req_id_nr + "] identity " +
		     (id->handle() || id->name() || id->id()) +
		     " perm " + PERM_DESC[ppdata->lastperm] +
		     " chained from " + PERM_DESC[REQIDPERM (req_id)] +
		     " to " + PERM_DESC[perm]);
      REQIDPERM (req_id) = perm;
    }
    else {
      id = REQIDID (req_id) = _db->passive_auth (ppdata->pp, req_id);
      ppdata->idid = id->id();
      ppdata->lastperm = REQIDPERM (req_id) = perm = _db->perm (id, ppdata->pp);
      MAC_DEBUG_MSG (access_chain_try,
		     "[req:" + ppdata->req_id_nr + "] resolved identity " +
		     (id->handle() || id->name() || id->id()) +
		     " perm " + PERM_DESC[perm]);
    }
    return perm;
  }

  int access_chain_perm (int|string ppoint, object req_id)
  // The chained (see access_chain_try()) version of access_perm(). It
  // differs from access_perm() in that it takes the currently
  // resolved permission for the request into account, just like
  // access_chain_try() does, and returns the least privileged of the
  // two permissions. The main difference from access_chain_try() is
  // that it doesn't permanently lower the permission for the request.
  {
    THISLOCK (thislock);

#ifdef MAC_DEBUG
    if (!req_id)
      throw (({"ModuleAC: Didn't get a request ID object\n", backtrace()}));
#endif

    if (!req_id->misc->common) req_id->misc->common = ([]);

    object(PPData) ppdata = get_ppdata (ppoint);
    if (ppdata) {
      object(Identity)|object(FakeIdentity) id;
      if (ppdata->req_id_nr == REQIDNR (req_id) &&
	  _misc_id_key && (id = REQIDID (req_id)) && ppdata->idid == id->id()) {
	int perm = min (REQIDPERM (req_id), ppdata->lastperm);
	MAC_DEBUG_MSG (access_chain_perm,
		       "[req:" + ppdata->req_id_nr + "] " +
		       (id->db ?
			"identity " + (id->handle() || id->name() || id->id()) :
			"unidentified request") +
		       " cached perm " + PERM_DESC[ppdata->lastperm] +
		       " chained from " + PERM_DESC[REQIDPERM (req_id)] +
		       " to " + PERM_DESC[perm]);
	return perm;
      }
    }

    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (access_chain_perm, "couldn't connect to ACDB");
      return PERM_NONE;
    }

    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (access_chain_perm, "unknown ppoint");
      return PERM_NONE;
    }

    ppdata->req_id_nr = REQIDNR (req_id) || (REQIDNR (req_id) = NEWREQIDNR);

    int perm;
    object(Identity)|object(FakeIdentity) id = REQIDID (req_id);
    if (id && id->db) {
      ppdata->idid = id->id();
      if (!id->table()) id = _db->identities[ppdata->idid];
    }
    else id = 0;
    if (id) {
      ppdata->lastperm = _db->perm (id, ppdata->pp);
      perm = min (REQIDPERM (req_id), ppdata->lastperm);
      MAC_DEBUG_MSG (access_chain_perm,
		     "[req:" + ppdata->req_id_nr + "] identity " +
		     (id->handle() || id->name() || id->id()) +
		     " perm " + PERM_DESC[ppdata->lastperm] +
		     " chained from " + PERM_DESC[REQIDPERM (req_id)] +
		     " to " + PERM_DESC[perm]);
    }
    else {
      id = REQIDID (req_id) = FakeIdentity();
      ppdata->idid = id->id();
      ppdata->lastperm = perm = PERM_NONE;
      MAC_DEBUG_MSG (access_chain_perm,
		     "[req:" + ppdata->req_id_nr + "] unidentified request");
    }
    return perm;
  }

  void access_set (int|string ppoint, object req_id, int perm)
  // Fakes the permission for the request req_id on the given
  // protection point in the cache. As long as access_try(),
  // access_perm() and access_chain_try() only consults the cache,
  // they will return perm.
  {
    THISLOCK (thislock);

#ifdef MAC_DEBUG
    if (!req_id)
      throw (({"ModuleAC: Didn't get a request ID object\n", backtrace()}));
#endif

    object(PPData) ppdata = get_ppdata (ppoint);

    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (access_set, "couldn't connect to ACDB");
      return;
    }

    if ((!ppdata || !ppdata->pp->table()) &&
	!(ppdata = make_ppdata (ppoint))) {
      MAC_DEBUG_MSG (access_set, "unknown ppoint");
      return;
    }

    if (!req_id->misc->common) req_id->misc->common = ([]);

    ppdata->req_id_nr = REQIDNR (req_id) || (REQIDNR (req_id) = NEWREQIDNR);

    object(Identity)|object(FakeIdentity) id = REQIDID (req_id);
    if (id && id->db) {
      ppdata->idid = id->id();
      MAC_DEBUG_MSG (access_set,
		     "[req:" + ppdata->req_id_nr + "] identity " +
		     (id->handle() || id->name() || id->id()) +
		     (ppdata->req_id_nr == REQIDNR (req_id) ?
		      " changed from " + PERM_DESC[ppdata->lastperm] + " to " :
		      " set to ") +
		     PERM_DESC[perm]);
    }
    else {
      id = REQIDID (req_id) = FakeIdentity();
      ppdata->idid = id->id();
      MAC_DEBUG_MSG (access_set,
		     "[req:" + ppdata->req_id_nr + "] unidentified request" +
		     (ppdata->req_id_nr == REQIDNR (req_id) ?
		      " changed from " + PERM_DESC[ppdata->lastperm] + " to " :
		      " set to ") +
		     PERM_DESC[perm]);
    }

    ppdata->lastperm = REQIDPERM (req_id) = perm;
  }

  // Accessing identities.

  int id_create (string name, int type, void|string handle,
		 void|mapping perms, void|string cryptwd,
		 void|multiset(int) members)
  // Adds a new identity to the database. If handle is given and an
  // identity already exists with that handle, nothing is changed.
  // perms specifies the initial permissions for the identity. It's
  // indexed on either ppoint handles/IDs or array tuples of ppoint
  // handles/IDs and the ModuleAC objects they belong to, and has the
  // permissions as values. If cryptwd is given, it's used as the
  // secret in an AUTH_PASSWD_CRYPT record in a newly created IdAuth
  // object belonging to an auth module that supports that method
  // (typically acintern_passwd). If members is given, this identity
  // must be a group. All identities whose IDs are in members become
  // members of this group. Returns 0 if there was an error accessing
  // the database, the ID of the identity otherwise (regardless
  // whether one was created or not).
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (id_create, "couldn't connect to ACDB");
      return 0;
    }

    object(Identity) id = handle && _db->identities->find_by_handle (handle);
    if (id) return id->id();

    if (members && type != ID_TYPE_GROUP)
      throw (({"Can't add members to a non-group\n", backtrace()}));

    object(Identity) id = _db->identities->new (type);
    id->set_name (name);
    id->set_handle (handle);
    _db->identities->add (id);
    _db->identities->init_auth_methods (id);
    MAC_DEBUG_MSG (id_create, "new identity " + id->id() + " created");

    if (perms)
      foreach (indices (perms), int|string|array ppoint) {
	object(ProtPoint) pp;
	int perm = perms[ppoint];
	if (arrayp (ppoint)) {
	  object mac = ppoint[1];
	  ppoint = ppoint[0];
	  if (stringp (ppoint))
	    pp = _db->ppoints->find_by_handle (mac->module(), ppoint);
	  else {
	    pp = _db->ppoints[ppoint];
	    if (pp && pp->own_module() != mac->module()) pp = 0;
	  }
	}
	else {
	  object(PPData) ppdata = get_ppdata (ppoint);
	  if (ppdata && ppdata->pp->table() || (ppdata = make_ppdata (ppoint)))
	    pp = ppdata->pp;
	}
	if (pp) {
	  pp->pclass()->set_perm (id, perm);
	  MAC_DEBUG_MSG (id_create, "set " + PERM_DESC[perm] + " permission on "
			 "pclass " + pp->pclass()->id());
	}
	else MAC_DEBUG_MSG (id_create, "unknown ppoint " + ppoint);
      }

    if (cryptwd) {
      // The trickery here to get to the authmod to set the cryptwd is
      // disgusting.. :P
      int ok = 0;
      foreach (id->auths(), object(IdAuth) ia) {
	object authmod = ia->auth_module();
	if (authmod && authmod->id_auth_set_cryptwd) {
	  authmod->id_auth_set_cryptwd (ia, cryptwd);
	  MAC_DEBUG_MSG (id_create, "set password in existing IdAuth " + ia->id());
	  ok = 1;
	  break;
	}
      }
      if (!ok)
	foreach (values (_db->identities->init_authmods) * ({}), object authmod)
	  if (authmod) {
	    object ia =
	      _db->try_call (authmod, "id_auth_init", _db->id_auths->new (authmod, id));
	    if (ia && !zero_type (_db->try_call (authmod, "id_auth_set_cryptwd",
						 ia, cryptwd))) {
	      _db->id_auths->add (ia);
	      MAC_DEBUG_MSG (id_create, "set password in added IdAuth " + ia->id());
	    }
	  }
    }

    if (members)
      foreach (indices (members), int idid) {
	object memid = _db->identities[idid];
	if (memid) memid->add_incl_group (id);
	else MAC_DEBUG_MSG (id_create, "no identity " + idid + " to add as member");
      }

    _db->commit();
    return id->id();
  }

  // These functions access the identity the given request is
  // currently authenticated as, or any identity specified by record
  // ID or handle. When used with a request id it means that, with the
  // exception of id_is_known(), they should only be used after
  // access_try() or access_chain_try().

  int id_is_known (object req_id)
  // Returns nonzero iff the request has an authenticated identity.
  {
    if (!_misc_id_key && !db_connect()) {
      MAC_DEBUG_MSG (id_is_known, "couldn't connect to ACDB");
      return 0;
    }

    if (!req_id->misc->common) req_id->misc->common = ([]);

    object id = REQIDID (req_id);
    return objectp (id) && !id->fake;
  }

#define ID_FROM_IDSPEC(id, id_spec, thislockstat, dblock, errval, func) \
  if (objectp (id_spec)) { \
    if (!_misc_id_key && !db_connect()) { \
      MAC_DEBUG_MSG (func, "couldn't connect to ACDB"); \
      return errval; \
    } \
    if (!(id_spec)->misc->common) (id_spec)->misc->common = ([]); \
    id = REQIDID (id_spec); \
    if (!objectp (id) || id->fake) { \
      MAC_DEBUG_MSG (func, "unidentified request"); \
      return errval; \
    } \
  } \
  else { \
    thislockstat; \
    dblock = DBLOCK; \
    if (!dblock) { \
      MAC_DEBUG_MSG (func, "couldn't connect to ACDB"); \
      return errval; \
    } \
    id = stringp (id_spec) ? \
      _db->identities->find_by_handle (id_spec) : _db->identities[id_spec]; \
    if (!id) return errval; \
  }

  int id_get_id (object|int|string id_spec)
  // Returns the record ID of the identity, or 0 on error.
  {
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, 0, id_get_id);

    // Access without locking possible because it's atomic.
    return id->id();
  }

  int id_is_everyone (object|int|string id_spec)
  // Returns nonzero iff the identity is the special everyone identity.
  {
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, 0, id_get_id);

    // Access without locking possible because it's atomic.
    return id == id->table()->everyone();
  }

  int id_is_privileged (object|int|string id_spec)
  // Returns nonzero iff the identity is the special privileged identity.
  {
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, 0, id_get_id);

    // Access without locking possible because it's atomic.
    return id == id->table()->privileged();
  }

  string id_get_name (object|int|string id_spec)
  // Returns the name of the identity (i.e. the fullname of the user),
  // or 0 on error.
  {
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, 0, id_get_name);

    // Access without locking possible because it's atomic.
    return id->name();
  }

  int id_get_type (object|int|string id_spec)
  // Returns the type of the identity, or ID_TYPE_N_A on error.
  {
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, ID_TYPE_N_A, id_get_type);

    // Access without locking possible because it's atomic.
    return id->type();
  }

  string id_get_handle (object|int|string id_spec)
  // Returns the handle of the identity (i.e. the username if any), or
  // 0 on error.
  {
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, 0, id_get_handle);

    // Access without locking possible because it's atomic.
    return id->handle();
  }

  array(int) id_get_memberships (object|int|string id_spec)
  // Returns the IDs of all (direct or indirect) memberships of the
  // given identity.
  {
    THISLOCK (thislock);
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, 0, id_get_memberships);

    return id && Array.map (indices (id->groups()), "id");
  }

  array(int) id_get_members (object|int|string id_spec)
  // Returns the IDs of all (direct or indirect) members of the given
  // group. Returns the empty array for users.
  {
    THISLOCK (thislock);
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, 0, dblock, 0, id_get_members);

    return id && Array.map (indices (id->members()), "id");
  }

  mixed id_get_data (object|int|string id_spec)
  // Returns the arbitrary data stored in the identity by this module,
  // or 0 on error. It's not thread safe to change this destructively.
  // If changed, id_set_data() must be called afterwards.
  {
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, 0, id_get_data);
    if (!id->db && !(id = _db->identities[id->id()])) {
      MAC_DEBUG_MSG (id_get_data, "identity has been deleted");
      return 0;
    }

    // Access without locking possible because it's atomic (for us
    // anyway; further locking is up to the caller).
    return id->data (_module);
  }

  int id_set_data (mixed data, object|int|string id_spec)
  // Stores the given data in the identity for later access by this
  // module. The data may be anything except objects, functions or
  // programs. Returns 0 if the data couldn't be saved for some reason
  // (e.g. the identity doesn't exist any more).
  {
    object id;
    mixed dblock;
    ID_FROM_IDSPEC (id, id_spec, THISLOCK (thislock), dblock, 0, id_set_data);
    if (!id->db && !(id = _db->identities[id->id()])) {
      MAC_DEBUG_MSG (id_get_data, "identity has been deleted");
      return 0;
    }

    if (!id->set_data (_module, data)) {
      MAC_DEBUG_MSG (id_set_data, "failed");
      return 0;
    }
    _db->commit();

    MAC_DEBUG_MSG (id_set_data, "ok");
    return 1;
  }

  array(int) id_get_list (int type)
  // Returns all identities of the specified type.
  {
    THISLOCK (thislock);
    mixed dblock = DBLOCK;
    if (!dblock) {
      MAC_DEBUG_MSG (id_get_list, "couldn't connect to ACDB");
      return 0;
    }

    return column (_db->identities->find_by_type (type), "id")();
  }
}

class ModuleACDB
// This is obsolete; its functionality is implemented directly in
// ModuleAC now.
{
  inherit ModuleAC;
}


// The AC database wrapped together.

class AC_DB
{
#ifdef THREADS
  inherit Thread.Mutex;
  // None of the service functions lock this by themselves.
//   class debuglock {
//     private object l;
//     void create (object _l) {l = _l; report_debug ("AC_DB lock\n");}
//     void destroy() {report_debug ("AC_DB unlock\n"); l = 0;}
//   }
//   object lock() {return debuglock (::lock());}
#endif

  class Backdoor
  {
    string user, cryptwd;
    object req_id;
    int count;			// Number of connects left.
  }

  static object(Backdoor) *_backdoors = ({});
  // Currently open backdoors.

  object(Identity) check_backdoor (object req_id);

  object(ReadableDecode) readable = ReadableDecode();

  int initialized = 0;

  object(IdAuths) id_auths;
  object(Identities) identities;
  object(ProtClasses) pclasses;
  object(ProtPoints) ppoints;

  array(function(int:void)) begin_funs = ({});
  // Called before data is accessed. They should ensure that the db is
  // up-to-date. If the argument is nonzero, the database is empty and
  // should be imported from scratch.

  array(function(mapping(string:multiset(int)):int)) commit_funs = ({});
  // Called after data has been changed. They get a mapping of
  // multisets of the dirty records in every table. Should return
  // nonzero if the db was successfully saved (the dirty bits for the
  // backend are cleared in that case).

  static array(mapping(string:multiset(int))) _dirty_records = ({});
  // The accumulated dirty records for each function in commit_funs.

  function(void:int) superuser = lambda() {return 0;};

  object config;

  static int _nr;
  // This integer counts a sequence that is unique for a constant
  // state of any running AC database, i.e. it's increased at least
  // every time the database change. It's never zero.

  void create (void|object conf)
  {
    config = conf;
    _nr = ++all_constants()->_ac_db_object_id_nr;
  }

  int nr() {return _nr;}

  void add_backend (function(int:void) begin_fun,
		    function(mapping(string:multiset(int)):int) commit_fun)
  {
    begin_funs += ({begin_fun});
    commit_funs += ({commit_fun});
    _dirty_records += ({([
      "id_auths": (<>),
      "identities": (<>),
      "pclasses": (<>),
      "ppoints": (<>)
    ])});
  }

  void remove_backend (function(int:void) begin_fun,
		       function(mapping(string:multiset(int)):int) commit_fun)
  {
    begin_funs -= ({begin_fun});
    int i = search (commit_funs, commit_fun);
    if (i >= 0) {
      commit_funs = commit_funs[..i-1] + commit_funs[i+1..];
      _dirty_records = _dirty_records[..i-1] + _dirty_records[i+1..];
    }
  }

  void begin (void|function exclude, void|int fromscratch)
  // Must be called before data is accessed. exclude may be a function
  // that should be excluded from the begin. If fromscratch is
  // nonzero, the database is reinitialized and imported from scratch.
  {
    if (fromscratch) initialize();
    if (exclude) (begin_funs - ({exclude})) (fromscratch);
    else begin_funs (fromscratch);
  }

  void commit (void|function exclude)
  // Must be called after data has changed. exclude may be a function
  // that should be excluded from the commit; the dirty records will
  // be cleared for that backend anyway.
  {
    mapping(string:multiset(int)) new_dirty = ([
      "id_auths": id_auths->dirty,
      "identities": identities->dirty,
      "pclasses": pclasses->dirty,
      "ppoints": ppoints->dirty
    ]);
    int n_size = `+(@Array.map (values (new_dirty), sizeof));

    clear_dirty();
    int cleaned = 0;

    for (int i = 0; i < sizeof (commit_funs); i++) {
      mapping(string:multiset(int)) old_dirty = _dirty_records[i];
      if (n_size)
	foreach (indices (new_dirty), string table)
	  old_dirty[table] |= new_dirty[table];

      function(mapping(string:multiset(int)):int) func = commit_funs[i];
      if (`+(@Array.map (values (old_dirty), sizeof)) &&
	  (func == exclude || func (old_dirty))) {
	cleaned = 1;
	foreach (indices (old_dirty), string table) old_dirty[table] = (<>);
      }
    }

    if (cleaned) _nr = ++all_constants()->_ac_db_object_id_nr;
  }

  object *conn_mods = ({});
  // The modules which have ModuleAC objects connected to this database.

  int possible_auth (object(Identity) identity,
		     object(ProtClass)|object(ProtPoint) prot)
  // Returns whether it is possible for the given identity to be
  // authenticated on the given pclass/ppoint prot.
  {
    if (!prot->active_authmods) prot = prot->pclass();
    return
      sizeof(prot->active_authmods() &
	     identity->active_authmods()) ||
      sizeof(prot->active_authmods()->get_auth_types()*({}) &
	     identity->active_authmods()->get_auth_types()*({}));
  }

  int perm (object(Identity) identity,
	    object(ProtClass)|object(ProtPoint) prot)
  // Returns the real permission. Only takes the permission settings
  // into account, which means the result is accurate only if the
  // given identity is truly authenticated for the given pclass/ppoint
  // prot.
  {
    if (identity == identities->privileged()) return PERM_WRITE;

    object(ProtClass) pc = prot->perm ? prot : prot->pclass();
    int perm = pc->perm (identity);

    if (perm == -1) {
      perm = PERM_NONE;
      foreach (indices (identity->groups()), object(Identity) id) {
	int p = pc->perm (id);
	if (p > perm) perm = p;
      }
    }

    return perm;
  }

  int simulated_perm (object(Identity) identity,
		      object(ProtClass)|object(ProtPoint) prot)
  // Returns the maximal simulated permission, i.e. the permission a
  // user would get if she were to authenticate herself as the given
  // identity or as any group the identity belongs to. This takes into
  // account the active auth modules for both the identity and the
  // given pclass/ppoint prot.
  {
    if (identity == identities->privileged()) return PERM_WRITE;

    object(ProtClass) pc = prot->perm ? prot : prot->pclass();

    if (identity->type() != ID_TYPE_GROUP && !possible_auth (identity, pc)) {
      int pmax = PERM_NONE;
      foreach (indices (identity->groups()), object(Identity) id)
	if (possible_auth (id, pc)) {
	  int p = perm (id, pc);
	  if (p > pmax) pmax = p;
	}
      return pmax;
    }
    else return perm (identity, pc);
  }

  int direct_perm (object(Identity) identity, object(ProtClass) pclass)
  // Returns the direct permission setting between the given identity
  // and pclass, or -1 if no direct permission is set.
  {
    return pclass->perm (identity);
  }

  void set_perm (object(Identity) identity, object(ProtClass) pclass, int perm)
  // If perm == -1 the permission is removed.
  {
    pclass->set_perm (identity, perm);
  }

  object(Identity) passive_auth (object(ProtPoint) ppoint, object req_id)
  // Performs passive authentication with ppoint and returns the
  // resulting identity.
  {
    foreach (ppoint->pclass()->active_authmods(), object mod)
      if (mod->passive_auth)
	foreach (mod->get_protocols(), string prot)
	  if (prot == req_id->prot[..sizeof(prot)-1]) {
	    object(Identity) id;
	    if (objectp (mod->passive_auth)) { // Remote module.
	      if (zero_type (id = try_call (mod, "passive_auth", ppoint, req_id)))
		break;
	    }
	    else id = mod->passive_auth (ppoint, req_id);
	    if (!id) return check_backdoor (req_id) || identities->everyone();
	    else if (id != identities->everyone()) {
	      if (mod->update_userinfo) mod->update_userinfo (id);
	      return id;
	    }
	    break;
	  }

    return check_backdoor (req_id) || identities->everyone();
  }

  mixed active_auth (object(ProtPoint) ppoint, object req_id,
		     void|string msg, void|string realm)
  // Performs active authentication. The result should be returned as
  // a server response. If no module responds, 0 is returned.
  {
    if (!realm) realm = ppoint->pclass()->realm();
    if (!realm) {
      report_error ("AClib: No realm for authentication set in protection class: " +
		    (ppoint->pclass()->name() ? ppoint->pclass()->name() : "noname") +
		    "\n");
      realm = "restricted area"; // Best we can do..
    }

    foreach (ppoint->pclass()->active_authmods(), object mod)
      if (mod->active_auth)
	foreach (mod->get_protocols(), string prot)
	  if (prot == req_id->prot[..sizeof(prot)-1]) {
	    mixed r;
	    if (objectp (mod->active_auth)) // Remote module.
	      r = try_call (mod, "active_auth", ppoint, req_id, realm, msg);
	    else r = mod->active_auth (ppoint, req_id, realm, msg);
	    if (r) return r;
	    break;
	  }

    return 0;
  }

  string open_backdoor (int count)
  // Returns a random "user:passwd" string that gives PERM_WRITE
  // permission on any ppoint for count accesses.
  {
#if constant (Crypto.rc4)
    string rnd = Crypto.randomness.reasonably_random()->read (16);
#else
    // For lobotomized Roxen: Do essentially what the function above
    // currently does (CVS ver 1.4).
    object rnd_file = Stdio.File ("/dev/urandom", "r");
    string rnd = rnd_file && rnd_file->read (16);
#endif
    if (!rnd) {
      report_debug ("AClib: Resorting to weak random source.\n");
      random_seed (time (1) ^
		   getpid() + getppid() << 8 ^
		   sizeof (all_constants()) << 16 ^
		   sizeof (id_auths) << 24);
      rnd = "";
      for (int i = 0; i < 16; i++) rnd += sprintf ("%c", random (256));
    }

    int r;
    string user = "", passwd = "";

    // Buglet: The modulus operations below causes nonuniform
    // distribution. The amount of random characters ought to make up
    // for this weakness.
    string user_safe_chars =
      "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
    for (r = 0; r < 8; r++)
      user += sprintf ("%c", user_safe_chars[rnd[r] % sizeof (user_safe_chars)]);
    for (; r < 16; r++)
      passwd += sprintf ("%c", (rnd[r]%('~'-'!')-':'+'!' || '~'-':') + ':');

    object(Backdoor) b = Backdoor();
    b->user = user;
    b->cryptwd = crypt (passwd);
    b->count = count;
    _backdoors += ({b});

    return user + ":" + passwd;
  }

  object(Identity) check_backdoor (object req_id)
  // Checks backdoors with basic http authentication.
  {
    if (sizeof (_backdoors)) {
      string user, passwd;
      if (req_id->realauth && sscanf (req_id->realauth, "%s:%s", user, passwd) == 2)
	foreach (_backdoors, object(Backdoor) b) 
	  if (user == b->user && crypt (passwd, b->cryptwd)) {
	    if (!objectp (req_id->misc->common["ac_req_id_" + _nr]) && --b->count <= 0)
	      _backdoors -= ({b});
	    else b->req_id = req_id;
	    return identities->privileged();
	  }
    }
    return 0;
  }

  int close_backdoor (string userpass)
  // Closes the backdoor opened by open_backdoor().
  {
    string *in = userpass / ":";
    foreach (_backdoors, object(Backdoor) b)
      if (in[0] == b->user && crypt (in[1], b->cryptwd)) {
	_backdoors -= ({b});
	return 1;
      }
    return 0;
  }

  void module_is_removed (object mod)
  // Should be called whenever a module is about to be removed. All
  // records in the database that refers to it are updated or removed
  // as appropriate.
  {
    foreach (id_auths->this_values(), object(IdAuth) r)
      r->module_is_removed (mod);
    foreach (ppoints->this_values(), object(ProtPoint) r)
      r->module_is_removed (mod);
    foreach (pclasses->this_values(), object(ProtClass) r)
      r->module_is_removed (mod);
  }

  void purge_missing_modules()
  // Searches through the database and removes all records that are
  // associated for modules that can't be found.
  {
    foreach (id_auths->this_values(), object(IdAuth) r)
      r->purge_missing_modules();
    foreach (ppoints->this_values(), object(ProtPoint) r)
      r->purge_missing_modules();
    foreach (pclasses->this_values(), object(ProtClass) r)
      r->purge_missing_modules();
  }

  void flush_all_caches()
  {
    id_auths->flush_caches();
    identities->flush_caches();
    pclasses->flush_caches();
    ppoints->flush_caches();
  }

  object disable_flush()
  {
    return class {
      private static array(object) _objs;
      void create (object ... objs) {_objs = objs;}
    } (id_auths->disable_flush(),
       identities->disable_flush(),
       pclasses->disable_flush(),
       ppoints->disable_flush()
      );
  }

  void destroy()
  {
    if (id_auths) destruct (id_auths);
    if (identities) destruct (identities);
    if (pclasses) destruct (pclasses);
    if (ppoints) destruct (ppoints);
  }

  void erase()
  {
    destroy();
    initialized = 0;
  }

  void initialize()
  {
    destroy();

    id_auths = IdAuths (this_object());
    identities = Identities (this_object());
    pclasses = ProtClasses (this_object());
    ppoints = ProtPoints (this_object());
    initialized = 1;
  }

  int last_read_binary;

  constant ACDB_COOKIE_TEXT1 = "ACDB text v1";

  constant ACDB_COOKIE_TEXT1_1 = "ACDB text v1.1";
  // Difference since v1: IdAuths no longer allowed on groups.

  constant ACDB_COOKIE_TEXT1_2 = "ACDB text v1.2";
  constant ACDB_COOKIE_BINARY1_2 = "ACDB binary v1.2";
  // Difference since v1.1: auth_data tuples may contain more than two elements.

  mapping(string:mapping(int:array)) dump_values (void|mapping(string:multiset(int)) ids)
  {
    if (ids) {
      mapping dump = ([]);
      foreach (indices (ids), string table)
	dump[table] = this_object()[table]->dump_values (indices (ids[table]));
      return dump;
    }
    else return (["identities": identities->dump_values(),
		  "id_auths": id_auths->dump_values(),
		  "pclasses": pclasses->dump_values(),
		  "ppoints": ppoints->dump_values()]);
  }

  void assign_values (mapping(string:mapping(int:array)) val)
  {
    object df = disable_flush();
    if (val->identities) identities->assign_values (val->identities);
    if (val->id_auths) id_auths->assign_values (val->id_auths);
    if (val->pclasses) pclasses->assign_values (val->pclasses);
    if (val->ppoints) ppoints->assign_values (val->ppoints);
  }

  string read_file (object file)
  // Returns an error message if there was a parse error, 0 otherwise.
  {
    int line = 1;
    mixed s = file->gets();

    if (!(<ACDB_COOKIE_TEXT1_1, ACDB_COOKIE_TEXT1_2, ACDB_COOKIE_BINARY1_2>)[s])
      return line + ": Incompatible AC database version\n";
    else file->gets(), line++;

    if (s == ACDB_COOKIE_BINARY1_2) {
      if (catch (assign_values (decode_value (file->read()))))
	return "Error decoding binary data\n";
      last_read_binary = 1;
      return 0;
    }

    object df = disable_flush();
    s = identities->read_readable (file, line);
    if (stringp (s)) return s;
    s = id_auths->read_readable (file, s);
    if (stringp (s)) return s;
    s = pclasses->read_readable (file, s);
    if (stringp (s)) return s;
    s = ppoints->read_readable (file, s);
    if (stringp (s)) return s;

    line = s + 1;
    s = file->gets();
    if (s != "EOF") return line + ": Unexpected end of file\n";
    last_read_binary = 0;

    return 0;
  }

  void write_file (object file, void|int binary)
  {
    if (binary) {
      file->write (ACDB_COOKIE_BINARY1_2 + "\n\n");
      file->write (encode_value (dump_values()));
    }

    else {
      file->write (ACDB_COOKIE_TEXT1_2 + "\n\n");
      identities->write_readable (file);
      file->write ("\n");
      id_auths->write_readable (file);
      file->write ("\n");
      pclasses->write_readable (file);
      file->write ("\n");
      ppoints->write_readable (file);
      file->write ("\nEOF\n");
    }
  }

  // Compatibility measures.
  string read_readable (object file) {return read_file (file);}
  void write_readable (object file) {write_file (file);}

  object dirty()
  {
    if (sizeof (id_auths->dirty)) return id_auths;
    if (sizeof (identities->dirty)) return identities;
    if (sizeof (pclasses->dirty)) return pclasses;
    if (sizeof (ppoints->dirty)) return ppoints;
    return 0;
  }

  void clear_dirty()
  {
    id_auths->dirty = (<>);
    identities->dirty = (<>);
    pclasses->dirty = (<>);
    ppoints->dirty = (<>);
  }

  // Module lookup etc.

  // Format:
  // o  Local forms:
  //    /modname#copy			in the current config
  //    config/modname#copy		in the local server
  // o  Remote forms:
  //    //remote-id/config/modname#copy	in the given remote server
  //    //.//modname#copy		a replicable module.
  // A module name is as often as possible normalized to local form.

  function(string|object,void|int:object|string) remote_resolve;
  function(void:string) local_id;

#if constant (roxen)
  inherit "roxenlib";

  static mapping(string|object:object|string) _lookup = ([]);

#ifdef AC_REMOTE_DEBUG
#  define DEBUGMSG(msg) roxen_perror (msg)
#else
#  define DEBUGMSG(msg)
#endif

  array try_call_2 (string modname, string func, array args)
  // Used remotely by try_call().
  {
    object mod = findmodule (modname);
    function fn = mod[func];
    return fn ? ({mod, fn (@args)}) : ({mod});
  }

  mixed try_call (object obj, string func, mixed... args)
  // Used to call a function in a module if it exists. ([])[0] is
  // returned either if the function doesn't exist or there's an error
  // in the remote communication.
  {
    function|object fn = obj[func];
    if (objectp (fn) && fn->objectid) {	// Remote call.
      mixed val;
      array err;
      if (err = catch (val = fn (@args))) {
	string modname = _lookup[obj];
	if (!modname || modname[..1] != "//") throw (err);
	array(string) parts = modname[2..] / "/";
	object rem = remote_resolve && remote_resolve (parts[0]);
	if (!rem) return ([])[0];
	object rem_acdb;
	catch (rem_acdb = rem->get_acdb());
	if (!rem_acdb) return ([])[0];
	// I/O error unlikely now.
	val = rem_acdb->try_call_2 (parts[1..] * "/", func, args);
	if (val) {
	  if (val[0] != obj) {
	    destruct (obj);
	    _lookup[val[0]] = modname;
	    _lookup[modname] = val[0];
	  }
	  return val[1];
	}
	return ([])[0];
      }
      return val;
    }
    else if (fn) return fn (@args);
    return ([])[0];
  }

  int has_function (string|object module, string func)
  {
    string modname;
    if (stringp (module)) modname = module, module = 0;
    else modname = modulename (module);
    if (!modname) return 0;
    if (modname[..1] == "//") {
      array(string) parts = modname[2..] / "/";
      object rem = remote_resolve && remote_resolve (parts[0]);
      if (!rem) return 0;
      object rem_acdb;
      catch (rem_acdb = rem->get_acdb());
      if (!rem_acdb) return 0;
      int res;
      if (catch (res = rem_acdb->has_function (parts[1..] * "/", func))) return 0;
      return res;
    }
    else {
      if (!module) module = findmodule (modname);
      return module && functionp (module[func]);
    }
  }

  object findmodule (string modname)
  {
    object mod = _lookup[modname];
    if (mod) return mod;

    if (modname[..1] == "//") {
      array(string) parts = modname[2..] / "/";
      if (sizeof (parts) < 2) {
	DEBUGMSG ("findmodule: " + modname + " malformed\n");
	return 0;
      }
      object rem = remote_resolve && remote_resolve (parts[0]);
      if (!rem) {
	//DEBUGMSG ("findmodule: Couldn't find remote object for " + modname + "\n");
	return 0;
      }
      object rem_acdb;
      catch (rem_acdb = rem->get_acdb());
      if (!rem_acdb) return 0;
      if (rem_acdb == this_object())
	throw (({"AClib: Using remote syntax " + modname + " for local acdb\n",
		 backtrace()}));
      catch (mod = rem_acdb->findmodule (parts[1..] * "/"));
      if (mod) {
	DEBUGMSG ("findmodule: Module for " + modname + " found remote\n");
	_lookup[mod] = modname;
	return _lookup[modname] = mod;
      }
      //DEBUGMSG ("findmodule: Module for " + modname + " not found remote\n");
      return 0;
    }

    if (config && modname[..0] == "/")
      if (mod = _lookup[modname = config->name + modname])
	return mod;
    if (mod = get_module (modname)) {
      DEBUGMSG ("findmodule: Module for " + modname + " found locally\n");
      _lookup[mod] = modname;
      return _lookup[modname] = mod;
    }
    //DEBUGMSG ("findmodule: Module for " + modname + " not found locally\n");
    return 0;
  }

  string modulename (object module)
  {
    string modname = _lookup[module];

    if (!modname)
      if (objectp (module) && objectp (module->variables)) { // True for remote modules.
	string remname = remote_resolve && remote_resolve (module);
	if (!stringp (remname)) {
	  //DEBUGMSG ("modulename: Couldn't find remote object name\n");
	  return 0;
	}
	object rem = remote_resolve (remname);
	if (!rem) {
	  DEBUGMSG ("modulename: Couldn't find remote object for " + remname + "\n");
	  return 0;
	}
	object rem_acdb;
	catch (rem_acdb = rem->get_acdb());
	if (!rem_acdb) return 0;
	if (rem_acdb == this_object()) {
	  report_warning ("AClib: Got remote object for local module " + remname + "\n");
	  return 0;
	}
	catch (modname = rem_acdb->modulename (module));
	if (modname) {
	  DEBUGMSG ("modulename: //" + remname + "/" + modname + " found\n");
	  modname = "//" + remname + "/" + modname;
	  _lookup[modname] = module;
	  return _lookup[module] = modname;
	}
	//DEBUGMSG ("modulename: Module name in remote object " + remname +
	//	  " not found\n");
	return 0;
      }
      else if (modname = get_modname (module)) {
	DEBUGMSG ("modulename: " + modname + " found locally\n");
	_lookup[modname] = module;
	_lookup[module] = modname;
      }

    if (modname) {
      string cname, rest;
      if (config && modname &&
	  sscanf (modname, "%s/%s", cname, rest) == 2 && cname == config->name)
	modname = "/" + rest;
      return modname;
    }
    //DEBUGMSG ("modulename: Couldn't find local module name\n");
    return 0;
  }

  string modulefullname (object|string module)
  {
    if (stringp (module)) module = findmodule (modulelocalname (module));

    if (objectp (module) && objectp (module->variables)) { // True for remote modules.
      string remname = remote_resolve && remote_resolve (module);
      if (!stringp (remname)) {
	DEBUGMSG ("modulefullname: Couldn't find remote object name\n");
	return 0;
      }
      object rem = remote_resolve (remname);
      if (!rem) {
	DEBUGMSG ("modulefullname: Couldn't find remote object for " + remname + "\n");
	return 0;
      }
      object rem_acdb;
      catch (rem_acdb = rem->get_acdb());
      if (!rem_acdb) return 0;
      if (rem_acdb == this_object()) {
	report_warning ("AClib: Got remote object for local module " + remname + "\n");
	return 0;
      }
      string modname;
      catch (modname = rem_acdb->modulefullname (modulename (module)));
      if (modname) {
	//DEBUGMSG ("modulefullname: \"[" + rem->name() + "] " + modname +
	//	  "\" found remote\n");
	return "[" + rem->name() + "] " + modname;
      }
      //DEBUGMSG ("modulefullname: Module fullname in remote object " + remname +
      //	" not found\n");
      return 0;
    }
    else {
      string modname = get_modfullname (module);
      if (!modname) {
	DEBUGMSG ("modulefullname: Couldn't find local module fullname\n");
	return 0;
      }
      //DEBUGMSG ("modulefullname: \"" + modname + "\" found locally\n");
      return modname;
    }
  }

  string moduleremotename (string modname)
  {
    if (modname[..1] == "//") return modname;
    else {
      object mod = findmodule (modname);
      if (mod && mod->replicable) {
	if (modname[..0] != "/") modname = modname[1..search (modname, "/", 1)];
	modname = "//./" + modname;
      }
      else {
	if (modname[..0] == "/") {
#ifdef DEBUG
	  if (!config) throw (({"AClib: Can't convert " + modname + " to remote form "
				"without a config object\n", backtrace()}));
#endif
	  modname = config->name + modname;
	}
	if (local_id) modname = "//" + local_id() + "/" + modname;
      }
      return modname;
    }
  }

  string modulelocalname (string modname)
  {
    if (modname[..4] == "//.//") {
      modname = modname[4..];
#ifdef DEBUG
      if (!config) throw (({"AClib: Can't convert " + modname + " to local form "
			    "without a config object\n", backtrace()}));
#endif
      if (!findmodule (config->name + modname)) {
	DEBUGMSG ("modulelocalname: Adding local copy of replicable module " +
		  modname + "\n");
	config->add_modules (({config->name + modname}));
      }
      return modname;
    }
    if (modname[..1] == "//") {
      if (!local_id) return modname;
      array(string) parts = modname[2..] / "/";
      if (parts[0] == local_id()) modname = parts[1..] * "/";
      else return modname;
    }
    string cname, rest;
    if (config &&
	sscanf (modname, "%s/%s", cname, rest) == 2 && cname == config->name)
      modname = "/" + rest;
    return modname;
  }

#undef DEBUGMSG

#else
  mixed try_call (object obj, string func, mixed... args)
    {throw (({"Needs Roxen\n", backtrace()}));}
  int has_function (string|object module, string func)
    {throw (({"Needs Roxen\n", backtrace()}));}
  object findmodule (string modname)
    {throw (({"Needs Roxen\n", backtrace()}));}
  string modulename (object module)
    {throw (({"Needs Roxen\n", backtrace()}));}
  string modulefullname (object|string module)
    {throw (({"Needs Roxen\n", backtrace()}));}
  string moduleremotename (string modname)
    {throw (({"Needs Roxen\n", backtrace()}));}
  string modulelocalname (string modname)
    {throw (({"Needs Roxen\n", backtrace()}));}
#endif
}

class ACDBBackend
{
  object(AC_DB) acdb;

  void begin();
  // Called before data is accessed. It should ensure that acdb is
  // up-to-date.

  int commit();
  // Called after data has been changed. Should return nonzero if acdb
  // was successfully saved (the dirty bits are cleared in that case).

  void create (object(AC_DB) _acdb)
  {
    acdb = _acdb;
    acdb->add_backend (begin, commit);
  }

  void destroy()
  {
    if (acdb) acdb->remove_backend (begin, commit);
  }
}
