// SiteBuilder content editor module.

constant cvs_version = "$Id: content_editor.pike,v 1.92 1999/10/21 13:45:49 mast Exp $";

constant thread_safe = 1;

#include <module.h>

inherit "module";
inherit "roxenlib";

#if __VERSION__ < 0.6
import sb_addpath;
#else
import .sb_addpath;
#endif

import Sitebuilder;

#if constant(ErrorContainer)

#define INHIBIT_COMPILE_ERRORS object __errcon; \
 master()->set_inhibit_compile_errors((__errcon = ErrorContainer())->got_error)
#define UNINHIBIT_COMPILE_ERRORS master()->set_inhibit_compile_errors(0)
#define COMPILE_ERRORS __errcon->get()

#else

#define INHIBIT_COMPILE_ERRORS master()->set_inhibit_compile_errors("")
#define UNINHIBIT_COMPILE_ERRORS master()->set_inhibit_compile_errors(0)
#define COMPILE_ERRORS master()->errors

#endif


// Global variables.

// #ifndef __NT__
constant tabs_dir = combine_path (__FILE__, "../../../tabs/");
constant templates_dir = combine_path (__FILE__, "../../../templates/");
constant img_dir = combine_path (__FILE__, "../../../img/");
// #else
// constant file_=replace(__FILE__,"\\","/");
// constant tabs_dir = combine_path (file_, "../../../tabs/");
// constant templates_dir = combine_path (file_, "../../../templates/");
// constant img_dir = combine_path (file_, "../../../img/");
// #endif

object config;

mapping tabs=([]);
array tabs_order;

int url_uniquifier = time (1);


// Connect to a sitebuilder module.

private static mapping(object:object(Site)) _sitebuilders;
private static object(Site) _site; // Can be destructed occasionally.

mapping(object:object(Site)) sitebuilders()
// Returns a mapping from configs to Site objects.
{
  if (!_sitebuilders) _sitebuilders = roxen->query_var ("SiteBuilders");
  if (_sitebuilders) m_delete (_sitebuilders, 0);
  return _sitebuilders || ([]);
}

string sb_select_name (object conf)
{
  if(!conf) return "";
  return conf->name +
    (conf->query_name() != conf->name ? " (" + conf->query_name() + ")" : "");
}

array(string) sb_select_list()
{
  return ({"none"}) + sort (Array.map (indices (sitebuilders()), sb_select_name));
}

object(Site) site()
{
  if (_site) return _site;
  foreach (indices (sitebuilders()), object conf)
    if (sb_select_name (conf) == QUERY (sitebuilder))
      return _site = _sitebuilders[conf];
}


// Filesystems to view files in different modes.

object(VCFile) get_view_vcfile(string path, string sub, object id)
{
  object wa = id->misc->wa;
  if (!wa) return 0;

  object(VCFile) vcfile =
    id->misc->wa->vcobj (path, id, id->misc->undelete_mode == "1",
			 sub[..2] == "rev" && sub[3..],
			 sub[..3] == "date" && sub[4..]);

  return vcfile && vcfile->isfile && vcfile;
}

mapping rawview (string path, string sub, object id)
{
  object vcfile=get_view_vcfile(path,sub,id);
  if (!vcfile || !vcfile->isfile) return 0;
  if (!id->misc->wa->rawview_access (id)) return id->misc->wa->rawview_access_throw (id);
  int undel = id->misc->undelete_mode == "1";

  string|object data;
  mapping md;
  mapping|object vcres;
  switch (sub) {
    case "old":
      if (objectp (vcres = vcfile->update (id)))
	data = vcres->old_data, md = vcres->old_md;
      break;
    case "loc":
      if (objectp (vcres = vcfile->update (id)))
	data = vcres->a_data, md = vcres->a_md;
      break;
    case "rep":
      if (objectp (vcres = vcfile->update (id)))
	data = vcres->b_data, md = vcres->b_md;
      break;
    default:
      if (objectp (vcres = vcfile->view (id, undel)))
	data = vcres->fd, md = vcres->md;
  }
  if (mappingp (vcres)) return http_low_answer (
    200, "<h1>" + html_encode_string (error_msg (vcres)) + "</h1>\n");
  if (!data || !md || !sizeof (md)) return 0;
  if (!id->misc->sb->ct_data_for (md["http-content-type"])->downloadp) return 0;

  mapping res;
  if (objectp (data)) {
    int length = vcfile->stat (0, undel, 1)[1];
    if (!length) data = "";
    else res = http_file_answer (data, md["http-content-type"], length);
  }
  if (stringp (data)) res = http_string_answer (data, md["http-content-type"]);
  if (!res->extra_heads) res->extra_heads = ([]);
  res->extra_heads->Expires = http_date (0);
  return res;
}


mapping view(string path, string sub, object id)
{
  array(object) wa_s =
    id->conf->get_providers("sitebuilder_workarea_filesystem");

#ifdef SB_DEBUG
  werror("CE: #wa's: " + sizeof(wa_s) + "\n");
#endif /* SB_DEBUG */

  wa_s = Array.filter(wa_s, lambda(object c) {
#ifdef SB_DEBUG
			      werror("CE: download_mode: " +
				     c->query("download_mode") + "\n");
#endif
			      return(c->query("download_mode") == "View");
			    });

  if (!sizeof(wa_s)) {
    // No suitable work area filesystem.
    return(0);
  }

  sort(Array.map(wa_s, lambda(object c) {
			 return sizeof(c->query("location"));
		       }), wa_s);

  object wa_fs = wa_s[0];

  string location = wa_fs->query("location");

  if ((path != "") && (path[0] == '/')) {
    path = path[1..];
  }
  mapping res = http_redirect(location + path + "?__sb_force_userid=" +
			      id->misc->sb->mac->id_get_id (id) +
			      "&" + url_uniquifier++, id);

  string cookieid;

  if (!(cookieid = id->cookies->SiteBuilder)) {
    cookieid = sprintf("%d:%d", getpid(), wa_fs->next_cookie++);
    res->extra_heads["Set-Cookie"] =
      sprintf("SiteBuilder=%s; "/*"expires=%s; "*/"path=/",
	      cookieid/*, http_date(time(1) + 1*60*60)*/);
  }

  mapping info;
  if (!(info = wa_fs->cookies[cookieid])) {
    info = wa_fs->cookies[cookieid] = (["revisions":([])]);
  }
  info->timestamp = time(1) + 1*60*60;
  info->undelete_mode = id->misc->undelete_mode == "1";

  // The current work area
  info->workarea = id->misc->wa;

  // sub is "revision/date"
  array a = sub/"/";
  if (a[0] != "") {
    // Revision
    info->revisions[path] = a[0];
  } else {
    m_delete(info->revisions, path);
  }
  if ((sizeof(a) > 1) && (a[1] != "")) {
    // Date

    // Note: changes '/' to '-'.
    info->date = a[1..]*"-";
  } else {
    // The date-tag is short-lived.
    m_delete(info, "date");
  }
  return(res);
}


mapping revert(string path, string sub, object id)
{
  mapping res = http_redirect(id->misc->build_url(id, 0, "files"), id);

  object(VCOldFile)|object(VCOldDir) vcobj =
    id->misc->wa->vcobj(path, id, 0, sub);

  if (!vcobj) {
    return(0);
  }

  mapping err = vcobj->revert(id);

  // FIXME: Check for error-return here!

  return(res);
}

// Content Editor UI.

class Tab
{
  inherit "wizard";

  import Sitebuilder;

  string dir,errmsg,name;
  object o,parent;

  string make_tablist(array(object) tabs, object id)
  {
    string res_tabs="";
    foreach(tabs, object tab)
    {
      if(!tab->visible(id))
	continue;
      mapping args = ([]);
      args->bgcolor="white";
      args->href = id->misc->build_url(id,0,tab->name,0);
      if(this_object()==tab)
	args->selected = "selected";
      res_tabs += make_container( "tab", args, tab->title() );
    }
    return
      "<tablist>"+
      make_container("config_tablist",([]), res_tabs)+
      "</tablist>";
  }

  
  string gtext_button(string href, string title, int enabled,
		      string|void target)
  {
    return enabled?
      ("<gtext xsize=390 xspacing=4 notrans bevel=4 spacing=3 fg=black "
       "bg=#f3e07d nfont=lucida scale=0.36 "
       "magic='"+title+"' magic_fg=#550a0d href='"+ href+"'"+
       (target? " target='"+replace(target, "'", "'\"'\"'")+"'":"")+
       ">"+title+"</gtext><br>"):
      ("<gtext xsize=390 xspacing=4 notrans bevel=4 spacing=3 fg=#666666 "
       "bg=#cccccc nfont=lucida scale=0.36>"+
       title+"</gtext><br>");
  }

  object compile()
  {
    INHIBIT_COMPILE_ERRORS;
    mixed err = catch
    {
      o = compile_file(dir+"/index.pike")(parent);
    };
    errmsg = COMPILE_ERRORS;
    UNINHIBIT_COMPILE_ERRORS;
    if(err && errmsg != "")
      werror("Errors while compiling "+dir+"/index.pike\n"+errmsg+"\n");
    return o;
  }

  string|mapping show(object id, string sub, array tabs_order, mapping defines, object parent)
  {
    string res;
    string|mapping|array tmp;
    array wanted_buttons;

    // recompile upon "reload"      
    if(!o || id->pragma["no-cache"])
      compile();

    if(!o) 
      return "Compilation of \""+dir+"/index.pike"+"\" failed:\n <pre>"
	+ errmsg + "</pre>\n";

    mapping tags=([]), containers=([]);
    array extra_wizard_args =
      (functionp(o->extra_wizard_args)? o->extra_wizard_args(id) : ({ }));
    foreach(glob("tag_*",indices(o)), string tag)
      tags["sb-internal-"+replace(tag[4..], "_", "-")]=o[tag];
    foreach(glob("container_*",indices(o)), string container)
      containers["sb-internal-"+replace(container[10..], "_", "-")]=
	o[container];

    if(search(sub,"wizard")==0)
    {
      id->misc->extra_wizard_tags=tags;
      id->misc->extra_wizard_container=containers;
      INHIBIT_COMPILE_ERRORS;
      mixed e = catch
      {
	tmp=wizard_menu(id,dir+"/wizards/",
			id->misc->build_url(id), @extra_wizard_args);
      };
      if(e)
      {
	werror("show compile wizard failed:\n"+COMPILE_ERRORS+
		master()->describe_backtrace(e));
	tmp = "<pre>" + COMPILE_ERRORS + "\n"
	  + master()->describe_backtrace(e)+"</pre>";
      }
      UNINHIBIT_COMPILE_ERRORS;
      if(mappingp(tmp))
	return tmp;
      res=
	"<template name=" + parent->QUERY (location) + "__sbinternal__templates/content_editor.tmpl>"+
	"<content>"+tmp+"</content></template>";
    }
    else
    {
      string res_buttons;
      if(file_stat(dir+"wizards"))
      {
	id->misc->raw_wizard_actions=1;
	tmp = get_actions(id, id->misc->build_url(id)+"/wizard/",
			  dir+"wizards/", ({ }));
	mapping buttons=([]);
	if(mappingp(tmp))
	  foreach(indices(tmp), string groupname)
	    foreach(tmp[groupname], mapping group)
	      foreach(values(group),array wizard)
	      {
		object wiz = get_wizard(wizard[2]->action,dir+"wizards/");
		if(!buttons[groupname])
		  buttons[groupname]=({});
		buttons[groupname] += ({
		  ([ "name":wizard[0],
		     "enabled":wiz->button_enabled(id,@extra_wizard_args),
		     "action":wizard[2]->action,
		     "target":wiz->wizard_target &&
		     wiz->wizard_target(id,@extra_wizard_args),
		  ])
		});
	      }
	res_buttons="";
	foreach(o->group_order+sort(indices(buttons)-o->group_order),
		string groupname)
	  if(buttons[groupname])
	  {
	    string group = "";
	    array extra_wizards = buttons[groupname]->action;
	    sort(buttons[groupname]->name, extra_wizards);
	    foreach(o->wizard_order+(extra_wizards-o->wizard_order),
		    string wizard_name)
	      foreach(buttons[groupname], mapping button)
		if(button->action==wizard_name && button->enabled!=-1)
		  group += gtext_button(
					id->misc->build_url(id,"/wizards/?action="+
							    button->action),
					button->name,
					button->enabled,
					button->target);
	    if(sizeof(group))
	      res_buttons += "<p>" + groupname + "<br>" + group;
	  }
      }
      id->misc->res_buttons=res_buttons;
      tags["sb-internal-buttons"]=lambda(string tagname, mapping args, object id)
				  {
				    return id->misc->res_buttons;
				  };
      string contents=Stdio.read_bytes(dir+"/index.html")||"";
      res=
	"<template name=" + parent->QUERY (location) + "__sbinternal__templates/content_editor.tmpl>"+
	"<content>"+contents+"</content>"+
	make_tablist(tabs_order,id)+
	"</template>";
    }
    // Add any external rxml tags
    res=parse_rxml(res,id,0,defines);
    // Add the internal tags
    res=parse_html(res,tags,containers,id);
    // Add any external rxml tags from the internal tags
    res=parse_rxml(res,id,0,defines);
    if(id->misc->sb_result_mapping)
      return id->misc->sb_result_mapping;
    return res;
  }

  int visible(object id)
  {
    if (!o) compile();
    if (!o) return 0;
    if (!id->misc->real_user && !o->backdoor_accessible) return 0;
    return !o->visible || o->visible(id);
  }

  string title()
  {
    if(!o) return 0;
    return o->title;
  }

  int order()
  {
    if(!o) return 0;
    return o->order;
  }
  
  void create(string _dir, string _name, object _parent)
  {
     dir=_dir;
     name=_name;
     parent=_parent;
     compile();
  }
}


// Initialization and standard module stuff.

void create (object conf)
{
  if (!conf) return;

  array(string) sb_sel_list = sb_select_list();
  string sb_sel = sb_select_name (conf) || "none";

  // Kludge to shut up the "Passing value not present in the misc array" warning.
  if (search (sb_sel_list, sb_sel) < 0) sb_sel_list += ({sb_sel});

  defvar ("location", "/admin/",
	  "Content Editor location", TYPE_LOCATION,
	  "The location of the Content Editor.");
  defvar ("sitebuilder", sb_sel,
	  "SiteBuilder", TYPE_STRING_LIST,
	  "The SiteBuilder that this Content Editor should handle.",
	  sb_sel_list);
}

void check_variable (string name, mixed value)
{
  if (name == "sitebuilder") _site = 0;
}

string status()
{
  variables["sitebuilder"][VAR_MISC] = sb_select_list();
  return site() ?
    "Connected to " + sb_select_name (search (_sitebuilders, _site)) : "Not active";
}

array register_module()
{
  return ({MODULE_LOCATION, "SiteBuilder Content Editor",
	   "Provides a user interface for administration and editing of "
	   "a SiteBuilder site.",
	   0, 1});
}

void start (int when, object conf)
{
  if (conf) config = conf;

  module_dependencies (config, ({"configtablist", "htmlparse", "tablify", "obox"}));
  object tab;
  string title;
  foreach(get_dir(tabs_dir)-({".","..","CVS"}),
	  string tabname)
  {
    tab=Tab(tabs_dir + tabname+"/",
	    tabname,
	    this_object());
    if(tab&&tab->title())
      tabs[tabname]=tab;
  }

  // Sort the tabs
  tabs_order=values(tabs);
  sort(Array.map(tabs_order,lambda(object t) { return t->order(); }),
       tabs_order);

  site();
}

array(string) list_images()
{
  return sort(get_dir (img_dir + "types/")-({".","..","CVS"}));
}

//   MODULE_LOCATION callbacks.

string build_url(object id, void|string rest, void|string tab, void|string path,
                 void|string wa_tag, void|string undelete_mode, void|multiset(string) unfolded)
{
  return sprintf("%s!%s!%s!%s!%s!%s!%d/%s",
		 query("location"),
		 wa_tag||id->misc->wa_tag||"",
		 path||id->misc->path,
		 indices(unfolded||id->misc->unfolded)*"&",
		 tab||id->misc->tab,
		 undelete_mode||id->misc->undelete_mode,
		 url_uniquifier++,
		 rest || "");
}

array stat_file (string f, object id)
// weblayout caches templates by calling stat_file, so we let it cache
// the content editor template too.
{
  string template;
  if(sscanf(f, "__sbinternal__templates/%s", template))
    return file_stat (templates_dir + (template-".."));
}

string|mapping find_file (string f, object id)
{
  object(Site) sb = site();
  if (!sb) return 0;

  // the special templates "filesystem".
  string template;
  if(sscanf(f, "__sbinternal__templates/%s", template))
    return http_string_answer(Stdio.read_bytes(templates_dir + (template-"..")));

  // and the same for images...
  string img;
  if(sscanf(f, "__sbinternal__img/%s", img))
    return http_string_answer(Stdio.read_bytes(img_dir + (img-"..")), "image/gif");

  // and some magic wing needed
  if(sscanf(f, "__sbinternal__regfile/%*s"))
     return http_string_answer( id->variables->regfile,
				"application/octet-stream" );

  // Check access. AC backdoor takes precedence.
  object ac_module = sb->get_ac_module();
  if (ac_module) ac_module->root_access (id);
  id->misc->backdoor = id->misc->acadm_backdoor;
  if (sb->ce_access_try (id) < PERM_READ && !id->misc->backdoor)
    return sb->ce_access_throw (id);
  id->misc->real_user = sb->mac->id_get_type (id) == sb->mac->ID_TYPE_USER;

  // User change?
  if (id->variables->__sb_force_login) {
    int myid = id->misc->backdoor? -17:sb->mac->id_get_id(id);
    int notid = (int) id->variables->__sb_force_login;
    if (myid == notid || sb->mac->id_is_everyone (id))
      if (sizeof (f))
	return http_redirect (QUERY (location) +
			      "?f=" + f + "&__sb_force_login=" + notid);
      else {
	m_delete (id->variables, "__sb_force_login");
	// It's really nice to have a good fallback page.
	mixed res = find_file (id->variables->f || f, id);
	if (!mappingp (res)) return sb->ce_access_throw (id);
	else {
	  mapping hdrs = sb->ce_access_throw (id);
	  if (hdrs->error) res->error = hdrs->error;
	  if (!res->extra_heads) res->extra_heads = ([]);
	  if (hdrs->extra_heads) res->extra_heads += hdrs->extra_heads;
	  if (hdrs->data && !sizeof (res->data || ""))
	    res->data = hdrs->data, res->len = hdrs->len;
	  return res;
	}
      }
    else
      if (id->variables->f) {
	id->raw_url = id->not_query = f = id->variables->f;
	m_delete (id->variables, "__sb_force_login");
	m_delete (id->variables, "f");
	return find_file (f, id);
      }
  }

  // Check if all plugin ppoints are created, and create them if needed
  if(sb->plugin_get_num_pp(id)!=sizeof(id->conf->get_providers("sitebuilder_plugin")))
  {
    foreach(id->conf->get_providers("sitebuilder_plugin"), object plugin)
      if(plugin->sb_plugin_name &&
	 !sb->plugin_pp_exists(plugin->sb_plugin_name,id))
	sb->plugin_pp_create(plugin->sb_plugin_name,id);
  }

  string sub="", unf="";
  int score=sscanf(f,"!%s!%s!%s!%s!%s!%*d/%s",
		   id->misc->wa_tag,
		   id->misc->path,
		   unf,
		   id->misc->tab,
		   id->misc->undelete_mode,
		   sub);

  if (!id->misc->wa_tag)
  {
    mapping uservars = (sb->mac->id_get_data(id)||([]));

    id->misc->wa_tag = (uservars->default_workarea
                       || sb->get_default_workarea());
  }
  id->misc->wa = sb->wa_lookup(id->misc->wa_tag, id);
  if (id->misc->wa) {
    id->misc->wa_perm = id->misc->wa->access_try (id);
    id->variables->wa_tag = id->misc->wa_tag = id->misc->wa->tag;
  }

  if(!id->misc->path) id->misc->path="/";
  if(!id->misc->tab)  id->misc->tab = id->misc->real_user ? "files" : "ac";
  if(!id->misc->undelete_mode) id->misc->undelete_mode="0";
  id->variables->undelete_mode= id->misc->undelete_mode == "1";
  id->misc->unfolded = ((unf && sizeof(unf))? mkmultiset(unf/"&") : (<>));

  // A small hack: go to /admin/[page] and you will be redirected to the
  // page about that file.  /js
  if(score == 0)
    return http_redirect(build_url(id,0,0,f),id);

  id->misc->sub = sub;
  id->misc->build_url=build_url;
  id->misc->sb=sb;
  id->misc->ce=this_object();
  id->misc->img_url = query("location")+"__sbinternal__img/";

  if (id->misc->wa) {
    switch(id->misc->tab)
    {
    case "rawview":
      // View files unparsed.
      return rawview(id->misc->path, sub, id);
    case "viewsource":
    {
      object vcf=get_view_vcfile(id->misc->path, sub, id);
      if(!vcf)
	return 0;
      return id->misc->sb->handler_for(vcf)->viewsource(vcf,id);
    }
    case "view":
      return view(id->misc->path, sub, id);
    case "revert":
      return revert(id->misc->path, sub, id);
    }
    if (id->cookies->SiteBuilder) {
      // Clear any cookies that have been set for this user.

      array(object) wa_s =
	id->conf->get_providers("sitebuilder_workarea_filesystem");

#ifdef SB_DEBUG
      werror("CE: #wa's: " + sizeof(wa_s) + "\n");
#endif /* SB_DEBUG */

      foreach(wa_s, object wa_fs) {
	m_delete(wa_fs->cookies, id->cookies->SiteBuilder);
      }
    }
  }

  // Add the content editor template
  string res;
  mapping defines = ([" _error": 200, " _extra_heads": ([])]);
  id->misc->defines = defines;

  if(!tabs[id->misc->tab])
    res = "No such tab! ("+id->misc->tab+")";
  else
  {
    mapping|string tmp = tabs[id->misc->tab]->show(id,sub,tabs_order,defines,this_object());
    if(mappingp(tmp))
      return tmp;
    else
      res = tmp;
  }

  return http_string_answer(res) + ([
    "stat": defines[" _stat"],
    "error": defines[" _error"],
    "rettext": defines[" _rettext"],
    "extra_heads": defines[" _extra_heads"],
  ]);
}
