/*
 * $Id: plain_filesystem.pike,v 1.7 1999/08/23 19:59:32 mast Exp $
 *
 * Workarea filesystem that uses plain files (just AC, no CVS).
 *
 * Henrik Grubbstrm 1999-06-04
 */

#include <module.h>

inherit "modules/filesystems/filesystem.pike";

//---- AC -------------------------------------------------------------------

// Code ripped from workarea.pike

object mac;

int PERM_NONE, PERM_READ, PERM_WRITE, ID_TYPE_USER;

void ac_init(object conf)
{
#ifdef SBDEBUG
  werror("ac_init() in workarea_filesystem.pike\n");
#endif

  if (!roxen->query_var ("AC"))
    error ("No AC module enabled.\n");
  else {
    mac = mac || roxen->query_var("AC")["ModuleAC"](conf, this_object());
    PERM_NONE = mac->PERM_NONE;
    PERM_READ = mac->PERM_READ;
    PERM_WRITE = mac->PERM_WRITE;
    ID_TYPE_USER = mac->ID_TYPE_USER;
  }

  mac->pp_create(QUERY(protection_point),
		 "Plain filesystem: " + query_location());

  foreach(mac->pp_get_ids()||({}), int ppid) {
    if (QUERY(protection_point) != mac->pp_get_handle(ppid)) {
      mac->pp_remove(ppid);
    }
  }
}

/*** Misc glue ******************************************************/
void start(int n, object c)
{
  if (c) {
    ac_init(c);
  }
  ::start();
}

void create(object c)
{
  ::create();

  defvar("protection_point", "plain_fs", "Protection point", TYPE_STRING,
	 "Unique name for the protection point. ");
}

array register_module()
{
  return ({ MODULE_LOCATION, "Plain filesystem",
	    "Plain filesystem, with AC control.", 0, 0 });
}

/*** function called as filesystem module ***************************/

mixed find_file(string f,object id)
{
  if (!mac) ac_init(id->conf);
  if (mac->access_try(QUERY(protection_point), id ) == PERM_NONE)
    return mac->access_throw (QUERY(protection_point), id);
  if (id->variables->_force_login &&
      mac->id_get_id (id) == (int) id->variables->_force_login)
  {
    return mac->access_throw (QUERY(protection_point), id);
  }
  if (id->variables->__user_id
      && id->variables->__user_id != mac->id_get_handle( id ))
    return mac->access_throw(QUERY(protection_point), id );

  if ((< "GET", "HEAD", >)[id->method]) {
    /********* Perform the access *********/
    return ::find_file(f, id);
  }
  return 0;
}

int dir_filter_function( string f, object id, string path )
{
  if (f[0]=='.'
      || f[-1]=='~'
      || f[0]=='#') return 0;
  if (f=="CVS") return 0;
  if (f == "__info")
    return 0;

  /***** CHECK for read access ******/

  return 1;
}

array find_dir( string f, object id )
{
  if (!mac) ac_init(id->conf);
  if (mac->access_try(QUERY(protection_point), id) == PERM_NONE)
    return 0;
  if(id->prot=="FTP"&&!id->misc->ftp_find_dir)
    return 0;

  array dir = ::find_dir(f, id);

  dir=Array.filter(dir, dir_filter_function, id,
		   sizeof( f ) && f[-1] == '/' ? f : f + "/" );

  return dir; 
}

array stat_file( string f, object id )
{
  if (!mac) ac_init(id->conf);
  if (mac->access_try(QUERY(protection_point), id) == PERM_NONE)
    return 0;

  if (f && strlen( f ) && f[0] == '/')
    f = f[1..];
  f = replace( f, "//", "/" );

  /****** Perform the stat if read access *******/

  return(::stat_file(f, id));
}
