// Loader for the SiteBuilder core.

constant cvs_version = "$Id: sitebuilder.pike,v 1.30 1999/08/23 19:58:52 mast Exp $";

constant thread_safe = 1;

#if constant (roxen)
#include <module.h>
inherit "module";
inherit "roxenlib";
#endif

#if __VERSION__ < 0.6
import sb_addpath;
#else
import .sb_addpath;
#endif

import Sitebuilder;


// Global variables.

object config;
private static object(Site) site;


// AC setup.

int save_config = 0;

int ac_is_hooked()
{
  object ac = config->modules["acadm"] && config->modules["acadm"]->enabled;
  return ac && ac->query ("mountpoint") == "";
}

object/*(ModuleAC)*/ get_mac (string acdbfile)
{
#if constant (roxen)
  mapping acvar = roxen->query_var ("AC");
  object ac = acvar && acvar->loaders && acvar->loaders[config];
  if (!ac) throw (({"No Access Control module enabled\n", backtrace()}));

  // You suddenly feel a strange nausea and loose focus for the next 20 lines.
  if (!(<"", acdbfile>)[ac->query ("db_filename")] && ac->query ("mountpoint") != "")
    // This throw will be catched and handled gracefully by Site->initialize().
    throw (({"The Access Control module seems to already be in use. "
	     "Please make sure it's free and delete its 'Database "
	     "file' setting.\n", backtrace()}));
  else if (ac->query ("db_filename") != acdbfile || ac->query ("mountpoint") != "") {
    ac->set ("db_filename", acdbfile);
    ac->set ("mountpoint", "");
    save_config = 1;
    ac->start (2, config);
  }

  ac->variables["db_filename"][VAR_CONFIGURABLE] = 0;
  ac->variables["mountpoint"][VAR_CONFIGURABLE] = 0;

  return acvar->ModuleAC (config, this_object());
#else
  return 0;
#endif
}


#if constant (roxen)

// Module connections.

array(object) find_content_editors()
{
  array(object) ces =
    (column (roxen->configurations->modules, "content_editor") - ({0}))->enabled;
  array(object) our_ces = ({});
  foreach (ces, object ce)
    if (ce->site() == site) our_ces += ({ce});
  return our_ces;
}

array(object) find_workarea_filesystems()
{
  multiset(object(Workarea)) was = mkmultiset (site->wa_list());
  array(object) wfs =
    Array.map ((column (roxen->configurations->modules, "workarea") - ({0}))->copies,
	       values) * ({});
  array(object) our_wfs = ({});
  foreach (wfs, object wf)
    if (wf->site() == site || was[wf->workarea()]) our_wfs += ({wf});
  return our_wfs;
}

string wa_name (object/*(ProtPoint)*/ ppoint)
// Called by the AC interface to get the name of a work area ppoint.
{
  if (!site) return ppoint->handle(); // Shouldn't happen.
  string tag = ppoint->handle()[9..];
  object(Workarea) wa = site->wa_lookup (tag);
  return wa ? "Work area: " + wa->name : "Removed work area (" + tag + ")";
}


// Initialization and standard module stuff.

private static string errormsg, headermsg;

void initialize (void|int create_storage)
{
  if (site) destruct (site);
  site = Site();
  errormsg = site->initialize (get_mac, QUERY (storage), create_storage);
  if (stringp (errormsg))
    headermsg = "SiteBuilder not started due to error:<br>" +
      replace (errormsg, "\n", " ") + // Don't want message broken with "More doc".
      "<p>SiteBuilder will try to create a new storage directory "
      "if you reload this module.";
  else {
    headermsg = "";
    mapping(object:object(Site)) sbs = roxen->query_var ("SiteBuilders");
    if (!sbs) {			// Race risk here. Not much we can do about it.
      sbs = ([]);
      roxen->set_var ("SiteBuilders", sbs);
    }
    sbs[config] = site;
  }
}

string status()
{
  if (!site || !site->is_initialized())
    if (errormsg) return "<strong>Not running due to error:</strong> " + errormsg;
    else return "<strong>Not running due to unknown error.</strong>\n";

  string res = "<table cellspacing=3 cellpadding=2 border=0>\n";

  res += "<tr><td valign=top>Work areas:</td><td>";
  array(object(Workarea)) was = site->wa_list();
  if (!sizeof (was)) res += "<strong>none</strong>";
  else res += Array.map (was->name, html_encode_string) * "<br>";
  res += "</td></tr>\n";

  res += "<tr><td valign=top>Content editor:</td><td>";
  array(object) ces = find_content_editors();
  if (!sizeof (ces)) {
    headermsg = "No content editor is enabled.";
    res += "<strong>none</strong>";
  }
  else {
    array(array(string)|string) url_lists = Array.map (ces->config, server_urls);
    for (int i = 0; i < sizeof (url_lists); i++) {
      array(string) urls = url_lists[i];
      string loc = ces[i]->query ("location");
      if (loc[..0] == "/") loc = loc[1..];
      for (int j = 0; j < sizeof (urls); j++)
	if (urls[j][..3] == "http")
	  urls[j] = "<a target=server_view href=\"" + urls[j] + loc + "\">" +
	    html_encode_string (urls[j] + loc) + "</a>";
	else urls[j] = 0;
      url_lists[i] = (urls - ({0})) * ", ";
    }

    if (sizeof (url_lists - ({""})))
      headermsg = "Open content editor: " + (url_lists - ({""})) * ", ";
    else
      headermsg = "No content editor is open for this site. See status for details.";

    for (int i = 0; i < sizeof (url_lists); i++)
      if (!sizeof (url_lists[i]))
	url_lists[i] = "Content editor in " +
	  html_encode_string (ces[i]->config->name) + " (no open ports)";
    res += url_lists * "<br>";
  }
  res += "</td></tr>\n";

  res += "<tr><td valign=top>Work area filesystems:</td><td>";
  array(object) wfs = find_workarea_filesystems();
  if (!sizeof (wfs)) res += "<strong>none</strong>";
  else {
    array(array(string)|string) url_lists = Array.map (wfs->config, server_urls);
    for (int i = 0; i < sizeof (url_lists); i++) {
      array(string) urls = url_lists[i];
      string loc = wfs[i]->query ("location");
      if (loc[..0] == "/") loc = loc[1..];
      for (int j = 0; j < sizeof (urls); j++)
	urls[j] = "<a target=server_view href=\"" + urls[j] + loc + "\">" +
	  html_encode_string (urls[j] + loc) + "</a>";
      if (wfs[i]->query ("download_mode") == "All sites")
	url_lists[i] = "Edit all sites: ";
      else if (wfs[i]->workarea())
	url_lists[i] = wfs[i]->query ("download_mode") +
	  " work area " + wfs[i]->workarea()->name + ": ";
      else url_lists[i] = "Edit all work areas: ";
      url_lists[i] += urls * ", ";
    }
    res += sort (url_lists) * "<br>";
  }
  res += "</td></tr>\n";

  res += "<tr><td valign=top>CVS version:</td><td>";
  if(catch{
    object cvs = VC.CVS.CVS(".", ".");
    string ver = cvs->command("--version");
    res += (ver/"\n"-({""}))[0];
  })
    res += "<strong>Not available</strong>";
  res += "</td></tr>\n";

  res += "</table>\n";
  return res;
}

void create (object conf)
{
  defvar ("storage", "../" + (conf ? filename_mangle (conf->name) : "storage") + ".sb/",
	  "SiteBuilder storage", TYPE_STRING,
	  "The directory where SiteBuilder will store the site contents. "
	  "This module needs to be reloaded for a change here to take "
	  "effect.\n"
	  "<p>Note that it's not possible to use several SiteBuilders "
	  "with the same storage.");
}

array register_module()
{
  return ({0, "SiteBuilder",
	   "This is the main SiteBuilder module." + (headermsg ? "<p>" + headermsg : ""),
	   0, 1});
}

void start (int when, object conf)
{
  if (conf) config = conf;

  string prevheadermsg = roxen->query_var ("SiteBuilder headermsg");
  string nextheadermsg = 0;

  module_dependencies (config, ({"acadm", "weblayout"}));

  if (!when) {			// Initial load or second after reload.
    if (!ac_is_hooked())
      if (prevheadermsg)
	headermsg = prevheadermsg, prevheadermsg = 0;
      else
	headermsg =
	  "It looks like you're installing a new SiteBuilder. Please "
	  "check that the SiteBuilder storage path points to a good "
	  "place. Unless you're importing an existing SiteBuilder site, "
	  "this should not be an existing directory. When you press "
	  "\"Save\" or reload this module, SiteBuilder will try to "
	  "start up a site in the specified directory, or create and "
	  "initialize a new site in it if it doesn't exist. This can "
	  "take some time; please be patient. Watch out for any messages "
	  "in this spot.";
    else initialize();
  }
  else {			// First after reload.
    if (ac_is_hooked()) initialize();
    else errormsg = "";
    if (errormsg) {
      initialize (1);
      if (site->is_initialized()) {
	// Got a new storage running. Set up a default site.
	module_dependencies (config, ({
	  "content_editor",
	  "workarea",
	  "sitebuilder_tags",
	  "weblayout",
	  "navigation"
	}));

	nextheadermsg =
	  "SiteBuilder is successfully initialized. Use the link below "
	  "to go to the content editor. The username and password is "
	  "the same as in this configuration interface.";
      }
      else
	// Avoids the "new SiteBuilder" blurb on reload when AC isn't hooked.
	nextheadermsg = headermsg;
    }
  }

  if (site && site->is_initialized() && ac_is_hooked()) {
    // Fix weblayout settings. Nobody asked, nobody told. Note: Keep
    // the sitebuilder_view server template in synch with this.
    mapping moddata = config->modules["weblayout"];
    object weblayout = moddata && moddata->enabled;
    if (weblayout) {
      if (weblayout->query ("tmpl_storage_path") != "/templates/")
	weblayout->set ("tmpl_storage_path", "/templates/"), save_config = 1;
      if (weblayout->query ("tmpl_config_dir") != "")
	weblayout->set ("tmpl_config_dir", ""), save_config = 1;
      weblayout->variables["tmpl_storage_path"][VAR_CONFIGURABLE] = 0;
      weblayout->variables["tmpl_config_dir"][VAR_CONFIGURABLE] = 0;
    }

    if (save_config) save_config = 0, config->save();
    else status();
  }

  if (prevheadermsg) headermsg = prevheadermsg + (headermsg ? "<p>" + headermsg : "");
  roxen->set_var ("SiteBuilder headermsg", nextheadermsg);
}

void stop()
{
  if (site) destruct (site);
}

#else // constant (roxen)

int main (int argc, array(string) argv)
{
  Site, Workarea, MetaData;
  VCFile, VCFileData, VCFileConflict, VCDir, VCOldFile, VCOldDir;
  werror ("Seems like Sitebuilder.pmod compiles ok.\n");
}

#endif // constant (roxen)
