// SiteBuilder filesystem module for accessing a work area.

constant cvs_version = "$Id: workarea.pike,v 1.80 1999/10/28 22:24:17 jhs Exp $";

constant thread_safe = 1;

#include <module.h>

inherit "module";
inherit "roxenlib";

#if __VERSION__ < 0.6
import sb_addpath;
#else
import .sb_addpath;
#endif

import Sitebuilder;


// Global variables.

object config;			// Used by sitebuilder.pike.

// Used by content_editor.pike to override the settings:
// Cookies used to specify session-specific settings.
//	"timestamp":int
//	"workarea":object(Workarea)
//	"revisions":mapping(string:string)
//	"date":string
//	"userid":int
mapping(string:mapping(string:mapping(string:string)|
               object(Workarea)|string|int)) cookies=([]);
int next_cookie;


// Connect to a sitebuilder module.

private static mapping(object:object(Site)) _sitebuilders; // Indexed on configs.
private static mapping(string:object(Workarea)) _wa_select_list = ([]);
private static object(Workarea) _workarea; // Can be destructed occasionally.
private static object(Site) _site; // Can be destructed occasionally.

int wa_not_selectable();
int sb_not_selectable();

mapping(object:object(Site)) sitebuilders()
// Returns a mapping from configs to Site objects.
{
  if (!_sitebuilders) _sitebuilders = roxen->query_var ("SiteBuilders");
  if (_sitebuilders) m_delete (_sitebuilders, 0);
  return _sitebuilders || ([]);
}

string sb_select_name (object conf)
{
  if(!conf) return "";
  return conf->name +
    (conf->query_name() != conf->name ? " (" + conf->query_name() + ")" : "");
}

array(string) sb_select_list()
{
  return ({"none"}) + sort (Array.map (indices (sitebuilders()), sb_select_name));
}

object(Site) site()
{
  if (sb_not_selectable()) return 0;
  if (_site) return _site;
  foreach (indices (sitebuilders()), object conf)
    if (sb_select_name (conf) == QUERY (sitebuilder))
      return _site = _sitebuilders[conf];
  return 0;
}

array(string) wa_select_list()
{
  array(string) sels = ({});
  _wa_select_list = ([]);
  foreach (indices (sitebuilders()), object config) {
    object(Site) site = _sitebuilders[config];
    foreach (site->wa_list(), object(Workarea) wa) {
      // FIXME: Ensure this string is unique.
      string sel = sb_select_name (config) + ": " + (wa->name ? wa->name : "nameless");
      sels += ({sel});
      _wa_select_list[sel] = wa;
    }
  }
  return ({"none"}) + sort (sels);
}

object(Workarea) workarea()
{
  if (wa_not_selectable()) return 0;
  if (!_workarea)
    wa_select_list(), _workarea = _wa_select_list[QUERY(workarea)];
  return _workarea;
}

mapping(string:object(Site)) sb_dir()
{
  mapping(object:object(Site)) sbs = sitebuilders();
  // FIXME: The names from filename_mangle are not guaranteed unique.
  return mkmapping (Array.map (indices(sbs)->query_name(), filename_mangle),
		    values (sbs));
}

mapping(string:object(Workarea)) wa_dir(object(Site)|void s)
{
  if (!s) s = site();
  if (!s) return 0;
  array(object(Workarea)) was = s->wa_list();
  mapping(string:object(Workarea)) dir =
    mkmapping (Array.map (was->name, filename_mangle), was);
  // FIXME: The names from filename_mangle are not guaranteed unique.
  return dir;
}

// workarea_path()
// [ Site, Workarea, newf ]
// 0			- Bad filename.
// [ 0, 0, 0 ]		- Top-level directory.
// [ Site, 0, 0 ]	- Site found.
// [ Site, WA, fname ]	- Site and Workarea found.
array(object(Site)|object(Workarea)|string) workarea_path(string f, object id)
{
  object(Site) sb;
  object(Workarea) wa = workarea();
  string newf = f;

  if (wa) {
    sb = wa->site;
  } else {
    if (QUERY(download_mode) == "All sites") {
      if ((< "", "/" >)[f]) {
	// Top-level directory.
	return(({ 0, 0, 0 }));
      }
      string sb_name;
      newf = "";
      if (f[0] == '/') {
	sscanf(f, "/%[^/]/%s", sb_name, newf);
      } else {
	sscanf(f, "%[^/]/%s", sb_name, newf);
      }
      sb = sb_dir()[sb_name];
      f = newf;
    } else {
      sb = site();
    }

    if (!sb) {
      return(0);
    }

    if ((< "", "/" >)[f]) {
      // Top-level directory.
      return({ sb, 0, 0 });
    }
    string wa_name;
    newf = "";
    if (f[0] == '/') {
      sscanf(f, "/%[^/]/%s", wa_name, newf);
    } else {
      sscanf(f, "%[^/]/%s", wa_name, newf);
    }
    wa = wa_dir(sb)[wa_name];

    if (!wa) {
      return(0);
    }
  }
  // Assert that newf doesn't begin with '/'.
  if ((newf != "") && (newf[0] == '/')) {
    string oldf = newf;
    newf = "";
    sscanf(oldf, "%*[/]%s", newf);
  }
  return ({ sb, wa, newf });
}


// Initialization and standard module stuff.

int wa_not_selectable() {return !((< "View", "Edit" >)[QUERY(download_mode)]);}
int sb_not_selectable() {return !(QUERY(download_mode) == "All work areas");}

void create (object conf)
{
  if (!conf) return;

  array(object(Workarea)) was =
    conf && sitebuilders()[conf] && _sitebuilders[conf]->wa_list();
  object(Workarea) wa = was && sizeof (was) && was[0];
  array(string) wa_sel_list = wa_select_list();
  string wa_sel = search (_wa_select_list || ([]), wa) || "none";

  array(string) sb_sel_list = sb_select_list();
  string sb_sel = sb_select_name (conf) || "none";

  // Kludge to shut up the "Passing value not present in the misc array" warning.
  if (search (wa_sel_list, wa_sel) < 0) wa_sel_list += ({wa_sel});
  if (search (sb_sel_list, sb_sel) < 0) sb_sel_list += ({sb_sel});

  defvar ("location", "/",
	  "Location", TYPE_LOCATION,
	  "The location of the work area filesystem. Note that setting this "
	  "to anything other than '/' for a filesystem in view mode will "
	  "probably break the site, unless relative paths are used in all "
	  "links.");
  defvar ("workarea", wa_sel,
	  "Work area", TYPE_STRING_LIST,
	  "The work area, in any running SiteBuilder on this server, that "
	  "this filesystem should access.",
	  wa_sel_list, wa_not_selectable);
  defvar ("sitebuilder", sb_sel,
	  "SiteBuilder", TYPE_STRING_LIST,
	  "The SiteBuilder whose work areas this filesystem should access.",
	  sb_sel_list, sb_not_selectable);
  defvar ("download_mode", "View", "Download mode", TYPE_STRING_LIST,
	  "The mode that files are downloaded in.<br>\n"
	  "<dl><dt>View</dt>\n"
	  "<dd>Files are sent through the RXML parser (default).</dd>\n"
	  "<dt>Edit</dt>\n"
	  "<dd>Files are sent as is, and the version remembered. Useful "
	  "for editing via eg FTP.</dd>\n"
	  "<dt>All work areas</dt>\n"
	  "<dd>Normally a work area filesystem accesses a single work area in "
	  "a SiteBuilder, but when this is active it will access all "
	  "work areas, with a subdirectory for each. That's typically useful "
	  "when you want to set up site editing through FTP. Implies "
	  "<b>Edit</b> mode.</dd>\n"
	  "<dt>All sites</dt>\n"
	  "<dd>Like <b>All work areas</b>, but with a subdirectory for "
	  "each of the sites. This is typically useful when you want "
	  "to set up site editing through FTP on a machine that uses "
	  "IP-less virtual hosting. Implies <b>Edit</b> mode.</dd>\n"
	  "</dl>",
	  ({ "View", "Edit", "All work areas", "All sites" }));
}

void check_variable (string name, mixed value)
{
  switch (name) {
    case "workarea": _workarea = 0; break;
    case "sitebuilder": _site = 0; break;
  }
}

string status()
{
  variables["workarea"][VAR_MISC] = wa_select_list();
  variables["sitebuilder"][VAR_MISC] = sb_select_list();
  if (workarea())
    return html_encode_string ("Connected to " +
			       sb_select_name (search (_sitebuilders, _workarea->site)) +
			       ": " + (_workarea->name ? _workarea->name : "nameless"));
  else if (site())
    return html_encode_string ("Connected to " +
			       sb_select_name (search (_sitebuilders, _site)));
  else
    return "Not active";
}

string query_location();

string query_name()
{
  string loc = " mounted on " + query_location();
  if (QUERY (download_mode) == "All sites") return "Edit: All sites" + loc;
  else if (workarea())
    return QUERY (download_mode) + ": " + QUERY (workarea) + loc;
  else if (site()) return "Edit: " + QUERY (sitebuilder) + ": All work areas" + loc;
  else return "Unassigned filesystem";
}

string query_provides()
{
  return("sitebuilder_workarea_filesystem");
}

array register_module()
{
  return ({MODULE_LOCATION | MODULE_PROVIDER,
	   "SiteBuilder Work Area Filesystem",
	   "Handles requests to work areas in SiteBuilder. A work area "
	   "filesystem can be used either in view or edit mode.\n"
	   "\n<ul>\n"
	   "<p><li>In view mode, all normal file parsing is performed. "
	   "Read access is required on both the work area and the file. "
	   "Uploads are never allowed.\n"
	   "<p><li>In edit mode, all files are returned \"as is\" so "
	   "that they may be edited. Write access is always required on "
	   "the work area. Read access is required on a file to download "
	   "it, write access to upload.\n"
	   "</ul>\n"
	   "\n<p>If you want to edit through FTP, named FTP must be "
	   "allowed in the server variables section, and the shell "
	   "database check must be disabled (you need to be in the "
	   "\"More options\" mode to see these variables). This has "
	   "however already been taken cared of if the \"SiteBuilder FTP "
	   "edit\" server type was chosen when this virtual server was "
	   "created.",
	   0, 0});
}

void start (int when, object conf)
{
  if (conf) config = conf;

  workarea();
  site();
}


// Error handling.

mixed vcerror_response (mapping error, object id, void|string msg)
// Takes an error mapping returned by most VCFile and VCDir functions
// and returns a suitable error response.
{
  if (!error) {
    return(http_low_answer(200, msg || ""));
  }

  int code;
  switch (error->type) {
    case "access denied":
      if (msg) error->reason = msg + (error->reason ? ": " + error->reason : "");
      if (workarea()) return workarea()->access_throw (id, error);
      else return site()->mac->access_throw (error->ppoint, error->reason);

    case "invalid file":
    case "invalid directory":
      code = 404;
      break;

    default:
      code = 200;
  }

  return http_low_answer (code, "<h1>" + html_encode_string (
    (msg ? msg + ": " : "") + error_msg (error)) + "</h1>\n");
}


// Async upload program

static class StartAsyncUpload
{
  import Sitebuilder;

  static int ignore_len;
  static object id;
  static object(VCFile) vcfile;
  static object(VCFileData) fd;
  static object my_fd;
  static function vcerror_response;

  // FIXME: Should done() and fail() be the same function,
  // that just looks at id->misc->len?
  static void done()
  {
    catch {
      my_fd->set_blocking();
    };

    mapping res;

    object handler = vcfile->wa->site->handler_for (vcfile, id);
    if (handler->upload (vcfile, id, fd)) res = fd->upload_end (0);
    else res = fd->upload_end (1);

    res = vcerror_response(res, id);
    if (id->send_result) {
      if (!id->my_fd) {
	// Kludge!
	id->my_fd = my_fd;
      }
      id->send_result(res);
    } else {
      // Obsolete in Roxen 1.4
      my_fd->write("HTTP/1.0 200 Transfer Complete.\r\n"
		       "Content-Length: 0\r\n"
		       "\r\n");
      my_fd->close();
    }
    fd = 0;
    id = 0;
    my_fd = 0;
  }

  static void fail()
  {
#ifdef WA_DEBUG
    werror(sprintf("StartAsyncUpload::fail():\n"
		   "%s\n", describe_backtrace(backtrace())));
#endif /* WA_DEBUG */
    catch {
      my_fd->set_blocking();
    };
    mapping res = vcerror_response(fd->upload_end(1), id);
    if (id->send_result) {
      if (!id->my_fd) {
	// Kludge!
	id->my_fd = my_fd;
      }
      id->send_result(res);
    } else {
      // Obsolete in Roxen 1.4
      my_fd->write("HTTP/1.0 403 Failed.\r\n"
		   "Content-Length: 0\r\n"
		   "\r\n");
      my_fd->close();
    }
    fd = 0;
    id = 0;
    my_fd = 0;
  }

  static void got_data(mixed ignored, string data)
  {
    if (sizeof(data)) {
      mixed err = catch {
	string d = data[..id->misc->len-1];
	int res = fd->fd->write(d);
	if (res < 0) {
	  // Some error...
	  fail();
	  return;
	} else if (res < sizeof(d)) {
	  // Disk full?
	  fail();
	  return;
	} else if (id->misc->len == sizeof(d)) {
	  // We're done!
	  id->data = data[id->misc->len..];
	  if (!sizeof(data)) {
	    id->data = 0;
	  }
	  id->misc->len = 0;
	  done();
	  return;
	}
	id->misc->len -= sizeof(data);
      };
      if (err) {
#ifdef WA_DEBUG
	werror(sprintf("StartAsyncUpload::got_data()\n"
		       "%s\n", describe_backtrace((mixed *) err)));
#endif
	fail();
      }
      return;
    }
    // Does this case exist?
    if (!ignore_len && id->misc->len) {
      // Fail.
      fail();
    } else {
      done();
    }
  }

  static void con_closed(mixed ignored)
  {
    if (!ignore_len && id->misc->len) {
      // Connection broken.
#ifdef WA_DEBUG
      werror(sprintf("StartAsyncUpload::con_closed()\n"
		     "Still %d bytes left to save!\n", id->misc->len));
#endif /* WA_DEBUG */
      fail();
    } else {
      // OK.
      done();
    }
  }

  static void start()
  {
#ifdef WA_DEBUG
    werror(sprintf("StartAsyncUpload::start()\n"
		   "Will send %d bytes\n", id->misc->len));
#endif /* WA_DEBUG */

    ignore_len = id->misc->len == 0x7fffffff;
    
    if (!id->misc->len) {
      done();
    }

    // Write any data already read by the protocol module
    if (id->data && sizeof(id->data)) {
      fd->fd->write(id->data[..id->misc->len - 1]);
      if (sizeof(id->data) < id->misc->len) {
	id->misc->len -= sizeof(id->data);
	id->data = 0;
      } else {
	id->data = id->data[id->misc->len..];
	done();
      }
    }

    // Ready to receive data callback.
    if (id->ready_to_receive) {
      id->ready_to_receive();
    } else {
      // Obsolete in Roxen 1.4
      if (id->clientprot == "HTTP/1.1") {
	my_fd->write("HTTP/1.1 100 Continue\r\n");
      }
    }

    my_fd->set_nonblocking(got_data, 0, con_closed);
  }

  void create(object(VCFileData) fd_, object id_,
	      function vcerror_response_, object(VCFile) vcfile_)
  {
    fd = fd_;
    id = id_;
    // Needed since my_fd in the id object can be zapped by the protocol
    // module.
    my_fd = id->my_fd;
    vcerror_response = vcerror_response_;
    vcfile = vcfile_;

    // To allow the caller to return.
    call_out(start, 0);
  }
}

static object(VCFile)|object(VCDir) get_vcobj(string f, object id, object(Workarea) wa)
{
  if (id->cookies->SiteBuilder) {
#ifdef SBDEBUG_COOKIE
    werror("WA: Cookie: " + id->cookies->SiteBuilder + ".\n");
#endif /* SBDEBUG_COOKIE */
    mapping(string:mapping(string:string)|object(Workarea)|string|int) cookie =
      cookies[id->cookies->SiteBuilder];

#ifdef SBDEBUG_COOKIE
    werror(sprintf("WA: Cookie:%O\n", cookie));
#endif /* SBDEBUG_COOKIE */
    if (cookie && (time(1) < cookie->timestamp)) {
#ifdef SBDEBUG_COOKIE
      werror("WA: Cookie valid.\n");
#endif /* SBDEBUG_COOKIE */

      id->misc->wa = wa = cookie->workarea;
      id->misc->sb = wa->site;
      id->misc->undelete_mode = cookie->undelete_mode;

      if (cookie->revisions && cookie->revisions[f]) {
	string revision = cookie->revisions[f];
#ifdef SBDEBUG_COOKIE
	werror("WA: Revision: "+revision+"\n");
#endif /* SBDEBUG_COOKIE */
	return(wa->vcobj(f, id, 0, revision));
      } else {
	string date = cookie->date;
#ifdef SBDEBUG_COOKIE
	werror("WA: Date: "+date+"\n");
#endif /* SBDEBUG_COOKIE */
	return(wa->vcobj(f, id, 0, 0, date));
      }
    } else if (cookie) {
#ifdef SBDEBUG_COOKIE
      werror(sprintf("WA: cookie has timed out: %d < %d\n",
		     time(1), cookie->timestamp));
#endif /* SBDEBUG_COOKIE */
      m_delete(cookies, id->cookies->SiteBuilder);
    }
  }

  return(wa->vcobj(f, id));
}

// MODULE_LOCATION interface.

string query_location()
{
  return QUERY (location);
}

array stat_file(string f, object id)
{
  //werror ("stat_file " + f + "\n");
  array res = workarea_path(f, id);

  if (!res) {
    return(0);
  }

  if (!res[1]) {
    // Top-level directory
    return(({ 0555, -3, 0, 0, 0, 0, 0, 0, 0, 0, 0 }));
  }

  object(Workarea) wa = res[1];
  f = res[2];

  if (!wa || wa->access_try (id) <
      (QUERY (download_mode) == "View" ? PERM_READ : PERM_WRITE)) return 0;

  object(VCFile)|object(VCDir) obj = get_vcobj(f, id, wa);

  if (!obj) {
    return(0);
  }

  array stat = obj->stat (id);
  if (stat && QUERY(download_mode) == "View") stat[0] &= 0555;
  return stat;
}

string real_file(string f, object id)
{
  // We don't exist in the real filesystem.
  return(0);
}

mixed find_dir(string f, object id)
{
  //werror ("find_dir " + f + "\n");
  int reqperm = QUERY (download_mode) == "View" ? PERM_READ : PERM_WRITE;

  array res = workarea_path(f, id);
  if (!res) {
    return(0);
  }

  if (!res[0]) {
    // FIXME: Check permissions?
    return indices (sb_dir());
  } else if (!res[1]) {
    array(string) wa_dirlist = ({});
    mapping(string:object(Workarea)) walist = wa_dir (res[0]);
    foreach (indices (walist), string name) {
      if (walist[name]->access_try (id) >= reqperm) wa_dirlist += ({name});
    }
    return wa_dirlist;
  }

  object(Workarea) wa = res[1];
  f = res[2];

  if (!wa || wa->access_try (id) < reqperm) return 0;

  object(VCFile)|object(VCDir) obj = get_vcobj(f, id, wa);

  if (!obj || obj->isfile) {
    return(0);
  }

  mapping|array(object) list = obj->list (id);
  if (mappingp(list)) return 0;

  // FIXME: Filtering?
  // FIXME: Filter on downloadp?
  return(list->name(id) - ({0}));
}

mixed find_file (string path, object id)
{
  //werror ("find_file " + path + "\n");
  array res = workarea_path(path, id);

  if (!res) {
    return(0);
  }
  if (!res[1]) {
    // Top-level directory.
    return(-1);
  }

  object(Workarea) wa = res[1];
  string f = res[2];

  if (!wa) return 0;
  if (wa->access_try (id) <
      (QUERY (download_mode) == "View" ? PERM_READ : PERM_WRITE))
    return wa->access_throw (id);

  {
    int myid = wa->site->mac->id_get_id (id);

    if (id->variables->__sb_force_login) {
      int notid = (int) id->variables->__sb_force_login;
      if (myid == notid || wa->site->mac->id_is_everyone (id))
	if (sizeof (f))
	  return http_redirect (QUERY (location) +
				"?path=" + path +
				"&__sb_force_login=" + notid + "&" + random (1<<30));
	else {
	  m_delete (id->variables, "__sb_force_login");
	  // It's really nice to have a good fallback page.
	  mixed res = find_file (id->variables->path || path, id);
	  if (!mappingp (res)) return wa->access_throw (id);
	  else {
	    mapping hdrs = wa->access_throw (id);
	    if (hdrs->error) res->error = hdrs->error;
	    if (!res->extra_heads) res->extra_heads = ([]);
	    if (hdrs->extra_heads) res->extra_heads += hdrs->extra_heads;
	    return res;
	  }
	}
      else
      	if (id->variables->path) {
	  id->raw_url = id->not_query = path = id->variables->path;
	  m_delete (id->variables, "__sb_force_login");
	  m_delete (id->variables, "path");
	  return find_file (path, id);
	}
    }

    else if (string force = id->variables->__sb_force_userid
			  || id->cookies->__sb_force_userid )
      if (myid != (int)force) return wa->access_throw (id);
  }

  string fname = combine_path("/", QUERY (location) + "/" + f);

  id->misc->wa = wa;
  id->misc->sb = wa->site;
  id->misc->localpath = f;

  switch(id->method) {
  case "GET":
  case "HEAD":
  case "POST":
    if (id->cookies->SiteBuilder)
      // Disable all internal caches if the cookie is active.
      id->pragma["no-cache"] = 1;

    object(VCFile)|object(VCDir) obj = get_vcobj(f, id, wa);

    if (!obj) {
      // File not found.
      return(0);
    }
    if (obj->perm (id) < PERM_READ)
      return vcerror_response (obj->access_denied (id, obj->isdir ? "list" : "view"),
			       id, "Couldn't get " + fname);
    if (obj->isdir) {
      return(-1);
    }
    object(VCFileData) fd;

    int view_mode =
      QUERY(download_mode) == "View" || obj->perm (id) < PERM_WRITE || obj->isoldfile;

    // Downloading of this file allowed?
    mapping ct_data = wa->site->ct_data_for(obj);
    if (!ct_data->downloadp) {
      // Downloading not allowed.
      // FIXME: Permission denied?
      return(0);
    } else {
      if (view_mode) {
	fd = obj->is_user_op (id, "view", id->misc->undelete_mode) &&
	  obj->view(id, id->misc->undelete_mode);
      } else {
	fd = obj->is_user_op (id, "download", id->misc->undelete_mode) &&
	  obj->download(id, id->misc->undelete_mode);
      }
    }
    if (!fd) return 0;
    if (mappingp (fd))
      return vcerror_response ((mapping) fd, id, "Couldn't get " + fname);

    mapping headers = ([]);
    foreach(indices(fd->md), string md) {
      if (md[..4] == "http-") {
	headers[md[5..]] = fd->md[md];
      }
    }

    object|string|mapping res = fd->fd;

    object handler = wa->site->handler_for(headers["content-type"]);

    if (QUERY (download_mode) == "View") {
      // View mode.
      // Call any associated handlers.
      res = handler->view(res, headers, id, fd->md, obj);
      if (headers["content-type"] && headers["content-type"][..5] == "roxen/")
	// Special content type that should be handled by other roxen
	// modules after sitebuilder.
	return res;
    } else {
      // Download mode.
      // Call any associated handlers.
      res = handler->download(res, headers, id, fd->md, obj);
    }

    if(objectp(res)) {
      array st = res->stat();
      if (!st) {
	throw(({ "File disappeared\n", backtrace() }));
      } else if (st[1] < 0) {
	throw(({ sprintf("Unexpected file length: %d\n", st[1]),
		 backtrace() }));
      }
      int len = st[1] - res->tell();

      if (len < 0) {
	// Error
	throw(({ sprintf("Negative file length: %d\n", len), backtrace() }));
      }

      if (len > 0) {
	res = http_file_answer(res, headers["content-type"], len);
      } else {
	// Special case to avoid broken recalculation of the reply length.
	res = http_string_answer("", headers["content-type"]);
      }
    }

    if(stringp(res))
      res = http_string_answer(res, headers["content-type"]);      

    if(mappingp(res) && !res->extra_heads) {
      // Return the rest of the http-* headers.
      m_delete(headers, "content-type");
      res->extra_heads = headers;

      // Try to make sure we aren't cached if the cookie is active.
      if (id->cookies->SiteBuilder) {
	res->since = res->extra_heads->Expires = http_date(0);
      }
    }

    if (res->type[0..4] == "roxen" || res->type[0..6] == "unknown")
      m_delete( res, "type" );

    return(res);
    break;

  case "DELETE":
    /* Not needed and breaks XEmacs. Remove it for the time being.
    if (QUERY (download_mode) == "View") break;

    object(VCFile)|object(VCDir) obj = get_vcobj(f, id, wa);

    if (!obj) {
      // File not found.
      return(0);
    }
    mapping res = obj->delete(id);
    if (mappingp (res)) {
      if (!obj->exists (id)) return 0;
      return vcerror_response (res, id, "Failed to DELETE " + fname);
    }
    return(http_low_answer(200, sprintf("<h1>%O DELETED from server</h1>\n",
					fname)));
    */
    break;

  case "MKDIR":
    if (QUERY (download_mode) == "View") break;

    // NOTE: Ignores any cookies.
    object(VCDir) obj = wa->vcdir(f, id);

    mixed created = obj->add(id, "");
    if (mappingp (created))
      return vcerror_response (res, id, "Failed to create directory " + fname);
    return(http_low_answer(200, sprintf("<h1>Directory %O created</h1>",
					fname))); 
    break;

  case "PUT":
    if (QUERY (download_mode) == "View") break;

    // NOTE: Ignores any cookies.
    object(VCFile) obj = wa->vcfile(f, id);

    // FIXME: Check downloadp?

    mapping md=([]);
    mapping|object(VCFileData) fd;
    object|mapping vcdata=obj->metadata(id);
    
    if(mappingp(vcdata))
    {
      string content_type = wa->site->find_content_type_from_filename(f);
      md["http-content-type"] = content_type;
      if(content_type=="text/html")
	md["template"]="default.tmpl";
      fd = obj->upload(id,md);
    }
    else
      fd = obj->upload(id);
    
    if (mappingp(fd)) {
      // Failure.
      return(vcerror_response(fd, id, "Failed to upload " + fname));
    }

    StartAsyncUpload(fd, id, vcerror_response, obj);

    return(http_pipe_in_progress());
    break;

  case "MV":
  case "MOVE":
  default:
    if (QUERY (download_mode) == "View") break;
    return 0;
  }
  return 0;
}
