// Metadata handling class for SiteBuilder.

// $Id: MetaData.pmod,v 1.14 1999/05/12 01:15:42 js Exp $

#if constant (roxen)

inherit "wizard";

#else

// Some dummies to be able to compile outside roxen.
string html_encode_string (string s)
  {throw (({"Outside roxen\n", backtrace()}));}
string parse_wizard_page (string f, object i, string w, void|string p,
			  void|string base)
  {throw (({"Outside roxen\n", backtrace()}));}
string html_table (array(string) s, array(array(string)) t, mapping|void o)
  {throw (({"Outside roxen\n", backtrace()}));}
string html_error (string s, object i)
  {throw (({"Outside roxen\n", backtrace()}));}

#endif

import Sitebuilder;

array get_vars(object site, string content_type, void|object id)
{
  array lang_ext,lang_name, lang=site->get_languages();
  lang_ext=indices(lang);
  lang_name=values(lang)->name;
  sort(lang_name,lang_ext);
  
  return
  ({ ([ "title":      "Type",
	"name":       "http-content-type",
	"doc":        "This is the type of the file.",
	"default":    "HTML",
	"defaults":   site->list_content_type_names(), // mapping
	"type":       "select",
	"apply":      1,
	"applicable": 1 ]),

     ([ "title":      "Template",
	"name":       "template",
	"doc":        "This is the template used on this page.",
	"default":    "default.tmpl",
	"defaults":   ({"__0:No template"})+(id ? site->list_templates(id) : ({})),
	"type":       "select",
	"applicable": content_type=="text/html",
     ]),

     ([ "title":      "Language",
	"name":       "language",
	"doc":        "The language the document is written in.",
	"default":    "",
	"defaults":    mkmapping(lang_name,lang_ext),
	"type":       "select",
	"applicable": 1 ]),

     ([ "title":      "Title",
	"name":       "title",
	"doc":        "This is the title of the page. Make sure"
                      " that it accurately describes it.",
	"default":    "",
	"type":       "string",
	"applicable": 1 ]),

     ([ "title":      "Keywords",
	"name":       "keywords",
	"doc":        "Comma separated list of document keywords. These are primarily used "
	              "when search-engines are indexing the site.",
	"default":    "",
	"type":       "string",
	"applicable": 1 ]),

     ([ "title":      "Description",
	"name":       "description",
	"doc":        "Document description. This is also primarily "
	              "used when search-engines are indexing the site.",
	"default":    "",
	"type":       "text",
	"args":       "wrap=on",
	"applicable": 1 ])
  });
}

mapping complete_metadata (object site, mapping md)
// Fills in the defaults for the missing metadata fields in md.
{
  if (!md["http-content-type"]) md["http-content-type"] = "application/octet-stream";
  foreach (get_vars (site, md["http-content-type"]), mapping var) {
    if (var->applicable && zero_type (md[var->name]))
      // FIXME: var["default"] might not always contain the thing we want here.. :P
      md[var->name] = var["default"];
  }
  return md;
}

#define fix(x) replace(x,"'","'\"'\"'")

array describe_metadata_var(mapping in, int|void no_docs)
{
  if(!in->applicable)
    return 0;
  array a=
  ({ "<font size=+1><b>"+in->title+"</b></font>",
     "<var name='meta_"+in->name+
     "' default='"+in["default"]+"' "+
     (in->defaults?("choices='"+
		    (arrayp(in->defaults)?
		     fix(in->defaults*","):
		     Array.map(sort(indices(in->defaults)),
			       lambda(string n,
				      mapping(string:string) cts)
			       {
				 return fix(cts[n]+":"+n);
			       }, in->defaults)*",")
		    +"' "):"")+
      " type='"+in->type + "'" + (in->args ? " " + in->args : "") + ">"+
     (in->apply?"<input type=submit name=apply value='Change'>":"") });
  if(!no_docs)
    a+= ({ ({ "<font size=-1><i>"+in->doc+"</i></font>" }) });
  return a;
}

array describe_metadata_var_view(mapping in, object id)
{
  return in->applicable?
  ({ in->title,
     html_encode_string(lambda(mapping in, string val) {
			  return mappingp(in->defaults)?
			    search(in->defaults, val)||val : val;
			}(in, id->variables["meta_"+in->name]||""))
  }) : 0;
}

string edit(object|void vcfile, object id, void|multiset metadata_list, void|int view)
{
  mapping metadata;
  if (!vcfile) metadata = ([]);
  else {
    mapping|object vcfd = view ?
      vcfile->metadata (id, id->misc->undelete_mode == "1") :
      vcfile->download (id);
    if (objectp (vcfd)) metadata = copy_value (vcfd->md);
  }
  if (!metadata)
    return "<b>Metadata is not accessible.</b>";

  if(!id->variables->inited)
    foreach(indices(metadata), string datum)
      id->variables["meta_"+datum]=metadata[datum];
  string res=
    "<input type=hidden name=inited value=1>"
    "<b>Metadata</b><br>\n";

  foreach(get_vars(id->misc->sb, metadata["http-content-type"],id), mapping var)
  {
    string x;
    if(x=id->variables["meta_"+var->name])
      metadata[var->name]=x;
  }

  array vars=get_vars(id->misc->sb, metadata["http-content-type"],id);
  if(metadata_list)
    vars=Array.filter(vars,lambda(mapping m, multiset mdlist)
			   {
			     return mdlist[m->name];
			   },metadata_list);

  if(view)
    res += html_table( ({ "Item", "Value" }),
		       Array.map( vars,
				  describe_metadata_var_view,id) - ({ 0 }));
  else
    res += html_table( ({ "Item", "Value" }),
		       Array.map( vars,
				  describe_metadata_var,id,1) - ({ 0 }));
  res = replace (res, "<td nowrap>", "<td>");
  return res;//+sprintf("md: <pre>%O</pre>",metadata);
}

string view(object vcfile, object id, void|multiset metadata_list)
{
  return edit(vcfile,id,metadata_list,1);
}

string edit_done(object vcfile, object id)
{
  if(!vcfile)
    return parse_wizard_page(html_error("Internal error, file went away.",
					id),
			     id,
			     "done");
  mapping fault;
  array(string) metas = glob("meta_*", indices(id->variables));

  if(fault = vcfile->set_metadata(id, mkmapping(Array.map(metas, `[], 5,
							  0x7fffffff),
						rows(id->variables, metas))))
    return parse_wizard_page(html_error(Sitebuilder.error_msg (fault), id),
			     id,
			     "done");

  return 0;
}


// Conflict resolution.

constant color_a = "#d0e0d0";
constant color_b = "#d0d0e0";
constant color_old = "#d4d4d4";
constant color_conflict = "#e0d0d0";

int merge_content_type_ok (object(VCFileConflict) vcfc)
{
  return vcfc->a_md["http-content-type"] == vcfc->b_md["http-content-type"] ||
    vcfc->state->meta_content_type;
}

void merge_forget_content_type (object(VCFileConflict) vcfc)
{
  m_delete (vcfc->state, "meta_content_type");
}

int merge_analyze (object(VCFileConflict) vcfc, void|object id)
// Return values: 0: no differences, 1: differences but no conflicts,
// 2: only decided conflicts, 3: undecided conflicts.
{
#ifdef SBDEBUG
  if (!mappingp (vcfc->state)) throw (({"vcfc->state not initialized\n", backtrace()}));
#endif
  mapping(string:mixed) state = vcfc->state;
  mapping(string:mixed) a = vcfc->a_md, b = vcfc->b_md, old = vcfc->old_md;

  mapping(string:string) variables = id ? id->variables : ([]);
  state->meta_missing = ({});
  if (!state->meta_sel) state->meta_sel = ([]);
  mapping(string:string) sel = state->meta_sel;

  if (a["http-content-type"] == b["http-content-type"])
    state->meta_content_type = a["http-content-type"];
  else if ((<"a", "b">)[variables["meta_http-content-type"]]) {
    sel["http-content-type"] = variables["meta_http-content-type"];
    switch (sel["http-content-type"]) {
      case "a": state->meta_content_type = a["http-content-type"]; break;
      case "b": state->meta_content_type = b["http-content-type"]; break;
    }
  }

  int res = 0;
  foreach (get_vars (vcfc->this_vcfile->wa->site, state->meta_content_type), mapping item)
    if (item->applicable) {
      string var = item->name;
      if (a[var] != b[var] || zero_type (a[var]) != zero_type (b[var])) {
	if (res < 1) res = 1;
	if (zero_type (b[var]) || old && b[var] == old[var])
	  {if (!sel[var]) sel[var] = "a";}
	else if (zero_type (a[var]) || old && a[var] == old[var])
	  {if (!sel[var]) sel[var] = "b";}
	else if (res < 2) res = 2;
	if ((<"a", "b">)[variables["meta_" + var]])
	  sel[var] = variables["meta_" + var];
	if (!sel[var]) {
	  state->meta_missing += ({lower_case (item->title)});
	  if (res < 3) res = 3;
	}
	else variables["meta_" + var] = sel[var];
      }
    }
  return res;
}

#define VALUE(val, tmp) (zero_type (val) ? "<i>no value</i>" : \
			 (tmp = (string) (val), \
			  sizeof (tmp) ? html_encode_string (tmp) : "&nbsp;"))

string merge_form (object(VCFileConflict) vcfc, object id)
{
#ifdef SBDEBUG
  if (!mappingp (vcfc->state) || !vcfc->state->meta_sel)
    throw (({"vcfc not analyzed or has no differences\n", backtrace()}));
#endif
  mapping(string:mixed) a = vcfc->a_md, b = vcfc->b_md, old = vcfc->old_md;
  string content_type = vcfc->state->meta_content_type;
  mapping(string:string) sel = vcfc->state->meta_sel;

  string res =
    "<table width=100% cellspacing=0 cellpadding=0 border=0 bgcolor=#000000><tr><td>"
    "<table width=100% cellspacing=0 cellpadding=3 border=0 bgcolor=#eeeeee>\n"
    "<tr valign=bottom><th align=left>Item</th>"
    "<th bgcolor=" + color_a + ">" + vcfc->state->msg_a_column + "</th>"
    "<th bgcolor=" + color_b + ">" + vcfc->state->msg_b_column + "</th>"
    //"<th><img src=/internal-roxen-unit alt='' width=1 height=1></th>"
    "<th align=left width=99%>Value</th></tr>\n"
    "<tr></tr>\n";

  foreach (get_vars (vcfc->this_vcfile->wa->site, content_type, id), mapping item)
    if (item->applicable) {
      string var = item->name, tmp;
      if (a[var] == b[var] && zero_type (a[var]) == zero_type (b[var]) ||
	  var == "http-content-type" && content_type)
	res += "<tr>"
	  "<td>" + replace (html_encode_string (item->title), " ", "&nbsp;") + "</td>"
	  "<td bgcolor=" + color_a + ">&nbsp;</td>"
	  "<td bgcolor=" + color_b + ">&nbsp;</td>"
	  //"<td><img src=/internal-roxen-unit alt='' width=1 height=1></td>\n"
	  "<td>" + (var == "http-content-type" ?
		    VALUE (content_type, tmp) : VALUE (a[var], tmp)) + "</td></tr>\n";
      else {
	string cl = old && a[var] == old[var] ? color_old : color_a;
	string cr = old && b[var] == old[var] ? color_old : color_b;
	string cn = cl != color_old && cr != color_old && color_conflict;
	res += "<tr>"
	  "<td" + (cn ? " bgcolor=" + cn : "") + " rowspan=2>"
	  "<a name='meta_" + var + "'>" +
	  replace (html_encode_string (item->title), " ", "&nbsp;") + "</a></td>"
	  "<td bgcolor=" + cl + ">"
	  "<var type=radio name=meta_" + var + " value=a>" +
	  (cl == color_old ? "&nbsp;Old&nbsp;" : "") + "</td>"
	  "<td bgcolor=" + cl + ">&nbsp;</td>"
	  //"<td><img src=/internal-roxen-unit alt='' width=1 height=1></td>\n"
	  "<td bgcolor=" + cl + ">" + VALUE (a[var], tmp) + "</td></tr>\n"
	  "<tr>"
	  "<td bgcolor=" + cr + ">&nbsp;</td>"
	  "<td bgcolor=" + cr + ">"
	  "<var type=radio name=meta_" + var + " value=b>" +
	  (cr == color_old ? "&nbsp;Old&nbsp;" : "") + "</td>"
	  //"<td><img src=/internal-roxen-unit alt='' width=1 height=1></td>\n"
	  "<td bgcolor=" + cr + ">" + VALUE (b[var], tmp) + "</td></tr>\n";
      }
      res += "<tr></tr>\n";
    }

  return res + "</table></td></tr></table>\n";
}

string merge_verify (object(VCFileConflict) vcfc, object id)
{
  switch (merge_analyze (vcfc, id)) {
    case 0..2: return 0;
    case 3:
      if (sizeof (vcfc->state->meta_missing) == 1)
	return "The conflict on " + vcfc->state->meta_missing[0] + " must be decided.";
      else
	return "The conflicts on " + String.implode_nicely (vcfc->state->meta_missing) +
	  " must be decided.";
    case 4:
      return "The content type must be decided.";
    default:
      throw (({"Unknown return value from merge_analyze()\n", backtrace()}));
  }
}

mapping(string:mixed) merge_done (object(VCFileConflict) vcfc, object id)
{
  if (merge_analyze (vcfc, id) > 2) return 0;

  mapping(string:mixed) new = ([]);
  mapping(string:mixed) a = vcfc->a_md, b = vcfc->b_md, old = vcfc->old_md;
  mapping(string:string) sel = vcfc->state->meta_sel;

  foreach (get_vars (vcfc->this_vcfile->wa->site, vcfc->state->meta_content_type, id),
	   mapping item)
    if (item->applicable) {
      string var = item->name;
      if (a[var] == b[var]) new[var] = a[var];
      else switch (sel[var]) {
	case "a": new[var] = a[var]; break;
	case "b": new[var] = b[var]; break;
	default: throw (({"Undecided conflict\n", backtrace()}));
      }
    }

  return new;
}
