// Handles an old revision of a directory tree. Is in all relevant
// aspects identical to a VCDir.

// $Id: VCOldDir.pike,v 1.14 1999/08/29 18:09:20 mast Exp $

import Sitebuilder;
import VC.CVS;

constant isdir = 1;
constant isolddir = 1;
object isfilep() {return 0;}
object isdirp() {return this_object();}

object/*(Workarea)*/ wa;


// Debugging tools.

#ifdef SBDEBUG
#define OPMSG(_msg) sbdebug ("VCOldDir ?:" + (wa && wa->tag) + \
			     ":/" + vcdir->real_abspath() + " rev " + \
			     revision + " date '" + date + "' " + _msg + "\n")
#else
#define OPMSG(_msg)
#endif


// Identity.

private static object/*(VCDir)*/ vcdir;
private static string revision, date;
private static object(versionedfile) vinfofile;
private static mapping(string:mixed) md;

void create (object/*(Workarea)*/ _wa, object/*(VCDir)*/ _vcdir,
	     string _revision, void|string _date)
{
  wa = _wa, vcdir = _vcdir, revision = _revision, date = _date;

  // FIXME: Handle an incorrect revision more gracefully?

  vinfofile = versionedfile (_vcdir->getinfofile()->file, revision, date);
  md = vinfofile->metadata();
  OPMSG ("create");
}


// Error types.

#define ERR_ACCESS_DENIED(_ppoint) \
  (["type": "access denied", "ppoint": (_ppoint), "vcobj": this_object()])
#define ERR_INVALID_OP_REMOVED(_op) \
  (["type": "invalid operation", "operation": (_op), "state": "N/R", \
    "reason": "Directory is removed.", \
    "vcobj": this_object()])
#define ERR_INVALID_OP_N_A(_op) \
  (["type": "invalid operation", "operation": (_op), "state": "N/E", \
    "reason": "Not applicable on a directory with an old version.", \
    "vcobj": this_object()])


// Caches.

private static int|object/*(VCOldDir)*/ parent;

private static nomask object/*(VCOldDir)*/ update_parent()
{
  if (parent) return intp (parent) ? 0 : parent;
  object/*(VCDir)*/ updir = vcdir->ascend (0, 1, 1);
  if (!updir) {			// At root.
    parent = -1;
    return 0;
  }
  // FIXME: Convert revision to date here.
  return parent = wa->_vcoldobj (updir, revision, date);
}



// Permission handling.

nomask int perm (object id) {return vcdir->perm (id);}
nomask object(Perm) permobj (object id) {return vcdir->permobj (id);}
object/*(VCOldDir)*/ has_ppoint() {return vcdir->has_ppoint() && this_object();}


// State queries.

private static constant valid_ops_normal = (<
  "ascend", "descend", "metadata", "list"
>);

private static constant op_perms = ([
  "ascend": PERM_NONE, "descend": PERM_NONE, "metadata": PERM_READ, "list": PERM_READ,
  "purge": PERM_WRITE,
]);

nomask string userstat (object id, void|int undelete_mode, void|int superuser)
// Returns "none", "exists" or "removed" (only in undelete mode).
// Returns 0 if this directory shouldn't be shown at all.
{
  if (!superuser && vcdir->perm (id) == PERM_NONE) return 0;
  return md->__removed ? undelete_mode ? "removed" : "none" : "exists";
}

nomask string userstat_ea (object id, void|int undelete_mode, void|int superuser)
  {return userstat (id, undelete_mode, superuser);}
nomask string userstat_rep (object id, void|int undelete_mode, void|int superuser)
  {return userstat (id, undelete_mode, superuser);}

nomask object/*(VCOldDir)*/ is_valid_op (object id, string operation,
					 void|int undelete_mode, void|int superuser)
{
  if (operation == "ascend")
    return update_parent() && parent->exists (id, undelete_mode, superuser);
  if (!superuser && vcdir->perm (id) < op_perms[operation]) return 0;
  if (!undelete_mode && md->__removed) return 0;

  switch (operation) {
    case "purge":
      return (superuser || wa->site->purge_access (id)) && this_object();

    case "add_ppoint": case "remove_ppoint":
      return (superuser || wa->site->ppoint_edit_access (id)) && this_object();
  }

  return valid_ops_normal[operation] && this_object();
}

nomask object/*(VCOldDir)*/ is_user_op (object id, string operation,
					void|int undelete_mode, void|int superuser)
{
  return is_valid_op (id, operation, undelete_mode, superuser);
}

nomask object/*(VCOldDir)*/ accessible (object id)
{
  return vcdir->perm (id) != PERM_NONE && this_object();
}

nomask object/*(VCOldDir)*/ exists (object id, void|int undelete_mode, void|int superuser)
{
  if (!superuser && vcdir->perm (id) == PERM_NONE) return 0;
  return (undelete_mode || !md->__removed) && this_object();
}

nomask string abspath (object id, void|int undelete_mode, void|int superuser)
{
  return exists (id, undelete_mode, superuser) && vcdir->real_abspath();
}

nomask string name (object id, void|int undelete_mode, void|int superuser)
{
  return exists (id, undelete_mode, superuser) && vcdir->real_name();
}

nomask array stat (object id, void|int undelete_mode, void|int superuser)
{
  if (!exists (id, undelete_mode, superuser)) return 0;
  array(int|string) s = vcdir->getdir()->stat();
  if (!s) return 0;
  s[0] = 0555;
  s[5] = s[6] = md->__removed ? undelete_mode ? "removed" : "none" : "exists";
  return s;
}


// Directory operations.

#define PERM_CHECK(_id, _needperm, _superuser) \
  if (!_superuser) { \
    object(Perm) pobj = vcdir->permobj (_id); \
    if (pobj->perm < (_needperm)) return ERR_ACCESS_DENIED (pobj->ppid); \
  }

// Protection point handling.

mapping|void add_ppoint (object id, void|int superuser)
  {return vcdir->add_ppoint (id, superuser);}
mapping|void remove_ppoint (object id, void|int superuser)
  {return vcdir->remove_ppoint (id, superuser);}

// Nondestructive operations (do not lock the VCDir object).

mapping|object/*(VCOldDir)*/ ascend (object id, void|int undelete_mode,
				     void|int superuser)
{
  object/*(VCOldDir)*/ up = update_parent();
  if (!up) return 0;
  if (!up->exists (id, undelete_mode, superuser)) {
    object(Perm) pobj = up->permobj (id);
    return pobj->perm == PERM_NONE ?
      ERR_ACCESS_DENIED (pobj->ppid) : ERR_INVALID_OP_REMOVED ("ascend");
  }
  return up;
}

mapping|object/*(VCOldFile|VCOldDir)*/ descend (
  object id, string name, void|int undelete_mode, void|int superuser)
{
  object vcobj = vcdir->descend (id, name, 1, 1);
  if (!vcobj) return 0;
  // FIXME: Convert revision to date here.
  vcobj = wa->_vcoldobj (vcobj, revision, date);
  if (!vcobj->exists (id, undelete_mode, superuser)) {
    object(Perm) pobj = vcobj->permobj (id);
    return pobj->perm == PERM_NONE ?
      ERR_ACCESS_DENIED (pobj->ppid) : ERR_INVALID_OP_REMOVED ("ascend");
  }
  return vcobj;
}

mapping|object(VCFileData) metadata (object id, void|int undelete_mode,
				     void|int superuser)
{
  PERM_CHECK (id, PERM_READ, superuser);
  mapping _md = md;
  if (_md->__removed) {
    if (!undelete_mode) return ERR_INVALID_OP_REMOVED ("metadata");
    _md -= (["__removed": 1]);
  }
  return VCFileData (_md, 0);
}

mapping|array(object/*(VCFile|VCDir)*/) list (object id, void|int undelete_mode,
					      void|int superuser)
{
  PERM_CHECK (id, PERM_READ, superuser);
  mapping|array(object) list = vcdir->list (id, 1, 1);
  if (mappingp (list)) return list;
  // FIXME: Convert revision to date here.
  return Array.map (list, wa->_vcoldobj, revision, date);
}

// Destructive operations (lock the VCDir object).

// mapping|void revert (object id, int warning, int superuser, object(VCOldDir) vcodir)

mapping|void purge (object id, void|int superuser)
{
  OPMSG ("purge");
  return vcdir->purge (id, superuser);
}

#if constant (cpp)
#define INVALID_FUN(_name) \
  mapping _name() {return ERR_INVALID_OP_N_A (#_name);}
#else
#define INVALID_FUN(_name) \
  mapping _name() {return ERR_INVALID_OP_N_A ("_name");}
#endif

INVALID_FUN (add);
INVALID_FUN (set_metadata);
INVALID_FUN (discard);
INVALID_FUN (update);
INVALID_FUN (delete);
INVALID_FUN (undelete);
INVALID_FUN (commit);
INVALID_FUN (copy);		// FIXME?
INVALID_FUN (move);
INVALID_FUN (join);
