// Handles an old revision of a file. Is in all relevant aspects
// identical to a VCFile.

// $Id: VCOldFile.pike,v 1.16 1999/08/29 18:12:28 mast Exp $

import Sitebuilder;
import VC.CVS;

constant isfile = 1;
constant isoldfile = 1;
object isfilep() {return this_object();}
object isdirp() {return 0;}

object/*(Workarea)*/ wa;


// Debugging tools.

#ifdef SBDEBUG
#define OPMSG(_msg) sbdebug ("VCOldFile ?:" + (wa && wa->tag) + \
			     ":/" + vcfile->real_abspath() + " rev " + \
			     revision + " date '" + date + "' " + _msg + "\n")
#else
#define OPMSG(_msg)
#endif


// Identity.

private static object/*(VCFile)*/ vcfile;
private static string revision, date;
private static object(versionedfile) vfile;
private static mapping(string:mixed) md;

void create (object/*(Workarea)*/ _wa, object/*(VCFile)*/ _vcfile,
	     string _revision, void|string _date)
{
  wa = _wa, vcfile = _vcfile, revision = _revision, date = _date;

  // FIXME: Handle an incorrect revision more gracefully?

  object (cachefile) cfile = _vcfile->getfile();
  // versionedfile doesn't always work when there are modified local
  // files (e.g. in the A/M state), so try to use the parent
  // (view area) file if we have one.
  if (cfile->parent) cfile = cfile->parent;

  vfile = versionedfile (cfile->file, revision, date);
  md = vfile->metadata();
  OPMSG ("create");
}


// Error types.

#define ERR_ACCESS_DENIED(_ppoint) \
  (["type": "access denied", "ppoint": (_ppoint), "vcobj": this_object()])
#define ERR_INVALID_OP_REMOVED(_op) \
  (["type": "invalid operation", "operation": (_op), "state": "N/R", \
    "reason": "File is removed.", \
    "vcobj": this_object()])
#define ERR_INVALID_OP_N_A(_op) \
  (["type": "invalid operation", "operation": (_op), "state": "N/E", \
    "reason": "Not applicable on a file with an old version.", \
    "vcobj": this_object()])


// Caches.

private static object/*(VCOldDir)*/ parent;

private static nomask object/*(VCOldDir)*/ update_parent()
{
  if (parent) return parent;
  // FIXME: Convert revision to date here.
  return parent = wa->_vcoldobj (vcfile->ascend (0, 1, 1), date, revision);
}



// Permission handling.

nomask int perm (object id) {return vcfile->perm (id);}
nomask object(Perm) permobj (object id) {return vcfile->permobj (id);}
object/*(VCOldFile)*/ has_ppoint() {return vcfile->has_ppoint() && this_object();}


// State queries.

private static constant valid_ops_normal = (<
  "ascend", "metadata", "view", "revert", "copy"
>);

private static constant op_perms = ([
  "ascend": PERM_NONE, "metadata": PERM_READ, "view": PERM_READ, "revert": PERM_WRITE,
  "copy": PERM_READ, "purge": PERM_WRITE,
]);

nomask string userstat_ea (object id, void|int undelete_mode, void|int superuser)
// Returns either "none" or 0 if this file shouldn't be shown at all.
{
  if (!superuser && vcfile->perm (id) == PERM_NONE) return 0;
  return "none";
}

nomask string userstat_rep (object id, void|int undelete_mode, void|int superuser)
// Returns "none", "exists" or "removed" (only in undelete mode).
// Returns 0 if this file shouldn't be shown at all.
{
  if (!superuser && vcfile->perm (id) == PERM_NONE) return 0;
  return md->__removed ? undelete_mode ? "removed" : "none" : "exists";
}

nomask object/*(VCOldFile)*/ is_valid_op (object id, string operation,
					  void|int undelete_mode, void|int superuser)
{
  if (operation == "ascend")
    return update_parent()->exists (id, undelete_mode, superuser);
  if (!superuser && vcfile->perm (id) < op_perms[operation]) return 0;
  if (!undelete_mode && md->__removed) return 0;

  switch (operation) {
    case "purge":
      return (superuser || wa->site->purge_access (id)) && this_object();

    case "add_ppoint": case "remove_ppoint":
      return (superuser || wa->site->ppoint_edit_access (id)) && this_object();
  }

  return valid_ops_normal[operation] && this_object();
}

nomask object/*(VCOldFile)*/ is_user_op (object id, string operation,
					 void|int undelete_mode, void|int superuser)
{
  return is_valid_op (id, operation, undelete_mode, superuser);
}

nomask object/*(VCOldFile)*/ accessible (object id)
{
  return vcfile->perm (id) != PERM_NONE && this_object();
}

nomask object/*(VCOldFile)*/ exists (object id, void|int undelete_mode,
				     void|int superuser)
{
  if (!superuser && vcfile->perm (id) == PERM_NONE) return 0;
  return (undelete_mode || !md->__removed) && this_object();
}

nomask string abspath (object id, void|int undelete_mode, void|int superuser)
{
  return exists (id, undelete_mode, superuser) && vcfile->real_abspath();
}

nomask string name (object id, void|int undelete_mode, void|int superuser)
{
  return exists (id, undelete_mode, superuser) && vcfile->real_name();
}

nomask array stat (object id, void|int undelete_mode, void|int superuser)
{
  if (!exists (id, undelete_mode, superuser)) return 0;
  array(int|string) s = vfile->stat();
  if (!s) return 0;
  s[0] = 0444;
  s[5] = "none";
  s[6] = md->__removed ? undelete_mode ? "removed" : "none" : "exists";
  return s;
}


// File operations.

#define PERM_CHECK(_id, _needperm, _superuser) \
  if (!_superuser) { \
    object(Perm) pobj = vcfile->permobj (_id); \
    if (pobj->perm < (_needperm)) return ERR_ACCESS_DENIED (pobj->ppid); \
  }

// Protection point handling.

mapping|void add_ppoint (object id, void|int superuser)
  {return vcfile->add_ppoint (id, superuser);}
mapping|void remove_ppoint (object id, void|int superuser)
  {return vcfile->remove_ppoint (id, superuser);}

// Nondestructive operations (do not lock the VCFile object).

mapping|object/*(VCOldDir)*/ ascend (object id, void|int undelete_mode,
				     void|int superuser)
{
  if (!update_parent()->exists (id, undelete_mode, superuser)) {
    object(Perm) pobj = parent->permobj (id);
    return pobj->perm == PERM_NONE ?
      ERR_ACCESS_DENIED (pobj->ppid) : ERR_INVALID_OP_REMOVED ("ascend");
  }
  return parent;
}

mapping|object(VCFileData) metadata (object id, void|int undelete_mode,
				     void|int superuser)
{
  PERM_CHECK (id, PERM_READ, superuser);
  mapping _md = md;
  if (_md->__removed) {
    if (!undelete_mode) return ERR_INVALID_OP_REMOVED ("metadata");
    _md -= (["__removed": 1]);
  }
  return VCFileData (_md, 0);
}

mapping|object(VCFileData) view (object id, void|int undelete_mode,
				 void|int superuser)
{
  PERM_CHECK (id, PERM_READ, superuser);
  mapping _md = md;
  if (_md->__removed) {
    if (!undelete_mode) return ERR_INVALID_OP_REMOVED ("view");
    _md -= (["__removed": 1]);
  }
  return VCFileData (_md, vfile->openfile ("r"));
}

// Destructive operations (lock the VCFile object).

mapping|void revert (object id, void|int warning, void|int superuser)
// Returns an error of type "removed revision" if the revision is
// removed.
{
  OPMSG ("revert");
  return vcfile->revert (id, warning, superuser, this_object());
}

mapping|void copy (object id, void|int warning, void|int superuser)
{
  OPMSG ("copy");
  return vcfile->copy (id, warning, superuser, this_object());
}

mapping|void purge (object id, void|int superuser)
{
  OPMSG ("purge");
  return vcfile->purge (id, superuser);
}

#if constant (cpp)
#define INVALID_FUN(_name) \
  mapping _name() {return ERR_INVALID_OP_N_A (#_name);}
#else
#define INVALID_FUN(_name) \
  mapping _name() {return ERR_INVALID_OP_N_A ("_name");}
#endif

INVALID_FUN (add);
INVALID_FUN (download);
INVALID_FUN (set_metadata);
INVALID_FUN (upload);
INVALID_FUN (discard);
INVALID_FUN (update);
INVALID_FUN (delete);
INVALID_FUN (undelete);
INVALID_FUN (commit);
INVALID_FUN (move);
INVALID_FUN (join);
