string cvs_cmd;
string cvs_wd;
string cvs_root;

private static string locate_binary(array path, string name)
{
   string dir;
   array info;
   foreach(path, dir)
   {
      string fname = dir + "/" + name;
#if constant(readlink)
      while(!catch(fname = readlink(fname)));
#endif
      if ((info = file_stat(fname))
	  && info[1]>=0
#ifndef __NT__	  
	  && (info[0] & 73)  // 0111 (8) == 73 (10)
#endif	  
	  )
	 return fname;
   }
   return 0;
}

private static string find_binaries(array path, array|void extra)
{
  cvs_cmd =
    cvs_cmd ||
    locate_binary(path, "cvs.exe") ||
    locate_binary(path, "cvs");
  
  if (!cvs_cmd)
    return "No cvs program found.";
  return 0;
}

void create(string wd, string root, void|string cvscmd)
{
   cvs_cmd=cvscmd;

   if(string tmp = find_binaries(getenv("PATH")/
#ifdef __NT__
				 ";"
#else				 
				 ":"
#endif				 
				 -({""})))
      throw(({tmp, backtrace()}));

   cvs_wd=wd;
   cvs_root=combine_path(getcwd(),root); // no relative path
}

string command(string cmd, string ... args)
{
#ifdef VCDEBUG
  werror("cvs does: "+cmd+" "+
	 replace(args*"",({" ",""}),({"\\ "," "}))+" in "
	 +cvs_wd+"\n");
#endif
  
#ifdef SBDEBUG
  array stat = file_stat (cvs_wd + "/#cvs.lock");
  if (stat)
    werror ("Hmm, there's a #cvs.lock " + (stat[1] >= 0 ? "file" : "dir") +
	    " here..\n");
#endif
  
  object p;
  object out=Stdio.File();
  int rc;
  
#ifndef VCDEBUG
#ifdef __NT__
  object err = Stdio.File ("NUL:", "w"); // VMS rulez
#else   
  object err = Stdio.File ("/dev/null", "w");
#endif
#endif
  
  p=Process.create_process(({cvs_cmd,"-q","-d:local:"+cvs_root,cmd})+
			   (args||({})),
			   ([ 
			     "stdout":out->pipe(),
#ifndef VCDEBUG
			     "stderr": err,
#endif
			     "cwd":cvs_wd
			   ]));
  
  string res = out->read();
  
#ifndef VCDEBUG
  if(err) err->set_id(0);
  err=0;
#endif
   
  if (!stringp (res))
    throw (({"No output from cvs command. errno: " +
	     out->errno() + "\n", backtrace()}));
  
//   // create_process.wait is buggy in 0.6
//   if ((rc=p->wait()))
//     throw(({ sprintf("FAILURE: CVS returned error code %d.\n"
// 		     "Failed call was Process.create_process(%O, %O)\n",
// 		     rc,
// 		     ({cvs_cmd,"-q","-d:local:"+cvs_root,cmd})+
// 		     (args||({})),
// 		     ([ "cwd":cvs_wd ])), backtrace() }));
  

#ifdef __NT__
  res=replace(res,"\r\n","\n");
#endif   

//    werror("cvs says: "+replace(res,"\n","\ncvs says: ")+"\n");
   
   return res;
}

array async_command(string cmd, string ... args)
{
#ifdef VCDEBUG
   werror("cvs does async: "+cmd+" "+
	  replace(args*"",({" ",""}),({"\\ "," "}))+" in "
	  +cvs_wd+"\n");
#endif

   object p;
   object out=Stdio.File();

#ifndef VCDEBUG
#ifdef __NT__
   object err = Stdio.File ("NUL:", "w"); // VMS rulez
#else   
   object err = Stdio.File ("/dev/null", "w");
#endif
#endif

   p=Process.create_process(({cvs_cmd,"-q","-d"+cvs_root,cmd})+
			    (args||({})),
			    ([ 
			       "stdout":out->pipe(),
#ifndef VCDEBUG
			       "stderr": err,
#endif
			       "cwd":cvs_wd
			    ]));

#ifndef VCDEBUG
   if(err) err->set_id(0);
   err=0;
#endif

   return ({p,out});
}

#undefine VCDEBUG
