//!
//! module vc
//! class cachedir
//!	class for handling of a directory
//!	of version-controlled files, 
//!	files are checkout'ed upon need
//!

object dir;
object parent;
string vpath; // virtual, from top cachedir

static mapping updating=([]); // options

mapping filec=([]); // cache of cachefile objects
mapping prefilec=([]); // cache of prestarted cachefile objects
mapping subdirc=([]); // cache of subdirs

// "front": 1    - keep files in front (automatic update to latest)
// "date": date  - keep files at this date
// "fdate": date - checkout files at this date upon unfake

//!
//! method void create(string _dir);
//! method void create(string _dir,object(VC.CVS.cachedir) parent);
//!     make and setup a cached directory
//!

void create(object|string _dir,
	    void|object _parent,
	    void|int|mapping uo,
	    void|string _path,
	    void|string _root)
{
   if (stringp(_dir)) dir=VC.CVS.dir(_dir, _root);
   else dir=_dir;

   if (!stringp(dir->wd)) throw(({"dummer!\n",backtrace()}));

   parent=_parent;
   if (!dir->file)
      throw(({"Initializing cachedir with other then dir",backtrace()}));
   
   if (mappingp(uo)) updating=uo;
   else if (intp(uo) && uo) updating->front=1;

   vpath=_path||"/";
}

mapping updating_options(mapping|void opt)
{
   if (!mappingp(opt)) return updating;
   else return updating=opt;
}

string virtual_path()
{
   return vpath;
}

object maybe_new_dir(string dirname)
{
   array a=file_stat(dir->repository+"/"+dirname);
   if (a && a[1]==-2) 
   {
      // make the directory, 
      // make dirname/CVS/Root,
      //      dirname/CVS/Entries,
      //      dirname/CVS/Repository,
      //      dirname/CVS/Tag
      // update CVS/Entries

      // CVS will do this if we make the directory
      // and add it...

      // ...and this is done by dir->mkdir.

      return dir->mkdir(dirname);
   }
   return 0;
}

object maybe_new_file(string filename)
{
#ifdef VCDEBUG
   werror("maybe_new_file "+dir->wd+"/"+filename+"\n");
#endif

   array a =
      (file_stat(dir->repository+"/"+filename+",v")
       || file_stat(dir->repository+"/Attic/"+filename+",v"));
   if (a)
   {
      // get a status of that one...
#ifdef VCDEBUG
      werror("maybe_new_file "+dir->wd+"/"+filename+" is in repos\n");
#endif

      object o=dir->file(filename,1); 
      object status=o->status();

#ifdef VCDEBUG
      werror(sprintf("maybe_new_file "+dir->wd+"/"+filename+
		     " obj: %O status: %O [%s]\n",
		     o,status,VC.CVS.nice_flags(status->status)));
#endif

      if (!status) 
      {
	 destruct(o);
	 return 0;
      }

      if ((status->status & VC.CVS.RMODIFIED) || !status->working_version)
	 return o;

      if (status->status & VC.CVS.ILLEGAL)
      {
	 // clean up ?
	 return o;
      }

      destruct(o);
   }
#ifdef VCDEBUG
   a =
      (file_stat(dir->repository+"/"+filename+",v")
       || file_stat(dir->repository+"/Attic/"+filename+",v"));

   if (a)
      werror("CVS.cachedir internal confusion:\n"
	     "found "+dir->repository+"/"+filename+",v or "+
	     dir->repository+"/Attic/"+filename+",v"+""+" second try\n");
#endif
}

object find_dir(array(string)|string path)
{
   object r;

   if (stringp(path)) path/="/";
   if (sizeof(path)==0) return this_object();
   if (path[0]=="") return find_dir(path[1..]);

   r=subdirc[path[0]];
   if (!r || !objectp(r))
   {
      object pd;
      object d;

      d=dir->subdir(path[0]);
      if (!d) d=maybe_new_dir(path[0]);
      if (!d) return 0;

      if (parent) 
      {
	 pd=parent->find_dir(path[0]);
	 if (!pd)
	    throw(({"failed to find parent dir "
		    "       "+dir->wd+" / "+path[0]+"\n"
		    "       "+parent->dir->wd+" / "+path[0]+"\n",
		    backtrace()}));
      }

      subdirc[path[0]]= r =
	 object_program(this_object())(d,pd,updating,vpath+path[0]+"/");
   }
   
   if (sizeof(path)>1)
      r=r->find_dir(path[1..]);

   return r;
}

object find_file(string name, void|int fake, string|void revision)
{
   object o;
   object pf;
   int f;
   array a;

   if(revision) {
     o = dir->file(name,fake);
     return o && 
       VC.CVS["cachefile"](o,this_object(),1,
			   VC.CVS["versionedfile"](o, revision),updating);
   }

   // do we have that file already?   

   o=filec[name];
   if (o && objectp(o)) return o;

   o=prefilec[name];
   if (o)
   {
      if (fake==2) return o; // prefile loading from child
      m_delete(prefilec,name);
      o->wake_up();
      if (objectp(o->file)) return filec[name]=o;
      destruct(o);
   }

   // ok, make it...

   o=dir->file(name,fake);

   if (!o)
   {
      o=maybe_new_file(name);	 // i don't have it. Maybe CVS does?
      f=1;
   }
   else
   {
      if ((!fake) && o->status()->status&VC.CVS.UNKNOWN)
	o=0;
   }

#ifdef VCDEBUG
   werror(sprintf("find_file "+name+" in "+dir->wd+": %O\n",o));
#endif

   if (!o) return 0;

   if (parent) pf=parent->find_file(name,fake); 

#ifdef VCDEBUG
   werror(sprintf("find_file "+name+" in "+dir->wd+": %O (file ok)\n",o));
#endif

   return 
      filec[name] =
      VC.CVS["cachefile"](o,this_object(),f||fake,pf,updating);
}

void purge(string name)
{
  object o=prefilec[name];
  if (o) {
    m_delete(prefilec,name);
    o->wake_up();
    destruct(o);
  }
  m_delete(filec,name);
  m_delete(subdirc,name);
  file_removed(name);
}

class AsyncStatus
{
   object obj;
   object pipe;

   array fnames;
   array fobjs;

   void wait();

   void init(object cadir,
	     array(object) _fobjs)
   {
      fobjs=_fobjs;
      array fnames=fobjs->file->file;
      array ac=fobjs[0]->file->cvs->async_command("status","-v",@fnames);
      pipe=ac[1];
      obj=ac[0];
      call_out(wait,1,this_object()); // awake
   }

   void wait()
   {
      remove_call_out(wait);
//       int p=gethrtime();
      object o=pipe;
      pipe=0;
      if (!o)
      {
	 if (obj) destruct(obj);
	 destruct(this_object());
	 return;
      }
      string data=o->read();
//       p=gethrtime()-p;
//       werror(p*0.000001+"s\n");
      if (obj) destruct(obj);

      int i;        
      array a=data/"==================================================================="-({""});
      if (sizeof(a)!=sizeof(fobjs))
      {
	 werror("VC.CVS.cachedir: asyncstatus array size differ, bailing\n");
	 return;
      }
      for (i=0; i<sizeof(a); i++)
      {
	 fobjs[i]->file->calculate_status("================"+a[i]);
	 fobjs[i]->file->async_status_pipe=0;
	 fobjs[i]->file->async_status_object=0;
      }
      destruct(this_object());
   }
}

int mtime,mrtime;
array dircache,rdircache;

void file_added(string name)
{
  mtime = mrtime = 0;
}

void file_removed(string name)
{
  mtime = mrtime = 0;
}

void dir_added(string name)
{
  mtime = mrtime = 0;
}

array get_dir()
{
   int mt,mrt;
   array cdir,cvsdir;
   int first=!dircache;

   mt=file_stat(dir->wd)[3];
   if (mt>mtime)
   {
      dircache=cdir=dir->filenames()|dir->subdirnames();
      mtime=mt;
   }
   else cdir=dircache;
   
   if (parent)
      cdir|=parent->get_dir();
   else
   {
      mrt=file_stat(dir->repository)[3];
      if (mrt>mrtime)
      {
	 mrtime=mrt;
	 cvsdir=
	    predef::get_dir(dir->repository)|
	    (predef::get_dir(dir->repository+"/Attic")||({}));

	 rdircache=
	   cvsdir=Array.map(cvsdir,
			    lambda(string s)
			    {
			      if (s[..4] != "#cvs." &&
				  s[strlen(s)-2..]==",v")
				return s[..strlen(s)-3];
			      else 
				return 0;
			    })-({"Attic", 0});
	 cdir|=cvsdir;
      }
      else cdir|=rdircache;
   }

   // instantiate and look for new files

   array newfiles=cdir-indices(prefilec)-indices(subdirc)-indices(filec);

   program Cachf=VC.CVS["cachefile"];
   foreach (newfiles,string name)
   {
      array a=file_stat(dir->wd+"/"+name);
	 
      if (!a || a[1]<0)  // directory
	 continue;
	 
      array a=
	 (file_stat(dir->repository+"/"+name+",v")
	  || file_stat(dir->repository+"/Attic/"+name+",v"));

      if (!a || a[1]<0)  // directory
	 continue;

      object pf=0;

      if (parent)
      {
	 pf=parent->find_file(name,2);
	 if (!pf) continue;
      }
      prefilec[name]=
	 Cachf(dir->file(name,1),this_object(),2,pf,updating);
      // 	    ->file->async_status();
   }

   /*
   array(object) to_status=({});
   object dir_status=AsyncStatus();
   foreach (values(prefilec)|values(filec), object f)
      if (f && !f->file->have_status() && !f->file->async_status_object) 
      {
	 to_status+=({f});
	 f->file->async_status_object=dir_status;
      }

   if (sizeof(to_status))
      dir_status->init(dir,to_status);
   else
      destruct(dir_status);
   */

   return cdir;
}

array stat()
{
   // ({ mode,size,atime,mtime,ctime,uid,gid })
   return dir->stat();
}

object mkdir(string name)
{
   if (parent)
   {
      parent->mkdir(name);
      return find_dir(name);
   }
   if (dir->mkdir(name)) {
      dir_added(name);
      return find_dir(name);
   }
   return 0;
}


object clone(string path)
{
   // install this repository in the 'path' directory
   // return a new cachedir object for that path

   VC.CVS.checkout_repository(dir->root,
			      dir->repository,
			      path,
			      dir->gettag());
   return object_program(this_object())(path,this_object(),0,vpath);
}

array(array(string)) cleanup_locks()
{
  return dir->cleanup_locks();
}
