//!
//! module vc
//! class cachedir
//!	class for handling of a directory
//!	of version-controlled files, 
//!	files are checkout'ed upon need
//!

import spider; // need parse_html
import VC.CVS;

int fake;      // fake file, still in cvs
int front;     // automatic update upon RMODIFIED

object file;   // CVS.file object
object parent; // CVS.cachefile parent (for read operations)

object dir;    // CVS.cachedir owner

mapping(string:mixed) metadata_cache;
int metadata_date, metadata_length=-1;

mapping updating=([]);

#ifdef SAVE_BACKUPS
#define SCRAP(file) rm ((file) + "~"); mv ((file), (file) + "~")
#else
#define SCRAP(file) rm (file)
#endif

void create(object _file, object _dir, void|int _fake, 
	    void|object _parent, void|mapping uo)
{
   if (!stringp(_file->file)) throw(({"dummer!\n",backtrace()}));
   if (!objectp(_dir->dir)) throw(({"dummer!\n",backtrace()}));

#ifdef VCDEBUG
   werror ("cachefile create " + (_file && (_file->wd + "/" + _file->file)) +
	   " fake " + _fake + " parent " + !!_parent + "\n");
#endif

   dir=_dir;
   file=_file;
   fake=_fake;
   parent=_parent;
   updating=uo||([]);
}

string virtual_path()
{
   return dir->vpath+file->file;
}

void write_version(string version);

void wake_up() // wake up preloaded
{
   if (fake<2) return;

   if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
   if (parent) parent->wake_up();

   array a=file_stat(file->wd+"/"+file->file);

   if (a && a[1]>=0)  // local file, give it back
   {
      fake=0; 
      return;
   }

   array a=
      (file_stat(dir->dir->repository+"/"+file->file+",v")
       || file_stat(dir->dir->repository+"/Attic/"+file->file+",v"));

   if (a && a[1]>=0)  // repos file, good so far
   {
      object _s=file->status();

      if (_s)
      {
	 int s=_s->status;
	 if (s & (RMODIFIED | ILLEGAL)) 
	 {
	    fake=1;
	    return; // ok
	 }
      }
   }

   destruct(file);
   file=0;
   fake=0;
   destruct(this_object());
   return; 
}

void unfake()
{
   if (!fake) return;

   // make a real file

   if (fake==2) 
   {
      wake_up();
      if (!file)
	 throw( ({ "preloaded -nonexisting- object called unfake(): "+
		file->wd+"/"+file->file+"\n",
		backtrace() }) );
	 
//       throw( ({ "preloaded object called unfake(): "+
// 		file->wd+"/"+file->file+"\n",
// 		backtrace() }) );
   }

   if (updating->date) // figure out a good version
   {
      updating->version=file->version_tagdate(updating->date);
      if (!updating->version) // it didn't exist then
      {
	 file->remove(1);
	 fake=0;
	 dir->file_removed(file->file);
	 return;
      }
   }

   if (updating->version)
   {
#ifdef VCDEBUG
werror(file->wd+"/"+file->file+": push version "+updating->version+"\n");
#endif
      file->update(updating->version);
      fake=0;
      dir->file_added(file->file);
      return;
   }
   
   if (updating->fdate) // push that file
   {
#ifdef VCDEBUG
werror(file->wd+"/"+file->file+": push date "+updating->fdate+"\n");
#endif

      string v=file->version_tagdate(updating->fdate);

#ifdef VCDEBUG
werror(file->wd+"/"+file->file+": push version "+v+"\n");
#endif

      if (!v) // it didn't exist then
      {
	 file->update();
	 file->remove(1);
	 dir->file_removed(file->file);
      }
      else
      {
	 file->update();
	 fake=0; // Make sure write_version doesn't call us back...
	 write_version(v);
	 dir->file_added(file->file);
      }
      fake=0;
      return;
   }

   file->update();
   dir->file_added(file->file);

   fake=0;
}

int ununfake( void|int force )
{
   if (fake) return 1;
   if ((file->status()->status & LMODIFIED)
       && !force) return 0; // noncommited changes
   SCRAP (file->wd+"/"+file->file);
   if (file->status()->status & ILLEGAL) {
     // We get this when removing a newly added file since the added
     // cvs entry will be bogus. An update corrects that ("new-born
     // foo has disappeared").
     file->update();
     // Remove any file that may got checked out due to race.
     rm(file->wd+"/"+file->file);
   }
   fake=1;
   metadata_cache = 0;
   metadata_length = -1;
   return 1;
}

int check_metadata_length(object f)
{
  string r, data = "", tag = "\n</metadata>\n";
  int p, c=0;

  while((r=f->read(512))&&sizeof(r)) {
    data += r;
    if((p=search(data, tag))>=0)
      return c+p+strlen(tag);
    if(sizeof(data)>16) {
      c += sizeof(data)-16;
      data = data[sizeof(data)-16..];
    }
  }
  return -1;
}

array stat()
{
   if (fake)
   {
      if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
      if (parent) return parent->stat();
      unfake();
   }
   array res = file->stat();
   if(res[1]>0) {
     object f=Stdio.File();
     if(metadata_length >= 0) {
       if(res()[3]>metadata_date)
	 metadata_length = -1;
     }
     if(metadata_length < 0 && f->open(file->wd+"/"+file->file,"r")) {
       metadata_cache = 0;
       metadata_date = res[3];
       metadata_length = check_metadata_length(f);
       f->close();
     }
     if(metadata_length > 0 && res[1]>=metadata_length)
       res[1] -= metadata_length;
   }
   return res;
}

static int unsafe_char(int c)
{
  return c<' ' || c=='"' || c=='&';
}

static string format_metadatum(string name, mapping(string:mixed) meta)
{
  mixed value = meta[name];
  
  if(intp(value))
    return " <datum name="+name+" valuen="+value+">\n";
  if(stringp(value) && sizeof(Array.filter(values(value), unsafe_char))==0)
    return " <datum name="+name+" value=\""+value+"\">\n";
  string v = encode_value(value);
#if constant(Gz.deflate)
  string v2 = Gz.deflate()->deflate(v);
  if(sizeof(v2)<sizeof(v))
    return " <datum name="+name+" valuez64=\""+MIME.encode_base64(v2, 1)+
      "\">\n";
  else
    v2 = 0;
#endif
  return " <datum name="+name+" value64=\""+MIME.encode_base64(v, 1)+"\">\n";
}

string format_metadata(mapping(string:mixed) meta)
{
  return "<metadata>\n"+Array.map(sort(indices(meta)),
				  format_metadatum, meta)*""+
    "</metadata>\n";
}

static string parse_metadatum(string tag, mapping(string:string) args,
			      mapping(string:mixed) meta)
{
  string name;

  if(!(name = args->name))
    return 0;

  name = lower_case(name);

  if(args->valuen)
    meta[name]=(int)args->valuen;
  else if(args->value)
    meta[name]=args->value;
  else if(args->value64)
    meta[name]=decode_value(MIME.decode_base64(args->value64));
#if constant(Gz.inflate)
  else if(args->valuez64)
    meta[name]=decode_value(Gz.inflate()->
			    inflate(MIME.decode_base64(args->valuez64)));
#endif
  return 0;
}

static string parse_metadatum_ann(string tag,mapping(string:string) args,
				  mapping(string:mixed) meta_ann,
				  object oldann)
{
   mapping tmp=([]);
   parse_metadatum(tag,args,tmp);
   
   foreach (indices(tmp),string name)
   {
      object o=VC.CVS.Annotation(oldann);
      o->data=tmp[name];
      meta_ann[name]=o;
   }
   return 0;
}

static mapping make_metadata(string md)
{
   mapping something = ([]);
   parse_html(md, (["datum":parse_metadatum]), ([]), something);
   return something;
}

static array separate_metadata(string s)
{
   if (s[0..9]=="<metadata>")
   {
      int i=search(s,"\n</metadata>\n");
      return ({make_metadata(s[10..i]), s[i+13..]});
   }
   else 
      return ({([]),s});
}


// metadata() _must_ write to metadata_cache
mapping metadata()
{
   if (fake)
   {
      if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
      if (parent) return metadata_cache=parent->metadata();
      unfake();
   }

   if(metadata_cache) {
     array(int) s = file_stat(file->wd+"/"+file->file);
     if((!s) || s[3]>metadata_date) {
       metadata_cache = 0;
       metadata_length = -1;
     }
   }

   if (!file)
     return ([]); // this shouldn't happen. FIXME!

   if(!metadata_cache) {
     metadata_cache = ([]);

     object f=Stdio.File();
     if (f->open(file->wd+"/"+file->file,"r")) {
       int stat_date = file->stat()[3];

       if(stat_date > metadata_date)
	 metadata_length = -1;

       if(metadata_length < 0)
	 metadata_length = check_metadata_length(f);

       if(metadata_length > 0) {
	 string md;
	 f->seek(0);
	 md = f->read(metadata_length);
	 parse_html(md, (["datum":parse_metadatum]), ([]), metadata_cache);
	 metadata_date = stat_date;
       } else
	 metadata_date = 0;

       f->close();
     }
   }
   return metadata_cache;
}

mapping parent_metadata()
{
  if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
  if (parent)
    return parent->metadata();
  else
    return ([ ]);
}

object openfile(string flags)
{
   if (fake)
   {
      if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
      if (parent && flags=="r")
	 return parent->openfile("r");
      unfake();
   }

   if(search(flags, "w")>=0 && (search(flags,"t")>=0 || search(flags,"r")<0))
     metadata();

   object f=Stdio.File();
   if (!f->open(file->wd+"/"+file->file,flags))
   {
#if 0
      file->update(); // strange. Try this
      if (!f->open(file->wd+"/"+file->file,flags))
#endif
	 return 0; // didn't help
   }

   if(search(flags, "t")>=0) 
   {
      if(search(flags, "w")>=0) 
      {
	 string md = format_metadata(metadata_cache || ([]));
	 f->write(md);
	 metadata_length = strlen(md);
	 metadata_date = f->stat()[3];
      }
      else 
      {
	 metadata_cache = 0;
	 metadata_length = -1;
	 metadata_date = 0;
      }
   } 
   else
   {
      if(metadata_length >= 0) 
      {
	 if(f->stat()[3]>metadata_date)
	    metadata_length = -1;
      }
      if(metadata_length < 0) 
      {
	 metadata_cache = 0;
	 metadata_date = f->stat()[3];
	 if((metadata_length = check_metadata_length(f))<=0)
	    f->seek(0);
      }
      if(metadata_length > 0)
	 f->seek(metadata_length);
   }
   return f;
}

array(object|mapping) openfile_with_metadata(string flags)
{
   int write_metadata=0;
   array st;

   if (fake)
   {
      if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
      if (parent && flags=="r")
	 return parent->openfile_with_metadata("r");
      unfake();
   }

   if(search(flags, "w")>=0 && (search(flags,"t")>=0 || search(flags,"r")<0)) {
     metadata();
     write_metadata=1;
   }

   object f=Stdio.File();
   if (!f->open(file->wd+"/"+file->file,flags))
      return 0;

   if(write_metadata) {
     if(!metadata_cache)
       metadata_cache=([]);
     string md = format_metadata(metadata_cache);
     f->write(md);
     metadata_length = strlen(md);
     metadata_date = f->stat()[3];
   } else if((!metadata_cache)||metadata_length<0||
	     (st=f->stat())[3]>metadata_date) {

     metadata_date = (st||f->stat())[3];
     metadata_cache = ([]);
     metadata_length = check_metadata_length(f);
     f->seek(0);
     if(metadata_length>=0) {
       string md = f->read(metadata_length);
       parse_html(md, (["datum":parse_metadatum]), ([]), metadata_cache);
     }
   } else
     f->seek(metadata_length);

   return ({f,metadata_cache});
}

static class TempFile {

  inherit Stdio.File;

  static int odate;
  static object file;
  static mapping(string:mixed) metadata;

  int reopen (string flags)
  {
    if(!file)
      throw(({"tried to reopen file already comitted or rolled back.\n",
	      backtrace()}));
    catch (close());
    if (!open (file->file->wd + "/" + file->file->file + "~0~", flags))
      return 0;
    if(search(flags, "w")>=0 && (search(flags,"t")>=0 || search(flags,"r")<0))
      write (file->format_metadata (metadata));
    else {
      int length = file->check_metadata_length (this_object());
      seek (max (length, 0));
    }
    return 1;
  }

  void commit()
  {
    if(!file)
      throw(({"tried to commit file already comitted or rolled back.\n",
	      backtrace()}));
    close();
    array(int) s = file_stat(file->file->wd+"/"+file->file->file);
    if(s && s[3]>odate)
      throw(({"file "+file->file->wd+"/"+file->file->file+
	      " changed under my feet.\n", backtrace()}));
    SCRAP (file->file->wd+"/"+file->file->file);
    if(!mv(file->file->wd+"/"+file->file->file+"~0~",
	   file->file->wd+"/"+file->file->file))
      throw(({"couldn't replace "+file->file->wd+"/"+file->file->file+
	      " with tempfile!\n", backtrace()}));
    file->metadata_date = 0;
    file=0;
  }

  void rollback()
  {
    if(file) {
      close();
      rm(file->file->wd+"/"+file->file->file+"~0~");
      file = 0;
    }
  }

  void destroy()
  {
    if(file)
      rollback();
  }

  void create(object f, int d, mapping(string:mixed) md)
  {
    ::create();
    file = f;
    odate = d;
    metadata = md;
  }

};

object openfile_temp_w(mapping|void new_metadata)
{
   int odate;

   if (fake)
     unfake();

   if(!new_metadata) {
     metadata();
     odate = metadata_date;
   }
   if(!odate) {
     array(int) s = file_stat(file->wd+"/"+file->file);
     if(s)
       odate = s[3];
   }

   object f=TempFile(this_object(), odate, new_metadata||metadata_cache||([]));
   if (!f->reopen("cwt"))
      return 0;

   return f;
}

mapping write_metadata(mapping m)
{
   object of = openfile("r");

   object f = openfile_temp_w(m);

   if(!f) {
     metadata_cache = 0;
     metadata_length = -1;
     return 0;
   }
   if(of) {
     string s;
     while((s=of->read(8192))&&sizeof(s))
       f->write(s);
     of->close();
   }
   f->commit();
   return metadata();
}

void write_version(string version)
{
  if (fake)
    unfake();
   object f=Stdio.File();
   SCRAP (file->wd+"/"+file->file);
   if (!f->open(file->wd+"/"+file->file,"cwt"))
      throw( ({"Couldn't write "+file->wd+"/"+file->file+"\n",
	       backtrace()}) );
   f->write(file->update_stdout(version));
   f->close();
   metadata_cache=0;
   metadata_length=-1;
}

object status();
string getfile()
{
   if (fake)
   {
      if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
      if (parent) return parent->getfile();
      unfake();
   }
   if (updating->front) status(); // will update automatically

   object f=openfile("r");
   if (!f) throw(({"failed to open "+file->wd
		   +"/"+file->file+"?\n",
		   backtrace()}));
   return f->read(0x7fffffff);
}

int write(string data)
{
   if (fake) unfake();

   metadata();

#if 0
   // This nukes the current metadata since openfile calls metadata()
   // after the file has disappeared.
   SCRAP (file->wd+"/"+file->file);
#endif

   object f=openfile("ctw");
   if (!f) throw(({"failed to open "+file->wd
		   +"/"+file->file+" for writing?\n",
		   backtrace()}));

   int res=f->write(data);
   f->close();

   return res;
}

int new()
{
   if (!fake) 
      throw(({"can't create existing file (new called on non-fake)\n",
	      backtrace()}));
   fake=0;
   openfile("cw");
   int rc = file->add();
   dir->file_added(file->file);
   return rc;
}

int commit(string comment)
{
   if (fake) return 0;

   file->commit( comment );

   if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
   if (parent)
     parent->update();
   ununfake( 1 );
   return 1;
}

static class DelayedCommit {

  static string comment;
  static array(int) successes=({});
  static mixed exception;

  void create(string c)
  {
    comment = c;
  }

  object add_file(object o)
  {
    exception=exception||catch(successes+=({o->commit(comment)}));
    return this_object();
  }

  array(int) collect()
  {
    mixed x = exception;
    array(int) s = successes;
    destruct(this_object());
    if(x)
      throw(x);
    else
      return s;
  }

};

object delayed_commit(string|object context)
{
  if(stringp(context))
    context = DelayedCommit(context);
  if(!(objectp(context) && functionp(context->add_file)))
    throw(({"context missing in delayed_commit.\n", backtrace()}));
  return context->add_file(this_object());
}

static string _conflicts_rev=0;
static string _conflicts_newrev=0;
static array  _conflicts;

// method array conflicts(void|string newrevision)
//
// returns
// ({ ({ mapping       yours, new }),        - metadata
//    array(array(string)) yours_new_diff }) - text data
//              -- elements are pairwise comparable, like
//            ({ ({ "foo", "bar", ... }) ,    - conflict
//               ({ "qrw", "qrw", ... }) })   - no conflict
// 
//                               ^^^^^
//                  this is       old, yours, new
// 
// 
// something like this:
// 
// object(cachefile) f;
// 
// if (f->update()==2) // conflicts, nothing done
// {
//    string ver=f->status()->working_version;
//    f->update_soft();   // change version to latest
// 
//    array cf=f->conflicts(f->status()->working_versoin, ver); 
// 
// ...handle conflicts...
// 
//    f->write( ... );           <.
//    f->write_metadata( ... );  <- from your handling of conflicts
// }
// 
// 
//                                    yours      in repos
// arrays are comparable pairwise;      v            v
// ie, if sizeof(cf[1][0])==2 and cf[1][0][1]!=cf[1][1][1],
// that means that section in the          ^            ^
// files has a conflict.                   second section 
//
// text data arrays should be put together with a sum,
// or imploding with "". 

array(mixed) resolve_conflicts(mixed old,mixed new1,mixed new2)
{
   if (mappingp(old))
      return ({ (new2|new1) - (old-new1),
		(new1|new2) - (old-new2),
		old,new1,new2 });
   
   if (!arrayp(old)) 
      return ({new1,new2,old,new1,new2}); /* you're wierd */

   array res=({});

   res=
      Array.map(Array.transpose(({old,new1,new2})),
		lambda(array row)
		{
		   array q=({row[1],row[2],@row});
		   
		   if (row[2]==row[0]) q[1]=q[0]; /* new 1 win over 2 */
		   else if (row[1]==row[0]) q[0]=q[1]; /* new 2 win over 1 */
		   
		   return q;
		});

   return Array.transpose(res);
}

array conflicts(void|string newrevision,void|string oldrevision)
{
   string new,old,cur;
   if (!oldrevision) oldrevision=status()->working_version;
   if (!newrevision) newrevision=status()->repository_version;

   new=file->update_stdout(newrevision);
   old=file->update_stdout(oldrevision);  
   cur=getfile();
   
   array anew=separate_metadata(new);
   array aold=separate_metadata(old);

   mapping md_yours,md_new;
   mapping md2=anew[0],md=aold[0],md1=metadata();

   array d3;

   if (file->isbinary())
      d3=({ ({ ({ aold[1] }), ({cur }), ({ anew[1] }) }) });
   else
      d3=Array.diff3(aold[1]/"\n",cur/"\n",anew[1]/"\n");

   foreach (d3=Array.transpose(d3),array row)
   {
      if (sizeof(row[0])) row[0]=row[0]*"\n"+"\n"; else row[0]="";
      if (sizeof(row[1])) row[1]=row[1]*"\n"+"\n"; else row[1]="";
      if (sizeof(row[2])) row[2]=row[2]*"\n"+"\n"; else row[2]="";
   }

#if 0
   if (equal(md_yours,md_new) && equal(@Array.transpose(d3)))
      return 0;
#endif
      
   return ({ resolve_conflicts(md,md1,md2),
	     Array.transpose(resolve_conflicts(@Array.transpose(d3))) });
}

int update_hard(void|string revision)
{
   SCRAP (file->wd+"/"+file->file);

   metadata_cache=0;
   metadata_length=-1;

   return file->update(revision);
}

int update_soft(void|string revision)
{
   rm(file->wd+"/"+file->file+"~2~");
   if (!mv(file->wd+"/"+file->file,file->wd+"/"+file->file+"~2~"))
      return 0; // make temporary file

   int rc = file->update(revision);

   if (rc) {
     // The empty file created above produced a conflict. This only
     // occurs when the same file has been added from two places.
     rm(file->wd+"/"+file->file);
     rc = file->update(revision);
   }

   if(!rc && !file->stat())
     // We didn't get any file.  Perhaps it was locally added.
     // In that case we need to try again.
     rc = file->update(revision);

   if (!rc) {
     rm(file->wd+"/"+file->file);
     if (!mv(file->wd+"/"+file->file+"~2~",file->wd+"/"+file->file))
       throw( ({"Update_soft fatal! Can't rename back kept file!\n",
		backtrace()}) );
     // Ho hum.
     file->add();
     dir->file_added(file->file);
     return 0;
   }

   SCRAP (file->wd+"/"+file->file);

   object of = Stdio.File(), f = Stdio.File();
   if (!of->open (file->wd + "/" + file->file + "~2~", "r"))
     throw (({"Update_soft fatal! Can't open temporary file for read!\n", backtrace()}));
   if (!f->open (file->wd + "/" + file->file, "cwt"))
     throw (({"Update_soft fatal! Can't open file for write!\n", backtrace()}));
   string s;
   while((s=of->read(8192))&&sizeof(s)) f->write(s);
   of->close();
   f->close();
   rm (file->wd + "/" + file->file + "~2~");
   // Have to copy the file back here to workaround a cvs timestamp
   // quirk: When cvs update checks out a nonexisting file it doesn't
   // sleep before writing the file, only afterwards. This means that
   // it'll write the current time in CVS/Entries, which might very
   // well be the same time as the file we're moving back later. CVS
   // will then report it as up-to-date without diffing, since the
   // timestamps match. Copying the file is the only way to reliably
   // ensure that the mtime gets updated.

   metadata_cache=0;
   metadata_length=-1;
   
   return 1;  // if these latest mv's and stuff doesn't work
              // we're out bad, anyway
}

int update(void|string revision)
{
   if (fake) unfake();

   if ( ( file->status()->status & (LMODIFIED|RMODIFIED) ) ==
	    (LMODIFIED|RMODIFIED)) // conflict possible
      if (conflicts(revision)) return 2;

   _conflicts=0;
   _conflicts_rev=0;
   metadata_cache=0;
   metadata_length=-1;

   if ( file->status()->status & LMODIFIED )
     return /*file->*/update_soft(revision); // only local

   return /*file->*/update_hard(revision); // only repos
}

object status()
{
   object s;

   if (fake)
   {
      if (!parent && dir->parent) parent = dir->parent->find_file (file->file, 1);
      if (parent)
      {
	 s=parent->status();
	 if (fake) s->status=(s->status&~RMODIFIED)|FAKE;
	 return s;
      }
      s=file->status();
      if (!(s->status&INCOMPLETE))
      {
	 s->working_version=s->repository_version;
	 s->status=(s->status&~RMODIFIED)|FAKE;
	 return s;
      }
      unfake();
   }

   s=file->status();

   if (updating->front)
   {
      if ( s->status&RMODIFIED ) 
      {
#ifdef VCDEBUG
	 werror( "CACHEFILE "+file->wd+"/"+file->file+" to front...\n");
#endif
	 update_hard();
      }
      
      return file->status();
   }

   return s;
}

void async_status() // prepare for a status() call
{
   file->async_status();
}

int isbinary(int ...set)
{
   return file->isbinary(@set);
}

array(mapping(string:object(VC.CVS.Annotation))|array(object(VC.CVS.Annotation))) annotate(void|string revision)
{
   array ann=file->annotate(revision || 
			    status()->working_version);

   int meta=0;
   mapping mdann=([]);

   int i,pos;
   string t;

   for (i=0; i<sizeof(ann); i++)
   {
      switch(meta) 
      {
	 case 0:
	    if(ann[i]->data[..10]=="<metadata>\n")
	       meta++;
	    else
	       break; // escape 
	    ann[i]->data=ann[i]->data[11..];

	 case 1:
	    if(ann[i]->data[..11]=="</metadata>\n")
	    {
	       meta++;
	       ann[i]->data=ann[i]->data[12..];
	       if (ann[i]->data=="") i++; 
	       break;
	    }

	    if ((pos=search(ann[i]->data,"\n</metadata>\n"))!=-1)
	    {
	       t=ann[i]->data[..pos];
	       ann[i]->data=ann[i]->data[pos+1..];
	    }
	    else
	    {
	       t=ann[i]->data;
	       ann[i]->data="";
	    }
	    
	    parse_html(t, 
		       (["datum":parse_metadatum_ann]), 
		       ([]), 
		       mdann,ann[i]);

	    if (ann[i]->data!="") i--;
	    continue;
      }
      return ({mdann,ann[i..]});
   }
}

array (mapping) log(string ...args)
{
  return file->log(@args);
  // FIX ME 
  // We should probably sort the log entries in some way, though
  // not like this
  //  result = Array.sort_array( result, lambda (mapping a, mapping b)
  //				     {
  //				       return a->revision < b->revision;
  //				     } );
}

// returns
// ({ ({ mapping       yours, new }),        - metadata
//    array(array(string)) yours_new_diff }) - text data
//              -- elements are pairwise comparable, like
//            ({ ({ "foo", "bar" }) ,    - diff
//               ({ "qrw", "qrw" }) })   - no diff


array(array(array(string)|mapping)) 
   diff(void|string revision,void|string revision2)
{
   if (!revision2 && fake) unfake();
      
   if (!revision) revision=status()->working_revision;

   string file1,file2;
   mapping md1,md2;

   if (revision2)
   {
      array a1,a2;
      a1=separate_metadata(file->update_stdout(revision));
      a2=separate_metadata(file->update_stdout(revision2));
      md1=a1[0];
      md2=a2[0];
      file1=a1[1];
      file2=a2[1];
   }
   else
   {
      file1=getfile();
      md1=metadata();
      array a2=separate_metadata(file->update_stdout(revision));
      md2=a2[0];
      file2=a2[1];
   }
   
   if (file->isbinary())
      return ({ ({ md1, md2 }),
		({ ({ file1, file2 }) }) });
   else
   {
      array a;
      a=Array.transpose(Array.diff(file1/"\n",file2/"\n"));
      return ({ ({ md1, md2 }), a });
   }
}

array(array(array(string)|mapping)) 
   diff3(string revisionnew,string revisionold,
			   void|string revisioncur)
{
   string file1,file2,file3;
   mapping md1,md2,md3;
   array a1,a2,a3;

   a1=separate_metadata(file->update_stdout(revisionold));
   a2=separate_metadata(file->update_stdout(revisionnew));
   md1=a1[0];
   md3=a2[0];
   file1=a1[1];
   file3=a2[1];

   if (revisioncur)
   {
      a2=separate_metadata(file->update_stdout(revisioncur));
      file2=a2[1];
      md2=a2[0];
   }
   else
   {
      file2=getfile();
      md2=metadata();
   }

   if (file->isbinary())
      return ({ ({ md1, md2, md3 }),
		({ ({ file1, file2, file3 }) }) });
   else
   {
      array a;
      a=Array.transpose(Array.diff3(file1/"\n",file2/"\n",file3/"\n"));
      return ({ ({ md1, md2, md3 }), a });
   }
}


string split_version(string ... tags)
{
   return file->split_version(@tags);
}
             // orig (split), others (source), yours (destination)
array try_join(void|string tag0, string tag1, void|string tag2)
{
   string old,new1,new2;
   mapping omd,md1,md2;
   array a;
   string tag0;

   if (!tag2)
   {
      new2=getfile();
      md2=metadata();
      tag2=status()->working_version;
   }
   else
   {
      a=separate_metadata(file->update_stdout(tag2));
      new2=a[1];
      md2=a[0];
   }

   if (!tag0) tag0=split_version(tag1,tag2);

   a=separate_metadata(file->update_stdout(tag1));
   new1=a[1];
   md1=a[0];

   if (tag0) 
   {
      a=separate_metadata(file->update_stdout(tag0));
      old=a[1];
      omd=a[0];
   }
   else
   {
      old="";
      omd=([]);
   }

   mapping md_yours = (md2|md1)	- (omd-md1);
   mapping md_new =   (md1|md2)	- (omd-md2);
   
   array d3;

   if (file->isbinary())
      d3=({ ({ ({ old }), ({ new1 }), ({ new2 }) }) });
   else
      d3=Array.diff3(old/"\n",new1/"\n",new2/"\n");

   foreach (d3=Array.transpose(d3),array row)
   {
      if (sizeof(row[0])) row[0]=row[0]*"\n"+"\n"; else row[0]="";
      if (sizeof(row[1])) row[1]=row[1]*"\n"+"\n"; else row[1]="";
      if (sizeof(row[2])) row[2]=row[2]*"\n"+"\n"; else row[2]="";
   }

   return ({ resolve_conflicts(omd,md1,md2),
	     Array.transpose(resolve_conflicts(@Array.transpose(d3))) });
}

int remove()
{
  dir->purge(file->file);
  if (fake)
    return file->remove(0);
  else
    return file->remove(1);
}


