// $Id: dir.pike,v 1.10 1999/03/20 00:25:46 js Exp $

//!
//! module vc
//! class dir
//!	class for handling of a directory
//!	of version-controlled files
//! 
//! method void create(string dir)
//!	Make a new version-controlling object.
//! arg string dir
//!	which directory to control
//!

import VC.CVS;

string wd;
string root;
string repository;
object cvs;

mapping fileobjcache=([]);
mapping dirobjcache=([]);

#define argerror(X) (throw(({X,backtrace()}))) /* argument error */
#define funerror(X) (throw(({X,backtrace()}))) /* function error */

void create(string _dir, string|void _root)
{
   if (!stringp(_dir)) 
      argerror("vc.dir->create: illegal argument 1\n");
   
   wd=_dir;
   if (wd=="" || wd=="/") wd="/";
   else if (wd[-1]=='/') wd=wd[0..strlen(wd)-2];

   array a=file_stat(wd);
   if (!a || a[1]!=-2) argerror("not a directory: "+wd+"\n");

   object f=Stdio.File();
   if (stringp(_root))
      root = _root;
   else
      if (f->open(wd+"/CVS/Root","r"))
      {
	 root=f->read(10000)-"\r"-"\n";
	 f->close();
      }
      else
	 funerror("no root information ("+wd+"/CVS/Root) found.\n");

   f=Stdio.File();
   if (f->open(wd+"/CVS/Repository","r"))
   {
      repository=combine_path(root, f->read(10000)-"\r"-"\n");
      f->close();
   }
   else
      funerror("no repository information ("+wd+"/CVS/Repository) found.\n");

   cvs=CVS(wd,root);
}

//!
//! method array allfiles()
//!    
//! returns all files (and subdirs, and links) in directory
//!
//! note
//!	this does a plain get_dir...
//!

array allfiles()
{
   return get_dir(wd);
}

//!
//! method array(string) filenames()
//! method array(string) subdirnames()
//!     returns names of version controlled files/subdirectories in the directory
//! see also: allfilenames
//!

array filenames()
{
   array(string) files=({});
   object f=Stdio.File();
   if (f->open(wd+"/CVS/Entries","r"))
   {
      foreach (f->read(1000000)/"\n",string s)
	 if (s!="" && s[0]=='/')
	    sscanf(s,"/%s/",s),files+=({s});
      f->close();
   }
   return files;
}

array subdirnames()
{
   array(string) subdirs=({});
   object f=Stdio.File();
   if (f->open(wd+"/CVS/Entries","r"))
   {
      foreach (f->read(1000000)/"\n",string s)
	 if (s!="" && s[0..1]=="D/")
	    sscanf(s,"D/%s/",s),subdirs+=({s});
      f->close();
   }
   return subdirs;
}

//!
//! method object subdir(string subdir)
//!    Get a VC.dir object handling a subdirectory.
//! returns the new VC.dir object.
//! note
//!	A '/' in a subdir or the subdir ".." is illegal.
//!	It will give an error if the subdir is not a directory.
//!

object subdir(string subdir)
{
   if (search(subdir,"/")!=-1 ||
       subdir=="..")  
      return 0;

   if (dirobjcache[subdir]) return dirobjcache[subdir];
   
   array a=file_stat(wd+"/"+subdir);
   if (!a || a[1]!=-2) return 0;

   return (dirobjcache[subdir]=VC.CVS["dir"](wd+"/"+subdir, root));
}

//!
//! method object file(string file)
//!    Get a VC.file object handling a file.
//! returns the new VC.file object.
//! note
//!	A '/' in a subdir or the subdir ".." is illegal.
//!	It will give an error if the subdir is not a directory.
//!

object file(string file,void|int iknowwhatimdoing)
{
   if (iknowwhatimdoing)
      return (fileobjcache[file] || 
	      (VC.CVS["file"](this_object(),file)));

   if (fileobjcache[file]) return fileobjcache[file];
   
   array a=file_stat(wd+"/"+file);
   if (!a || a[1]<0) return 0;

   return fileobjcache[file]=VC.CVS["file"](this_object(),file);
}

//!
//! method array(object(VC.dir)) subdirs()
//!    Get an array of all version controlled subdirs
//!    in the directory.
//! returns an array of VC.dir objects.
//!

array(object) subdirs()
{
   return Array.map(subdirnames(),subdir);
}

//!
//! method array(object(VC.file)) files()
//!    Get an array of all version controlled files
//!    in the directory.
//! returns an array of VC.file objects.
//!

array(object) files()
{
   return Array.map(filenames(),file);
}

//!
//! method array ignored()
//!	Get information about which files are ignored.
//! returns array of files and/or wildcards ignored in this directory
//! see also: ignore_add, ignore_remove, filterignored
//!
//!
//! method array filterignored(array files)
//!	filter out ignored files
//! returns array of filenames not ignored
//! see also: ignored
//!
//!
//! method void ignore_add(string file)
//! method void ignore_add(array files)
//! method void ignore_remove(string file)
//! method void ignore_remove(array files)
//! 	Add/remove files and/or wildcards to ignore.
//!
//! see also: ignored
//!

array filesignored=0;
int modfilesignored=0;

array defaultignored=({"*~","*#",".#*","*.o","*.a","CVS",".",".."});

array ignored()
{
   array a;
   if (!filesignored
      || ((a=file_stat(wd+"/.cvsignore")) && a[3]>modfilesignored))
   {
      object f=Stdio.File();
      if (f->open(wd+"/.cvsignore","r"))
      {
	 if(a || (a=f->stat()))
	    modfilesignored = a[3];
	 filesignored=(f->read(1000000)-"\r")/"\n";
	 f->close();
      }
      else filesignored=({});
   }
   return filesignored|defaultignored;
}

array filterignored(array files)
{
   foreach (ignored(),string gl) 
      files-=glob(gl,files);

   return files;
}

void ignore_add(string|array files)
{
   if (stringp(files)) files=({files});
   array ign=ignored();
   ign|=files;
   ign-=defaultignored;
   object f=Stdio.File();
   if (f->open(wd+"/.cvsignore","wtc"))
   {
      f->write(ign*"\n"+"\n");
      f->close();
   }
   else funerror("can't create/write "+wd+"/.cvsignore\n");
}

void ignore_remove(string|array files)
{
   if (stringp(files)) files=({files});
   array ign=ignored();
   ign-=files;
   ign-=defaultignored;
   object f=Stdio.File();
   if (f->open(wd+"/.cvsignore","wtc"))
   {
      f->write(ign*"\n"+"\n");
      f->close();
   }
   else funerror("can't create/write "+wd+"/.cvsignore\n");
}

//! 
//! method int|object(VC.dir) mkdir(string name)
//!	create a subdirectory and add it to the repository
//! returns a VC.dir object of the subdirectory on success, 0 upon failure
//! note
//!	'/' in the directory name is not allowed
//!

object mkdir(string name)
{
   if (search(name,"/")!=-1) 
      argerror("not a valid subdir name: "+name+"\n");

   if ((<"Attic", "#cvs.lock">)[name]) return 0; // no way

   predef::mkdir(wd+"/"+name);

   array a=file_stat(wd+"/"+name);
   if (!a || a[1]!=-2) return 0; // failed

   if(catch{string res=cvs->command("add",name);})
     return 0;

   return subdir(name); // success
}

//!
//! method int update(void|string file, string|void revision, void|int dirs, void|int recursive)
//!	Update (create) (all) file(s) from the repository.
//!
//! returns 1 on success, 0 on failure, 2 if conflict(s) was created
//! note
//!	will fail if the local file has conflicts
//!

int update(void|string file,void|string revision, 
	   void|int dirs, void|int recursive)
{
   string res;
   array args=({});

   if (revision) args+=({"-r",revision});
   if (dirs) args+=({"-d"});
   if (recursive) args+=({"-R"});
   if (file) args+=({file});

   res=cvs->command("update",@args);

   if (res=="") return 0;

   if (res[0]=='C' || -1!=search(res,"\nC")) return 2; // conflicts

   return 1; // ok
}

//!
//! method array stat()
//!

array stat()
{
   array a=file_stat(wd);
   return a;
}

//!
//! method string gettag()
//!    returns tag if existing
//!

string gettag()
{
   object f=Stdio.File();
   if (!f->open(wd+"/CVS/Tag","r"))
      return 0;
   
   if (f->read(1)=="T")
      return f->read(1000000);
   return 0;
}

//!
//! method array(array(string)) cleanup_locks()
//!    removes leftover CVS locks, returns list of filenames and owners
//!

static array(array(string)) cleanup_locks_rec(string dir,
					      array(array(string)) res,
					      mapping(int:string) user)
{
  foreach(get_dir(dir)||({}), string fn) {
    array s = file_stat(dir+"/"+fn);
    if(fn=="#cvs.lock") {
      string u =
#ifdef __NT__
	"???"; // NTFIXME
#else
	(s && (user[s[5]] || (user[s[5]]=(getpwuid(s[5])||({ "???" }))[0]))) ||
	"???";
#endif	  
      if(rm(dir+"/"+fn))
	res += ({ ({ dir+"/"+fn, u }) });
      else
	res += ({ ({ dir+"/"+fn, u, strerror(errno()) }) });
    } else if(s && s[1]==-2)
      res = cleanup_locks_rec(dir+"/"+fn, res, user);
  }
  return res;
}

array(array(string)) cleanup_locks()
{
  if(!repository)
    return 0;
  return cleanup_locks_rec(repository, ({}), ([]));
}
