// $Id: file.pike,v 1.56 1999/08/13 21:03:28 mast Exp $

//!
//! module vc
//! class dir
//!	class for handling of a directory
//!	of version-controlled files
//! 
//! method void create(string dir)
//! method void create(string dir,string root)
//!	Make a new version-controlling object.
//! arg string dir
//!	which directory to control
//! arg string root
//!	which root directory to use
//!

import VC.CVS;

string file;
string wd;
object cvs;
object dir;

int binary=0;

constant isfile=1;

#define argerror(X) (throw(({X,backtrace()}))) /* argument error */
#define funerror(X) (throw(({X,backtrace()}))) /* function error */

void create(object(VC.CVS.dir) _dir,string _file)
{
   if (!objectp(_dir)) 
      argerror("vc.file->create: illegal argument 1\n");

   dir=_dir;
   wd=dir->wd;
   cvs=dir->cvs;

   if (!wd || !cvs) 
      argerror("vc.file->create: illegal argument 1\n");

   file=_file;

   if (search(file,"/")!=-1)
      argerror("vc.file->create: illegal argument 2 (contains '/')\n");

// skip this... i need to do both update and status without 
// checkout'ed files...
//    array a=file_stat(wd+"/"+file);
//    if (!a || a[1]<0) argerror("no such file: "+wd+"/"+file+"\n");
}


//!
//! method int ignored()
//! returns 1 if this file is ignored by the VC system, else 0
//! see also: VC.dir->ignored, VC.dir->ignore_add, VC.dir->ignore_remove
//!

int ignored()
{
   return search(dir->ignored(), file)>=0;
}

//!
//! method object status()
//!    figure out current status 
//!
//!    the structure returned is a status object, containing:
//!    <pre>
//!    class 
//!    {
//! 	  int status;                  - status flags
//! 	  string working_version;      - base version for local file
//! 	  string repository_version;   - latest version in repository
//! 	  string branchtag;            - tag for this branch 
//! 	  string branchver;            - branch version base
//! 	  array(string) tags;          - tags on this file
//! 	  array(string) tags_versions; - ... and corresponding versions
//!    }
//!    </pre>
//!    status flags is a bitfield, containing these flags:
//!    <pre>
//!      UNKNOWN   - not known to vc system
//!      ILLEGAL   - illegal (error, non-existant, etc)
//!      IGNORED   - ignored
//!     
//!      RMODIFIED - remote/repository modified
//!      LMODIFIED - locally modified
//!      LREMOVED  - locally removed
//!      LADDED    - locally added
//!     	  
//!      CONFLICT  - has conflict
//!      BINARY    - binary file
//!    </pre>
//!    thus (status&(RMODIFIED|LMODIFIED))==0 means the file is
//!    up-to-date and does not need to be commited or updated.
//!
//!    branchtag and branchver is zero if this is the main branch.
//!
//! returns the object described above
//!

static object _status=class status_class
{
   int status;
   string working_version;
   string repository_version;
   string branchtag; // "sticky tag"   
   string branchver; // these two are zero if main branch is zero
   array(string) tags;
   array(string) tags_versions; // corresponds to 'tags'
   int local_atime;
   int local_mtime;
   int local_ctime;
   int repository_atime;
   int repository_mtime;
   int repository_ctime;

   mapping|string cast(string to)
   {
      if (to=="mapping") return mkmapping(indices(this_object()),
					   values(this_object()));
      else if (to=="string") return VC.CVS.nice_flags(status);
      else return 0;
   }
}();
static int _status_mtime=-1;
static string repository_file=0;
static int _sreps=0,_slocs=0;

static object calculate_status(string cvsoutput)
{
   string cstatus,cwrev,cwdate,crrev,crfile,cstag,csdate,csopt,tags;
   string cfile;

//    werror(" | "+cvsoutput/"\n"*"\n | "+"\n");

   object s=_status;
   array a;

   tags="";

   int scan=
      sscanf(cvsoutput,
	  "%*[= \t\n]File: %sStatus: %s\n"
	  "%*[ \t\n]Working revision:%*[ \t\n]%s%*[ \t\n]%s\n"
	  "%*[ \t\n]Repository revision:%*[ \t\n]%s%*[ \t\n]%s\n" // scan <= 13
	  "%*[ \t\n]Sticky Tag:%*[ \t\n]%s\n"
	  "%*[ \t\n]Sticky Date:%*[ \t\n]%s\n"
	  "%*[ \t\n]Sticky Options:%*[ \t\n]%s\n" // scan <= 22
	  "%*[ \t\n]Existing Tags:%*s\n"
	  "%s",
	  cfile,cstatus,
	  cwrev,cwdate,
	  crrev,crfile,
	  cstag,csdate,csopt,
	  tags);

   sscanf(cfile||"no file ","no file %s",cfile);
   if ((cfile||"")[..strlen(file)-1]!=file)
      werror("VC.CVS.status(): WARNING! file names differ! "+file+"  "+cfile+"\n");

   if (scan==0)
   {
      werror("VC.CVS.file->status(): Unexpected form for cvs status:\n"+cvsoutput+"\n");
      s->status=UNKNOWN;
      return s;
   }

   if (crfile!=repository_file &&
       (a=file_stat(crfile)))
   {
      s->repository_atime=a[2];
      s->repository_mtime=a[3];
      s->repository_ctime=a[4];

      repository_file=crfile;
   }
   else
      s->repository_atime=0;

   int new_status_mtime =
      (s->local_mtime<s->repository_mtime)
      ?s->repository_mtime:s->local_mtime;

   // Caveat Emptor!  If new_status_mtime == time(0) the file
   // might change status within the same second without us
   // noticing.  Better not update _status_mtime in this case.
   if (new_status_mtime < time(0))
     _status_mtime=new_status_mtime;

   s->tags=({});
   s->tags_versions=({});

   if (scan==14   // unknown, but existing, probably
       && (<"Unknown", "Up-to-date">)[cstatus]) // yes
     // In a branch, cvs says "Unknown" for files that doesn't exist
     // in it. In the main trunk, it says "Up-to-date" instead. :P
   {
      if (ignored()) s->status=IGNORED;
      else s->status=UNKNOWN;
      return s;
   }
   if (csopt=="-kb") binary=1;
   s->status=ILLEGAL;
   s->status=([ "Locally Added":LADDED|LMODIFIED,
		"Locally Modified":LMODIFIED,
		"Locally Removed":LREMOVED|LMODIFIED,
		"Needs Checkout":RMODIFIED,
		"Needs Merge":LMODIFIED|RMODIFIED,
		"Needs Patch":RMODIFIED,
		"Unknown":UNKNOWN,
		"Entry Invalid":ILLEGAL,
		"Unresolved Conflict":CONFLICT|LMODIFIED,
		"Up-to-date":0 /* UPTODATE */,
             ])[cstatus];

   if(cwrev && cwrev!="No")
     s->working_version=cwrev;
   if(crrev && crrev!="No")
     s->repository_version=crrev;

   if (scan<23) // error, give no flags
   {
      object f=Stdio.File();
      if (f->open(wd+"/CVS/Tag","r"))
      {
	 s->branchtag=f->read()[1..]-"\n"-"\r"; // kludgy... *sigh* /mirar
	 f->close();
      }
      s->status|=INCOMPLETE;
      return s;
   }

   if (cstag!="(none)")
   {
      sscanf(cstag,"%s%*[\t ](%*s:%*[ \t]%s)%*s",
	     s->branchtag,s->branchver);
   }

   foreach (tags/"\n"-({""}),string tag)
   {
      string t,r;
      sscanf(tag,"%*[ \t]%s%*[ \t](%*s:%*[ \t]%s)",t,r);
      s->tags+=({t});
      s->tags_versions+=({r});
   }
}

object async_status_pipe=0;
object async_status_object=0;

static object have_status()
{
   object s=_status;
   array a;
   int mtime1=0,mtime2=-1;

   if (repository_file)
   {
      if (a=file_stat(repository_file))
      {
	 s->repository_atime=a[2];
	 s->repository_mtime=a[3];
	 s->repository_ctime=a[4];
	 mtime1=a[3];
      }
   }
   else
   {
      // No repository file.  We can't check whether one has appeared
      // without actually running "cvs status".  Help!
      return 0;
   }
   if (a=file_stat(wd+"/"+file)) 
   {
      s->local_atime=a[2];
      s->local_mtime=a[3];
      s->local_ctime=a[4];
      mtime2=a[3];
   }
   else 
   {
      if (s->local_atime) 
	 return 0;
   }

   if (mtime2>mtime1) mtime1=mtime2;
   if (mtime1==_status_mtime)
      return s;

   return 0;
}

void async_status()
{
   if (async_status_object) return;
   if (have_status()) return;

   array a=cvs->async_command("-n","status","-v",file);	// -n: see below.
   async_status_pipe=a[1];
   async_status_object=a[0];
}

object status()
{
   if (async_status_pipe)
   {
      calculate_status(async_status_pipe->read());
      async_status_pipe=0;
      destruct(async_status_object); 
      async_status_object=0;
      return _status;
   }
   if (async_status_object)
      async_status_object->wait();

   if (have_status()) return _status;

   // -n here keeps cvs status from meddling with the Entries file,
   // which it will do e.g. when the mtime has changed but the
   // contents hasn't. It will then report up-to-date and update the
   // timestamp in Entries, but it will not sleep to ensure that a
   // write immediately afterwards changes the mtime.
   calculate_status(cvs->command("-n","status","-v",file));

   return _status;
}


//!
//! method int commit(string comment)
//!	commit a file with the given comment to the
//!	version control repository
//! returns true (1) on success, throws exception on error
//! note
//!	will fail if the file has conflicts, or
//!	if the file is not up-to-date.
//!

int commit(string comment)
{
   string res=cvs->command("commit","-m",comment,file);
   if (res=="")
     throw (({"Didn't get any output from cvs commit\n", backtrace()}));
   return 1;
}

//!
//! method int add(void|string comment)
//! method int remove(void|int dounlink)
//!	Add a file to version control/
//!	remove a file from version control.
//!
//!	A file cannot be removed if it exists
//!	in the filesystem; the dounlink flag
//!	make the vc system remove the file.
//!
//! returns true (1) on success, 0 on failure 
//!

int add(void|string comment)
{
   catch{
      if (comment)
	 cvs->command("add",(binary?"-kb":"-ko"),"-m",comment,file);
      else
	 cvs->command("add",(binary?"-kb":"-ko"),file);
   };
   return !!(status()->status&LADDED);
}

int remove(void|int dounlink)
{
   if (dounlink) 
      rm(wd+"/"+file);
   cvs->command("remove",file);
   return !!(status()->status&LREMOVED);
}

//!
//! method int update(string|void revision)
//! method string update_stdout(string|void revision)
//! method int join(string revision)
//!	Update/join a file from the repository.
//!
//! returns 1 on success, 0 on failure, 2 if conflict(s) was created
//! note
//!	will fail if the local file has conflicts
//!

int update(void|string revision)
{
   string res;
   if (revision)
      res=cvs->command("update",@(binary?({"-kb"}):({})),"-r",revision,file);
   else 
      res=cvs->command("update",@(binary?({"-kb"}):({})),file);

   if (res=="") return 0;

   _status_mtime = -1;
   
   if (res[0]=="C") return 2;
   return 1;
}

string update_stdout(void|string revision, void|string date)
{
   return cvs->command("update","-p",@(binary?({"-kb"}):({})),
		       @(revision?({"-r",revision}):({})),
		       @(date?({"-D",date}):({})),file);
}

int join(string revision)
{
   string res;
   if (revision)
      res=cvs->command("update",@(binary?({"-kb"}):({})),"-j",revision,file);
   else
      res=cvs->command("update",@(binary?({"-kb"}):({})),file);

   if (res=="") return 0;
   
   if (res[0]=="C") return 2;
   return 1;
}

//!
//! method int tag(string tag,void|string revision)
//! method int branch(string tag,void|string revision)
//! method int branchoff(string tag,void|string revision)
//!	Tag/branch file.
//!	If the tag already exists, it will be moved.
//!	If revision is given, the tagging/branching will be 
//!	on that revision and not the local file.
//!
//!	The difference between <ref>branch</ref> and 
//!	<ref>branchoff</ref> is that the latter will
//!	keep you in the old branch and the previous
//!	will change the working branch.
//!
//!	<ref>tag</ref> will not branch at all, but tag files.
//!
//! returns true (1) on success, 0 on failure 
//!

int tag(string tag,void|string revision)
{
   return 0;
}

int branch(string tag,void|string revision)
{
   return 0;
}

// local file needs update to get new branch tag, or some other magic

//!
//! method array diff()
//! method array diff(string revision)
//! method array diff(string revision1,string revision2)
//!	diff a file,
//!	between current (left) and last version (right),
//!	between current (left) and given revision (right),
//!	or between two revisions (first on left, second on right).
//!
//!	Result is written to an array of strings or array (tupels)
//!	of array where there is a difference:
//!
//!	<pre>
//!     ({"foo\n",({"bar\ngazonk\n","hey\n"})})
//!	  ^same^^   ^^^left^^^^^^^^ ^right^
//!	</pre>
//!
//! returns the array
//!

array(string|array(string)) diff(void|string rev,void|string rev2)
{
   return 0;
}

//!
//! method array conflicts()
//!	get conflicts in a file. The result is the
//!	same array as given from a <ref>diff</ref>.
//!	right is the revision version and left is the
//!	local version.
//!
//! returns the array
//!

array(string|array(string)) conflicts()
{
   return 0;
}

//!
//! method int isbinary()
//! method int isbinary(int set)
//!	get/set file-is-binary flag.
//! returns -1 upon error, 1 if the file is binary and 0 if it is a textfile
//!

int isbinary(int ...set)
{
   if (sizeof(set)) binary=!!set[0];
   return binary;
}

//!
//! method int hasconflict()
//! method int hasconflict(int set)
//!	Get/set file-has-conflict flag.
//! returns -1 upon error, 1 if the file is binary and 0 if it is a textfile
//!

int hasconflict(int ...set)
{
   return 0;
}

//!
//! method int move(string destination)
//!	Move a file/subdir to a new location.
//!
//!     NOT IMPLEMENTED AND IMPOSSIBLE TO IMPLEMENT USING CVS
//!


//!
//! method array stat()
//!

array stat()
{
   array a=file_stat(wd+"/"+file);
   return a;
}

//!
//! metod array (mapping) log(void|string revision,void|string dates)
//!	get the cvs log from a file
//!
//!    	revision: revision "," revision
//!    	dates:    dates "," dates
//!    		  or < date
//!    	       or > date
//!    	       or date < date
//!    	    where date = yyyy-mm-dd, ie "1998-02-02"
//!            or        yyyy-mm-dd hh:mm (ie "2003-07-19 19:59")
//!
//!	Result is written to an array of Logentry objects.
//!

class Logentry
{
  string revision;		// Revision of this entry
  string date;			// Date of this entry
  string author;		// Author
  string message;		// Commit message
  array(string) branches;	// Branches
}

array (object) log(void|string revision,void|string dates)
{
  string in = cvs->command( "log",
			    @ (( dates? ({"-d"+dates}): ({})) +
			       ( revision?
				 ({(revision==""? "-b":"-r"+revision)}):({}))),
			    file );
  array entries;
  array (object(Logentry)) result = ({ });

  in -= "==================================================================="
  "==========";
  entries = (in / "\n----------------------------\n")[1..];
  foreach (entries, string e)
  {
    array arr = e / "\n";
    string s;
    object le=Logentry();
    string state = 0;

    while (sizeof( arr ) && arr[-1] == "")
      arr = arr[0..sizeof( arr )-2];
    if (sizeof( arr ) < 2
	|| sscanf( arr[0], "revision %s", le->revision ) != 1
	|| sscanf( arr[1], "date: %s;  author: %s;  state: %s;%*s",
		   le->date, le->author, state ) < 2)
    {
       throw( ({ "VC.CVS.file:log Format error\n", backtrace() }) );
    }

    // cvs log prints the date in UTC (without specifying that time
    // zone), but other cvs commands reads it in local time. It should
    // be safe to just add the timezone to the date to make it work
    // correctly when fed to other commands. FIXME: Convert it to
    // local time instead.
    le->date += " UTC";

    if (state == "dead") {
      // Suppress this version
      continue;
    }
    if (sscanf(arr[2],"branches:  %s", s))
    {
       le->branches=s/";  "-({""});

       le->message = arr[3..] * "\n";
    }
    else le->message = arr[2..] * "\n";

    result += ({ le });
  }
  return result;
}

//!
//! method array annotate(string|void revision)
//!	annotate a file.
//!
//!	Result is written to an array of Annotation objects,
//!	containing the fields
//!     revision, user, date and data.
//!
//! note
//!	The result can be put together to the whole file
//!     by something like this: (annotate()->data)*"".
//!
//! returns an array of Annotation objects
//!

array(object(VC.CVS.Annotation)) annotate(void|string revision)
{
   string res;
   if (revision)
      res=cvs->command("annotate","-r",revision,file);
   else
      res=cvs->command("annotate",file);

   array ann=({});
   object current;

   string revision,last;
   foreach((res/"\n"), string line) 
   {
      if(2==sscanf(line, "%[^:]: %s", revision, line)) 
      {
	 if(current && last == revision)
	    current->data+=line+"\n";
	 else  
	 {
	    last=revision;
	    current=VC.CVS.Annotation();
	    ann += ({ current });

	    string who, when;
	    sscanf((replace(revision,"\t", " ")/" "-({""}))*" ",
          	    "%s (%s %s)", revision, who, when);
	    current->revision=revision;
	    current->user=who;
	    current->date=when;
	    current->data=line+"\n";
	 }
      }
   }
   return ann;
}





//! method string split_revision(string ... tags)

string split_version(string ... tags)
{
#ifdef VCDEBUG
  werror( sprintf( "split version: %O\n", tags ) );
#endif
  
   if (!sizeof(tags)) return 0; // no
    
   object s=status();
   mapping ts=mkmapping(s->tags,s->tags_versions);
   
   int len=12381;
   array vers=({});
   foreach (tags, string tag)
   {
      vers+=({ ( ts[tag] || tag )/"." });
      if (sizeof(vers[-1])&1) vers[-1]+=({0});
      if (sizeof(vers[-1])<len) len=sizeof(vers[-1]);
   }

   int i,j;

   for (i=0; i<len; )
   {
      mixed n=vers[0][i];
      for (j=1; j<sizeof(vers) && n==vers[j][i]; j++);
      if (j<sizeof(vers)) break;
      i+=2;
      n=vers[0][i-1];
      for (j=1; j<sizeof(vers) && n==vers[j][i-1]; j++);
      if (j<sizeof(vers)) break;
   }

   if (i==0) return 0; // no common version
   
   string|int min=9999;
   
   for (j=0; j<sizeof(vers); j++)
      if ((int)min>(int)vers[j][i-1]) min=vers[j][i-1];
   
   if (min==0) // last is a non-version
      if (i==2) return 0; // no common version (shouldn't happen)
      else return vers[0][..i-2]*"."+".1"; // first version?
   
   return vers[0][..i-2]*"."+"."+min;
}


string version_tagdate(string date,void|string tag)
{
   // date is yyyy-mm-dd[ hh:mm] (ie "1999-03-13" [+" 19:59"])
   // figure out the version from tag (default is the sticky) and date

   // returns version (string) or 0 

   if (!tag) 
      if (!(tag=status()->branchtag))
      {
	 throw( ({"file has no tag: "+wd+"/"+file+"\n",
		  backtrace()}) );
      }
  
#ifdef VCDEBUG
   werror("version_tagdate: "+date+" tag="+tag+"\n");
#endif
   
   // need log for these

   if (search(date," ")==-1) date+=" 23:59";

   array l=log(tag,"<"+date);
   
   if (sizeof(l))
   {
      // first entry is what we want (the latest before the end of that date)
      return l[0]->revision;
   }
   else // branch may be new, and has no own log entries
   {
#ifdef VCDEBUG
   werror("version_tagdate: "+date+" tag="+tag+" loop init...\n");
#endif
      string v=split_version(tag);
      while (v && v!="")
      {
#ifdef VCDEBUG
   werror("version_tagdate: "+date+" tag="+tag+" loop on v="+v+"\n");
#endif
	 l=log(v,"<"+date);
	 if (sizeof(l)) return l[0]->revision;
	 array nv=v/".";
	 v=nv[..sizeof(nv)-3]*".";
      }
      return 0;
   }
}
