
// file status flags

constant UPTODATE =  0; // ok

constant UNKNOWN=    1; // not known to vc system
constant ILLEGAL=    2; // illegal (error, non-existant, etc)
constant IGNORED=    4; // ignored

constant RMODIFIED=  8; // remote/repository modified
constant LMODIFIED= 16; // locally modified
constant LREMOVED = 32; // locally removed
constant LADDED   = 64; // locally added

constant CONFLICT =128; // has conflict

constant BINARY   =256; // binary file

constant FAKE     =512; // this file is faked (cachefile)
constant INCOMPLETE=1024; // status is incomplete (is, but not in CVS/Entries)


constant flagnames= ({
   "UNKNOWN","ILLEGAL","IGNORED","RMODIFIED","LMODIFIED",
   "LREMOVED","LADDED","CONFLICT","BINARY","FAKE","INCOMPLETE"});

class Annotation
{
   string revision;
   string user;
   string date;
   mixed data;

   void create(void|object old)
   {
      if (old)
      {
	 revision=old->revision;
	 user=old->user;
	 date=old->date;
	 data=old->data;
      }
   }
}

//!
//! method int init(string cvsroot)
//! 	initiates a cvs root directory
//!

int init(string cvsroot)
{
   object cvs=VC["CVS"]["CVS"](".",cvsroot);
   cvs->command("init");
   return !!file_stat(cvsroot+"/CVSROOT");
}

int new_repository(string cvsroot, 
		   string repository)
{
   return mkdir(cvsroot+"/"+repository);
}

int checkout_repository(string cvsroot,
			string repository,
			string dir,
			void|string tag,
			void|int time)
{
   function make_file=
      lambda(string file,string data)
      {
	 object f=Stdio.File(file,"wtc");
	 f->write(data);
	 f->close();
      };
   
   cvsroot=combine_path(getcwd(),cvsroot);
   repository=combine_path(cvsroot,repository);
   if (repository[..strlen(cvsroot)-1] == cvsroot) {
      repository = repository[strlen(cvsroot)..];
      while (repository[..0] == "/")
	repository = repository[1..];
      if (!strlen(repository))
	repository = ".";
   }

   mkdir(dir);
   mkdir(dir+"/CVS");
   make_file(dir+"/CVS/Root",cvsroot+"\n");
   make_file(dir+"/CVS/Repository",repository+"\n");
   make_file(dir+"/CVS/Entries","D\n");
   if (tag)
      make_file(dir+"/CVS/Tag","T"+tag+"\n");
   else if (time)
   {
      mapping tm=localtime(time);
      make_file(dir+"/CVS/Tag",
		sprintf("D%02d.%02d.%02d.%02d.%02d.%02d\n",
			tm->year+(tm->year<100?0:1900), // yes...
			tm->month+1,
			tm->mday,
			tm->hour,
			tm->min,
			tm->sec));
   }
}

int rtag(string cvsroot,
	 string tag,
	 string repository,
	 void|string oldtag,
	 void|string fordate)
{
   object cvs=VC["CVS"]["CVS"](".",cvsroot);
   if (fordate)
      cvs->command("rtag","-b","-F","-D",fordate,tag,repository);
   else if (oldtag)
      cvs->command("rtag","-b","-F","-r",oldtag,tag,repository);
   else
      cvs->command("rtag","-b","-F",tag,repository);
   return 1;
}

int rtag_delete(string cvsroot,
		string tag,
		string repository,
		void|string oldtag,
		void|string fordate)
{
   object cvs=VC["CVS"]["CVS"](".",cvsroot);
   if (fordate)
      cvs->command("rtag","-d","-F","-D",fordate,tag,repository);
   else if (oldtag)
      cvs->command("rtag","-d","-F","-r",oldtag,tag,repository);
   else
      cvs->command("rtag","-d","-F",tag,repository);
   return 1;
}

// helper (debug) function
string nice_flags(int fs)
{
   array flags=({});

   foreach (flagnames,string s)
      if (fs&this_object()[s]) flags+=({s});
   return flags*"|";
}

