inherit "../base.pike";

import Sitebuilder;

constant title = "Files";
constant order = 2;
constant group_order = ({ "View",
			  "Edit",
			  "File",
			  "Version Control",
			  "Access Control",
			  "Undelete" });
constant wizard_order = ({ "edit_file.pike",
			   "create_file.pike",
			   "edit_metadata.pike",
			   "download_file.pike",
			   "upload_file.pike",
			   "upload_new_file.pike",
			   "create_dir.pike",
			   "copy.pike",
			   "move.pike",
			   "delete.pike",
			   "view.pike",
			   "view_source.pike",
			   "log.pike",
			   "annotate.pike",
			   "discard.pike",
			   "update.pike",
			   "commit.pike",
			   "join.pike",
			   "add_ppoint.pike",
			   "remove_ppoint.pike",
			   "enter_normal_mode.pike",
			   "enter_undelete_mode.pike",
			   "undelete.pike" });

array extra_wizard_args(object id)
{
  if(id->misc->wa)
    id->misc->vcobj=id->misc->wa->vcobj(id->misc->path, id);
  return ({ id->misc->vcobj });
}

string tag_metadata_viewer(string tagname, mapping args, object id)
{
  object vcfile = id->misc->wa && id->misc->wa->vcfile(id->misc->path, id);
  return MetaData.view(vcfile,id);
}

string tag_previewer(string tagname, mapping args, object id)
{
  object vcfile = id->misc->wa && id->misc->wa->vcfile(id->misc->path, id);
  if(id->misc->sb->handler_for(vcfile)->preview)
    return
      "<b>Preview</b><br>"+
      id->misc->sb->handler_for(vcfile)->preview(vcfile,id);
  else
    return "";
}


string tag_metadata_editor(string tagname, mapping args, object id)
{
  object vcfile;
  if(args->read_from)
    vcfile = id->misc->wa && id->misc->wa->vcfile(args->read_from, id);
  multiset(string) metadata_list = (args->metavars?
				    (< @(args->metavars/",") >) : 0);
  return MetaData.edit(vcfile, id, metadata_list);
}

string tag_status_icon(string tag, mapping args, object id)
{
  string loc = args["local"]||"none", rem = args->remote||"none";
  string alt, src;

  if(loc=="none" && rem!="removed") {
    // If the file doesn't have a local status, make a blank icon
    src = "/internal-roxen-unit";
    alt = "                   ";
  } else {
    src = id->misc->img_url+loc+"_"+rem+".gif";
    alt = sprintf("[%|8s/%|8s]", loc, rem);
  }

  return "<img border=0 width=30 height=20 src=\""+src+"\" alt=\""+
    replace(alt, " ", "&nbsp;")+"\">";
}

string tag_menu_icon(string tag, mapping args, object id)
{
  if(args->what=="menu"||args->what=="menu-back")
    return ("<img border=0 width=20 height=20 src=\""+id->misc->img_url+
	    args->what+".gif\" alt=''>");
  else if(args->what=="menu-open")
    return ("<img border=0 width=20 height=20 src=\""+id->misc->img_url+
	    args->what+".gif\" alt=''>");
}

static string select_icon_for_metadata(mapping meta, object id)
{
  string ctype = meta["http-content-type"] || "unknown";
  return id->misc->img_url+"types/"+
    (id->misc->sb->ct_data_for(ctype)->img ||
     id->misc->sb->ct_data_for("application/octet-stream")->img ||
     "unknown.gif");
}

static string add_navigation(object id, string ty, string o,
			     string path, string txt, int|void discard,
			     string|void extra)
{
  if(o) {
    multiset(string) opts = mkmultiset(o/",");
    if(!opts["no"+ty])
      if(opts["submit"])
	return parse_html(replace(txt,"\0","&nbsp;"),
			  (["img":lambda(string tag, mapping args,
					 object id, string name){
				    args->name = name;
				    args->type = "image";
				    return make_tag("input", args);
				  },
			    "sb-internal-status-icon":tag_status_icon,
			    "sb-internal-menu-icon":tag_menu_icon]), ([]),
			  id, "_nav_"+sprintf("%@02x", values(path)))+
	(extra||"");
      else
      {
	string ahref="<a href=\""+
	  http_encode_string(id->misc->build_url(id, 0, 0, "/"+path))+
	  "\">";
	return ahref+replace(txt,
			     ({ "\000", "\001" }),
			     ({ "</a>&nbsp;"+ahref,
				"</a>&nbsp;&nbsp;"+ahref }))
	  +"</a>"+(extra||"");
      }
  }
  return (discard? "":txt);
}

static string filelisting_file(object(VCFile) vcfile,
			       mapping args, object id, int level, int undel)
{
  mapping|object meta = vcfile->metadata(id, undel);
  meta = objectp (meta) ? meta->md : ([]);
  string res =
    "<img border=0 width=25 height=20 src=/internal-roxen-unit alt='  '>"+
    "<img border=0 src=\""+
    http_encode_string(select_icon_for_metadata(meta,id))+"\" alt=''>\0"
    "<sb-internal-status-icon local="+
    vcfile->userstat_ea(id, undel)+
    " remote="+vcfile->userstat_rep(id, undel)+">";
  if(level==0) res+="<b>";
  string fn=
    (level==0?("/"+vcfile->abspath(id, undel)):
     vcfile->name(id, undel) ) ||
    "Mystery file";
  res += "\001<tt>"+replace(html_encode_string(fn)," ","&nbsp;")+"</tt>";
  if(level==0) res+="</b>";
  string abspath;
  if(abspath=vcfile->abspath(id, undel))
    res = add_navigation(id, "file", args->navigate, abspath, res);
  res += "<tt>"+String.strmult("&nbsp;",18-sizeof(fn)) + "</tt>";
  array ss;
  if(args->filesize && (ss=vcfile->stat(id, undel)) && sizeof(ss)>1) {
    float s = (float)ss[1];
    int xp=0;
    while(s>=999.95) {
      s /= 1024.0;
      xp++;
    }
    res += "&nbsp;<tt>"+replace(sprintf((xp? "%5.1f%c":"%5d%c"), (xp? s:(int)s),
					" KMGTPEZY"[xp])," ", "&nbsp;")+"</tt>";
  }
  if(meta->title)
    res +=
      "&nbsp;&nbsp;<i>"+replace(html_encode_string(meta->title)," ","&nbsp;")
      +"</i>";
  if(abspath && args->name)
    res = "<var name="+args->name+" type=checkbox "+
      (args->unchecked?"un":"")+"checked value='/"+
      replace(abspath, "'", "'\"'\"'")+"'>"+res;
  if(level==0)
    res +=
      "<br><br>"
      "<sb-internal-metadata-viewer>"
      "<br>"
      "<sb-internal-previewer>";
  return res+"<br>\n";
}

static string make_flik(mapping args, object id,
			multiset(string) unfolded, string abspath)
{
  int u = unfolded && unfolded[abspath];
  string imgbody = "width=20 height=20 src=\""+id->misc->img_url+
    (u?"":"un")+"fold.gif\" border=0 alt='"+(u?"\\/":"--")+"'";

  return (args->flik=="submit"?
	  "<input type=image "+imgbody+" name=_fl_"+(u?"f":"u")+"_"+
	  sprintf("%@02x", values(abspath))+">":
	  "<a href=\""+
	  http_encode_string(id->misc->build_url(id, 0, 0, 0, 0, 0,
						 (u? unfolded-(<abspath>) :
						  (unfolded||(<>))+
						  (<abspath>))))+
	  "\"><img "+imgbody+"></a> ");
}

static string filelisting_dir(object(VCDir) vcdir,
			      mapping args, object id, int level, int undel,
			      multiset(string) unfolded)
{
  string res =
    "<sb-internal-menu-icon what=menu"+(level==0?"-open":"")+">\0"+
    (level!=0?
     ("\0<sb-internal-status-icon local="+
      vcdir->userstat_ea(id, undel)+
      " remote="+vcdir->userstat_rep(id, undel)+">\0"):"")+
     
//      "<img border=0 src=/internal-roxen-unit alt='"+
//       replace("                   ", " ", "&nbsp;")+
//       "' width=31 height=20>\0":""+
    replace(sprintf("\0<tt>%s</tt>",
		    replace((level?"":"<b>")+
			    html_encode_string(( level?vcdir->name(id, undel)||"":
						 ("/"+vcdir->abspath(id, undel))
						 ||"")+"/")+
			    (level?"":"</b>")
			    ,
			    "//","/")),
		    " ", "&nbsp;");

  string abspath;
  if(abspath=vcdir->abspath(id, undel))
    res = add_navigation(id, "dir", args->navigate, abspath, res);
  res += "<br>\n";
  if(args->flik && level>0 && abspath)
    res = make_flik(args, id, unfolded, abspath)+res;
  if(abspath && args->name && args->checkdirs)
    res = "<var name="+args->name+" type=checkbox "+
      (args->unchecked?"un":"")+"checked value='/"+
      replace(abspath, "'", "'\"'\"'")+"'>"+res;
  res = "<dt>" + res;

  if(level==0 || args->recursive ||
     (args->flik && unfolded && unfolded[abspath]))
  {
    array|mapping list = vcdir->list(id, undel);
    if(mappingp(list))
      return "<b>"+ html_encode_string (error_msg (list)) +"</b>";
    list = Array.filter (list, "exists", id, undel);
    array dirs = Array.filter (list, "isdirp"), files = Array.filter (list, "isfilep");
    sort (dirs->name (id, undel), dirs);
    files=Array.sort_array(files,
			   lambda(object a, object b)
			   {
			     if(a->name(0,1,1)=="index.html")
			       return 0;
			     if(b->name(0,1,1)=="index.html")
			       return 1;
			     string
			       ct_a=(string)a->metadata(0,1,1)->md["http-content-type"],
			       ct_b=(string)b->metadata(0,1,1)->md["http-content-type"];
			     string
			       mt_a=(ct_a/"/")[0],
			       mt_b=(ct_b/"/")[0],
			       st_a=(ct_a/"/")[-1],
			       st_b=(ct_b/"/")[-1];
			     if(ct_a==ct_b)
			       return a->name(0,1,1) > b->name(0,1,1);
			     else if(mt_a==mt_b)
			       return st_a > st_b;
			     else
			     {
			       if(mt_a=="text")
				 return 0;
			       if(mt_b=="text")
				 return 1;
			       return ct_a > ct_b;
			     }
			   });
//  sort (files->name (id, undel), files);
    list = dirs + files;
    res +=
      "<dl>"+ (Array.map(
	list,
	lambda(object f, mapping args, object id, int level,
	       int undel, multiset(string) unf) {
	  if (args->is_user_op ? f->is_user_op (id, args->is_user_op, undel) :
	      args->is_valid_op ? f->is_valid_op (id, args->is_valid_op, undel) :
	      f->exists(id, undel))
	    return "<dt>"+(f->isdir?
			   filelisting_dir(f, args, id, level, undel, unf) :
			   filelisting_file(f, args, id, level, undel));
	  else return "";
	},
	args, id, level+1, undel, unfolded
      )-({0}))*""+((level==0&&!sizeof(list))?"<br><br>Directory is empty!":"")+
      "</dl>";
  }
  return res;
}

string tag_filelisting(string tagname, mapping args, object id)
{
  int undel = id->misc->undelete_mode=="1";
  string res = "";

  if(args->undelete_mode)
    undel = !!(int)(args->undelete_mode);

  multiset(string) unfolded = id->misc->unfolded||(<>);
  if(args->flik=="submit" && unfolded) {
    if(id->variables->_fl)
      unfolded = mkmultiset(id->variables->_fl/"&");
    foreach(glob("_fl_*.x", indices(id->variables)), string xx) {
      int i, action;
      string fn;
      if(2==sscanf(xx, "_fl_%c_%[^.].x", action, fn)) {
	for(i=0; i<strlen(fn); i++) {
	  int h;
	  sscanf(fn[i..i+1], "%x", h);
	  fn = fn[..i-1]+sprintf("%c", h)+fn[i+2..];
	}
	if(unfolded[fn]!=(action=='u'))
	  unfolded[fn]=(action=='u');
      }
    }
    id->variables->cancel_url = id->misc->build_url(id, 0, 0, 0, 0, 0,
						    unfolded);
    res += "<input type=hidden name=_fl value='"+indices(unfolded)*"&"+"'>\n";
  }

  if(args->files)
    foreach(args->files/"\0", string fn) {
      object vcobj = id->misc->wa && id->misc->wa->vcobj(fn, id, undel);
      if(!vcobj || !vcobj->exists(id, undel)) {
	if (vcobj)
	  werror ("Path not accessible 1: " +
		  (vcobj->wa && vcobj->wa->tag) + ":/" + vcobj->real_abspath() + " " +
		  vcobj->userstat_ea (0, 1, 1) + " " +
		  vcobj->userstat_rep (0, 1, 1) + "\n");
	else werror ("Path not accessible 1: no vcobj\n");
	res += "<b>Path "+fn+" is not accessible.</b><br>\n";
      }
      else
	res += (vcobj->isdir?
		filelisting_dir(vcobj, args, id, 1, undel, unfolded) :
		filelisting_file(vcobj, args, id, 1, undel));
    }
  else {
    string path = args->file || id->misc->path;
    object vcobj = (args->file?
		    id->misc->wa && id->misc->wa->vcobj(args->file, id, undel):
		    id->misc->vcobj);
    if(!vcobj || !vcobj->exists(id, undel)) {
      if (vcobj)
	werror ("Path not accessible 2: " +
		(vcobj->wa && vcobj->wa->tag) + ":/" + vcobj->real_abspath() + " " +
		vcobj->userstat_ea (0, 1, 1) + " " +
		vcobj->userstat_rep (0, 1, 1) + "\n");
      else werror ("Path not accessible 2: no vcobj\n");
      return "<b>Path "+path+" is not accessible.</b><br>\n"+
	add_navigation(id, "dir", path != "/" && args->navigate, "",
		       "<img border=0 src=internal-gopher-menu alt=''>"
		       "Back to /", 1);
    }
    res += (vcobj->isdir?
	    filelisting_dir(vcobj, args, id, 0, undel, unfolded) :
	    filelisting_file(vcobj, args, id, 0, undel));
    string abspath;
    if(objectp(vcobj->ascend(id, undel)) &&
       (abspath=vcobj->ascend(id, undel)->abspath(id, undel)))
      res = add_navigation(id, "dir", args->navigate, abspath,
 			   "<sb-internal-menu-icon what=menu-back>\001"+
			   (vcobj->isdir?
			    "Up to parent directory":
			    "Back to current directory"), 1, "<p>\n")+res;
  }
  
  if(args->flik=="submit")
    foreach(glob("_fl*", indices(id->variables)), string v)
      m_delete(id->variables, v);

  return res;
}

string tag_dirselector (string tagname, mapping args, object id)
{
  string dir = args["default"];
  if(id->variables[args->name])
    dir = id->variables[args->name];
  foreach(glob("_nav_*.x", indices(id->variables)), string xx)
    if(1==sscanf(xx, "_nav_%[^.].x", dir))
      for(int i=0; i<strlen(dir); i++) {
	int h;
	sscanf(dir[i..i+1], "%x", h);
	dir = dir[..i-1]+sprintf("%c", h)+dir[i+2..];
      }
  id->variables[args->name] = dir;
  foreach(glob("_nav_*", indices(id->variables)), string v)
    m_delete(id->variables, v);
  return "<sb-internal-filelisting flik=submit navigate=submit,nofile file='"+
    replace(dir, "'", "'\"'\"'")+"'>";
}

string container_rawview (string name, mapping args, string text, object id)
{
  string plainres = args->hide || text;

  object wa = id->misc->wa;
  if (!wa) return plainres;

  object vcfile = wa->vcobj (args->file || "", id,
			     (int) (args->undelete_mode || id->misc->undelete_mode),
			     args->revision, args->date);
  if (!wa->site->ct_data_for (vcfile)->downloadp) return plainres;

  string url;
  if (args->revision)
    url = id->misc->build_url (id, "rev" + args->revision, "rawview", args->file);
  else if (args->date)
    url = id->misc->build_url (id, "date" + args->date, "rawview", args->file);
  else if (args->update)
    url = id->misc->build_url (id, args->update, "rawview", args->file);

  return "<a href=\"" + http_encode_string (url) + "\" target=" +
    html_encode_tag_value ("RoxenSBView:"+id->misc->wa_tag) + ">" + text + "</a>";
}

string container_viewsource (string name, mapping args, string text, object id)
{
  string plainres = args->hide || text;

  object wa = id->misc->wa;
  if (!wa) return plainres;

  object vcfile = wa->vcobj (args->file || "", id,
			     (int) (args->undelete_mode || id->misc->undelete_mode),
			     args->revision, args->date);
  if (!wa->site->ct_data_for (vcfile)->downloadp) return plainres;

  string url;
  if (args->revision)
    url = id->misc->build_url (id, "rev" + args->revision, "viewsource", args->file);
  else if (args->date)
    url = id->misc->build_url (id, "date" + args->date, "viewsource", args->file);
  else if (args->update)
    url = id->misc->build_url (id, args->update, "viewsource", args->file);

  return "<a href=\"" + http_encode_string (url) + "\" target=" +
    html_encode_tag_value ("RoxenSBView:"+id->misc->wa_tag) + ">" + text + "</a>";
}

string container_view(string name, mapping args, string text, object id)
{
  return("<a href=\"" +
	 http_encode_string(id->misc->build_url(id, (args->revision||"")+
						"/"+(args->date||""),
						"view")) + "\" target=" +
	 html_encode_tag_value ("RoxenSBView:"+id->misc->wa_tag) + ">" + text + "</a>");
}

string container_revert(string name, mapping args, string text, object id)
{
  if (!args->revision) {
    return(text);
  }
  return("<a href=\"" +
	 http_encode_string(id->misc->build_url(id, args->revision, "revert"))+
	 "\">" + text + "</a>");
}
