// Generic conflict resolution wizard. Used as base class for update
// and join.

// Note that state info is stored in both id->variables and
// vcfc->state in this wizard.

// $Id: _merge.pike,v 1.13 1999/05/07 04:39:04 mast Exp $

inherit "_wizardbase";

import Sitebuilder;

#define ERRORMSG(text) \
  (text ? "<p><font color=darkred>" + text + "</font>\n" : "")


// Callbacks for inheriting wizards.

int is_valid_op (object(VCFile), object id);

mapping|object(VCFileConflict) get_vcfc (object(VCFile) vcfile, object id);
void init_state (object(VCFileConflict) vcfc);

mapping|void finish_use_a (object(VCFileConflict) vcfc);
mapping|void finish_use_b (object(VCFileConflict) vcfc);
mapping|void finish_use_new (object(VCFileConflict) vcfc,
			     mapping new_md, string new_data);


// Internals.

private static mapping|string merge_done (object(VCFileConflict) vcfc, object id)
{
  mapping res;

  if (!vcfc->a_md || !vcfc->b_md)
    switch (vcfc->state->merge_use_choice) {
      case "a": res = finish_use_a (vcfc); break;
      case "b": res = finish_use_b (vcfc); break;
      default: return "Undecided choice.";
    }

  else {
    mapping(string:mixed) new_md = 0;
    string new_data = 0;
    if (vcfc->state->merge_resolve_md) {
      new_md = MetaData.merge_done (vcfc, id);
      if (!new_md) return "Error merging metadata."; // Shouldn't happen.
    }
    if (vcfc->state->merge_resolve_data) {
      new_data = vcfc->state->merge_handler->merge_done (vcfc, id);
      if (!new_data) return "Error merging contents."; // Shouldn't happen.
    }
    res = finish_use_new (vcfc, new_md, new_data);
  }

  return res && error_msg (res);
}


// Wizard overrides.

string make_title (object id)
{
  array(string) parts = ::make_title (id) / "";
  if (sizeof (parts) > 1 && id->variables->merge_file)
    parts[-1] = "</b> &nbsp; /" + html_encode_string (id->variables->merge_file);
  return parts * "";
}


// Handle a single file.

string dispatch_merge_file (object id, string page, object vcobj)
{
  if (!vcobj) return "cancel";
  object(VCFile) vcfile;
  if (vcobj->isfile) vcfile = vcobj;
  else {
    if (!id->variables->merge_file) return "cancel"; // Bogus variables if this happens.
    vcfile = vcobj->wa->vcobj (id->variables->merge_file, id);
    if (!vcfile || !vcfile->isfile) return "cancel";
  }

  mapping|object(VCFileConflict) vcfc = get_vcfc (vcfile, id);
  if (!vcfc || mappingp (vcfc)) {
    if (mappingp (vcfc))
      id->misc->wizard_merge_fatal = html_error (error_msg (vcfc), id);
    if (!vcfc)
      id->misc->wizard_merge_fatal = "There are no changes in this file.";
    return "done";		// Just report the error in the done page.
  }
  id->misc->wizard_merge_vcfc = vcfc;

  multiset(string) init_files = id->variables->merge_init_files;
  if (!init_files) init_files = id->variables->merge_init_files = (<>);
  string abspath = vcfile->real_abspath();
  if (!vcfc->state || !init_files[abspath]) {
    init_state (vcfc);
    if (init_files[abspath]) return "file_interrupted";
    else init_files[abspath] = 1;
  }

  if (!vcfc->a_md || !vcfc->b_md) { // Either file is removed.
#ifdef SBDEBUG
    if (!vcfc->a_md && !vcfc->b_md)
      throw (({"Neither a nor b file exists\n", backtrace()}));
#endif
    if (!id->variables->merge_use_choice)
      id->variables->merge_use_choice = vcfc->state->merge_use_choice;
    if (page == "conflict") return "removed_conflict";
  }

  else {
    int ct_ok = MetaData.merge_content_type_ok (vcfc);
    vcfc->state->merge_ct_ok = ct_ok;
    int analyze_md = MetaData.merge_analyze (vcfc, id);
    vcfc->state->merge_analyze_md = analyze_md;
    if (!ct_ok) {
      vcfc->state->merge_analyze = 4;
      return "type_conflict";
    }
    else {
      object handler = vcfc->state->merge_handler;
      if (!handler)
	handler = vcfc->state->merge_handler = id->misc->sb->handler_for (vcfile);
      int analyze_data = handler->merge_analyze (vcfc, id);
      vcfc->state->merge_analyze_data = analyze_data;
      vcfc->state->merge_analyze = max (analyze_md, analyze_data);
      if (vcfc->state->merge_analyze >= 3 && page != "type_conflict") return "conflict";
    }
  }

  return page;
}

string page_type_conflict (object id, object(VCFile) vcfile)
{
  object(VCFileConflict) vcfc = id->misc->wizard_merge_vcfc;

  string res = "<var type=hidden name=merge_revisit value=1>\n" +
    ERRORMSG (id->misc->wizard_error);

  if (!id->variables->merge_guide_msg)
    res += "<var type=hidden name=merge_guide_msg value=1>\n"
      "<p>The content types differs. You have to decide on that "
      "first, since it controls how the conflict resolution is done.\n";

  vcfc->state->merge_type_conflict = 1;
  vcfc->state->merge_resolve_md = 1;
  MetaData.merge_forget_content_type (vcfc);
  res += ERRORMSG (id->misc->wizard_error_md) +
    "\n<h3>Metadata: Different content types</h3>\n"
    "\n<p>" + MetaData.merge_form (vcfc, id);

  return res;
}

int verify_type_conflict (object id, object(VCFile) vcfile)
{
  object(VCFileConflict) vcfc = id->misc->wizard_merge_vcfc;
  int ct_ok = MetaData.merge_content_type_ok (vcfc);
  vcfc->state->merge_ct_ok = ct_ok; // Need to update the state here.
  if (ct_ok) return 0;
  id->misc->wizard_error_md = MetaData.merge_verify (vcfc, id);
  return !!id->misc->wizard_error_md;
}

string page_conflict (object id, object(VCFile) vcfile)
{
  object(VCFileConflict) vcfc = id->misc->wizard_merge_vcfc;

  string res = "<var type=hidden name=merge_revisit value=1>\n" +
    ERRORMSG (id->misc->wizard_error);

  // FIXME: Show log messages between work area revision and repository
  // revision in update wizard.

  // Metadata.

  res += "\n<h3>Metadata: " + ({
    "No differences",
    "Conflict-free differences",
    "Resolved conflicts",
    "Unresolved conflicts",
  })[vcfc->state->merge_analyze_md] + "</h3>\n";

  if (id->variables->merge_guide_msg == "1") {
    id->variables->merge_guide_msg = "2";
    res += "\n<p>The content type has been chosen. Please continue "
      "with the rest of the conflict resolution.\n";
  }

  res += ERRORMSG (id->misc->wizard_error_md);
  if (vcfc->state->merge_analyze_md) {
    vcfc->state->merge_resolve_md = 1;
    res += "\n<p>" + MetaData.merge_form (vcfc, id);
  }

  // Contents.

  res += "\n<h3>Contents: " + ({
    "No differences",
    "Conflict-free differences",
    "Resolved conflicts",
    "Unresolved conflicts"
  })[vcfc->state->merge_analyze_data] + "</h3>\n";

  res += ERRORMSG (id->misc->wizard_error_data);
  if (vcfc->state->merge_analyze_data) {
    vcfc->state->merge_resolve_data = 1;
    res += "\n<p>" + vcfc->state->merge_handler->merge_form (vcfc, id);
  }

  return res;
}

int verify_conflict (object id, object(VCFile) vcfile)
{
  object(VCFileConflict) vcfc = id->misc->wizard_merge_vcfc;

  if (vcfc->state->merge_resolve_md)
    id->misc->wizard_error_md = MetaData.merge_verify (vcfc, id);
  if (vcfc->state->merge_resolve_data)
    id->misc->wizard_error_data = vcfc->state->merge_handler->merge_verify (vcfc, id);
  return !!(id->misc->wizard_error_md || id->misc->wizard_error_data);
}

string page_removed_conflict (object id, object(VCFile) vcfile)
{
  object(VCFileConflict) vcfc = id->misc->wizard_merge_vcfc;

  string res = "<var type=hidden name=merge_revisit value=1>\n" +
    ERRORMSG (id->misc->wizard_error);

  // FIXME: Show log messages between work area revision and repository
  // revision in update wizard.

  // FIXME: Report moved away files better.
  res += "\n<p>" + vcfc->state->msg_removed + " You can either\n"
    "<dl>\n"
    "<dt><var type=radio name=merge_use_choice value=a> " +
    vcfc->state->msg_use_a + ", or<dd>\n"
    "<dt><var type=radio name=merge_use_choice value=b> " +
    vcfc->state->msg_use_b + ".<dd>\n"
    "</dl>\n";

  return res;
}

int verify_removed_conflict (object id, object(VCFile) vcfile)
{
  object(VCFileConflict) vcfc = id->misc->wizard_merge_vcfc;

  if ((<"a", "b">)[id->variables->merge_use_choice]) {
    vcfc->state->merge_use_choice = id->variables->merge_use_choice;
    return 0;
  }
  id->misc->wizard_error = vcfc->state->msg_no_choice_err;
  return 1;
}

string page_file_interrupted (object id, object(VCFile) vcfile)
{
  return html_error (
    "The operation has been interrupted. Normally this occurs when "
    "someone else commits a new revision, or when you have done "
    "something with this file from another session. You have to start "
    "over.", id);
}


// Handle several files in a directory tree.

string dispatch_merge_dir (object id, string page, object vcobj);

string dispatch_merge_dir_base (object id, array(object(VCFile)) vcfiles,
				void|array(mapping|object(VCFileConflict)) vcfcs)
// Run from dispatch_merge_dir in the inheriting class.
{
  {
    array(string) merge_files = glob ("merge_file_*.?", indices (id->variables));
    foreach (merge_files, string var) m_delete (id->variables, var);
    if (sizeof (merge_files) == 2) {
      int i;
      if (sscanf (merge_files[0], "merge_file_%d", i) != 1)
	return "cancel";	// Bogus variables if this happens.
      array(string) filenames = (id->variables->merge_files || "") / "\0";
      if (sizeof (filenames) <= i) return "cancel";
      foreach (glob ("meta_*", indices (id->variables)), string var)
	m_delete (id->variables, var);
      foreach (glob ("data_*", indices (id->variables)), string var)
	m_delete (id->variables, var);
      m_delete (id->variables, "merge_use_choice");
      id->variables->merge_file = filenames[i];
      return "conflict";
    }
  }
  m_delete (id->variables, "merge_file");

  multiset(string) init_files = id->variables->merge_init_files;
  if (!init_files) init_files = id->variables->merge_init_files = (<>);

  if (!vcfcs) vcfcs = Array.map (vcfiles, get_vcfc, id);
  array(string) statuses = allocate (sizeof (vcfcs));
  array(string) filenames = vcfiles->real_abspath();
  array(string) lostfiles = ({});

  for (int i = 0; i < sizeof (vcfcs); i++) {
    mapping|object(VCFileConflict) vcfc = vcfcs[i];
    if (objectp (vcfc)) {
      if (!vcfc->state || !init_files[filenames[i]]) {
	init_state (vcfc);
	if (init_files[filenames[i]]) lostfiles += ({"/" + filenames[i]});
	else init_files[filenames[i]] = 1;
      }

      if (!vcfc->a_md || !vcfc->b_md) {
	vcfc->state->merge_analyze = 2;
	switch (vcfc->state->merge_use_choice) {
	  case "a": statuses[i] = vcfc->state->status_use_a; break;
	  case "b": statuses[i] = vcfc->state->status_use_b; break;
	  default:
	    statuses[i] = vcfc->state->status_rem_undecided;
	    vcfc->state->merge_analyze = 3;
	}
      }

      else {
	int ct_ok = vcfc->state->merge_ct_ok ||
	  MetaData.merge_content_type_ok (vcfc);
	vcfc->state->merge_ct_ok = ct_ok;
	if (!ct_ok) {
	  vcfc->state->merge_analyze = 4;
	  statuses[i] = "<font color=darkred>Different content types</font>";
	}
	else {
	  object handler = vcfc->state->merge_handler;
	  if (!handler) handler = vcfc->state->merge_handler =
			  id->misc->sb->handler_for (vcfiles[i]);
	  int analyze_md = vcfc->state->merge_analyze_md ||
	    MetaData.merge_analyze (vcfc); // No id here.
	  vcfc->state->merge_analyze_md = analyze_md;
	  int analyze_data = vcfc->state->merge_analyze_data ||
	    handler->merge_analyze (vcfc); // No id here.
	  vcfc->state->merge_analyze_data = analyze_data;
	  vcfc->state->merge_analyze = max (analyze_md, analyze_data);
	  statuses[i] = ({
	    "No differences",
	    "Conflict-free differences",
	    "Resolved conflicts",
	    "<font color=darkred>Unresolved conflicts</font>",
	  })[vcfc->state->merge_analyze];
	}
      }
    }
  }

  if (sizeof (lostfiles)) {
    id->variables->merge_lostfiles = lostfiles * "\0";
    return "dir_interrupted";
  }

  id->misc->wizard_merge_files = ({vcfiles, vcfcs, statuses, filenames});
  return 0;
}

string|mapping page_file_list (object id, object(VCDir) vcdir)
{
  array(object(VCFile)) vcfiles = id->misc->wizard_merge_files[0];
  array(string|mapping|object(VCFileConflict)) vcfcs = id->misc->wizard_merge_files[1];
  array(string|int) statuses = id->misc->wizard_merge_files[2];
  array(string) filenames = id->misc->wizard_merge_files[3];
  sort (filenames, vcfiles, vcfcs, statuses);

  string res = "<var type=hidden name=merge_file_list value=1>\n"
    "<h3>File status summary</h3>\n" +
    ERRORMSG (id->misc->wizard_error) +
    "<p>"
    "<table width=100% cellspacing=0 cellpadding=0 border=0 bgcolor=#000000>\n<tr><td>"
    "<table width=100% cellspacing=0 cellpadding=3 border=0 bgcolor=#eeeeee>\n" +
    "<tr><th>&nbsp</th><th width=99% align=left>File</th>"
    "<th align=left>Status</th></tr>\n"
    "<tr></tr>\n";

  for (int i = 0; i < sizeof (vcfcs); i++)
    if (vcfcs[i]) {
      res += "<tr valign=bottom><td>" +
	(objectp (vcfcs[i]) && vcfcs[i]->state->merge_analyze ?
	 "<input type=image name=merge_file_" + i +
	 " width=20 height=20 border=0 src='" + id->misc->img_url + "unfold.gif'>" :
	 "&nbsp;") + "</td>\n"
	"<td>" + html_encode_string ("/" + filenames[i]);
      if (mappingp (vcfcs[i]) || stringp (vcfcs[i])) {
	res += "\n<br><font color=darkred>Error: " +
	  html_encode_string (mappingp (vcfcs[i]) ? error_msg (vcfcs[i]) : vcfcs[i]) +
	  "</font>";
	filenames[i] = "0";
      }
      res += "</td>\n"
	"<td nowrap>" + (stringp (statuses[i]) ? statuses[i] : "&nbsp;") + "</td></tr>\n";
    }
    else filenames[i] = "0";

  id->variables->merge_files = filenames * "\0";

  return res + "</table></td></tr></table>\n";
}

int verify_file_list (object id, object(VCDir) vcdir)
{
  array(mapping|object(VCFileConflict)) vcfcs = id->misc->wizard_merge_files[1];

  int go_back = 0;
  foreach (vcfcs, mapping|object(VCFileConflict) vcfc)
    if (vcfc)
      if (mappingp (vcfc) || !vcfc->state) go_back = 1;
      else {
	int|string res = -1;	// -1 used as flag.
	if (!vcfc->a_md || !vcfc->b_md) {
	  if ((<"a", "b">)[vcfc->state->merge_use_choice])
	    res = merge_done (vcfc, id);
	}
	else
	  if (vcfc->state->merge_analyze < 3)
	    res = merge_done (vcfc, id);
	if (res == -1) {
	  id->misc->wizard_error = "There are still unresolved conflicts.";
	  go_back = 1;
	}
	else if (stringp (res)) go_back = 1;
      }
  id->misc->wizard_merge_dir_ok = !go_back;
  return go_back;
}

string page_dir_interrupted (object id, object vcobj)
{
  array(string) lostfiles = sort ((id->variables->merge_lostfiles || "") / "\0");
  string files = sizeof (lostfiles) == 1 ? "file" : "files";
  return html_error (
    "The operation was interrupted for the following " + files + ":\n"
    "<p><dl>\n"
    "<dt>" + lostfiles * "<dd>\n<dt>" + "<dd>\n"
    "</dl>\n" +
    "<p>Normally this occurs for a file when someone else commits a new "
    "revision of it, or when you have done something with it from "
    "another session. You have to start over with the " + files + ".", id);
}


// The finish state for both files and dirs.

string dispatch_wizard_done (object id, string page, object vcobj)
{
  if (!vcobj) return "cancel";
  if (vcobj->isfile || id->variables->merge_file) {
    if (zero_type (id->misc->wizard_merge_vcfc)) {
      string redirect = dispatch_merge_file (id, page, vcobj);
      if (redirect != page) return redirect;
    }

    mapping|object(VCFileConflict) vcfc = id->misc->wizard_merge_vcfc;
    if (!vcfc) return 0;
    if (objectp (vcfc) && !vcfc->state) return "file_interrupted";

    string res = merge_done (vcfc, id);
    if (res) {
      id->misc->wizard_merge_fatal = html_error (res, id);
      return "done";
    }
  }

  else {
    if (!id->misc->wizard_merge_files) {
      string redirect = dispatch_merge_dir (id, page, vcobj);
      if (redirect != page) return redirect;
    }

    if (zero_type (id->misc->wizard_merge_dir_ok)) verify_file_list (id, vcobj);
    if (!id->misc->wizard_merge_dir_ok) return "file_list";

    return 0;
  }
}

string wizard_done (object id, object vcobj)
{
  return id->misc->wizard_merge_fatal;
}


// The wizard page automaton (not complete).

string conflict_prev (object id, string page, object vcobj)
{
  return id->variables->merge_file_list && "file_list";
}

string conflict_next (object id, string page, object vcobj)
{
  return id->variables->merge_file_list && "file_list";
}

mapping(string:array) wizard_automaton = ([
  "type_conflict": ({		// Page to resolve a content type difference.
    lambda (object id, string page, object vcobj)
      {return dispatch_merge_file (id, page, vcobj);},
    conflict_prev,
    "conflict",
  }),

  "conflict": ({		// Normal metadata/contents conflict resolution page.
    lambda (object id, string page, object vcobj)
      {return dispatch_merge_file (id, page, vcobj);},
    lambda (object id, string page, object vcobj)
    {
      mapping|object(VCFileConflict) vcfc = id->misc->wizard_merge_vcfc;
      if (vcfc && vcfc->state && vcfc->state->merge_type_conflict)
	return "type_conflict";
      return conflict_prev (id, page, vcobj);
    },
    conflict_next,
  }),

  "removed_conflict": ({	// Page if either file is removed.
    lambda (object id, string page, object vcobj)
      {return dispatch_merge_file (id, page, vcobj);},
    conflict_prev,
    conflict_next,
  }),

  "file_interrupted": ({	// State lost page for files.
    0,
    "conflict",
    -1,
  }),

  "file_list": ({		// Page to summarize the merge state of several files.
    dispatch_merge_dir,
    0,
    0,
  }),

  "dir_interrupted": ({		// State lost page for dirs.
    0,
    "file_list",
    -1,
  }),

  "done": ({dispatch_wizard_done}), // Wizard done page.
]);
