/*
 * $Id: annotate.pike,v 1.18 1999/05/17 01:07:48 mast Exp $
 *
 * SiteBuilder wizard to display an annotated listing of a file.
 *
 * Henrik Grubbstrm 1998-12-15
 */

inherit "_wizardbase";

constant name = "View//Annotate";
constant doc = "Show an annotated listing of a file";
constant operation = "log";
constant view_mode = 1;

#define error id->misc->wizard_error
#define ERROR (error?"<font color=darkred>" + error + "</font><p>\n":"")

int button_enabled(object id, void|object vcobj)
{
  int up = ::button_enabled (id, vcobj);
  if (up <= 0) return up;
  return vcobj->annotate && 1;
}

string page_0(object id, object vcobj)
{
  if (button_enabled (id, vcobj) < 1) {
    return(0);
  }

  string path = Sitebuilder.nice_abspath (vcobj);
  mapping|array(array(object(VC.CVS.Annotation))) ann =
    vcobj->annotate(id, id->misc->undelete_mode == "1");

  if (mappingp(ann)) {
    return html_error(Sitebuilder.error_msg(ann), id);
  }

  string ct = id->misc->sb->content_type_for(vcobj, id);
  mapping ct_data = id->misc->sb->ct_data_for(ct, id);

  string res;

  if ((ann[0]["http-content-type"]) && ct_data) {
    // Clean up the output.
    ann[0]["http-content-type"]->data = ct_data->name;
  }

  res = "<b>Metadata</b><br>\n";

  // Used to present the metadata.
  array(mapping(string:mixed)) md_var_info =
    Sitebuilder.MetaData.get_vars(id->misc->sb, ct);

  // First show the known meta-data items.
  array(array(string)) md_tab =
    Array.map(md_var_info,
	      lambda(mapping(string:mixed) var_info,
		     mapping(string:object(VC.CVS.Annotation)) md) {
		object(VC.CVS.Annotation) e;
		if ((e = md[var_info->name])) {
		  return(({ e->date,
			    html_encode_string(e->user || ""),
			    html_encode_string(var_info->title || ""),
			    html_encode_string(e->data || "") }));
		}
		return(0);
	      }, ann[0]) - ({ 0 });

  // And then any additional items.
  md_tab += Array.map(sort(indices(ann[0]) - md_var_info->name),
		      lambda(string item,
			     mapping(string:object) md) {
			object(VC.CVS.Annotation) e = md[item];
			return(({ e->date,
				  html_encode_string(e->user || ""),
				  html_encode_string(item || ""),
				  html_encode_string(e->data || "") }));
		      }, ann[0]);

  res += html_table(({ "Date", "User", "Item", "Value" }),
		    md_tab);

  if (sizeof(ann) > 1) {
    // Annotation exists for the contents of this file too.

    if (ct_data && ct_data->downloadp) {
      // Users are allowed to download the file.

      object handler = id->misc->sb->handler_for(ct, id);

      if (!handler->binaryp()) {
	// Contents aren't binary.

	res += "<br>\n"
	  "<b>Contents</b><br>\n";

	res += html_table(({ "Date", "User", "Lines" }),
			  Array.map(ann[1],
				    lambda(object(VC.CVS.Annotation) e) {
				      return(({ e->date,
						html_encode_string(e->user || ""),
						"<pre>" +
						html_encode_string(e->data || "") +
						"</pre>",
				      }));
				    }));
      }
    }
  }
  
  return(res);
}
