inherit "_wizardbase";

constant name= "Version Control//Commit";
constant doc = ("Commit this file or directory");
constant operation = "commit";

#define error id->misc->wizard_error
#define ERROR (error?"<font color=darkred>"+error+"</font><p>":"")


string page_0(object id, object vcobj)
{
  if (!vcobj) return 0;

  if (vcobj->isfile) {
    if (!vcobj->is_user_op (id, "commit"))
      return "This file is not in a committable state.\n";
    return 0;
  }
  /* Recursive operation is currently too expensive.
  else {
    mapping|array list = vcobj->list_recursive (id);
    if (mappingp (list) ||
	!sizeof ((list->isfilep() - ({0}))->is_valid_op (id, "commit") - ({0})))
      return "There are no committable files here.";
  }
  */

  return ERROR+
    "<sb-internal-filelisting flik=submit name=files is_valid_op=commit>";
}

int verify_0(object id, object vcobj)
{
  if (!vcobj) return 0;

  if(vcobj->isdir) {
    if (!(id->variables->files && sizeof(id->variables->files/"\0"-({"0"})))) {
      error = "You must select at least one file to commit.  If you do not "
	"want to commit any files, use <b>Cancel</b> to leave the wizard.\n";
      return 1;
    }
  }
  else if (!vcobj->is_user_op (id, "commit")) {
    error = "This file is not in a committable state.\n";
    return 1;
  }
}

string page_1(object id, object vcobj)
{
  if (!vcobj) return 0;

  array(string) files;

  if(vcobj->isfile)
    // Use the superuser flag since sanity checks are not relevant here.
    files = ({ "/"+vcobj->abspath(id, 0, 1) });
  else
    files = id->variables->files/"\0"-({"0"});

  return ERROR+
    (sizeof(files)==1? "This file is":"These files are")+
    " being committed:<p><sb-internal-filelisting files='"+
    replace(files*"\0", "'", "'\"'\"'")+"'></p>"
    "<p>Enter a log message:<br>\n<var name=msg></p>\n"
    "Then click the <b>Ok</b> button to actually commit the file"+
    (sizeof(files)==1? "":"s")+".\n";
}

int verify_1(object id, object vcobj)
{
  if (!vcobj) return 0;

  if (!id->variables->msg || id->variables->msg=="") {
    error = "You must enter a log message.";
    return 1;
  }
}

string|mapping wizard_done(object id, object vcobj)
{
  if (!vcobj) return 0;

  string msg = id->variables->msg;
  array(object(Sitebuilder.VCFile)) files;

  if(vcobj->isfile)
    files = ({ vcobj });
  else
    files = Array.map(id->variables->files/"\0"-({"0"}),
		      id->misc->wa->vcfile, id);

  if(!files || !sizeof(files) || !msg || !sizeof(msg))
    return parse_wizard_page(html_error("Internal error, wizard constraints "
					"violated.", id), id, "done");

  foreach(files, object(Sitebuilder.VCFile) file)
  {
    if(!file)
      return parse_wizard_page(html_error("Internal error, file went away.",
					  id), id, "done");
    mapping fault;
    if(fault = file->commit(id, msg))
      return parse_wizard_page(html_error(Sitebuilder.error_msg (fault), id), id, "done");
  }

  if (!vcobj->exists (id)) {
    id->misc->path=combine_path(id->misc->path,"../");
    return http_redirect(id->misc->build_url(id),id);
  }

  return 0;
}
