inherit "_wizardbase";

import Sitebuilder;

constant name= "File//Copy";
constant doc = ("Copy this file or directory");
constant operation = "copy";

#define error id->misc->wizard_error
#define ERROR (error?"<font color=darkred>"+error+"</font><p>":"")

#define quote(X) (replace((X)||"", "'", "'\"'\"'"))


string page_0(object id, object vcobj)
{
  if (!vcobj) return 0;

  mapping|object parent = vcobj->ascend(id);

  if (mappingp(parent)) return html_error (error_msg(parent), id);

  if (!parent)
    return html_error("Can't copy the root directory!", id);

  return ERROR+
    "Select a destination directory where to put the copy of <b>/"+
    html_encode_string(vcobj->abspath(id))+"</b>:\n"
    "<p><sb-internal-dirselector default='"+quote(parent->abspath(id))+
    "' name=destdir></p>\n"
    "Select a "+(vcobj->isdir? "dir":"file")+"name for the copy: "
    "<var name=destname type=string size=40 default='"+
    quote(vcobj->name(id))+"'>";
}

int verify_0(object id, object vcobj)
{
  if (!vcobj) return 0;

  if (!id->variables->destdir) {
    error = "You must select a destination directory.";
    return 1;
  }

  if (!id->variables->destname || id->variables->destname=="") {
    error = "You must select a "+(vcobj->isdir? "dir":"file")+"name.";
    return 1;
  }

  if(!valid_name(id->variables->destname))
  {
    error = "Invalid characters in "+(vcobj->isdir? "dir":"file")+"name.";
    return 1;
  }

  mapping|object ddir = vcobj->wa->vcdir(id->variables->destdir, id);

  if(!objectp(ddir)) {
    error = "Can't access destination directory <b>/"+id->variables->destdir+
      "</b>"+(mappingp(ddir)? ": "+error_msg(ddir):"");
    return 1;
  }

  if(!ddir->exists(id)) {
    error = "Destination directory <b>/"+id->variables->destdir+
      "</b> does not exist.";
    return 1;
  }

  if(vcobj->isdir) {
    // Check that ddir is not (a subdir of) vcobj
    object asc = ddir;
    do {
      if(asc->abspath(id) == vcobj->abspath(id)) {
	error = "The destination directory must not be a subdirectory of "
	  "<b>/"+html_encode_string(vcobj->abspath(id))+"</b>.";
	return 1;
      }
    } while(objectp(asc = asc->ascend(id)));
  }

  mapping|object dobj = ddir->descend(id, id->variables->destname);

  if(mappingp(dobj)) {
    error = "Can't access destination "+(vcobj->isdir? "directory":"file")+
      " <b>/"+id->variables->destdir+(id->variables->destdir==""?"":"/")+
      id->variables->destname+"</b>: "+error_msg(dobj);
    return 1;
  }

  if(dobj && dobj->abspath(id) == vcobj->abspath(id)) {
    error = "Can't copy <b>/"+html_encode_string(vcobj->abspath(id))+
      "</b> over itself.";
    return 1;
  }

  if(dobj && dobj->exists(id)) {
    if(vcobj->isdir) {
      error = "Can't copy the directory to <b>/"+
	html_encode_string(dobj->abspath(id))+"</b> as "
	"there already exists a "+(dobj->isfile? "file":"directory")+
	" by that name.";
      return 1;
    } else if(dobj->isdir) {
      error = "Can't copy the file to <b>/"+
	html_encode_string(dobj->abspath(id))+"</b> as "
	"there already exists a directory by that name.";
      return 1;
    }
  }

  id->misc->__ddir = ddir;
  id->misc->__dobj = dobj;
}

string page_1(object id, object vcobj)
{
  string res;

  if (!vcobj) return 0;

  verify_0(id, vcobj);

  res = ERROR+
    "The "+(vcobj->isdir? "directory":"file")+" <b>/"+
    html_encode_string(vcobj->abspath(id))+
    "</b> will be copied to the new location <b>/"+
    html_encode_string(id->variables->destdir)+
    (id->variables->destdir==""?"":"/")+
    html_encode_string(id->variables->destname)+"</b>.";

  if(id->misc->__dobj && id->misc->__dobj->exists(id))
    res += html_warning("The destination file already exists.  Its contents "
			"will be overwritten if you proceed.", id);

  if(vcobj->isdir)
    res += "<p>Enter a log message:<br>\n<var name=commitmsg></p>\n";

  return res;
}

int verify_1(object id, object vcobj)
{
  if(!vcobj)
    return 0;

  if(vcobj->isdir &&
     (!stringp(id->variables->commitmsg) || id->variables->commitmsg=="")) {
    error = "You must enter a log message.";
    return 1;
  }
}

string|mapping wizard_done(object id, object vcobj)
{
  if(!vcobj)
    return 0;

  if(verify_0(id, vcobj))
    return parse_wizard_page (html_error (error, id), id, "done");

  object ddir = id->misc->__ddir;
  mapping|object dobj = id->misc->__dobj;

  mapping err;
  int madefile=0;

  if(vcobj->isdir && !stringp(id->variables->commitmsg))
    return parse_wizard_page(html_error("Internal error, wizard constraints "
					"violated.", id), id, "done");

  if(dobj)
    if(dobj->exists(id))
      ;
    else
      err = dobj->add(id);
  else
    if(vcobj->isfile) {
      dobj = ddir->newfile(id, id->variables->destname);
      madefile = 1;
    } else
      dobj = ddir->newdir(id, id->variables->destname,
			  id->variables->commitmsg);

  if(mappingp(dobj))
    err = dobj;

  if(mappingp(err))
    return parse_wizard_page (html_error (error_msg(err), id), id, "done");

  if(vcobj->isdir) {
    // Copy dir
    err = vcobj->copy(id, dobj, id->variables->commitmsg);
  } else {
    // Copy file
    err = vcobj->copy(id, dobj);
  }

  if(mappingp(err)) {

    // If the destination was a new file, remove the empty file
    if(madefile)
      dobj->discard(id);

    return parse_wizard_page (html_error (error_msg(err), id), id, "done");
  }

  string path = dobj->abspath(id);
  if(path) {
    id->misc->path="/"+path;
    return http_redirect(id->misc->build_url(id),id);
  }
}
