inherit "_wizardbase";

constant name= "File//Create directory";
constant doc = ("Create a new directory in the current directory");
constant operation = "newdir";

#define error id->misc->wizard_error
#define ERROR (error?"<font color=darkred>"+error+"</font><p>":"")

string page_0(object id, object(Sitebuilder.VCDir) vcdir)
{
  if(!vcdir)
    return 0;

  return ERROR+"<b>Select directory name:</b><br>"+
    replace("/"+html_encode_string(vcdir->abspath(id))+"/", "//", "/")+
    "<var name=dirname type=string "
    "size=40 default=\"\">";
}

int verify_0(object id, object(Sitebuilder.VCDir) vcdir)
{
  if (!id->variables->dirname || id->variables->dirname=="") {
    error = "You must select a directory name.";
    return 1;
  }

  if (search(id->variables->dirname, "/")>=0) {
    error = "The directory name must not contain slashes.  If you intend to "
      "create a directoty in a subdirectory, you need to do so from a view of "
      "that directory.";
    return 1;
  }

  if(!Sitebuilder.valid_name(id->variables->dirname))
  {
    error = "Invalid characters in directory name.";
    return 1;
  }

  mapping|object f = vcdir->descend(id, id->variables->dirname, 0, 1);

  if (objectp(f))
  {
    error = "There already exists a "+(f->isdir? "directory":"file")+
      " with the name "+html_encode_string(id->variables->dirname)+
      " in this directory.  Please select a different name.";
    return 1;
  }
  else if (mappingp(f))
  {
    error = Sitebuilder.error_msg (f);
    return 1;
  }
  error = "";
}

string page_1(object id, object(Sitebuilder.VCDir) vcdir)
{
  if(!vcdir)
    return 0;

  return "You are about to create a new directory named <b>"+
    html_encode_string(id->variables->dirname)+"</b>.\n"
    "<p>The full pathname of this directory will be <b>"+
    html_encode_string(replace("/"+vcdir->abspath(id)+"/", "//", "/")+
                       id->variables->dirname)+
    "</b>\n<p>Click OK to confirm.\n";
}

string|mapping wizard_done(object id, object(Sitebuilder.VCDir) vcdir)
{
  if(!vcdir)
    return parse_wizard_page(html_error("Internal error, directory went away.",
					id), id, "done");

  mapping fault;

  object(Sitebuilder.VCDir)|mapping newd =
    vcdir->newdir(id, id->variables->dirname, ""); // FIXME: Specify commit message?

  if(mappingp(newd))
    fault = newd;

  if(fault)
    return parse_wizard_page(html_error(Sitebuilder.error_msg (fault), id), id, "done");

  id->misc->path=combine_path(id->misc->path, id->variables->dirname);
  return http_redirect(id->misc->build_url(id),id);
}
