// $Id: create_file.pike,v 1.22 1999/08/23 12:54:55 wing Exp $

inherit "_wizardbase";

constant name= "Edit//Edit new file";
constant doc = ("Create and edit a file in this directory");
constant operation = "newfile";

import Sitebuilder;

#define error id->misc->wizard_error
#define ERROR (error?"<font color=darkred>"+error+"</font><p>":"")

int button_enabled(object id, void|object vcobj)
{
  if(vcobj && !vcobj->isdir) return -1;
  return ::button_enabled (id, vcobj);
}

string page_0(object id, object(VCDir) vcdir)
{
  if(!vcdir)
    return 0;

  return ERROR+"<b>Select filename:</b><br>"+
    replace("/"+html_encode_string(vcdir->abspath(id))+"/", "//", "/")+
    "<var name=filename type=string size=40 default=\"\">";
}

int verify_0(object id, object(VCDir) vcdir)
{
  if (!vcdir) return 0;

  if (!id->variables->filename || id->variables->filename=="") {
    error = "You must select a filename.";
    return 1;
  }

  if (search(id->variables->filename, "/")>=0) {
    error = "The filename must not contain slashes.  If you intend to create "
      "a file in a subdirectory, you need to do so from a view of that "
      "directory.";
    return 1;
  }

  if(!valid_name(id->variables->filename))
  {
    error = "Invalid characters in filename.";
    return 1;
  }

  mapping|object f = vcdir->descend(id, id->variables->filename, 0, 1);

  if (objectp(f))
  {
    // Slight security risk: A user with write access in a dir can
    // this way discover the names of files (s)he isn't allowed to
    // see.
    error = "There already exists a "+(f->isdir? "directory":"file")+
      " with the name "+html_encode_string(id->variables->filename)+
      " in this directory.  Please select a different name.";
    return 1;
  }
  id->variables["meta_http-content-type"]=
    id->misc->sb->find_content_type_from_filename(id->variables->filename);

  error = "";
}

string page_1(object id, object(VCDir) vcdir)
{
  if(!vcdir)
    return 0;

  id->misc->wizard_title_path =
    vcdir->real_abspath() + "/" + id->variables->filename;

  return ERROR+"Edit the metadata for the new file:<br><br>\n"
    "<sb-internal-metadata-editor _metavars=http-content-type "
    "'>\n";
}

string page_2( object id, object (VCDir) vcdir )
{
  array def_files
    = id->misc->sb->get_default_file( id->variables["meta_http-content-type"],
				      id );
  mapping (string:string) m = ([ ]);

  id->misc->wizard_title_path =
    vcdir->real_abspath() + "/" + id->variables->filename;

  foreach (def_files, object vcfile)
    m[ (vcfile->metadata( id )->md->title
	&& vcfile->metadata( id )->md->title != ""
	? vcfile->metadata( id )->md->title
	: vcfile->name( id )) ] = vcfile->abspath( id );

  if (!sizeof( m ))
    return 0;
  if (sizeof( m ) == 1)
  {
    id->variables->default_file = values( m )[0];
    return 0;
  }
  return ERROR + "Choose a template file to create your new file from:<p>"
    + "<var name=default_file type=select choices='"
    + Array.map( sort( indices( m ) ),
		 lambda (string i, mapping m)
		 {
		   return m[ i ] + ":"
		     + replace( html_encode_string( i ), ",", " ");
		 }, m ) * ","
    + "'>";
}

class fake_fd
{
  string read()
  {
    return "";
  }
}

string page_3(object id, object(VCDir) vcdir)
{
  mapping uservars = (id->misc->sb->mac->id_get_data(id)||([]));
  object handler=id->misc->sb->
    handler_for(id->variables["meta_http-content-type"], id);

  id->misc->wizard_title_path =
    vcdir->real_abspath() + "/" + id->variables->filename;

  if (uservars->editor
      && id->misc->sb->editor_for( uservars->editor,
				   id->variables["meta_http-content-type"],
				   id ))
    return "Press OK to create the file and launch an external editor.";
  else if (handler->edit)
  {
    object fake;
    
    if (id->variables->default_file)
    {
      fake = id->misc->wa->vcobj( id->variables->default_file, id );
      if (mappingp( fake ))
	fake = 0;
      if (fake)
	fake = fake->view( id );
      if (mappingp( fake ))
	fake = 0;
      if (fake)
	fake = fake->fd;
    }
    if (!fake)
      fake = fake_fd();
    return
      "<font size=+1><b>"+html_encode_string(id->variables->filename)+
      ":</b></font><br><br>"+handler->edit(fake,id);
  }
  else
    return "Press OK to create this file.";
}

string wizard_done(object id, object(VCDir) vcdir)
{
  mapping fault;
  array(string) metas = glob("meta_*", indices(id->variables));
  object(VCFile)|mapping newf =
    vcdir->newfile(id, id->variables->filename);

  id->misc->wizard_title_path =
    vcdir->real_abspath() + "/" + id->variables->filename;

  if(mappingp(newf))
    fault = newf;
  else
    fault = newf->set_metadata(id,
			       mkmapping(Array.map(metas, `[], 5, 0x7fffffff),
					 rows(id->variables, metas)));
  if(fault)
    return parse_wizard_page(html_error(error_msg (fault), id), id, "done");
  
  object handler=id->misc->sb->handler_for(newf, id);
  mapping uservars = (id->misc->sb->mac->id_get_data(id)||([]));
  mapping|object vcfd=newf->upload(id);
  if(objectp(vcfd))
  {
    if (!(uservars->editor
	  && id->misc->sb->editor_for( uservars->editor,
				       id->variables["meta_http-content-type"],
				       id ))
	&& handler->edit_done)
      handler->edit_done(vcfd->fd,id);
    else if (id->variables->default_file)
    {
      object vcfile = id->misc->wa->vcobj( id->variables->default_file, id );
      object fd = vcfile && vcfile->view( id )->fd;

      vcfd->fd->write( fd->read() );
    }
    vcfd->upload_end(0);
  }
  else
    return parse_wizard_page(html_error(Sitebuilder.error_msg(vcfd),
					id),
			     id, "done");
  if (uservars->editor
      && id->misc->sb->editor_for( uservars->editor,
				   id->variables["meta_http-content-type"],
				   id ))
    return (id->misc->sb->get_editor_for( uservars->editor, newf, id ));
  return 0;
}

string parse_wizard_page(string form, object id, string wiz_name)
{
  // Big kludge. No shit?
  array(string) a = ::parse_wizard_page(form, id, wiz_name) / "method=get";
  return a[0] + "method=post" + a[1..] * "method=get";
}
