// Delete a file or directory.

// $Id: delete.pike,v 1.12 1999/10/11 21:28:29 wing Exp $

inherit "_wizardbase";

import Sitebuilder;

constant name= "File//Delete";
constant doc = "Delete this file or directory";
constant operation = "delete";

#define ERRORMSG(text) \
  (text ? "<p><font color=darkred>" + text + "</font>\n" : "")


// Wizard callbacks.

int button_enabled( object id, object vcobj )
{
  if (vcobj && vcobj->abspath() == "")
    return 0;
  return ::button_enabled( id, vcobj );
}

string dispatch (object id, string page, object vcobj)
{
  if (button_enabled (id, vcobj) < 1) return "cancel";
  switch(page)
  {
   case "start":
     if (vcobj->isdir)
     {
       if (!sizeof( vcobj->list( id ) ))
	 return "confirm";
       if (!vcobj->can_delete( id ))
       {
	 id->misc->wizard_delete_error
	   = html_error( "You do not have permission to delete the "
			 "directory /" + html_encode_string (vcobj->real_abspath())
			 + " and all files and directories within it.",
			 id );
	 return "confirm";
       }
       id->misc->wizard_start_page = 1;
     }
     else
       return "confirm";
     break;
  }
  return 0;
}

mapping(string:array) wizard_automaton = ([
  "start": ({			// Commit message when deleting a dir.
    dispatch,
    0,
    "confirm",
  }),

  "confirm": ({			// Page that tries to delete with warning.
    0,
    lambda (object id, string page, object vcobj)
    { if (id->misc->wizard_start_page) return "start"; else return 0; },
    0,
  }),
]);

string page_start (object id, object vcobj)
{
  return ERRORMSG( id->misc->wizard_error ) +
    "\n<p>" + html_warning( "The directory /" +
			    html_encode_string (vcobj->real_abspath()) +
                            " is not empty. Are you sure you want to delete "
			    " it and all files and diretories within it?",
			    id );
}

string page_confirm (object id, object vcobj)
{
  string result;

  m_delete (id->variables, "delete_warn");
  if (id->misc->wizard_delete_error)
    return 0;
  result = ERRORMSG (id->misc->wizard_error) + "\n<p>";
  if (vcobj->isdir)
  {
    if (sizeof( vcobj->list( id ) ))
      result += html_warning("Warning: You have selected to delete the "
			     "directory /"+html_encode_string (vcobj->real_abspath())+
			     " and all files and directories within it.", id) +
	"\n<p>The following files and directories will be deleted:\n"
	"<p><sb-internal-filelisting recursive><p>Undoing this operation "
	"will be very awkward.  Please confirm that you really want to "
	"perform this operation:\n<p>"
	"<var name=nconfirm type=checkbox value=yes> "
	"I still want do perform a recursive delete of /"
	+html_encode_string (vcobj->real_abspath())+".\n";
    else
      result += "The directory /" + html_encode_string (vcobj->abspath()) +
	" is about to be deleted.";
    result += "<p>Enter a log message:\n<br><var name=delete_msg>\n<br>";
  }
  else
    result += " Press <b>Ok</b> if you want to delete the file /"
      + html_encode_string (vcobj->real_abspath()) + ".";
  return result;
}

int verify_confirm (object id, object vcobj)
{
  if (vcobj->isdir)
  {
    if (sizeof( vcobj->list( id ) ))
      if (search((id->variables->nconfirm||"")/"\0", "yes")<0)
      {
	id->misc->wizard_error
	  = "You must confirm your intention to perform a "
	  "recursive delete to be able to proceed.  If you wish to abort "
	  "deletion of the directory, select Cancel.";
	return 1;
      }
    if (!id->variables->delete_msg || !sizeof (id->variables->delete_msg))
    {
      id->misc->wizard_error = "You must enter a log message.";
      return 1;
    }
  }
  return 0;
}

string|mapping wizard_done (object id, object vcobj)
{
  if (id->misc->wizard_delete_error) return id->misc->wizard_delete_error;

  if (vcobj->is_valid_op (id, "delete")) {
    mapping res = vcobj->isdir ?
      vcobj->delete (id, id->variables->delete_msg) : vcobj->delete (id);
    if (res) return html_error (error_msg (res), id);
  }

  if (!vcobj->exists (id, id->misc->undelete_mode == "1"))
  {
    id->misc->path=combine_path(id->misc->path,"../");
    return http_redirect(id->misc->build_url(id),id);
  }

  return 0;
}
