// Discard a file or several files in a directory tree.

// $Id: discard.pike,v 1.12 1999/10/21 18:44:08 mast Exp $

inherit "_wizardbase";

constant name= "Version Control//Discard your changes";
constant doc = "Discard your changes";
constant operation = "discard";

import Sitebuilder;

#define ERRORMSG(text) \
  (text ? "<p><font color=darkred>" + text + "</font>\n" : "")


// Internals.

static string format_errors (array(mapping) errors, object id)
{
  if (sizeof (errors) == 1)
    return html_error (error_msg (errors[0]), id);
  else if (sizeof (errors) > 1)
    return html_error ("The following files couldn't be discarded due to errors:\n"
		       "<p><dl>\n"
		       "<dt>/" + errors->vcobj->real_abspath() * "<dd>\n<dt>/" + "<dd>\n"
		       "</dl>\n"
		       "<p>They have probably been changed from another session.", id);
  return 0;
}


// Wizard callbacks.

string dispatch (object id, string page, object vcobj)
{
  if (button_enabled (id, vcobj) < 1) return "cancel";

  if (vcobj->isdir) {
    array(string) filenames = (id->variables->discard_files || "0") / "\0" - ({"0"});
    array(object(VCFile)) vcfiles = (Array.map (filenames, vcobj->wa->vcobj, id) -
				     ({0}))->is_valid_op (id, "discard") - ({0});
    if (!sizeof (vcfiles)) {
      id->misc->wizard_discard_no_files = 1;
      return "start";
    }
    id->misc->wizard_discard_vcfiles = vcfiles;
  }
  else {
    id->misc->wizard_discard_vcfiles = ({vcobj});
    return "confirm";
  }

  return 0;
}

string dispatch_confirm (object id, string page, object vcobj)
{
  string redirect = dispatch (id, page, vcobj);
  if (redirect != page) return redirect;
}

mapping(string:array) wizard_automaton = ([
  "start": ({			// Select the files in a dir tree to discard.
    dispatch,
    0,
    "confirm",
  }),

  "confirm": ({			// Page that tries to discard with warning.
    dispatch,
    lambda (object id, string page, object vcobj)
      {return vcobj->isdir && "start";},
    0,
  }),
]);

string page_start (object id, object vcobj)
{
  return ERRORMSG (id->misc->wizard_error) +
    "\n<p>Select files to discard:"
    "\n<p><sb-internal-filelisting flik=submit name=discard_files is_valid_op=discard>";
}

int verify_start (object id, object vcobj)
{
  if (id->misc->wizard_discard_no_files) {
    id->misc->wizard_error = "Please select at least one file.";
    return 1;
  }
  return 0;
}

string page_confirm (object id, object vcobj)
{
  array(mapping) errs = id->misc->wizard_discard_vcfiles->discard (id, 1) - ({0});
  array(mapping) warnings = ({}), errors = ({});
  foreach (errs, mapping err)
    if (err->type == "warning") warnings += ({err}); else errors += ({err});

  string res = format_errors (errors, id);

  if (sizeof (warnings))
    return (res ? res + "\n<p>\n" : "") +
      (vcobj->isdir ?
       (sizeof (warnings) > 1 ?
	"The following files are locally modified:\n"
	"<p><dl>\n"
	"<dt>/" + warnings->vcobj->real_abspath() * "<dd>\n<dt>/" + "<dd>\n"
	"</dl>\n<p>" :
	"The file /" + warnings[0]->vcobj->real_abspath() + " is locally modified.") :
       "The file is locally modified.") +
      " Press <b>Ok</b> if you want to continue.";

  id->misc->wizard_discard_error = res;
  m_delete (id->misc, "wizard_discard_vcfiles");
  return 0;
}

string wizard_done (object id, object vcobj)
{
  if (id->misc->wizard_discard_error) return id->misc->wizard_discard_error;
  return id->misc->wizard_discard_vcfiles &&
    format_errors (id->misc->wizard_discard_vcfiles->discard (id) - ({0}), id);
}
