// Join one file or all files in a directory tree between two
// work areas.

// $Id: join.pike,v 1.12 1999/05/16 19:33:18 mast Exp $

inherit "_merge";

import Sitebuilder;

constant name = "Version Control//Join";
constant doc = "Join changes in two work areas";
constant operation = "join";

#define ERRORMSG(text) \
  (text ? "<p><font color=darkred>" + text + "</font>\n" : "")

int button_enabled (object id, void|object vcobj)
{
  int up = ::button_enabled (id, vcobj);
  if (up <= 0) return up;

  array(object(Workarea)) was = id->misc->sb->wa_list (id);
  foreach (was, object(Workarea) wa)
    if (wa != id->misc->wa && wa->access_perm (id) >= PERM_WRITE)
      return 1;
  return 0;
}


// Internals.

private static nomask object(Workarea) get_wa_other (object id)
{
  object(Workarea) wa_other =
    id->variables->select_workarea &&
    id->misc->sb->wa_lookup (id->variables->select_workarea);
  if (!wa_other || wa_other->access_try (id) < PERM_WRITE) return 0;
  id->misc->wizard_join_wa_other = wa_other;
  return wa_other;
}


// _merge callbacks.

int is_valid_op (object vcobj, object id)
{
  return vcobj->is_valid_op (id, "join");
}

mapping|object(VCFileConflict) get_vcfc (object(VCFile) vcfile, object id)
{
  object(Workarea) wa_other = id->misc->wizard_join_wa_other || get_wa_other (id);
  if (!wa_other || wa_other == vcfile->wa) return 0; // Bogus variables.
  return vcfile->join (id, wa_other);
}

static constant init_state_template = ([
  "msg_removed": ({"The file doesn't exist in work area $wathis.",
		   "The file doesn't exist in work area $waother."}),
  "msg_use_a": ({"remove the file in $waother", "add the file to $waother"}),
  "msg_use_b": ({"add the file to $wathis", "remove the file in $wathis"}),
  "msg_no_choice_err":
    "You must choose whether to add the file to both work areas "
    "or remove it from them.",
  "status_use_a": ({"Remove from $waother", "Add to $waother"}),
  "status_use_b": ({"Add to $wathis", "Remove from $wathis"}),
  "status_rem_undecided": ({
    "<font color=darkred>Doesn't exist in $wathis</font>",
    "<font color=darkred>Doesn't exist in $waother</font>"
  }),

  "msg_a_column": "&nbsp;&nbsp;<gtext fg=black bg=" + MetaData.color_a +
    " nfont=lucida 4 rotate=90>$wathis</gtext>&nbsp;&nbsp;",
  "msg_b_column": "&nbsp;&nbsp;<gtext fg=black bg=" + MetaData.color_b +
    " nfont=lucida 4 rotate=90>$waother</gtext>&nbsp;&nbsp;",

  "link_rawview_a": "the <sb-internal-rawview file='$abspath' revision=$thisrev>"
    "version in $wathis</sb-internal-rawview>",
  "link_rawview_b": "the <sb-internal-rawview file='$abspath' revision=$otherrev>"
    "version in $waother</sb-internal-rawview>",
  "link_rawview_old": "the <sb-internal-rawview file='$abspath' revision=$oldrev>"
    "common ancestor</sb-internal-rawview>",
]);

void init_state (object(VCFileConflict) vcfc)
{
  mapping(string:mixed) state = init_state_template + ([]);

  array(string) from = ({
    "$abspath",
    "$wathis",
    "$waother",
    "$thisrev",
    "$otherrev",
    "$oldrev",
  });
  array(string) to = ({
    vcfc->this_vcfile->real_abspath(),
    vcfc->this_vcfile->wa->name,
    vcfc->other_vcfile->wa->name,
    vcfc->this_rev || "0",
    vcfc->other_rev || "0",
    vcfc->old_rev || "0",
  });

  foreach (indices (state), string var) {
    if (arrayp (state[var]))
      if (!vcfc->a_md) state[var] = state[var][0];
      else if (!vcfc->b_md) state[var] = state[var][1];
      else m_delete (state, var);
    if (stringp (state[var])) state[var] = replace (state[var], from, to);
  }

  vcfc->state = state;
}

mapping|void finish_use_a (object(VCFileConflict) vcfc)
{
  // join_commit() uses vcfc->state->md and vcfc->state->data.
  vcfc->state->md = vcfc->a_md, vcfc->state->data = vcfc->a_data;
}

mapping|void finish_use_b (object(VCFileConflict) vcfc)
{
  vcfc->state->md = vcfc->b_md, vcfc->state->data = vcfc->b_data;
}

mapping|void finish_use_new (object(VCFileConflict) vcfc,
			     mapping new_md, string new_data)
{
  vcfc->state->md = new_md || vcfc->a_md;
  vcfc->state->data = new_data || vcfc->a_data;
}


// Choose the other work area.

string page_start (object id, object vcobj)
{
  m_delete (id->variables, "merge_file");
  m_delete (id->variables, "merge_file_list");
  m_delete (id->variables, "merge_init_files");

  return "Select the work area to join with: "
    "<sb-internal-select-workarea exclude='" + vcobj->wa->tag + "' write-only>\n";
}

int verify_start (object id, object vcobj)
{
  object(Workarea) wa_other = get_wa_other (id);
  if (!wa_other || wa_other == vcobj->wa) return 1;
  return 0;
}


// Manage all the files in the hierarchy for the join.

string dispatch_merge_dir (object id, string page, object vcobj)
{
  if (!vcobj || !vcobj->isdir) return "cancel"; // Bogus variables if this happens.
  object(Workarea) wa_other = get_wa_other (id);
  if (!wa_other || wa_other == vcobj->wa) return "cancel";

  mapping|array(object(VCFileConflict)) vcfcs = vcobj->join (id, wa_other);
  if (mappingp (vcfcs)) {
    id->misc->wizard_merge_fatal = html_error (error_msg (vcfcs), id);
    return "done";		// Just report the error in the done page.
  }
  vcfcs -= ({0});
  if (!sizeof (vcfcs)) {
    id->misc->wizard_merge_fatal = html_error (
      "There are no changes in either workarea.", id);
    return "done";
  }
  id->misc->wizard_join_vcfcs = vcfcs;

  array(object(VCFileConflict)) vcfcs = id->misc->wizard_join_vcfcs;
  array(object(VCFile)) vcfiles = vcfcs->this_vcfile;

  if (sizeof (vcfiles) == 1) {
    // Trig a direct redirect in dispatch_merge_dir_base to the
    // conflict pages if it's only a single file.
    id->variables["merge_file_0.x"] = id->variables["merge_file_0.y"] = 1;
    id->variables->merge_files = vcfiles[0]->real_abspath();
  }

  return dispatch_merge_dir_base (id, vcfiles, vcfcs);
}


// Ask for a commit message and finish.

string dispatch_commit_msg (object id, string page, object vcobj)
// Also used for the commit message page.
{
  if (!vcobj) return "cancel";
  if (id->misc->wizard_merge_fatal) return "done";

  string redirect =
    ::dispatch_wizard_done (id, page, vcobj); // Doesn't do anything final for join.
  if (redirect) return redirect;

  return 0;
}

string page_commit_msg (object id, object vcobj)
{
  return ERRORMSG (id->misc->wizard_error) +
    "\n<p>The " + (vcobj->isfile ? "file" : "directory tree") + " /" +
    html_encode_string (vcobj->real_abspath()) +
    " is being joined between work areas " +
    html_encode_string (vcobj->wa->name) + " and " +
    html_encode_string (id->misc->wizard_join_wa_other->name) + "."
    "\n<p>Enter a log message:<br>\n"
    "<var name=join_msg><br>\n"
    "Then click the <b>Ok</b> button to complete the join "
    "(this may take some time; please be patient).";
}

int verify_commit_msg (object id, object vcobj)
{
  if (!id->variables->join_msg || !sizeof (id->variables->join_msg)) {
    id->misc->wizard_error = "You must enter a log message.";
    return 1;
  }
  return 0;
}

string dispatch_wizard_done (object id, string page, object vcobj)
{
  page = dispatch_commit_msg (id, page, vcobj);
  if (page) return page;

  if (!id->variables->join_msg || !sizeof (id->variables->join_msg)) return "commit_msg";

  mapping err = vcobj->join_commit (id, id->misc->wizard_join_wa_other,
				    id->variables->join_msg);
  if (err) {
    id->misc->wizard_error = error_msg (err);
    return vcobj->isfile ? "conflict" : "file_list";
  }

  return 0;
}


// Complete the wizard automaton.

string conflict_prev (object id, string page, object vcobj)
{
  return id->variables->merge_file_list && "file_list" ||
    "start";			// May always go back to the other work area chooser.
}

string conflict_next (object id, string page, object vcobj)
{
  return id->variables->merge_file_list && "file_list" ||
    "commit_msg";		// Always go on to the commit message page.
}

mapping(string:array) wizard_automaton = ::wizard_automaton + ([
  "start": ({			// Chooser for the other work area.
    0,
    0,
    lambda (object id, string page, object vcobj)
      {return vcobj && vcobj->isfile ? "conflict" : "file_list";},
  }),

  "file_list": ({		// Page to summarize the merge state of several files.
    dispatch_merge_dir,
    "start",
    "commit_msg",
  }),

  "commit_msg": ({		// Asks for the commit message.
    dispatch_commit_msg,
    lambda (object id, string page, object vcobj)
      {return vcobj->isfile ? "conflict" : "file_list";},
    0,
  }),
]);
