/*
 * $Id: log.pike,v 1.26 1999/05/12 01:14:02 js Exp $
 *
 * SiteBuilder wizard to display the commit log of a file.
 *
 * Henrik Grubbstrm 1998-12-01
 */

inherit "_wizardbase";

constant name = "View//Log";
constant doc = "Show the revision log of a file";
constant operation = "log";
constant view_mode = 1;

#define error id->misc->wizard_error
#define ERROR (error?"<font color=darkred>" + error + "</font><p>\n":"")

string wizard_done(object id, object vcobj)
{
  if (button_enabled (id, vcobj) < 1) {
    return(0);
  }

  mapping|array(object(Sitebuilder.VCLogEntry)) log =
    vcobj->log(id, 0, 0, id->misc->undelete_mode == "1");

  if (mappingp(log))
    return parse_wizard_page(html_error(Sitebuilder.error_msg (log), id),
			     id,
			     "done",
			     0,
			     id->misc->build_url(id));
  string path = "'" + replace (Sitebuilder.nice_abspath (vcobj), "'", "'\"'\"'") + "'";

  // FIXME: Assumes that the content-type doesn't change.
  // Use undelete_mode to make sure we get the corrent content-type data.
  string ct = id->misc->sb->content_type_for(vcobj, id, 1);
  mapping ct_data = id->misc->sb->ct_data_for(ct, id);

  string res = "";
  int rawview = vcobj->isfile && vcobj->wa->rawview_access (id);
  int revert = vcobj->isfile &&
    vcobj->is_user_op (id, "revert", id->misc->undelete_mode == "1") && 1;

  if (vcobj->isfile) {
    if (ct_data->parsep) {
      // Parsed page.
      array(array(string)) list = ({});
      string nextdate = "now";
      foreach (log, object(Sitebuilder.VCLogEntry) e) {
	list += ({
	  ({e->action || "",
	    replace(e->date, " ", "&nbsp;"),
	    e->user || "",
	    html_encode_string(e->message||""),
	    (e->action == "delete")?"&nbsp;":
	    "<sb-internal-viewsource hide=&nbsp; file=" + path +
	    " revision=" + e->revision + ">Source</sb-internal-viewsource>"
	    "&nbsp;"
	    "<sb-internal-view hide=&nbsp; file=" + path +
	    " revision=" + e->revision + ">As reverted</sb-internal-view>"
	    "&nbsp;"
	    "<sb-internal-view hide=&nbsp; file=" + path +
	    " revision=" + e->revision +
	    " date='" + nextdate + "'>As was</sb-internal-view>"
	  }) +
	  (revert ? ({
	    (e->action == "delete")?"&nbsp;":
	    "<td><sb-internal-revert hide=&nbsp; file=" + path +
	    " revision=" + e->revision + ">Revert</sb-internal-revert>"
	  }) : ({}))
	});
	nextdate = e->date;
      }
      res += html_table(({ "Action", "Date", "User", "Message", "View" }) +
			(revert ? ({ "Revert" }) : ({})),
			list);
    } else {
      // Unparsed file.
      object h = id->misc->sb->handler_for(ct, id);
      array(array(string)) list = ({});
      foreach (log, object(Sitebuilder.VCLogEntry) e) {
	list += ({
	  ({e->action || "",
	    replace(e->date, " ", "&nbsp;"),
	    e->user || "",
	    html_encode_string(e->message||""),
	    (e->action == "delete")?"&nbsp;":
	    (h->noview ?
	     (h->viewsource ?
	      "<sb-internal-viewsource hide=&nbsp; file=" + path +
	      " revision=" + e->revision + ">Source</sb-internal-viewsource>" :
	      "&nbsp;") :
	     "<sb-internal-rawview hide=&nbsp; file=" + path +
	     " revision=" + e->revision + ">View</sb-internal-rawview>")
	  }) +
	  (revert ? ({
	    (e->action == "delete")?"&nbsp;":
	    "<td><sb-internal-revert hide=&nbsp; file=" + path +
	    " revision=" + e->revision + ">Revert</sb-internal-revert>"
	  }) : ({}))
	});
      }
      res += html_table(({ "Action", "Date", "User", "Message", "View" }) +
			(revert ? ({ "Revert" }) : ({})),
			list);
    }
  } else {
    // View and Revert aren't meaningful for directories.
    array(array(string)) list = ({});
    foreach (log, object(Sitebuilder.VCLogEntry) e) {
      list += ({
	({e->action || "",
	  replace(e->date, " ", "&nbsp;"),
	  e->user || "",
	  html_encode_string(e->message||"")
	})
      });
    }
    res += html_table(({ "Action", "Date", "User", "Message" }),
		      list);
  }
  return parse_wizard_page (res, id, "done", 0, id->misc->build_url(id));
}
