// Update one file or several files in a directory tree.

// $Id: update.pike,v 1.8 1999/05/07 04:37:15 mast Exp $

inherit "_merge";

import Sitebuilder;

constant name = "Version Control//Update";
constant doc = "Update and merge in changes in this file or directory";
constant operation = "update";

#define ERRORMSG(text) \
  (text ? "<p><font color=darkred>" + text + "</font>\n" : "")


// _merge callbacks.

int is_valid_op (object vcobj, object id)
{
  return vcobj->is_valid_op (id, "update");
}

object(VCFileConflict) get_vcfc (object(VCFile) vcfile, object id)
{
  return vcfile->update (id);
}

static constant init_state_template = ([
  "msg_removed": "The file has been removed from the site.",
  "msg_use_a": "add your copy to the site",
  "msg_use_b": "discard your copy",
  "msg_no_choice_err":
    "You must choose whether to add your copy to the site or discard it.",
  "status_use_a": "Add your file",
  "status_use_b": "Discard your file",
  "status_rem_undecided": "<font color=darkred>Removed from site</font>",

  "msg_a_column": "&nbsp;&nbsp;You&nbsp;&nbsp;",
  "msg_b_column": "&nbsp;&nbsp;Site&nbsp;&nbsp;",

  "link_rawview_a": "<sb-internal-rawview file='$abspath' update=loc>"
    "your version</sb-internal-rawview>",
  "link_rawview_b": "the <sb-internal-rawview file='$abspath' update=rep>"
    "site version</sb-internal-rawview>",
  "link_rawview_old": "the <sb-internal-rawview file='$abspath' update=old>"
    "common ancestor</sb-internal-rawview>",
]);

void init_state (object(VCFileConflict) vcfc)
{
  mapping(string:mixed) state = init_state_template + ([]);
  string abspath = vcfc->this_vcfile->abspath (0, 0, 1);
  foreach (indices (state), string var)
    if (stringp (state[var])) state[var] = replace (state[var], "$abspath", abspath);
  vcfc->state = state;
}

mapping|void finish_use_a (object(VCFileConflict) vcfc)
{
  return vcfc->update_finish (1);
}

mapping|void finish_use_b (object(VCFileConflict) vcfc)
{
  return vcfc->update_finish (0);
}

mapping|void finish_use_new (object(VCFileConflict) vcfc,
			     mapping new_md, string new_data)
{
  return vcfc->update_finish (1, new_md, new_data);
}


// Select and manage a list of files to update.

string page_file_chooser (object id, object vcobj)
{
  if (!vcobj || !vcobj->isdir) return 0;

  /* Recursive operation is currently too expensive.
  mapping|array list = vcobj->list_recursive (id);
  if (mappingp (list) ||
      !sizeof (Array.filter (Array.filter (list, "isfilep"), is_valid_op, id)))
    return err_no_files_msg;
  */

  m_delete (id->variables, "merge_file_list");
  m_delete (id->variables, "merge_file");

  return ERRORMSG (id->misc->wizard_error) +
    "\n<p>Select files to update:"
    "\n<p><sb-internal-filelisting flik=submit name=merge_files is_valid_op=update>";
}

int verify_file_chooser (object id, object vcobj)
{
  if (sizeof ((id->variables->merge_files || "0") / "\0" - ({"0"}))) return 0;
  id->misc->wizard_error = "Please select at least one file.";
  return 1;
}

string dispatch_merge_dir (object id, string page, object vcobj)
{
  if (!vcobj || !vcobj->isdir) return "cancel"; // Bogus variables if this happens.
  if (!id->variables->merge_files) return "file_chooser";

  array(string) filenames = (id->variables->merge_files || "0") / "\0" - ({"0"});
  array(object(VCFile)) vcfiles =
    Array.filter (Array.map (filenames, vcobj->wa->vcobj, id) - ({0}), "isfilep");

  if (sizeof (vcfiles) == 1) {
    // Trig a direct redirect in dispatch_merge_dir_base to the
    // conflict pages if it's only a single file.
    id->variables["merge_file_0.x"] = id->variables["merge_file_0.y"] = 1;
    id->variables->merge_files = vcfiles[0]->abspath (0, 0, 1);
  }

  page = dispatch_merge_dir_base (id, vcfiles);
  if (page) return page;
  if (sizeof (id->misc->wizard_merge_files[1]) &&
      !sizeof (id->misc->wizard_merge_files[1] - ({0})))
    return "done";		// No file produced a vcfc. Just report this and end.
  return 0;
}

string wizard_done (object id, object vcobj)
{
  if (id->misc->wizard_merge_files &&
      sizeof (id->misc->wizard_merge_files[1]) &&
      !sizeof (id->misc->wizard_merge_files[1] - ({0})))
    return "The file(s) have been updated. None required manual inspection.";

  return ::wizard_done (id, vcobj);
}


// Complete the wizard automaton.

string conflict_prev (object id, string page, object vcobj)
{
  return id->variables->merge_file_list && "file_list" ||
    // Back up to the file chooser if we went directly to one file.
    vcobj->isdir && "file_chooser";
}

mapping(string:array) wizard_automaton = ::wizard_automaton + ([
  "start": ({			// Only redirects to the start page for a file or dir.
    lambda (object id, string page, object vcobj)
      {return vcobj->isfile ? "conflict" : "file_chooser";},
  }),

  "file_chooser": ({		// Page to choose the files to update.
    0,
    0,
    "file_list",
  }),

  "file_list": ({		// Page to summarize the merge state of several files.
    dispatch_merge_dir,
    "file_chooser",
    0,
  }),
]);
