inherit "wizard";
constant name= "User preferences//User preferences";
constant doc = ("Edit your user preferences");

int button_enabled(object id)
{
  return id->misc->sb && id->misc->sb->userprefs_edit_access (id);
}

static constant uservars = ({
  "nl", "editrows", "editcols", "editor", "default_workarea", 
});


string uv_nl_name = "Text download convention";
string uv_editrows_name = "Edit window rows";
string uv_editcols_name = "Edit window columns";
string uv_default_workarea_name = "Default work area";
string uv_editor_name = "Editor profile";

string uv_nl_help = "Set this variable to the type of computer you use to "
"download pages on, and text files will get the proper line break.";
string uv_editrows_help = "The number of rows the Edit "
"wizard should have.";
string uv_editcols_help = "The number of columns the Edit "
"wizard should have.";
string uv_default_workarea_help = "Your default workarea. If you choose the "
"Site Default Work Area your administrator can change it.";
string uv_editor_help = "The editor profile determines if you will use a "
"web-based or local editor to edit files.";

string uv_nl_display(object id, string nl)
{
  array(array(string)) nls = ({
    ({"UNIX", "\012"}),
    ({"IBM PC (and compatibles)", "\015\012"}),
    ({"Macintosh", "\015"}),
  });
  return "<select name=nl>"+
    Array.map(nls, lambda(array(string) nn, string nl) {
		     return "<option value='"+
		       replace(nn[1], ({"\012","\015"}), ({"&#10;","&#13;"}))+
		       "'"+(nl==nn[1]? " selected":"")+">"+nn[0];
		   }, nl)*"\n"+
    "</select>\n";
}

string uv_editrows_display(object id, int v)
{
  return "<input type=string name=editrows value='"+(v||20)+"'>";
}

string uv_editcols_display(object id, int v)
{
  return "<input type=string name=editcols value='"+(v||80)+"'>";
}

string uv_default_workarea_display(object id, int v)
{
  mapping uservars = (id->misc->sb->mac->id_get_data(id)||([]));
  string default_workarea = uservars->default_workarea;
  string site_default_workarea = id->misc->sb->get_default_workarea();
  string site_default_workarea_name
    = id->misc->sb->wa_lookup( site_default_workarea )->name;

  return "<sb-internal-select-workarea args=write-only "
    "selected=" + (default_workarea||"__extra__") + " "
    "extra_value=__default__ extra_name='Site Default Work Area ("
    + site_default_workarea_name + ")'>";
}

string uv_editor_display( object id, string v )
{
  string res = "<default value='" + v + "'>\n<select name=editor>\n";

  foreach (({ "Builtin" }) + id->misc->sb->list_editors(), string e)
    res += "<option>" + e + "\n";
  res += "</select>\n</default>\n";
  return res;
}

string uv_nl_update(object id, string v)
{
  return v;
}

int uv_editrows_update(object id, string v)
{
  return (int)v;
}

int uv_editcols_update(object id, string v)
{
  return (int)v;
}

string uv_default_workarea_update( object id, string v )
{
  if (v == "__default__")
    return 0;
  else
    return v;
}

string uv_editor_update( object id, string v )
{
  if (v == "Builtin")
    return 0;
  else
    return v;
}

#define UV_GET(u, name) (lambda(mixed v, mixed ... args){ \
  return (functionp(v)? v(@args):v); \
}(this_object()["uv_"+(u)+"_"+name], id, vars[(u)]))

string page_0(object id)
{
  array res = ({});
  mapping vars = id->misc->sb->mac->id_get_data(id)||([]);

  foreach(uservars, string uv) {
    res += ({ ({ UV_GET(uv, "name"), UV_GET(uv, "display"),
		 UV_GET(uv, "help")||"" }) });
  }
  return replace(html_table(({"Variable", "Setting", "Help"}), res),
		 "td nowrap", "td valign=top");
}

string wizard_done(object id)
{
  mapping vv = copy_value(id->misc->sb->mac->id_get_data(id)||([]));
  mapping vars = id->variables;

  foreach(uservars, string uv)
    if(vars[uv])
      {
	mixed n = UV_GET(uv, "update");
	if(!zero_type(n))
	  vv[uv] = n;
      }
  if (vars->select_workarea)
    vv["default_workarea"]
      = uv_default_workarea_update( id, vars->select_workarea );

  if(!id->misc->sb->mac->id_set_data(vv, id))
    return parse_wizard_page(html_error("Failed to update your preferences.",
					id), id, "done");

  return 0;
}
